/***************************************************************************
                          accountsmanager.cpp -  description
                             -------------------
    begin                : Sat May 3 2008
    copyright            : (C) 2008 by Valerio Pilo
    email                : valerio@kmess.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "accountsmanager.h"

#include "utils/kmessconfig.h"
#include "currentaccount.h"
#include "kmess.h"
#include "kmessapplication.h"
#include "kmessdebug.h"

#include <QDir>

#include <KLocale>


#ifdef KMESSDEBUG_ACCOUNTSMANAGER
  #define KMESSDEBUG_ACCOUNTSMANAGER_KWALLET
#endif



// Initialize the singleton instance to zero
AccountsManager* AccountsManager::instance_(0);



/**
 * Constructor
 */
AccountsManager::AccountsManager()
: passwordManager_( 0 )
{
}



// Destructor
AccountsManager::~AccountsManager()
{
  // Delete the accounts
  qDeleteAll( accounts_ );
  accounts_.clear();

#ifdef KMESSDEBUG_ACCOUNTSMANAGER
  kDebug() << "DESTROYED.";
#endif
}



// Add a new account to the list
void AccountsManager::addAccount( Account *account )
{
  // Append account
  accounts_.append( account );

#ifdef KMESSDEBUG_ACCOUNTSMANAGER
  kDebug() << "New account created:" << account->getHandle() << ".";
#endif

  emit accountAdded( account );
}



// An account's settings have been changed
void AccountsManager::changeAccount( Account *account, QString oldHandle, QString oldFriendlyName )
{
  QString newHandle = account->getHandle();

  // We're saving settings for a new account
  if( ! accounts_.contains( account ) )
  {
#ifdef KMESSDEBUG_ACCOUNTSMANAGER
    kDebug() << "Saving settings for new account" << newHandle;
#endif

    // Add the account to the account list.
    addAccount( account );
  }
#ifdef KMESSDEBUG_ACCOUNTSMANAGER
  else
  {
    kDebug() << "Saving settings for " << oldHandle;
  }
#endif


  CurrentAccount *currentAccount = CurrentAccount::instance();

  // If it's the current account that has been changed, copy the account settings there
  if( newHandle == currentAccount->getHandle() )
  {
    // Copy the account settings to the current account
#ifdef KMESSDEBUG_ACCOUNTSMANAGER
    kDebug() << "Copying modified Account settings to CurrentAccount.";
#endif

    currentAccount->copyAccount( account );
  }

  // If the account is set to use autologin, make sure that all other accounts do not
  if( ! account->isDeleted() && account->getUseAutologin() )
  {
    foreach( Account *otherAccount, accounts_ )
    {
      if( otherAccount != account )
      {
        otherAccount->setUseAutologin( false );
      }
    }
  }

  // Rename the account's directory if the handle has changed
  if( oldHandle != newHandle )
  {
#ifdef KMESSDEBUG_ACCOUNTSMANAGER
    kDebug() << "Moving account dir from" << oldHandle << "to" << newHandle;
#endif

    QString accountsPath = KMessConfig::instance()->getAccountsDirectory();

    QDir accountsDir( accountsPath );

    if( accountsDir.exists( newHandle ) )
    {
      kWarning() << "Destination account directory" << newHandle << "already exists!";
    }

    if( accountsDir.rename( oldHandle, newHandle ) )
    {
      kWarning() << "Unable to move account directory" << oldHandle << "to" << newHandle << "!";
    }
  }

  // Commit configuration changes to disk, now
  // Makes sure the settings are not lost if KMess is terminated or crashes.
  account->saveProperties();

  // Emit a signal so that the UI can be updated
  emit accountChanged( account, oldHandle, oldFriendlyName );
}



// Verify is there is a saved account with the specified handle
bool AccountsManager::contains( Account *account ) const
{
  return accounts_.contains( account );
}



// Delete the given account
void AccountsManager::deleteAccount( Account *account )
{
  if( KMESS_NULL(account) ) return;

  emit accountDeleted( account );

  QString handle = account->getHandle();
  KMessApplication *kmessApp = static_cast<KMessApplication*>( kapp );

  if( ( handle == CurrentAccount::instance()->getHandle() ) && kmessApp->getContactListWindow()->isConnected() )
  {
#ifdef KMESSDEBUG_ACCOUNTSMANAGER
  kDebug() << "Not deleting account " << handle << " yet.";
#endif

    // Already delete account data.
    // NOTE: when more classes have an Account* object, consider creating a setDeleted(true)
    // method, and let deleteAccountData() fire a deleteAccount(this) signal.
    account->setDeleted();
    return;
  }

#ifdef KMESSDEBUG_ACCOUNTSMANAGER
  kDebug() << "Deleting account for " << handle << ".";
#endif

  // Remove the account from the list of accounts
  if( accounts_.removeAll( account ) == 0 )
  {
    kWarning() << "Account" << handle << "not found in collection!";
  }

  // Delete the account directory
  KMessConfig::instance()->destroyConfigDir( KMessConfig::instance()->getAccountsDirectory() + "/" + handle );

  // Delete the account's contents
  account->setDeleted();
  account->deleteLater();
}



/**
 * Delete the instance of the accounts manager
 */
void AccountsManager::destroy()
{
  delete instance_;
  instance_ = 0;
}



// Return the account for a given handle
Account *AccountsManager::getAccountByHandle( const QString &handle )
{
  foreach( Account *account, accounts_ )
  {
    if( account->getHandle() == handle )
    {
      return account;
    }
  }

  return 0;
}



// Get the list of all saved accounts
const QList<Account *> AccountsManager::getAccounts() const
{
  return accounts_;
}



//Initialize the KWallet password manager
void AccountsManager::initializePasswordManager( bool block )
{
  // Don't initialize if it is already
  if( passwordManager_ != 0 )
  {
#ifdef KMESSDEBUG_ACCOUNTSMANAGER_KWALLET
    kDebug() << "KWallet is already initialized.";
#endif
    return;
  }

  // Obtain the window ID of the main KMess window
  KMessApplication *kmessApp = static_cast<KMessApplication*>( kapp );
  WId wId = kmessApp->getContactListWindow()->window()->winId();

  if( block )
  {
    // Open KWallet (synchronously - blocks KMess until the password is given)
    passwordManager_ = KWallet::Wallet::openWallet( KWallet::Wallet::LocalWallet(), wId, KWallet::Wallet::Synchronous );
    if( passwordManager_ == 0 )
    {
      kWarning() << "Opening KWallet failed!";
      return;
    }
    slotWalletOpen( true );
    return;
  }
  else
  {
    // Open KWallet (asynchronously - the slot will be called when - and if - the wallet will be open)
    passwordManager_ = KWallet::Wallet::openWallet( KWallet::Wallet::LocalWallet(), wId, KWallet::Wallet::Asynchronous );
    connect( passwordManager_, SIGNAL(   walletOpened(bool) ),
             this,             SLOT  ( slotWalletOpen(bool) ) );
  }

  // An internal KWallet error has occurred
  if( passwordManager_ == 0 )
  {
    kWarning() << "Opening KWallet failed!";
  }
}



// Read the passwords from the just opened wallet
void AccountsManager::slotWalletOpen( bool success )
{
  if( ! success )
  {
#ifdef KMESSDEBUG_ACCOUNTSMANAGER_KWALLET
    kDebug() << "Wallet opening failed, aborting.";
#endif

    delete passwordManager_;
    passwordManager_ = 0;
    return;
  }

  // Opening the wallet has failed, due to the user entering a wrong wallet password for example
  if( passwordManager_ == 0 || ! passwordManager_->isOpen() )
  {
#ifdef KMESSDEBUG_ACCOUNTSMANAGER_KWALLET
    kDebug() << "Opening KWallet failed, ref is" << passwordManager_;
#endif
    return;
  }

  // Look for an existing kmess folder and create one if needed
  if( ! passwordManager_->hasFolder( "KMess" ) )
  {
    passwordManager_->createFolder( "KMess" );
  }

  // Set the current folder for password storing, it should't fail
  if( ! passwordManager_->setFolder( "KMess" ) )
  {
#ifdef KMESSDEBUG_ACCOUNTSMANAGER_KWALLET
    kWarning() << "Setting KWallet folder failed!";
#endif
    return;
  }

#ifdef KMESSDEBUG_ACCOUNTSMANAGER_KWALLET
  kDebug() << "KWallet successfully opened, reading passwords.";
#endif

  QString password;

  // Read the password for each account
  foreach( Account *account, accounts_ )
  {
    if( passwordManager_->readPassword( account->getHandle(), password ) != 0 || password.isEmpty() )
    {
#ifdef KMESSDEBUG_ACCOUNTSMANAGER_KWALLET
      kDebug() << "Couldn't read password for account" << account->getHandle();
#endif
      continue;
    }

    account->setPassword( password );
  }

#ifdef KMESSDEBUG_ACCOUNTSMANAGER_KWALLET
  kDebug() << "Passwords have been read, emitting ready signal.";
#endif
  emit passwordsReady();

  return;
}



/**
 * Return a singleton instance of the accounts manager
 */
AccountsManager* AccountsManager::instance()
{
  // If the instance is null, create a new accounts manager and return that.
  if ( instance_ == 0 )
  {
    instance_ = new AccountsManager();
  }
  return instance_;
}



/**
 * Load the accounts from disk
 */
void AccountsManager::readProperties()
{
  // Obtain the list of account directories in the KMess config directory
  QStringList accountList( KMessConfig::instance()->getAccountsList() );

  // Add each one of the saved accounts to the internal list
  foreach( QString handle, accountList )
  {
    if( Account::isValidEmail( handle ) )
    {
      Account *account = new Account();
      account->readProperties( handle );
      addAccount( account );
    }
  }
}



// Show the settings dialog for a given account
void AccountsManager::showAccountSettings( Account *account, QWidget *parentWindow, AccountSettingsDialog::Page startingPage )
{
  // No account has been specified, create one
  if( ! account )
  {
#ifdef KMESSDEBUG_ACCOUNTSMANAGER
    kDebug() << "Opening settings for a new account";
#endif

    // Create an account to store the profile information
    account = new Account();

    // Assume the user *does* want to save it's settings,
    // otherwise he/she would have used the main login dialog.
    account->setGuestAccount( false );
  }
#ifdef KMESSDEBUG_ACCOUNTSMANAGER
  else
  {
    kDebug() << "Opening settings for " << account->getHandle();
  }
#endif

  CurrentAccount *currentAccount = CurrentAccount::instance();
  KMess *kmess = static_cast<KMessApplication*>( kapp )->getContactListWindow();
  bool isCurrentAccount = ( account->getHandle() == currentAccount->getHandle() );
  bool isConnectedCurrentAccount = isCurrentAccount && ( kmess->isConnected() );

  // Show the settings dialog
  AccountSettingsDialog *accountSettingsDialog = AccountSettingsDialog::instance( parentWindow );

  connect( accountSettingsDialog, SIGNAL(   deleteAccount(Account*)                 ),
           this,                  SLOT  (   deleteAccount(Account*)                 ) );
  connect( accountSettingsDialog, SIGNAL( changedSettings(Account*,QString,QString) ),
           this,                  SLOT  (   changeAccount(Account*,QString,QString) ) );

  // If it's the current account, copy some UI settings from the current account to the account
  if( isConnectedCurrentAccount )
  {
#ifdef KMESSDEBUG_ACCOUNTSMANAGER
    kDebug() << "Copying UI settings from CurrentAccount to the Account object.";
#endif
    account->copyAccountUISettings( currentAccount );
  }

  accountSettingsDialog->loadSettings( account, isConnectedCurrentAccount, startingPage );
  accountSettingsDialog->show();
  accountSettingsDialog->raise();
}



// Store the password of all accounts to the user's KWallet
void AccountsManager::storeAccountsPasswords()
{
  QString handle;
  QString password;

  // If the user didn't activate the KWallet, do nothing as it isn't available now
  if( passwordManager_ == 0 )
  {
#ifdef KMESSDEBUG_ACCOUNTSMANAGER_KWALLET
    kDebug() << "KWallet is not ready, not saving passwords.";
#endif
    return;
  }

  // Store the password for each account
  foreach( Account *account, accounts_ )
  {
    handle   = account->getHandle();
    password = account->getPassword();

    // If the password is empty, delete the whole account entry from the wallet
    if( passwordManager_->hasEntry( handle ) && password.isEmpty() )
    {
#ifdef KMESSDEBUG_ACCOUNTSMANAGER_KWALLET
      kDebug() << "Account" << handle << "removed from kwallet";
#endif

      passwordManager_->removeEntry( handle );
      continue;
    }

    // Store the password
    if( passwordManager_->writePassword( handle, password ) != 0 )
    {
#ifdef KMESSDEBUG_ACCOUNTSMANAGER_KWALLET
      kDebug() << "Couldn't store password for account" << handle;
#endif
    }
  }
}



#include "accountsmanager.moc"
