/***************************************************************************
                          chat.cpp  -  description
                             -------------------
    begin                : Wed Jan 15 22:41:32 CST 2003
    copyright            : (C) 2003 by Mike K. Bennett
    email                : mkb137b@hotmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "chat.h"

#include "../contact/contactbase.h"
#include "../contact/contactlist.h"
#include "../network/chatinformation.h"
#include "../network/chatmessage.h"
#include "../network/msnswitchboardconnection.h"
#include "../network/specialgroups.h"
#include "../utils/richtextparser.h"
#include "../utils/xmlfunctions.h"
#include "../accountaction.h"
#include "../currentaccount.h"
#include "../kmessapplication.h"
#include "../kmessdebug.h"
#include "../kmess.h"
#include "chatmessageview.h"
#include "chatwindow.h"
#include "contactaction.h"
#include "contactsidebar.h"

#include <QApplication>
#include <QDir>
#include <QFile>
#include <QLayout>
#include <QPushButton>
#include <QRegExp>
#include <QTextCodec>
#include <QTextDocument>
#include <QToolBox>

#include <KAction>
#include <KActionMenu>
#include <KFileDialog>
#include <KGlobal>
#include <KLocale>
#include <KMenuBar>
#include <KMessageBox>
#include <KMenu>
#include <KNotification>
#include <KProcess>
#include <KStatusBar>
#include <KTextBrowser>
#include <KTextEdit>
#include <KToolBar>
#include <KWindowSystem>


#ifdef KMESSDEBUG_CHAT
  #define KMESSDEBUG_CHAT_GENERAL
  #define KMESSDEBUG_CHAT_CONTACTS
  #define KMESSDEBUG_CHAT_FILETRANSFER
//   #define KMESSDEBUG_CHAT_TYPING_MESSAGES
#endif


/**
 *  Define the maximum interval (in seconds) between each update of
 *  the status bar label, when displaying "contact is typing" messages.
 */
#define CHAT_MAXIMUM_TYPING_TIME         5



// The constructor
Chat::Chat( QWidget *parent )
: ChatView( parent )
, firstMessage_( true )
, initialized_( false )
, lastContact_( QString() )
, msnSwitchboardConnection_( 0 )
{
  hide();

  // Setup the contact typing information timer
  typingTimer_.setSingleShot( true );
  connect( &typingTimer_, SIGNAL(       timeout() ),
           this,          SLOT  ( contactTyping() ) );
}



// The destructor
Chat::~Chat()
{
#ifdef KMESSDEBUG_CHAT_GENERAL
  kDebug() << "DESTROYED.";
#endif
}



// Choose the contact to start an invitation with.
QString Chat::chooseContact()
{
  if( KMESS_NULL(msnSwitchboardConnection_) ) return QString();

  // Choose the contact.
  const QStringList &contactsInChat = msnSwitchboardConnection_->getContactsInChat();
  switch( contactsInChat.count() )
  {
    case 0:
      // No contacts in the chat
      if( msnSwitchboardConnection_->getLastContact().isEmpty() )
      {
        // Send invitation to the first contact that should appear in the chat.
        return msnSwitchboardConnection_->getFirstContact();
      }
      else
      {
        // Resume chat with last contact
        return msnSwitchboardConnection_->getLastContact();
      }

    case 1:
      // Choose the only contact available in the chat
      return contactsInChat.first();

    default:
      // Multiple contacts in the chat.
      // TODO: The official client opens a contact-choose dialog here, and starts a new chat with the selected contact.
      KMessageBox::sorry( this, i18n("You can't start this invitation because there are multiple contacts in this chat.") );
      return QString();
  }
}



// A contact joined the chat
void Chat::contactJoined(QString handle, QString friendlyName)
{
  ContactAction *contactAction;

  if( ! participants_.contains( handle ) )
  {
    participants_.prepend( handle );
  }

  lastContact_ = QString::null;

  contactAction = getContactActionByHandle( handle );
  if ( contactAction != 0 )
  {
    contactAction->setInChat( true );
  }

  // Show message when the user has the sidebar disabled,
  // or when a multi-chat has started.
  bool isSidebarVisible = contactSidebar_->isVisible();
  if( ( ! isSidebarVisible && ! currentAccount_->getShowSidebar() )
  ||  ( msnSwitchboardConnection_ && msnSwitchboardConnection_->getContactsInChat().size() > 1 ) )
  {
    showMessage( ChatMessage( ChatMessage::TYPE_NOTIFICATION,
                                         ChatMessage::CONTENT_NOTIFICATION_PRESENCE,
                                         false,
                                         i18n("%1 has joined the chat.", friendlyName),
                                         handle,
                                         friendlyName ) );
  }


  // Change the chat title
  emit chatInfoChanged();

  // Allow future changes in the nickname to update the chat title again
  ContactBase *contact = currentAccount_->getContactByHandle( handle );
  if( contact )
  {
    connect( contact, SIGNAL( changedFriendlyName() ), this, SIGNAL( chatInfoChanged() ) );
  }
}



// A contact left the chat
void Chat::contactLeft(QString handle, bool isChatIdle )
{
#ifdef KMESSDEBUG_CHAT_GENERAL
  kDebug() << "contact '" << handle << "' has left" << (isChatIdle ? " due to inactivity." : "") << ".";
#endif

  ContactAction *contactAction = getContactActionByHandle( handle );

  if ( contactAction != 0 )
  {
    contactAction->setInChat( false );
  }

  participants_.removeAll( handle );

  // This was the last participant, save it as the last being in so we can
  // still recognize this window as a chat with this contact
  if( participants_.isEmpty() )
  {
    lastContact_ = handle;
  }

  // Show message when the user has the sidebar disabled,
  // or a user left a multi-chat
  bool isSidebarVisible = contactSidebar_->isVisible();
  if( ( ! isSidebarVisible && ! currentAccount_->getShowSidebar() )
  ||  ( msnSwitchboardConnection_ && msnSwitchboardConnection_->getContactsInChat().size() > 0 ) )
  {
    QString contactName = currentAccount_->getContactFriendlyNameByHandle( handle );
    QString message;

    if( isChatIdle )
    {
      message = i18n( "The conversation went idle, %1 has left the chat.", contactName );
    }
    else
    {
      message = i18n( "%1 has left the chat.", contactName );
    }

    showMessage( ChatMessage( ChatMessage::TYPE_NOTIFICATION,
                                         ChatMessage::CONTENT_NOTIFICATION_PRESENCE,
                                         false,
                                         message,
                                         handle,
                                         contactName ) );
  }

  // Change the caption to remove the contact who has left
  emit chatInfoChanged();

  // Disconnect the contact to avoid it updating the chat title
  ContactBase *contact = currentAccount_->getContactByHandle( handle );
  if( contact )
  {
    disconnect( contact, SIGNAL( changedFriendlyName() ), this, SIGNAL( chatInfoChanged() ) );
  }
}



// A contact is typing
// If called without arguments, it will only update the list of typing contacts.
void Chat::contactTyping( const QString &handle, const QString &friendlyName, bool forceExpiration )
{
  int secsTo;
  int nextUpdate = CHAT_MAXIMUM_TYPING_TIME; // Allow a maximum number of seconds between each update

#ifdef KMESSDEBUG_CHAT_TYPING_MESSAGES
  kDebug() << "Called with handle=" << handle << ", friendlyName=" << friendlyName;
  kDebug() << "Current list:" << typingContacts_;
#endif

  // Do not add anything to the list if the friendly name is empty
  if( ! friendlyName.isEmpty() )
  {
    QTime startTime = QTime::currentTime();

    // If a contact is still typing, re-insert it in the list, so that the typing start time is updated
    if( typingContacts_.contains( friendlyName ) )
    {
      typingContacts_.remove( friendlyName );
    }

    // When we receive a message from this contact, we only remove it from the typing contacts
    if( ! forceExpiration )
    {
      // Add this contact to the list of contacts which are typing, and record when it has started writing
      startTime.start();
      typingContacts_.insert( friendlyName, startTime );
    }
  }

  // Update the list to remove expired typing events
  QMutableMapIterator<QString,QTime> it( typingContacts_ );
  while( it.hasNext() )
  {
    it.next();
    secsTo = (int) ( it.value().elapsed() * .001 ); // Only keep the seconds

    // If the typing event has expired, stop displaying it
    if( secsTo >= CHAT_MAXIMUM_TYPING_TIME )
    {
      typingContacts_.remove( it.key() );
    }

    // Find the typing event which will expire first, so we can next update when it expires
    if( secsTo && secsTo < nextUpdate )
    {
      nextUpdate = secsTo;
    }
  }

#ifdef KMESSDEBUG_CHAT_TYPING_MESSAGES
  kDebug() << "List after cleanup and update:" << typingContacts_;
#endif

  // The more time passes from the last update, the more we get close to the next update
  if( nextUpdate < 5 )
  {
    nextUpdate = CHAT_MAXIMUM_TYPING_TIME - nextUpdate;
  }

#ifdef KMESSDEBUG_CHAT_TYPING_MESSAGES
  kDebug() << "Next update in" << nextUpdate << "seconds.";
#endif

  // Update the user interface
  if( typingContacts_.contains( friendlyName ) )
  {
    contactSidebar_->contactTyping( handle, friendlyName );
  }
  else
  {
    contactSidebar_->messageReceived( handle );
  }

#ifdef KMESSDEBUG_CHAT_TYPING_MESSAGES
  kDebug() << "emitting signal from:" << this;
#endif

  // Let the other window components update, too
  emit gotTypingMessage( this );

  // Activate the timer for the next update if needed
  if( ! typingContacts_.isEmpty() )
  {
    typingTimer_.start( nextUpdate * 1000 );
  }
}



// Return a list of the contacts in the chat to be used as window caption
QString Chat::getCaption()
{
  QStringList activeContacts;

  // The chat has no switchboard connection.. it is not even a chat!
  if( msnSwitchboardConnection_ == 0 )
  {
    // If the last contact is empty, it's probably a test chat window.
    if( ! lastContact_.isEmpty() )
    {
      activeContacts << lastContact_;
    }
  }
  else
  {
    // Query the switchboard for the list of participants
    activeContacts = msnSwitchboardConnection_->getContactsInChat();
  }

  // How many people are there in here?
  switch( activeContacts.count() )
  {
    // The chat is empty
    case 0:
      return QString();

    // One contact is connected; or the only contact of the session has left
    case 1:
      return currentAccount_->getContactFriendlyNameByHandle( activeContacts[0], STRING_CLEANED );

    // Two contacts are connected
    case 2:
      return i18nc( "Name of a chat tab", "%1 and %2",
                    currentAccount_->getContactFriendlyNameByHandle( activeContacts[0], STRING_CLEANED ),
                    currentAccount_->getContactFriendlyNameByHandle( activeContacts[1], STRING_CLEANED )
                    );

    // Three or more contacts are connected
    default:
      return i18nc( "Name of a chat tab", "%1 et al.",
                    currentAccount_->getContactFriendlyNameByHandle( activeContacts[0], STRING_CLEANED )
                    );
  }
}



// Return the chat window which currently contains this Chat
ChatWindow *Chat::getChatWindow()
{
  return static_cast<ChatWindow*>( window() );
}



// Return the list of contacts in the chat
QStringList Chat::getParticipants()
{
  // If all contacts have left the chat, return the last contact as being in chat,
  // even if it's not technically true
  if( participants_.isEmpty() )
  {
    return QStringList() << lastContact_;
  }

  return participants_;
}



// Return the list of contacts which are typing
QStringList Chat::getTypingContacts()
{
  return typingContacts_.keys();
}



// Return the switchboard connection used by the chat window.
MsnSwitchboardConnection * Chat::getSwitchboardConnection() const
{
  return msnSwitchboardConnection_;
}



// Initialize the object
bool Chat::initialize( MsnSwitchboardConnection *switchboardConnection )
{
  if ( initialized_ )
  {
    kDebug() << "already initialized.";
    return false;
  }

  if( ! ChatView::initialize() )
  {
    kDebug() << "Couldn't setup the chat view widget.";
    return false;
  }

  // Connect the sidebar slots
  connect( contactSidebar_, SIGNAL( setContactAllowed(QString)      ),
           this,            SIGNAL( setContactAllowed(QString)      ) );
  connect( contactSidebar_, SIGNAL(   setContactAdded(QString,bool) ),
           this,            SIGNAL(   setContactAdded(QString,bool) ) );
  connect( contactSidebar_, SIGNAL( setContactBlocked(QString,bool) ),
           this,            SIGNAL( setContactBlocked(QString,bool) ) );

  // Save the switchboard connection
  setSwitchboardConnection( switchboardConnection );

  // Pre-register the contacts as participants to this chat.
  lastContact_  = switchboardConnection->getFirstContact();
  participants_ = switchboardConnection->getContactsInChat();

  // Save the date/time the chat started.
  // Use it to save a chatlog later.
  startDate_ = QDate::currentDate();
  startTime_ = QTime::currentTime();

  // Update the chat window caption with the contact name
  emit chatInfoChanged();

  initialized_ = true;
  return true;
}



// Invite a contact to the chat
void Chat::inviteContact( QString handle )
{
  if( msnSwitchboardConnection_ == 0 )
  {
    return;
  }

  msnSwitchboardConnection_->inviteContact( handle );
}



// Return whether or not the chat is at its first incoming message
bool Chat::isChatFirstMessage()
{
  return firstMessage_;
}



// Return whether or not the contact is in this chat.
bool Chat::isContactInChat( const QString &handle )
{
#ifdef KMESSDEBUG_CHAT_CONTACTS
  kDebug() << "Checking handle " << handle << " - participants: " << participants_.join(",") << " - lastContact: " << lastContact_;
#endif

  if( participants_.isEmpty() )
  {
#ifdef KMESSDEBUG_CHAT_CONTACTS
    kDebug() << "returning " << ( lastContact_ == handle ? "true" : "false" );
#endif
    return ( lastContact_ == handle );
  }

#ifdef KMESSDEBUG_CHAT_CONTACTS
  kDebug() << "returning " << ( participants_.contains( handle ) ? "true" : "false" );
#endif
  return participants_.contains( handle );
}



// Return whether or not this is an exclusive chat with the given contact
bool Chat::isExclusiveChatWithContact( const QString& handle )
{
#ifdef KMESSDEBUG_CHAT_CONTACTS
  kDebug() << "Checking handle " << handle << " - participants: " << participants_.join(",") << " - lastContact: " << lastContact_;
#endif
  if( participants_.isEmpty() )
  {
#ifdef KMESSDEBUG_CHAT_CONTACTS
    kDebug() << "returning " << ( lastContact_ == handle ? "true" : "false" );
#endif
    return ( lastContact_ == handle );
  }

#ifdef KMESSDEBUG_CHAT_CONTACTS
  kDebug() << "returning " << ( participants_.contains(handle) && participants_.count() < 2  ? "true" : "false" );
#endif
  return ( participants_.contains(handle) && participants_.count() < 2 );
}



// The chat is closing
void Chat::queryClose()
{
#ifdef KMESSDEBUG_CHAT_GENERAL
  kDebug() << "Saving chat log";
#endif

  // Notify ChatMaster
  emit closing( this );

  // Save the chat
  saveChatAutomatically();

  /**
   * TODO: Uncomment this part: it needs a _reliable_ method to check whether there are running file
   * transfers. The commented part uses "ApplicationList::isEmpty()" which I've removed, the new
   * method to check will probably be elsewhere. The main code of that method was this:
   * <code>
      // check if there are active applications
      if ( mimeApplications_.count() +
          p2pApplications_.count() -
          abortingApplications_.count() == 0)
        return true; else return false;
   * </code>
   */
  /*
  // If there are active applications show a warning to the user
  if( participants_.count() == 1 ) // Transfers are not possibile in multiple chats
  {
    // Check if there are active applications
    ContactBase *contact = currentAccount_->getContactByHandle( participants_.first() );
    if( contact != 0 && contact->getApplicationList() != 0
        && ! contact->getApplicationList()->isEmpty() )
    {
      int choice = KMessageBox::warningContinueCancel( this, i18n("Continue closing the chat window?\nActive transfers will be aborted!"));
      if( choice == KMessageBox::Cancel )
      {
        return false;
      }
    }
  }
  */
}



// The application is exiting
void Chat::queryExit()
{
#ifdef KMESSDEBUG_CHAT_GENERAL
  kDebug() << "Saving chat log";
#endif

  // Just save the chat before going down
  saveChatAutomatically();
}



// A message was received from a contact.
void Chat::receivedMessage(const ChatMessage &message)
{
#ifdef KMESSDEBUG_CHAT_GENERAL
  kDebug() << "Displaying message, internal type=" << message.getType();
  kDebug() << "First message = " << firstMessage_ << ".";
#endif

  // Find if it's the first message of the chat
  bool isFirstMessage = isEmpty();

  showMessage( message );

  // If this is the first message received...
  if( firstMessage_ )
  {
    // Change the first message property only when we're actually receiving the second message;
    // this way the notification balloons can work correctly. (and maybe future features requiring
    // correct detection of the first message!)
    if( ! isFirstMessage )
      firstMessage_ = false;
  }

  // Check if we should send an automatic reply to the chat message
  if( currentAccount_           != 0
  &&  msnSwitchboardConnection_ != 0
  &&  currentAccount_->getAutoreply()
  && ( lastSentAutoMessage_.isNull() || lastSentAutoMessage_.elapsed() > 30000 ) )
  {
    // Send an auto away message every 30 sec (if the contact keeps writing)

#ifdef KMESSDEBUG_CHAT_GENERAL
    kDebug() << "Sending auto reply to chat message.";
#endif

    QString awayMessage = i18nc( "Automatic reply message",
                                  "%1 (This message was sent automatically)",
                                  currentAccount_->getAutoreplyMessage() );

    // Send the autoreply message to the contact(s)
    msnSwitchboardConnection_->sendChatMessage( awayMessage );

    // Show the autoreply message
    ChatMessage autoMessage = ChatMessage( ChatMessage::TYPE_OUTGOING,
                                            ChatMessage::CONTENT_MESSAGE,
                                            false,
                                            awayMessage,
                                            currentAccount_->getHandle(),
                                            currentAccount_->getFriendlyName(),
                                            currentAccount_->getPicturePath(),
                                            currentAccount_->getFont(),
                                            currentAccount_->getFontColor() );
    showMessage( autoMessage );

    // Save the time the message was sent, to display the next after some time
    lastSentAutoMessage_.start();
  }

  // Remove this contact from the list of typing messages
  contactTyping( message.getContactHandle(), message.getContactName(), true );

  // Emit the new message signal so that the user will receive a notification
  emit gotChatMessage( message, this );

  // Put back the focus to the message edit box
  getMessageEditField()->setFocus();
}



// Save the chat if necessary
void Chat::saveChatAutomatically()
{
  QTime       currentTime;
  QDate       currentDate;
  QString     basePath, year, month, day, hour, minute, time, file, savePath;
  QDir        dir;
  int         saveStructure;


  // Check if "save chats" is enabled
  if ( ! currentAccount_->getSaveChats() )
  {
#ifdef KMESSDEBUG_CHAT_GENERAL
    kDebug() << "User has save chats disabled, not saving chat.";
#endif
    return;
  }


  // Check if the message area is not empty
  if(isEmpty() )
  {
#ifdef KMESSDEBUG_CHAT_GENERAL
    kDebug() << "Message area is empty, not saving chat.";
#endif
    return;
  }

  // Get the base path
  basePath = currentAccount_->getSaveChatPath();

  dir.setPath( basePath );
  if( ! dir.exists() )
  {
    // TODO: Add a message box to notify the chat log folder does not exist (and avoid displaying it when KMess is forced to quit).
    kWarning() << "Base directory '" << basePath << "' doesn't exist.  Check your settings.";
    return;
  }

#ifdef KMESSDEBUG_CHAT_GENERAL
  kDebug() << "Automatically saving the chat.";
#endif

  // Get date strings with numbers padded.
  year    .sprintf("%04d", startDate_.year()   );
  month   .sprintf("%02d", startDate_.month()  );
  day     .sprintf("%02d", startDate_.day()    );
  hour    .sprintf("%02d", startTime_.hour()   );
  minute  .sprintf("%02d", startTime_.minute() );
  time = hour + ":" + minute;

#ifdef KMESSDEBUG_CHAT_GENERAL
  kDebug() << "Date is " << year << "/" << month << "/" << day << " " << hour << ":" << minute << ".";
#endif

  // Determine the path to save
  saveStructure = currentAccount_->getSavedChatDirectoryStructure();

  switch(saveStructure)
  {
    case Account::BYYEAR :
    {
      // save as chatpath/yyyy/mmdd hh:mm contactemail.html
      if(! dir.exists(year)) dir.mkdir(year);
      dir.cd(year);
      file = "/" + month + "-" + day + " ";
      break;
    }
    case Account::BYMONTH :
    {
      // save as chatpath/yyyy/mm/dd hh:mm contactemail.html
      if(! dir.exists(year))  dir.mkdir(year);
      dir.cd(year);
      if(! dir.exists(month)) dir.mkdir(month);
      dir.cd(month);
      file = "/" + day + " ";
      break;
    }
    case Account::BYDAY :
    {
      // save as chatpath/yyyy/mm/dd/hh:mm contactemail.html
      if(! dir.exists(year))  dir.mkdir(year);
      dir.cd(year);
      if(! dir.exists(month)) dir.mkdir(month);
      dir.cd(month);
      if(! dir.exists(day))   dir.mkdir(day);
      dir.cd(day);
      file = "/";
      break;
    }
    case Account::SINGLEDIRECTORY :
    default :
    {
      // save as chatpath/yyyymmdd hh:mm contactemail.html
      file = "/" + year + "-" + month + "-" + day + " ";
    }
  }

  savePath = dir.absolutePath() + file + time + " ";
  if( participants_.isEmpty() )
  {
    savePath += lastContact_;
  }
  else
  {
    savePath += participants_.first();
  }


  // Make sure chat logs are not automatically overwritten.
  if( QFile::exists( savePath + ".html" ) )
  {
    kWarning() << "Chat log " << savePath << ".html already exists!";
    for( int i = 1; i < 100; i++ )
    {
      if( ! QFile::exists(savePath + "." + QString::number(i) + ".html") )
      {
        savePath += "." + QString::number(i);
        break;
      }
    }
  }

  // Save the chat
  saveChatToFile( savePath + ".html" );
}



// Enable or disable the parts of the window which allow user interaction
void Chat::setEnabled( bool isEnabled )
{
  // Also remove the switchboard connection when disabling
  if( ! isEnabled && msnSwitchboardConnection_ != 0 )
  {
    setSwitchboardConnection( 0 );
  }

  // Do not reactivate if there is no switchboard connection (ChatMaster will recreate it for us)
  if( isEnabled && msnSwitchboardConnection_ == 0 )
  {
    return;
  }

  ChatView::setEnabled( isEnabled );
}



// Change the switchboard connection used by the chat window.
void Chat::setSwitchboardConnection( MsnSwitchboardConnection *newConnection )
{
  QStringList participants;
  QStringList::Iterator it;

  // Everything is connected already.
#ifdef KMESSDEBUG_CHAT_GENERAL
  if( msnSwitchboardConnection_ == 0 )
  {
    kDebug() << "Switchboard was not set up.";
  }
  else
  {
    kDebug() << "Switchboard was "
              << ( msnSwitchboardConnection_ == newConnection ? "already set" : "different" ) << "." << endl;
  }
#endif

  if( msnSwitchboardConnection_ != 0 )
  {
    // Notify that all contacts have left the chat
    if( ! msnSwitchboardConnection_->isEmpty() )
    {
      // getContactsInChat() returns the last contact if the SB is empty.
      participants = msnSwitchboardConnection_->getContactsInChat();
#ifdef KMESSDEBUG_CHAT_GENERAL
      kDebug() << "Removing old switchboard's participants from the chat:" << participants;
#endif
      foreach( QString handle, participants )
      {
        contactSidebar_->contactLeft( handle, false );
      }
    }

    // Delete all signals between us and the old switchboard
    disconnect( this, 0, msnSwitchboardConnection_, 0 );
    disconnect( msnSwitchboardConnection_, 0, contactSidebar_, 0 );
    disconnect( msnSwitchboardConnection_, 0, this, 0 );

    // Also delete the old connection, if we're setting up a new one
    if( msnSwitchboardConnection_ != newConnection )
    {
      msnSwitchboardConnection_->deleteLater();
    }
  }

  msnSwitchboardConnection_ = newConnection;

  if( newConnection == 0 )
  {
       connect( this, SIGNAL( sendMessageToContact(QString) ),
                this, SLOT  (    slotSendingFailed(QString) ) );
    return;
  }
  else
  {
    disconnect( this, SIGNAL( sendMessageToContact(QString) ),
                this, SLOT  (    slotSendingFailed(QString) ) );
  }

  // Make the new signals connections
  connect( msnSwitchboardConnection_, SIGNAL(        contactJoinedChat(QString, QString)             ),
           contactSidebar_,           SLOT  (            contactJoined(QString, QString)             ) );
  connect( msnSwitchboardConnection_, SIGNAL(          contactLeftChat(QString,bool)                 ),
           contactSidebar_,           SLOT  (              contactLeft(QString,bool)                 ) );
  connect( msnSwitchboardConnection_, SIGNAL(            contactTyping(QString, QString)             ),
           contactSidebar_,           SLOT  (            contactTyping(QString, QString)             ) );

  connect( msnSwitchboardConnection_, SIGNAL(        contactJoinedChat(QString, QString)             ),
           this,                      SLOT  (            contactJoined(QString, QString)             ) );
  connect( msnSwitchboardConnection_, SIGNAL(          contactLeftChat(QString,bool)                 ),
           this,                      SLOT  (              contactLeft(QString,bool)                 ) );
  connect( msnSwitchboardConnection_, SIGNAL(            contactTyping(QString, QString)             ),
           this,                      SLOT  (            contactTyping(QString, QString)             ) );
  connect( msnSwitchboardConnection_, SIGNAL(              chatMessage(const ChatMessage&)           ),
           this,                      SLOT  (          receivedMessage(const ChatMessage&)           ) );
  connect( msnSwitchboardConnection_, SIGNAL(            receivedNudge(const QString&)               ),
           this,                      SLOT  (        slotReceivedNudge(const QString&)               ) );
  connect( msnSwitchboardConnection_, SIGNAL(            sendingFailed(QString)                      ),
           this,                      SLOT  (        slotSendingFailed(QString)                      ) );
  connect( msnSwitchboardConnection_, SIGNAL(                 deleteMe(MsnSwitchboardConnection*)    ),
           this,                      SLOT  ( setSwitchboardConnection()                             ) );

  connect( this,                      SIGNAL(     sendMessageToContact(QString)                      ),
           msnSwitchboardConnection_, SLOT  (          sendChatMessage(QString)                      ) );
  connect( this,                      SIGNAL(      sendInkToContact(QByteArray)                      ),
           msnSwitchboardConnection_, SLOT  (               sendInk(QByteArray)                      ) );
  connect( this,                      SIGNAL(             userIsTyping()                             ),
           msnSwitchboardConnection_, SLOT  (        sendTypingMessage()                             ) );


  // Add to the sidebar all the contacts that are present in the new switchboard
  participants = msnSwitchboardConnection_->getContactsInChat();
#ifdef KMESSDEBUG_CHAT_GENERAL
  kDebug() << "Adding new switchboard's participants to the chat:" << participants;
#endif
  foreach( QString handle, participants )
  {
    contactSidebar_->contactJoined( handle, QString() ); // The sidebar slot doesn't use the second parameter
  }
}



// Show a wink received by a contact
void Chat::showWink( const QString &handle, const QString &filename, const QString &/*animationName*/ )
{
#ifdef KMESSDEBUG_CHAT_GENERAL
  kDebug() << "Processing received wink data.";
#endif

  // TODO: move the processing of wink files to a special "WinkArchive" class.

  // Test whether the file exists, before cabextract fails.
  if( ! QFile::exists(filename) )
  {
    kWarning() << "file not found: " << filename << "!";
    return;
  }

  // Remove extension
  QString dirName = QString(filename).remove( QRegExp("\\.[a-z]+$") );

  QFile metaFile(dirName + "/content.xml");

  // See if the file was extracted already
  if( ! metaFile.exists() )
  {
    // Run cabextract to get the contents of the file
    // -q: quiet, only print errors, -d: output dir, -l: force lowercase (avoiding incompatibilities because win32 is case-insensitive)
    KProcess cabextract(this);
    cabextract.closeReadChannel( QProcess::StandardOutput );
    cabextract.closeReadChannel( QProcess::StandardError );
    cabextract << "cabextract" << "-q" << "-d" << dirName << filename;
    bool done = cabextract.execute( 5000 );
    // TODO: capture `cabextract` error messages, display nicely.

    if( done != 0 )
    {
      showMessage( ChatMessage( ChatMessage::TYPE_SYSTEM,
                                           ChatMessage::CONTENT_SYSTEM_ERROR,
                                           false,
                                           i18n("The wink could not be displayed. Make sure you have 'cabextract' installed."),
                                           handle ) );
      return;
    }
    if( cabextract.exitStatus() != QProcess::NormalExit || ! metaFile.exists() )
    {
      showMessage( ChatMessage( ChatMessage::TYPE_SYSTEM,
                                           ChatMessage::CONTENT_SYSTEM_ERROR,
                                           false,
                                           i18n("The wink could not be displayed. The data could not be read."),
                                           handle ) );
      return;
    }
  }

  // Open the file
  if( ! metaFile.open( QIODevice::ReadOnly ) )
  {
    kWarning() << "file could not be opened: " << metaFile.fileName() << "!";
    return;
  }

  // Parse the XML
  QString xmlErrorMessage;
  QDomDocument xml;
  if( ! xml.setContent(&metaFile, true, &xmlErrorMessage) )
  {
    kWarning() << "parsing of wink XML failed: " << xmlErrorMessage << "!";
    return;
  }

  // Get the package nodes
  QDomElement packageNode = xml.namedItem("package").toElement();
  if( packageNode.isNull() )
  {
    kWarning() << "XML element /package not found!";
    return;
  }

  if( packageNode.attribute("type") != "wink" )
  {
    kWarning() << "XML element /package/@type is not 'wink'!";
    return;
  }

  if( packageNode.attribute("contenttype") != "D" )
  {
    kWarning() << "XML element /package/@contenttype is not 'D'!";
  }

  // Get the child nodes
  QString thumbnail;
  QString animation;
  QString animationType;
  QDomNodeList packageItems = packageNode.childNodes();
  for(int i = 0; i < packageItems.count(); i++)
  {
    QDomElement child = packageItems.item(i).toElement();
    if( child.tagName() == "item" )
    {
      QString type = child.attribute("type");
      if( type == "animation" )
      {
        animation = child.attribute("file");
        animationType = child.attribute("mimetype");
      }
      else if( type == "thumbnail-show" || type == "thumbnail" )
      {
        thumbnail = child.attribute("file");
        // Type not stored. There are winks with mimetype="image/png" while they're sending a JPG.
      }
      else
      {
        kWarning() << "unknown wink element type: '" << type << "'!";
      }
    }
  }

  // TODO: clean up weird characters from wink file name.

  if( thumbnail.isEmpty() || animation.isEmpty() || animationType.isEmpty() )
  {
    kWarning() << "XML elements are missing, could not play wink!";
    return;
  }

  // Escape data for inclusion in HTML.
  animation     = Qt::escape(animation).replace("'", "&apos;");  // quotes are not escaped by QStyleSheet::escape().
  animationType = animationType.remove( QRegExp("[^a-zA-Z0-9\\-/]") );

  // Height of 150 is a lot, but most winks are designed to be full-screen, so they can hardly be displayed smaller.
  QString html = "<div class='winkContainer'><object width='90%' height='150' align='center' classid='clsid:D27CDB6E-AE6D-11cf-96B8-444553540000'>"
                 "<param name='movie' value='" + dirName + "/" + animation + "' />"
                 "<param name='quality' value='autohigh' />"
                 "<param name='play' value='true' />"
                 "<param name='loop' value='false' />"
                 "<param name='menu' value='false' />"
                 "<param name='align' value='center' />"
                 "<param name='bgcolor' value='#ffffff' />"
                 "<param name='wmode' value='transparent' />"
                 "<embed width='90%' height='150' align='center'"
                 " src='" + dirName + "/" + animation + "' type='" + animationType + "'"
                 " play='true' loop='false' menu='false'"
                 " bgcolor='#ffffff' quality='autohigh' wmode='transparent'></embed></object></div>";

  // Show a notification
  QString friendlyName = currentAccount_->getContactFriendlyNameByHandle(handle);
  receivedMessage( ChatMessage( ChatMessage::TYPE_NOTIFICATION,
                                ChatMessage::CONTENT_NOTIFICATION_WINK,
                                true,
                                i18n( "You have received a wink from %1", friendlyName ) + html,
                                handle,
                                friendlyName ) );
  // TODO: include contact meta data in ChatMessage() constructor to show wink nicer
}



// Notify the user of a received nudge
void Chat::slotReceivedNudge(const QString &handle)
{
#ifdef KMESSDEBUG_CHAT_GENERAL
  kDebug() << "Received a nudge.";
#endif

  // Get friendly name
  QString friendlyName = currentAccount_->getContactFriendlyNameByHandle(handle);
  receivedMessage( ChatMessage( ChatMessage::TYPE_NOTIFICATION,
                                ChatMessage::CONTENT_NOTIFICATION_NUDGE,
                                true,
                                i18n("You received a nudge from %1!", friendlyName),
                                handle,
                                friendlyName ) );

  // Start the buzz effect.
  emit gotNudge();
}



// Signal that a message could not be sent to the switchboard
void Chat::slotSendingFailed( QString message )
{
  message = ( message.length() > 16 ) ? ( message.left( 16 ) + "..." ) : message;

  // We're assuming the chat is empty or broken, so we won't query the switchboard for the contact handle
  showMessage( ChatMessage( ChatMessage::TYPE_SYSTEM,
                                       ChatMessage::CONTENT_SYSTEM_ERROR,
                                       false,
                                       i18n("The message '%1' could not be sent.", message ),
                                       lastContact_ ) );

}



// Send a file to a contact.
void Chat::startFileTransfer( QList<QUrl> fileList )
{
#ifdef KMESSDEBUG_CHAT_FILETRANSFER
  kDebug();
#endif
  if( KMESS_NULL(msnSwitchboardConnection_) ) return;

  // Ask for a contact
  QString handle = chooseContact();

  if( fileList.isEmpty() )
  {
#ifdef KMESSDEBUG_CHAT_FILETRANSFER
    kDebug() << "asking for a file to send";
#endif

    // If no file names were given, ask for a filename
    QString filename = KFileDialog::getOpenFileName( KUrl("kfiledialog:///:fileupload" ) );

    // Stop if no file was selected
    if( filename.isEmpty() )
    {
      return;
    }

    // Ask the ChatMaster to initiate the file transfer. If there are multiple contacts
    // in this chat, it needs to create a new chat session for that contact.
    emit requestFileTransfer( handle, filename );
  }
  else
  {
    foreach( QUrl file, fileList )
    {
#ifdef KMESSDEBUG_CHAT_FILETRANSFER
      kDebug() << "Sending file " << file;
#endif
      emit requestFileTransfer( handle, file.toLocalFile() );
    }
  }
}



// Send a nudge to a contact
void Chat::slotSendNudge()
{
  if( msnSwitchboardConnection_ == 0 )
  {
    return;
  }

  // Display message
  const QStringList &contactsInChat = msnSwitchboardConnection_->getContactsInChat();
  QString message;
  if( contactsInChat.count() == 1 && msnSwitchboardConnection_->isConnected() )
  {
    message = i18n( "You've sent a nudge to %1!",
                    currentAccount_->getContactFriendlyNameByHandle( contactsInChat.first() )
                  );
  }
  else
  {
    // Chat is either empty (should resume), or a multi-chat
    message = i18n( "You've sent a nudge!" );
  }

  showMessage( ChatMessage( ChatMessage::TYPE_NOTIFICATION,
                                       ChatMessage::CONTENT_NOTIFICATION_NUDGE,
                                       false,
                                       message,
                                       QString::null ) );

  // Send nudge
  msnSwitchboardConnection_->sendNudge();

  // Shake the window
  emit gotNudge();
}



// Start a chat
void Chat::startChat()
{
  if( KMESS_NULL(msnSwitchboardConnection_) )
  {
#ifdef KMESSDEBUG_CHAT_GENERAL
    kWarning() << "Tried to open a chat without a switchboard!";
#endif
    return;
  }

  if( ! msnSwitchboardConnection_->isInactive() )
  {
#ifdef KMESSDEBUG_CHAT_GENERAL
    kDebug() << "Opening a normal chat.";
#endif
    // Initialize an online chat
  }
  else
  {
#ifdef KMESSDEBUG_CHAT_GENERAL
    kDebug() << "Opening an offline chat.";
#endif
    // Initialize an offline chat
    QString handle = msnSwitchboardConnection_->getLastContact();
  }

  // Add the participant contacts to the chat, even if they're offline
  QStringList participants = msnSwitchboardConnection_->getContactsInChat();
#ifdef KMESSDEBUG_CHAT_GENERAL
  kDebug() << "Adding initial chat participants to the chat:" << participants;
#endif
  foreach( QString handle, participants )
  {
    contactSidebar_->contactJoined( handle, QString() ); // The sidebar slot doesn't use the second parameter
  }

  getMessageEditField()->setFocus();
}



//Start voice conversation
void Chat::startConversation()
{
  if( KMESS_NULL(msnSwitchboardConnection_) )
  {
#ifdef KMESSDEBUG_CHAT_GENERAL
    kWarning() << "Tried to open a conversation without a switchboard!";
#endif
    return;
  }

  kWarning() << "not implemented!";
  // msnSwitchboardConnection_->startApp( MsnSwitchboardConnection::VOICECONVERSATION );
}



// Start GnomeMeeting with a contact.
void Chat::startMeeting()
{
  // Choose the contact.
  QString handle = chooseContact();

  if( KMESS_NULL(msnSwitchboardConnection_) )
  {
#ifdef KMESSDEBUG_CHAT_GENERAL
    kWarning() << "Tried to open a meeting with \'" << handle << "\' without a switchboard!";
#endif
    return;
  }

  emit requestNetMeeting(handle);
}



#include "chat.moc"
