/***************************************************************************
                          chat.h  -  description
                             -------------------
    begin                : Wed Jan 15 22:41:32 CST 2003
    copyright            : (C) 2003 by Mike K. Bennett
    email                : mkb137b@hotmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef CHAT_H
#define CHAT_H

#include "chatview.h"

#include <QDateTime>
#include <QList>
#include <QMap>
#include <QStringList>
#include <QTimer>

#include <KApplication>


// Forward declarations
class ChatInformation;
class ChatMessage;
class ChatView;
class ChatWindow;
class ContactBase;
class ContactAction;
class KMessTest;
class MsnSwitchboardConnection;

class QDir;



/**
 * @brief Manage an active chat
 *
 * This class manages a chat. It handles the communication with the Switchboard connection;
 * and connects the Chat View signals and calls to make user interaction possible.
 *
 * It is meant to be inserted into a ChatWindow object, not to be used alone.
 *
 * @ingroup Chat
 */
class Chat : public ChatView
{
  Q_OBJECT

  friend class KMessTest;

  public:
    // The construtor
                       Chat( QWidget* parent = 0 );
    // The destructor
                      ~Chat();
    // Return a list of the contacts in the chat to be used as window caption
    QString            getCaption();
    // Return the chat window which currently contains this Chat
    ChatWindow        *getChatWindow();
    // Return the list of contacts in the chat
    QStringList        getParticipants();
    // Return the list of contacts which are typing
    QStringList        getTypingContacts();
    // Initialize the object
    bool               initialize( MsnSwitchboardConnection *switchboardConnection );
    // Return the switchboard connection used by the chat window.
    MsnSwitchboardConnection * getSwitchboardConnection() const;
    // Return whether or not the contact is in this chat.
    bool               isContactInChat( const QString &handle );
    // Return whether or not this is an exclusive chat with the given contact
    bool               isExclusiveChatWithContact( const QString& handle );
    // Return whether or not the chat is at its first incoming message
    bool               isChatFirstMessage();
    // The application is exiting
    void               queryExit();
    // The chat is closing
    void               queryClose();
    // Enable or disable the parts of the window which allow user interaction
    void               setEnabled( bool isEnabled );
    // Start a chat
    void               startChat();
    // Show a wink received by a contact
    void               showWink( const QString &handle, const QString &filename, const QString &animationName );

  public slots:
    // A message was received from a contact.
    void               receivedMessage(const ChatMessage &message);
    // Change the switchboard connection used by the chat window.
    void               setSwitchboardConnection( MsnSwitchboardConnection *newConnection = 0 );
    // Send a nudge
    void               slotSendNudge();
    // Send a file to a contact.
    void               startFileTransfer( QList<QUrl> fileList = QList<QUrl>() );
    //Voice conversation
    void               startConversation();
    // Start GnomeMeeting with a contact.
    void               startMeeting();

  private: // Private methods
    // Check if the user enabled an auto-reply, and send it
    void               checkAutoReply();
    // Choose the contact to start an invitation with.
    QString            chooseContact();
    // Save the chat if necessary
    void               saveChatAutomatically();

  private slots: // Private slots
    // A contact joined the chat
    void               contactJoined(QString handle, QString friendlyName);
    // A contact left the chat
    void               contactLeft(QString handle, bool isChatIdle);
    // A contact is typing
    void               contactTyping( const QString &handle = QString(), const QString &friendlyName = QString(), bool forceExpiration = false );
    // Invite a contact to the chat
    void               inviteContact(QString handle);
    // Notify the user of a received nudge
    void               slotReceivedNudge(const QString &handle);
    // Signal that a message could not be sent to the switchboard
    void               slotSendingFailed( QString message );

  private: // Private attributes
    // Whether or not the incoming message is the first incoming message
    bool               firstMessage_;
    // Whether or not the object was initialized
    bool               initialized_;
    // The last contact that left the chat.
    QString            lastContact_;
    // The last time an automessage has been sent
    QTime              lastSentAutoMessage_;
    // The object that connects to the switchboard server.
    MsnSwitchboardConnection *msnSwitchboardConnection_;
    // The list of participants to this chat
    QStringList        participants_;
    // The date/time the chat started
    QDate              startDate_;
    QTime              startTime_;
    // The list of users which are currently typing
    QMap<QString,QTime> typingContacts_;
    // A timer to update the typing information
    QTimer             typingTimer_;


  signals: // Public signals
    // Signal that some chat details (like the number of active contacts) have changed
    void               chatInfoChanged();
    // Signal that the chat is about to close.
    void               closing(QObject *object);
    // Signal the presence of a new incoming chat message
    void               gotChatMessage( const ChatMessage &message, Chat *chat );
    // Signal that a typing message has been received
    void               gotTypingMessage( Chat *chat );
    // Signal that a nudge has been received
    void               gotNudge();
    // Signal that the ChatMaster should start a filetransfer.
    void               requestFileTransfer(const QString &handle, const QString &filename);
    // Signal that the ChatMaster should start a newmeeting session.
    void               requestNetMeeting(const QString &handle);
    // The user wants to add or remove a contact.
    void               setContactAdded( QString handle, bool isAdded );
    // The user wants to allow a contact to see his/hers online status.
    void               setContactAllowed( QString handle );
    // The user wants to block or unblock a contact. 
    void               setContactBlocked( QString handle, bool isBlocked );
};

#endif
