/***************************************************************************
                          chatview.cpp  -  description
                             -------------------
    begin                : Wed Jan 15 2003
    copyright            : (C) 2003 by Mike K. Bennett
    email                : mkb137b@hotmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "chatview.h"

#include "../contact/contactbase.h"
#include "../contact/contactlist.h"
#include "../dialogs/addemoticondialog.h"
#include "../utils/kmessconfig.h"
#include "../utils/kmessshared.h"
#include "../currentaccount.h"
#include "../emoticonmanager.h"
#include "../kmess.h"
#include "../kmessapplication.h"
#include "../kmessdebug.h"
#include "contactaction.h"
#include "chatmessagestyle.h"
#include "chatmessageview.h"
#include "contactsidebar.h"
#include "inkedit.h"

#include <QClipboard>
#include <QColor>
#include <QDropEvent>
#include <QEvent>
#include <QHBoxLayout>
#include <QKeyEvent>
#include <QLayout>
#include <QPushButton>
#include <QRegExp>
#include <QStringList>
#include <QTextCodec>
#include <QToolBox>
#include <QVBoxLayout>

#include <KAction>
#include <KApplication>
#include <KFileDialog>
#include <KGlobal>
#include <KHTMLView>
#include <KLocale>
#include <KMenu>
#include <KMessageBox>
#include <KRun>
#include <KStandardAction>
#include <KStandardDirs>
#include <KTextBrowser>
#include <KTextEdit>
#include <KUrl>


#define TYPING_NOTIFICATION_TIME       4000



// The constructor
ChatView::ChatView( QWidget *parent )
: QWidget( parent )
, Ui::ChatView()
, contactSidebar_( 0 )
, currentAccount_( 0 )
, chatStyle_( 0 )
, currentZoom_( 100 )
, doSendTypingMessages_( true )
, initialized_( false )
{
  // Setup the UI as first thing
  setupUi( this );

  // Insert a KHTMLPart in the placeholder
  chatMessageView_ = new ChatMessageView( static_cast<QWidget *>( messageViewFrame_ ) );
  connect( chatMessageView_, SIGNAL(                popupMenu(const QString&,const QPoint&)  ),
           this,             SLOT  (      slotShowContextMenu(const QString&,const QPoint&)  ) );
  connect( chatMessageView_, SIGNAL(           openUrlRequest(const KUrl&)                   ),
           this,               SLOT(       slotOpenURLRequest(const KUrl&)                   ) );
  connect( this,             SIGNAL( updateApplicationMessage(const QString&,const QString&) ),
           chatMessageView_,   SLOT( updateApplicationMessage(const QString&,const QString&) ) );

  // Create a layout to maximize the KHTMLPart
  QBoxLayout *messageLayout = new QHBoxLayout( messageViewFrame_ );
  messageLayout->setSpacing( 0 );
  messageLayout->setContentsMargins( 0, 0, 0, 0 );
  messageLayout->addWidget( chatMessageView_->view() );   // Stretches widget

  // Create the ink editor and connect its signals
  inkEdit_ = new InkEdit();
  inkEdit_->installEventFilter(this);

  inkEdit_->hide();
  editorFrame_->layout()->addWidget( inkEdit_ );

  connect( inkEdit_, SIGNAL(     inkModified(bool)   ),
           this,       SLOT( slotInkModified(bool)   ) );

  // Connect the message edit so that if its displayed color changes,
  // it checked to match the user's chosen color.
  connect( messageEdit_, SIGNAL( currentColorChanged( const QColor & ) ),
           this,         SLOT  (  editorColorChanged( const QColor & ) ) );

  connect( messageEdit_, SIGNAL( textChanged()          ),
           this,         SLOT  ( messageTextChanged() ) );

  connect( sendButton_,  SIGNAL( clicked()              ),
           this,         SLOT  ( sendClicked()        ) );
  connect( switchButton_,SIGNAL( clicked()              ),
           this,         SLOT  ( slotSwitchEditor()   ) );

  // Set initial widget state
  sendButton_   ->setEnabled( false );
  sendButton_   ->setFocusPolicy( Qt::NoFocus );
  newLineButton_->setFocusPolicy( Qt::NoFocus );
  switchButton_ ->setFocusPolicy( Qt::NoFocus );
  switchButton_ ->setIcon( KIcon( "draw-freehand" ) );

  //Reset index for fast-retype
  indexSentences_ = 0;

  // Enable drag&drop of files to the message box
  messageEdit_->setAcceptDrops( true );
  messageEdit_->installEventFilter( this );
  messageEdit_->viewport()->installEventFilter( this );

  // And to the messages view
  chatMessageView_->view()->setAcceptDrops( true );
  chatMessageView_->view()->installEventFilter( this );

  // Lastly, initialize the "Typing notification" timer
  userTypingTimer_.setInterval( TYPING_NOTIFICATION_TIME );
  userTypingTimer_.setSingleShot( true );
}


// The destructor
ChatView::~ChatView()
{
  // Delete chat messages. Here we also delete the last contact messages which are links to items of chatMessages_
  qDeleteAll( chatMessages_ );
  chatMessages_.clear();
  lastContactMessages_.clear();

  // Delete the contact actions
  // Avoid crashing when the switchboard/contact fires a "contactLeftChat" signal.
  qDeleteAll( contactActions_ );
  contactActions_.clear();

  // Also erase the chat style parser
  delete chatStyle_;
}



// Set the zoom factor of the text
void ChatView::changeZoomFactor( bool increase )
{
  int newZoom = chatMessageView_->zoomFactor();

  if( increase )
  {
    newZoom += 25;
  }
  else
  {
    newZoom -= 25;
  }

  if( newZoom > 20 && newZoom < 300 )
  {
    currentZoom_ = newZoom;
    chatMessageView_->setZoomFactor( currentZoom_ );
  }
}



// Create the contact actions
void ChatView::createContactActions()
{
  ContactAction          *contactAction;
  QHashIterator<QString,Contact*>  it( currentAccount_->getContactList()->getContactList() );
  Contact                *contact;

  while ( it.hasNext() )
  {
    contact = it.next().value();
    contactAction = new ContactAction( contact, inviteMenu_ );
    connect( contactAction, SIGNAL(     activated(QString) ),
             this,          SLOT  ( inviteContact(QString) ) );
    contactActions_.append( contactAction );
  }
}



// Copy the currently selected text
void ChatView::editCopy()
{
  // Get the cursor reference
  QTextCursor cursor = messageEdit_->textCursor();

  if( cursor.hasSelection() )
  {
    messageEdit_->copy();
  }
  else if( chatMessageView_->hasSelection() )  // never has focus, only selection.
  {
    kapp->clipboard()->setText( chatMessageView_->selectedText() );
    // setData() for HTML version.
  }
}



// The color in the text box changed.
void ChatView::editorColorChanged(const QColor &color)
{
#ifdef KMESSTEST
  KMESS_ASSERT( currentAccount_ != 0 );
#endif

  // Whenever the QTextEdit's contents is entirely deleted, the font color spontaneously resets to black;
  // if this happens, set the color back to the user's color.
  // HACK: This possibly is a Qt <= 4.4 bug exposed.
  if ( color.name() != currentAccount_->getFontColor() )
  {
#ifdef KMESSDEBUG_CHATVIEW
    kDebug() << "Color was reset to" << color.name() << ". Restoring it to the account setting.";
#endif
    messageEdit_->setTextColor( QColor( currentAccount_->getFontColor() ) );
  }
}


// The user pressed return in the message editor, so send the message
void ChatView::enterPressed()
{
  // If there's no typed messages, avoid sending empty message
  if( messageEdit_->document()->isEmpty() )
  {
    return;
  }

  // Don't send any typing messages while preparing to send the message in the text box...
  doSendTypingMessages_ = false;

  // Send the message.
  sendMessage();

  // Now messages can again be sent
  doSendTypingMessages_ = true;
}



// Event filter to detect special actions in the message editor; also detect those of the message view. They get filtered
// here since the message view has a NoFocus policy, which prevents it from having keyboard shortcut events.
bool ChatView::eventFilter( QObject *obj, QEvent *event )
{
  QStringList fileList;
  QKeyEvent       *keyEvent   = 0;
  QDragEnterEvent *dragEvent  = 0;
  QDropEvent      *dropEvent  = 0;
  bool ctrl  = false;
  bool shift = false;
  int  key   = 0;

  if( obj == inkEdit_ )
  {
    // If key was pressed, check if it's return or
    // switch the editor to messages editor.
    if ( event->type() == QEvent::KeyPress )
    {
      keyEvent = static_cast<QKeyEvent*>( event );
      key      = keyEvent->key();
      if( key == Qt::Key_Return || key == Qt::Key_Enter )
      {
        sendClicked();
        return false;
      }

      slotSwitchEditor();

      // Grep the key pressed and put into editor to continue
      // the sentence started
      if ( QChar( key ).isLetterOrNumber() )
      {
        messageEdit_->insertPlainText ( QChar( key ) );
      }
    }
    return false;
  }

  switch( event->type() )
  {
    case QEvent::DragEnter:
      dragEvent = static_cast<QDragEnterEvent*>( event );

      if( dragEvent->mimeData()->hasUrls() )
      {
        dragEvent->acceptProposedAction();
      }
      break;

    case QEvent::WindowActivate:
      if( messageEdit_->isVisible() )
      {
        messageEdit_->setFocus();
      }
      else
      {
        inkEdit_ ->inkFrame_->setFocus();
      }

      break;

    case QEvent::FocusOut:
    case QEvent::Show:
      if( isVisible() && obj == messageEdit_ )
      {
        messageEdit_->setFocus( Qt::OtherFocusReason );
      }
      break;

    case QEvent::Drop:
      dropEvent = static_cast<QDropEvent*>( event );

      // Process the drop only if it contains file URLs and comes from out of our application.
      if( dropEvent->mimeData()->hasUrls() && dropEvent->source() == 0 )
      {
        #ifdef KMESSDEBUG_CHATVIEW
                kDebug() << "Drag'n'dropped files: " << dropEvent->mimeData()->urls();
        #endif

        // Send the files to the contact
        startFileTransfer( dropEvent->mimeData()->urls() );

        dropEvent->acceptProposedAction();
        return true;
      }
      break;

    case QEvent::KeyPress:
      // Get key information
      keyEvent = static_cast<QKeyEvent*>( event );
      ctrl     = keyEvent->modifiers() & Qt::ControlModifier;
      shift    = keyEvent->modifiers() & Qt::ShiftModifier;
      key      = keyEvent->key();

      // Detect enter press to send the message
      if( key == Qt::Key_Return || key == Qt::Key_Enter )
      {
        if( shift || ctrl )
        {
          // ctrl+enter or shift+enter, insert a new line instead.
          messageEdit_->insertPlainText( "\n" );
        }
        else
        {
          enterPressed();
        }

        // Blocks further handling of this event.
        keyEvent->ignore();
        return true;
      }
      // Detect text copying from the Message View
      else if( ctrl && key == Qt::Key_C )
      {
        editCopy();
      }
      // Detect 'save chat' command
      else if( ctrl && key == Qt::Key_S )
      {
        showSaveChatDialog();
      }
      // Detect fast-retype ctlr+UpArrow or ctrl+DownArrow
      else if( ctrl && key == Qt::Key_Up )
      {
        if( indexSentences_ == chatMessages_.size() )
        {
          // store the last sentence in writing
          lastSentence_ = messageEdit_->toPlainText();
        }

        while( indexSentences_ > 0  )
        {
          const ChatMessage *chatMessage = chatMessages_[ --indexSentences_ ];
          if( chatMessage->getType() == ChatMessage::TYPE_OUTGOING )
          {
            messageEdit_->setPlainText( chatMessage->getBody() );
            break;
          }
        }
      }
      else if( ctrl && key == Qt::Key_Down )
      {
        while( indexSentences_ < ( chatMessages_.size() - 1 ) )
        {
          const ChatMessage *chatMessage = chatMessages_[ ++indexSentences_ ];
          if( chatMessage->getType() == ChatMessage::TYPE_OUTGOING )
          {
            messageEdit_->setPlainText( chatMessage->getBody() );
            break;
          }
        }

        if( indexSentences_ == ( chatMessages_.size() - 1 ) )
        {
          messageEdit_->setPlainText( lastSentence_ );
          indexSentences_ = chatMessages_.size();
        }

      }
      break;

    default:
      break;
  }

  return false;
}



// Return the contact action with the given handle
ContactAction* ChatView::getContactActionByHandle( const QString& handle )
{
  foreach( ContactAction *contactAction, contactActions_ )
  {
    if ( contactAction->getHandle() == handle )
    {
      return contactAction;
    }
  }

  return 0;
}



// Return a pointer to this chat's Contact Sidebar
ContactSidebar *ChatView::getContactSidebar() const
{
  return contactSidebar_;
}



// Return the encoding of the HTML document.
QString ChatView::getEncoding() const
{
  return chatMessageView_->encoding();
}



// Return the HTML source of the page.
QString ChatView::getHtml() const
{
  return chatMessageView_->getHtml();
}



// Return a pointer to this chat's menu of inviteable contacts
KMenu *ChatView::getInviteMenu() const
{
  return inviteMenu_;
}



// Give access to the message edit area.
KTextEdit * ChatView::getMessageEditField() const
{
  return messageEdit_;
}



// Return the selected text in the message area.
QString ChatView::getSelectedText(bool asHtml) const
{
  if( asHtml )
  {
    return chatMessageView_->selectedTextAsHTML();
  }
  else
  {
    return chatMessageView_->selectedText();
  }
}



// Initialize the object
bool ChatView::initialize()
{
#ifdef KMESSDEBUG_CHATVIEW
  kDebug() << "initializing.";
#endif

  if ( initialized_ )
  {
    kDebug() << "already initialized.";
    return false;
  }
  if ( !initializeCurrentAccount() )
  {
    return false;
  }
  if ( !initializeContactSidebar() )
  {
    return false;
  }

  // Create a menu for the contact actions
  inviteMenu_ = new KMenu( i18n("&Invite") );

  // Create the contact actions
  createContactActions();

  // Get chat style
  chatStyle_ = new ChatMessageStyle();
  updateChatStyle();

  initialized_ = true;
  return true;
}



// Set up the contact sidebar
bool ChatView::initializeContactSidebar()
{
#ifdef KMESSTEST
  KMESS_ASSERT( contactSidebar_ == 0 );
#endif

  // Create the sidebar
  contactSidebar_ = new ContactSidebar( parentWidget() );
  // Initialize the sidebar
  if ( !contactSidebar_->initialize() )
  {
    kDebug() << "Couldn't initialize the contact sidebar.";
    return false;
  }

  // Hide the sidebar by default
  contactSidebar_->hide();

#ifdef KMESSTEST
  KMESS_ASSERT( contactSidebar_ != 0 );
#endif

  return true;
}



// Initialize the current account
bool ChatView::initializeCurrentAccount()
{
  currentAccount_ = CurrentAccount::instance();
  if ( currentAccount_ == 0 )
  {
    kDebug() << "Couldn't get the instance of the current account.";
    return false;
  }
  connect( currentAccount_, SIGNAL(      changedFontSettings() ),
           this,            SLOT  (         updateEditorFont() ) );
  connect( currentAccount_, SIGNAL( changedChatStyleSettings() ),
           this,            SLOT  (          updateChatStyle() ) );
  updateEditorFont();
  return true;
}



// Insert an emoticon into the message editor
void ChatView::insertEmoticon( QString emoticonText )
{
  // Insert the text at the cursor.
  messageEdit_->insertPlainText( emoticonText );
  messageEdit_->setFocus();
}



// Whether or not the message area is empty
bool ChatView::isEmpty() const
{
  return chatMessageView_->isEmpty();
}



// The message text changed, so the user is typing
void ChatView::messageTextChanged()
{
  // Disable or enable the send button depending on whether or not the message edit is empty
  if ( messageEdit_->document()->isEmpty() )
  {
    sendButton_->setEnabled( false );
    return;
  }

  // Something has been typed, enable the button so the user can send the message
  sendButton_->setEnabled( true );

  // There's some text, but the user doesn't want us to send typing notifications to the contacts
  if( ! doSendTypingMessages_ )
  {
    return;
  }

  // If the typing timer is already going, do nothing
  if( userTypingTimer_.isActive() )
  {
    return;
  }

  // If the last character of the message is a newline, sending the typing signal would
  //  cause the actual text message to be received twice, so check the last character...
  if( messageEdit_->toPlainText().right(1) == "\n" )
  {
    return;
  }

  emit userIsTyping();
  userTypingTimer_.start();
}



// The user clicked the new line button so insert a new line in the editor
void ChatView::newLineClicked()
{
  messageEdit_->insertPlainText( "\n" );
}



// Restore the window properties (called by the parent ChatWindow)
void ChatView::readProperties( const KConfigGroup &config )
{
  KConfigGroup group;

  // Choose the config group if it was not given
  if( &config == 0 )
  {
    group = KMessConfig::instance()->getAccountConfig( currentAccount_->getHandle(), "ChatView" );
  }
  else
  {
    group = config;
  }

#ifdef KMESSDEBUG_CHATVIEW
  kDebug() << "Reading saved properties.";
#endif

  // Set the default widget sizes. The message editor has a default read size of 1 so it'll be
  //  as small as possible, without becoming invisible
  int minChatViewHeight = messageViewFrame_->minimumSize().height();
  messageSplitter_->setSizes( group.readEntry( "MessageSplitterSizes", QList<int>() << minChatViewHeight << 1 ) );

  // Load the level of zoom
  currentZoom_ = group.readEntry( "CurrentZoom", chatMessageView_->zoomFactor() );
  // Restore it in the chat messages view
  chatMessageView_->setZoomFactor( currentZoom_ );
  // Read if the spelling check feature is enabled
  messageEdit_->setCheckSpellingEnabled( group.readEntry( "UseSpellCheck", false ) );
}



// Save the chat to the given file
void ChatView::saveChatToFile( const QString &path )
{
#ifdef KMESSDEBUG_CHATVIEW
  kDebug() << "saving chat to '" << path << "'.";
#endif

  // Create and open the file.
  QFile file( path );
  if( ! file.open( QIODevice::WriteOnly ) )
  {
    kWarning() << "File save failed - couldn't open file '" << path << "'.";
    KMessageBox::sorry( this, i18n("Could not save chat log. Make sure you have permission to write in the folder where logs are being saved.") );
    return;
  }

  // Get the text
  // Make the text show nicely when viewed in an editor.
  QString text = getHtml();
  text = text.replace( QRegExp("<br>"), "<br>\n" );

  // Output the HTML with the right text encoding
  QTextStream textStream( &file );
  textStream.setCodec( QTextCodec::codecForLocale() );
  textStream << text;
  file.close();
}



// Save the window properties (called by parent ChatWindow)
void ChatView::saveProperties( KConfigGroup &config )
{
  KConfigGroup group;

  // Choose the config group if it was not given
  if( &config == 0 )
  {
    group = KMessConfig::instance()->getAccountConfig( currentAccount_->getHandle(), "ChatView" );
  }
  else
  {
    group = config;
  }

#ifdef KMESSDEBUG_CHATVIEW
  kDebug() << "Saving properties.";
#endif

  group.writeEntry( "MessageSplitterSizes", messageSplitter_->sizes()            );
  group.writeEntry( "CurrentZoom",          currentZoom_                         );
  group.writeEntry( "UseSpellCheck",        messageEdit_->checkSpellingEnabled() );
}



// Scroll to the bottom of the chat browser
void ChatView::scrollChatToBottom()
{
  chatMessageView_->scrollChatToBottom();
}



// The user clicked send, so send the message
void ChatView::sendClicked()
{
  if( messageEdit_->isVisible() )
  {
    sendMessage();
  }
  else
  {
    sendInk();
  }

  // Ensure the typing or draw area always gets focus,
  // so the user can type or draw the next message directly.
  // This area looses focus when you do other things like choosing emoticons.
  if( messageEdit_->isVisible() )
  {
    messageEdit_->setFocus();
  }
  else
  {
    inkEdit_ ->inkFrame_->setFocus();
  }
}



// Send a ink via server
void ChatView::sendInk()
{
  // Get the image bytes and emit the signal
  QByteArray ink = inkEdit_->getImageBytes();

  emit sendInkToContact( ink );

  // Clear and reset the state of send button
  inkEdit_->clearImage();
  sendButton_->setEnabled( false );

  // Show the message in the browser window
  QString text = "<img src='data:image/png;base64," + ink.toBase64() + "'>";

  ChatMessage message( ChatMessage::TYPE_OUTGOING,
                       ChatMessage::CONTENT_MESSAGE_INK,
                       false,
                       text,
                       currentAccount_->getHandle(),
                       currentAccount_->getFriendlyName(),
                       currentAccount_->getPicturePath());

  showMessage(message);
}


// Send a message via the server
void ChatView::sendMessage()
{
  // Max length is fixed to 10 chunks ( multi-packet )
  int maxSendableMessageLength = 14000;

  // Get the message field's contents
  QString text = messageEdit_->toPlainText();

  // Don't send empty messages
  if( text.isEmpty() )
  {
    return;
  }

  // If the text is longer than the sendable amount, put the remainder back in the text edit.
  // Since the message will be sent as UTF8, it's the UTF8 length we have to consider.
  QByteArray utf8Text = text.toUtf8();
  if( utf8Text.length() > maxSendableMessageLength )
  {
    // If so, then divide the text into the first part and a remainder.
    QByteArray remainder = utf8Text.right( utf8Text.length() - maxSendableMessageLength );
    text                 = QString::fromUtf8( utf8Text.left( maxSendableMessageLength ) );

    // Return the remainder to the message edit.
    QString remainingText = QString::fromUtf8( remainder );
    messageEdit_->setText( remainingText );
    lastSentence_ = remainingText;
  }
  else
  {
    // Remove the message from the message field (allowing undo)
    lastSentence_.clear();

    QTextCursor cursor = messageEdit_->textCursor();
    messageEdit_->setTextCursor( cursor );

    cursor.beginEditBlock();
    cursor.select( QTextCursor::Document );
    cursor.removeSelectedText();
    cursor.endEditBlock();
  }

  // Replace "\n" with "\r\n" (to follow the Windows linebreak style followed by the MSN servers)
  // @TODO Move this server stuff to the network/protocol code
  text = text.replace( QRegExp("\r?\n"), "\r\n" );

  // Ask the server to send the message to the contact(s)
  emit sendMessageToContact( text );

  // Reset the typing timer, so it restarts correctly
  userTypingTimer_.stop();


  // Show the message in the browser window
  ChatMessage message(ChatMessage::TYPE_OUTGOING,
                      ChatMessage::CONTENT_MESSAGE,
                      false,
                      text,
                      currentAccount_->getHandle(),
                      currentAccount_->getFriendlyName( STRING_ORIGINAL ),
                      currentAccount_->getPicturePath(),
                      currentAccount_->getFont(),
                      currentAccount_->getFontColor() );

  showMessage(message);

  // Reset the last sentence counter.
  // showMessage() updates chatMessages_
  indexSentences_ = chatMessages_.size();
}



// Enable or disable the parts of the chat which allow user interaction
void ChatView::setEnabled( bool isEnabled )
{
  // Set the state of the main UI elements
  contactSidebar_->setEnabled( isEnabled );
  inviteMenu_    ->setEnabled( isEnabled );
  inkEdit_       ->setEnabled( isEnabled );
  messageEdit_   ->setEnabled( isEnabled );
  newLineButton_ ->setEnabled( isEnabled );
  sendButton_    ->setEnabled( isEnabled );
  switchButton_  ->setEnabled( isEnabled );

  // Clear the contact actions at the very moment they expire (on disconnection)
  qDeleteAll( contactActions_ );
  contactActions_.clear();

  // The chat is being enabled back, replace some stuff which changes
  // at every connection
  if( isEnabled )
  {
    // Recreate the contact actions
    createContactActions();

    // Re-read the current account pointer
    // because it likely is a different one
    currentAccount_ = CurrentAccount::instance();
  }
}



// Add the given message to the message browser.
void ChatView::showMessage(const ChatMessage &message)
{
  if(KMESS_NULL(chatStyle_)) return;

  // Clone the message so it can be stored in the local qlist objects.
  // @TODO-PORTING This shouldn't be this way...
  ChatMessage *chatMessage = message.clone();
  QString messageHtml;

  // See if the same contact sent the previous message too.
  // In that case, we combine both messages for the chat style.
  // Otherwise the lastContactMessages_ list is reset.
  if( chatMessage->isNormalMessage() && currentAccount_->getGroupFollowupMessages() )
  {
    if( ! lastContactMessages_.isEmpty() )
    {
      ChatMessage *lastContactMessage = lastContactMessages_.last();

      // Check for contact handle and message type, so offline messages won't be
      // grouped with normal incoming messages.
      if( lastContactMessage->getContactHandle() == chatMessage->getContactHandle()
      &&  lastContactMessage->getType()          == chatMessage->getType() )
      {
        lastContactMessages_.append(chatMessage);
      }
      else
      {
        lastContactMessages_.clear();
      }
    }
  }
  else
  {
    lastContactMessages_.clear();
  }

  // Convert the message, add to the browser.
  if( lastContactMessages_.count() > 1 )
  {
#ifdef KMESSDEBUG_CHATVIEW
    kDebug() << "replacing last contact message with new contents.";
#endif

    messageHtml = chatStyle_->convertMessageList(lastContactMessages_);
    chatMessageView_->replaceLastMessage(messageHtml);
  }
  else
  {
#ifdef KMESSDEBUG_CHATVIEW
    kDebug() << "appending new message.";
#endif

    messageHtml = chatStyle_->convertMessage(*chatMessage);
    chatMessageView_->addHtmlMessage(messageHtml);

    if( chatMessage->isNormalMessage() )
    {
      lastContactMessages_.append( chatMessage );
    }
  }


  // If the parser found custom emoticons, add them.
  // The pending list of chatStyle is erased with each new convertMessage() call.
  const QStringList &emoticonTags = chatStyle_->getPendingEmoticonTagIds();
  if( ! emoticonTags.isEmpty() )
  {
    // Copy to internal list.
    for(QStringList::const_iterator it = emoticonTags.begin(); it != emoticonTags.end(); ++it)
    {
      pendingEmoticonTags_.append(*it);
    }
  }


  // Add to memory, this is used for:
  // - changing chat styles
  // - regenerating messages to group them
  // - regenerating messages for received custom emoticons.
  chatMessages_.append( chatMessage );

  // Also notify the sidebar, so that it updates the right contact frame
  if( ! KMESS_NULL( contactSidebar_ ) )
  {
    contactSidebar_->messageReceived( chatMessage->getContactHandle() );
  }
}



// Show a dialog to save the chat.
void ChatView::showSaveChatDialog()
{
  QFileInfo targetFile;
  int overwriteAllowed;
  QString path = currentAccount_->getSaveChatPath() + "/kmess-chat.html"; // Default name for the saved chat

  // Repeat until a file is chosen or the operation is cancelled
  do
  {
    // Show a dialog to get a filename from the user.
    path = KFileDialog::getSaveFileName( path, "*.html" );

    // Verify if the user has canceled the command
    if( path.isEmpty() )
    {
      return;
    }

    // Default to accept the path name
    overwriteAllowed = KMessageBox::Ok;

    // Check if the file exists and if so, warn the user
    targetFile.setFile( path );
    if( targetFile.exists() )
    {
      overwriteAllowed = KMessageBox::warningContinueCancel(
                                0,
                                i18n( "The file '%1' already exists.\ndo you want to overwrite it?", targetFile.fileName() ),
                                i18n("Overwrite File"),
                                KGuiItem( i18n("Over&write") ) );
    }

  } while( overwriteAllowed == KMessageBox::Cancel );

  saveChatToFile( path );
}



/**
 * Open a dialog to add a new custom emoticon seen in the chat
 */
void ChatView::slotAddNewEmoticon()
{
  if( ! chatViewClickedUrl_.isValid() || chatViewClickedUrl_.protocol() != "kmess" || chatViewClickedUrl_.host() != "emoticon" )
  {
#ifdef KMESSDEBUG_CHATVIEW
    kDebug() << "Ignoring request for invalid URL: " << chatViewClickedUrl_;
#endif

    return;
  }

/*
 * KMess' internal emoticon addition URLs are in the form
 * <code>kmess://emoticon/contactHandle/urlEncodedShortcut/urlEncodedPicturePath</code>
 */

  QString path  = chatViewClickedUrl_.path().mid( 1 );  // remove first /

#ifdef KMESSTEST
  KMESS_ASSERT( ! path.isEmpty() );
#endif

  // Reset the url
  chatViewClickedUrl_.clear();

  QString handle   = path.section( "/", 0, 0 );   // First parameter: contact handle
  QString shortcut = path.section( "/", 1, 1 );   // Second parameter: emoticon shortcut
  QString picture  = path.section( "/", 2    );   // Third parameter: emoticon picture path

  const ContactBase *contact = CurrentAccount::instance()->getContactByHandle( handle );
  if( contact == 0 )
  {
#ifdef KMESSDEBUG_CHATVIEW
    kDebug() << "Contact handle not found: " << handle;
#endif

    return;
  }


  // URL-Decode the encoded strings
  picture  = KUrl::fromPercentEncoding( picture.toAscii()  );
  shortcut = KUrl::fromPercentEncoding( shortcut.toAscii() );

#ifdef KMESSDEBUG_CHATVIEW
  kDebug() << "Showing Add Emoticon dialog - shortcut=" << shortcut << " picture=" << picture;
#endif

  // Finally, show the dialog, preselecting the picture file and shortcut
  AddEmoticonDialog *addDialog = new AddEmoticonDialog( EmoticonManager::instance()->getTheme( true ), this );

  connect( addDialog,        SIGNAL( addedEmoticon(QString) ),
           chatMessageView_,   SLOT( addedEmoticon(QString) ) );

  addDialog->preSelect( picture, shortcut );

}



// The user are drawing a ink
void ChatView::slotInkModified( bool erasered )
{
  if( ! erasered )
  {
    if( ! sendButton_->isEnabled() )
    {
      sendButton_->setEnabled( true );
    }
  }
  else
  {
    sendButton_->setEnabled( false );
  }
}



// The user clicked the "copy address" or "copy email" option in the context menu
void ChatView::slotCopyAddress()
{
  QString url = chatViewClickedUrl_.url();

  if( ! chatViewClickedUrl_.isValid() )
  {
#ifdef KMESSDEBUG_CHATVIEW
    kDebug() << "Not copying invalid URL: " << url;
#endif

    return;
  }

#ifdef KMESSDEBUG_CHATVIEW
  kDebug() << "Copying URL: " << url;
#endif

  // Strip the pseudo-protocol 'mail to' from email addresses
  if( url.left( 7 ) == "mailto:" )
  {
    url = url.mid( 7 );
  }

  kapp->clipboard()->setText( url );

  // Reset the url
  chatViewClickedUrl_.clear();
}



// The user clicked the "copy text" option in the context menu.
void ChatView::slotCopyChatText()
{
  kapp->clipboard()->setText( this->getSelectedText( false ) );
}



// The user clicked the "find text" option in the context menu
void ChatView::slotFindChatText()
{
  chatMessageView_->findText();
}



// Open a new url clicked in the khtml widget
void ChatView::slotOpenURLRequest( const KUrl &url )
{
    chatViewClickedUrl_ = url;

    // Internal URLs form: kmess://call_type/parameters?more_parameters

    if( url.protocol() == "kmess" )
    {
      // Application URLs form: kmess://application/responseType/contactHandle?cookieId
      if( url.host() == "application" )
      {
        // Handle the applications input
        slotSendAppCommand();
      }
      // Emoticon URLs form: kmess://emoticon/contactHandle/urlEncodedShortcut/urlEncodedPictureTag
      else if( url.host() == "emoticon" )
      {
        // Add a new emoticon
        slotAddNewEmoticon();
      }
    }
    else if( url.protocol() == "file" )
    {
      // Obtain the widget of the main KMess window, to correctly link it to the opened app
      KMessApplication *kmessApp = static_cast<KMessApplication*>( kapp );
      QWidget *mainWindow = kmessApp->getContactListWindow()->window();

      // Execute the local file with the system's default association (KRun auto-deletes itself)
      new KRun( url, mainWindow, 0, true );
    }
    else
    {
      // Just execute the link
      slotVisitAddress();
    }
}



// The user clicked the "select all" option in the context menu.
void ChatView::slotSelectAllChatText()
{
  chatMessageView_->selectAll();
}



/**
 * The user clicked a kmess internal link in the ChatMessageView
 */
void ChatView::slotSendAppCommand()
{
  // Ignore non-internal links and non-application internal links
  if( ! chatViewClickedUrl_.isValid() || chatViewClickedUrl_.protocol() != "kmess" || chatViewClickedUrl_.host() != "application" )
  {
#ifdef KMESSDEBUG_CHATVIEW
    kDebug() << "Not sending invalid application link: " << chatViewClickedUrl_.url();
#endif

    return;
  }

/*
 * KMess' internal application URLs are in the form
 * kmess://application/responseType/accountHandle?cookieId
 */

  QString path  = chatViewClickedUrl_.path().mid( 1 );  // remove first /
  QString query = chatViewClickedUrl_.query().mid( 1 ); // Remove the ?

  QString method  = path.section( "/", 0, 0 );   // First parameter: response type
  QString contact = path.section( "/", 1, 1 );   // Second parameter: contact handle
  QString cookie  = query;                       // Third parameter: transfer cookie ID

#ifdef KMESSTEST
  KMESS_ASSERT( ! path.isEmpty() );
  KMESS_ASSERT( ! query.isEmpty() );
#endif

#ifdef KMESSDEBUG_CHATVIEW
  kDebug() << "Sending application link - method=" << method << " contact=" << contact << " cookie=" << cookie;
#endif

  emit appCommand( cookie, contact, method );
}



// The user right clicked at the KHTMLPart to show a popup.
void ChatView::slotShowContextMenu( const QString &clickedUrl, const QPoint &point )
{
  KAction *urlAction = 0;

  // Add items to this context menu
  KMenu *contextMenu = new KMenu( this );

  // Analyze incoming URL, if present
  if( ! clickedUrl.isEmpty() )
  {
    KUrl url( clickedUrl );

#ifdef KMESSDEBUG_CHATVIEW
    kDebug() << "Clicked URL: " << url.prettyUrl();
#endif

    if( url.protocol() == "kmess" && url.host() == "emoticon" )
    {
//       urlAction = new KAction( i18n("Add this &Emoticon..."),  "list-add",           0, this, SLOT(slotAddNewEmoticon()),  this, "addemoticon"  );
      urlAction = new KAction( KIcon( "list-add" ), i18n("Add this &Emoticon..."), this );
      connect( urlAction, SIGNAL(triggered(bool)), this, SLOT(slotAddNewEmoticon()) );
      contextMenu->addAction( urlAction );
    }
    else if( url.protocol().left( 6 ) == "mailto" )
    {
//       urlAction = new KAction( i18n("Send &Email"),            "mail",  0, this, SLOT(slotVisitAddress()),    this, "sendemail"    );
      urlAction = new KAction( KIcon( "mail" ), i18n("Send &Email"), this );
      connect( urlAction, SIGNAL(triggered(bool)), this, SLOT(slotVisitAddress()) );
      contextMenu->addAction( urlAction );
//       urlAction = new KAction( i18n("Copy E&mail"),            "copy",          0, this, SLOT(slotCopyAddress()),     this, "copyemail"    );
      urlAction = new KAction( KIcon( "copy" ), i18n("Copy E&mail"), this );
      connect( urlAction, SIGNAL(triggered(bool)), this, SLOT(slotCopyAddress()) );
      contextMenu->addAction( urlAction );
    }
    else
    {
//       urlAction = new KAction( i18n("Visit &Link"),            "launch",        0, this, SLOT(slotVisitAddress()),    this, "visiturl"     );
      urlAction = new KAction( KIcon( "launch" ), i18n("Visit &Link"), this );
      connect( urlAction, SIGNAL(triggered(bool)), this, SLOT(slotVisitAddress()) );
      contextMenu->addAction( urlAction );
//       urlAction = new KAction( i18n("Copy &Address"),          "copy",          0, this, SLOT(slotCopyAddress()),     this, "copyurl"      );
      urlAction = new KAction( KIcon( "edit-copy" ), i18n("Copy &Address"), this );
      connect( urlAction, SIGNAL(triggered(bool)), this, SLOT(slotCopyAddress()) );
      contextMenu->addAction( urlAction );
    }

    chatViewClickedUrl_ = url;
  }

  // Create items
  KAction *copyAction       = KStandardAction::copy( this, SLOT(slotCopyChatText()), 0 );  // , actionCollection() );
  KAction *selectAllAction  = KStandardAction::selectAll( this, SLOT(slotSelectAllChatText()), 0 ); // , actionCollection() );
  KAction *saveToFileAction = KStandardAction::save( this, SLOT(showSaveChatDialog()), 0 );  // , actionCollection() );
  KAction *findAction       = KStandardAction::find( this, SLOT(slotFindChatText()),   0 );

  // Update the labels a bit though
  copyAction      ->setText( i18n("&Copy text") );
  selectAllAction ->setText( i18n("Select &All") );
  findAction      ->setText( i18n("Find &Text" ) );
  saveToFileAction->setText( i18n("Save chat to &File") );

  // Add a separator to divide the context-depending entries from the rest of the menu
  if( urlAction != 0 )
  {
    contextMenu->addSeparator();
  }

  contextMenu->addAction( copyAction );
  contextMenu->addAction( selectAllAction );
  contextMenu->addAction( findAction );
  contextMenu->addSeparator();
  contextMenu->addAction( saveToFileAction );

  // Set items disabled, depending on the text selection
  if( chatMessageView_->hasSelection() )
  {
    saveToFileAction->setEnabled( false );
  }
  else
  {
    copyAction->setEnabled( false );
  }

  // Show the menu
  contextMenu->exec( point );
  delete contextMenu;
  delete copyAction;
}



// Switch the editor between ink editor and text editor
void ChatView::slotSwitchEditor()
{
  // TODO Create inkEdit only on request
  bool isMessageEditVisible =  messageEdit_->isVisible();
  messageEdit_->setVisible( ! isMessageEditVisible );
  inkEdit_    ->setVisible(   isMessageEditVisible );

  if( isMessageEditVisible )
  {
    // The inkEdit is visible
    if( inkEdit_->isEmpty() )
    {
      sendButton_->setEnabled( false );
    }
    else
    {
      sendButton_->setEnabled( true );
    }

    switchButton_->setIcon( KIcon( "preferences-desktop-font" ) );
    inkEdit_ ->inkFrame_->setFocus();
    newLineButton_->setEnabled( false);
  }
  else
  {
    // The messageEdit is visible
    if( messageEdit_->document()->isEmpty() )
    {
      sendButton_->setEnabled( false );
    }
    else
    {
      sendButton_->setEnabled( true );
    }

    messageEdit_ ->setFocus();
    switchButton_->setIcon( KIcon( "draw-freehand" ) );
    newLineButton_->setEnabled( true );
  }
}



// The user clicked the "visit address" or "send email" option in the context menu, or clicked a link in the ChatMessageView
void ChatView::slotVisitAddress()
{
  if( ! chatViewClickedUrl_.isValid() )
  {
#ifdef KMESSDEBUG_CHATVIEW
    kDebug() << "Not opening invalid URL: " << chatViewClickedUrl_;
#endif

    return;
  }

#ifdef KMESSDEBUG_CHATVIEW
  kDebug() << "Opening URL: " << chatViewClickedUrl_;
#endif

  // Launch the browser for the given URL
  KMessShared::openBrowser( chatViewClickedUrl_ );

  // Reset the url
  chatViewClickedUrl_.clear();
}



// Update the chat style
void ChatView::updateChatStyle()
{
#ifdef KMESSDEBUG_CHATVIEW
  kDebug() << "updating style.";
#endif

  if(KMESS_NULL(chatStyle_))      return;
  if(KMESS_NULL(currentAccount_)) return;

  // Load new settings in the chat style object
  chatStyle_->setStyle           ( currentAccount_->getChatStyle()         );
  chatStyle_->setContactFont     ( currentAccount_->getContactFont()       );
  chatStyle_->setContactFontColor( currentAccount_->getContactFontColor()  );
  chatStyle_->setShowTime        ( currentAccount_->getShowMessageTime()   );
  chatStyle_->setUseContactFont  ( currentAccount_->getUseContactFont()    );
  chatStyle_->setUseEmoticons    ( currentAccount_->getUseEmoticons()      );
  chatStyle_->setUseFontEffects  ( currentAccount_->getUseFontEffects()    );
  chatStyle_->setUseFormatting   ( currentAccount_->getUseChatFormatting() );

  // Support grouping of follow-up messages.
  bool groupFollowups = currentAccount_->getGroupFollowupMessages();
  lastContactMessages_.clear();

  // Replace the entire contents with the new style.
  QString newHtml;
  QListIterator<ChatMessage *> it(chatMessages_);
  while( it.hasNext() )
  {
    ChatMessage *chatMessage = it.next();

    if( groupFollowups && chatMessage->isNormalMessage() )
    {
      // See if the message should be added to the queue.
      // There could be another follow-up message later.
      if( ! lastContactMessages_.isEmpty() )
      {
        // Check whether the contact differs or it's a different type,
        // the handle to difference between "offline_incoming" and "incoming" messages.
        ChatMessage *lastMessage = lastContactMessages_.last();
        if( lastMessage->getType()          != chatMessage->getType()
        ||  lastMessage->getContactHandle() != chatMessage->getContactHandle() )
        {
          // New mesage should not be added to queue,
          // so previous mesage was last the one. Flush the queue.
          newHtml += "\n<div class=\"messageContainer messageListContainer\">"
                  + chatStyle_->convertMessageList(lastContactMessages_)
                  + "</div>\n";
          lastContactMessages_.clear();
        }
      }

      // Append new contact message to the queue.
      lastContactMessages_.append(chatMessage);
    }
    else
    {
      // Not a contact message.
      // First flush the queue, then add the new message.
      if( ! lastContactMessages_.isEmpty() )
      {
        newHtml += "\n<div class=\"messageContainer messageListContainer\">"
                 + chatStyle_->convertMessageList(lastContactMessages_)
                 + "</div>\n";
        lastContactMessages_.clear();
      }

      // Convert the current message.
      newHtml += "\n<div class=\"messageContainer\">"
               + chatStyle_->convertMessage(*chatMessage)
               + "</div>\n";
    }
  }

  QString messageRoot = chatStyle_->convertMessageRoot();
  if( messageRoot.isNull() )
  {
    // Assign standard HTML
    chatMessageView_->setStandardHtml( newHtml, chatStyle_->getCssFile(), chatStyle_->getBaseFolder() );
  }
  else
  {
    // Assign the custom HTML root
    chatMessageView_->setHtml( messageRoot, newHtml );
  }

  // If there are still grouped messages, add them after setHtml().
  // This allows the chatMessageView_ to replace them later with replaceLastMessage();
  if( ! lastContactMessages_.isEmpty() )
  {
    chatMessageView_->addHtmlMessage( chatStyle_->convertMessageList(lastContactMessages_) );
    // Don't clear queue here, so replaceLastMessage() will be called.
  }

  // Call the scroll function a bit later,
  // so Qt/kde get a chance to update the height before the scrolling starts.
  QTimer::singleShot(50, chatMessageView_, SLOT(scrollChatToBottom()));
}



// Update the messages which contain custom emoticons
void ChatView::updateCustomEmoticon( const QString &handle, const QString &code )
{
#ifdef KMESSTEST
  KMESS_ASSERT( ! code.isEmpty() );
#endif

  // Check for empty replacements.
  if( code.isEmpty() )
  {
    kWarning() << "can't update custom emoticon, emoticon code not given (contact=" << handle << ").";
    return;
  }

  // Get contact emoticon replacements.
  const ContactBase *contact = currentAccount_->getContactByHandle(handle);
  if(KMESS_NULL(contact)) return;

  // Get emoticon replacement, instruct chatMesasgeView to replace it.
  const QString &replacement = contact->getEmoticonReplacements()[code];
  chatMessageView_->updateCustomEmoticon( code, replacement, handle, pendingEmoticonTags_ );
}



// Update the editor's font to match the account's font
void ChatView::updateEditorFont()
{
#ifdef KMESSTEST
  KMESS_ASSERT( currentAccount_ != 0 );
#endif
  QColor color;

  if( currentAccount_ == 0 )
  {
    return;
  }

  // Change the color to the user's color
  color.setNamedColor( currentAccount_->getFontColor() );
  messageEdit_->setTextColor( color );
  messageEdit_->setFont( currentAccount_->getFont() );
  messageEdit_->setCurrentFont( currentAccount_->getFont() );
  messageEdit_->setFontPointSize( currentAccount_->getFont().pointSize() );

  // If the contact font settings changed, regenerate the messages.
  if( chatStyle_ != 0 )
  {
    if( currentAccount_->getUseContactFont()   != chatStyle_->getUseContactFont()
    ||  currentAccount_->getContactFont()      != chatStyle_->getContactFont()
    ||  currentAccount_->getContactFontColor() != chatStyle_->getContactFontColor() )
    {
      updateChatStyle();
    }
  }
}



#include "chatview.moc"

