/***************************************************************************
                          chatview.h  -  description
                             -------------------
    begin                : Wed Jan 15 2003
    copyright            : (C) 2003 by Mike K. Bennett
    email                : mkb137b@hotmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef CHATVIEW_H
#define CHATVIEW_H

#include "../network/chatmessage.h"

#include "ui_chatview.h"

#include <QPoint>
#include <QStringList>
#include <QTimer>

#include <KConfigGroup>
#include <KHTMLPart>
#include <KUrl>


// Forward declarations
class KMenu;

class ChatMessageStyle;
class ChatMessageView;
class ContactAction;
class ContactSidebar;
class CurrentAccount;
class InkEdit;
class KMessTest;
class KTextEdit;



/**
 * The view widget of a chat with a contact with message display and entry.
 *
 * @author Mike K. Bennett
 * @ingroup Chat
 */
class ChatView : public QWidget, private Ui::ChatView
{
  Q_OBJECT

  friend class KMessTest;

  public:
    // The constructor
                          ChatView( QWidget *parent=0 );
    // The destructor
    virtual              ~ChatView();
    // Copy the currently selected text
    void                  editCopy();
    // Initialize the object
    bool                  initialize();
    // Return the contact action with the given handle
    ContactAction        *getContactActionByHandle(const QString& handle);
    // Return a pointer to this chat's menu of inviteable contacts
    KMenu                *getInviteMenu() const;
    // Return a pointer to this chat's Contact Sidebar
    ContactSidebar       *getContactSidebar() const;
    // Return the encoding of the HTML document.
    QString               getEncoding() const;
    // Return the HTML source of the page.
    QString               getHtml() const;
    // Give access to the message edit area.
    KTextEdit            *getMessageEditField() const;
    // Return the selected text in the message area.
    QString               getSelectedText(bool asHtml) const;
    // Whether or not the message area is empty
    bool                  isEmpty() const;
    // Update the messages which contain custom emoticons
    void                  updateCustomEmoticon( const QString &handle, const QString &code );
    // Change the text zoom factor
    void                  changeZoomFactor( bool increase );
    // Restore the window properties (called by the parent ChatWindow)
    void                  readProperties( const KConfigGroup &config = *((const KConfigGroup*)0) );
    // Save the window properties (called by the parent ChatWindow)
    void                  saveProperties( KConfigGroup &config = *((KConfigGroup*)0) );

  public slots:
    // Insert an emoticon into the message editor
    void                  insertEmoticon( QString emoticonText );
    // Scroll to the bottom of the chat browser
    void                  scrollChatToBottom();
    // Enable or disable the parts of the chat which allow user interaction
    void                  setEnabled( bool isEnabled );
    // Add the given message to the message browser.
    void                  showMessage( const ChatMessage &message );
    // Show a dialog to save the chat.
    void                  showSaveChatDialog();
    // Save the chat to the given file
    void                  saveChatToFile(const QString &path );

  private: // Private methods
    // Create the contact actions
    void                  createContactActions();
    // Event filter to detect special actions in the message editor.
    bool                  eventFilter( QObject *obj, QEvent *event );
    // Set up the contact sidebar
    bool                  initializeContactSidebar();
    // Initialize the current account
    bool                  initializeCurrentAccount();
    // Send a ink via server
    void                  sendInk();
    // Send a message via the server
    void                  sendMessage();
    // Send a file to a contact.
    virtual void          startFileTransfer( QList<QUrl> fileList = QList<QUrl>() ) = 0;

  private slots: // Private slots
    // The user pressed return in the message editor, so send the message
    void                  enterPressed();
    // The message text changed, so the user is typing
    void                  messageTextChanged();
    // The user clicked the new line button so insert a new line in the editor
    void                  newLineClicked();
    // The user clicked send, so send the message
    void                  sendClicked();
    // Open a dialog to add a new custom emoticon seen in the chat
    void                  slotAddNewEmoticon();
    // The user are drawing a ink
    void                  slotInkModified( bool erasered );
    // The user clicked a kmess internal link in the ChatMessageView
    void                  slotSendAppCommand();
    // The user clicked the "copy address" or "copy email" option in the context menu
    void                  slotCopyAddress();
    // The user clicked the "copy text" option in the context menu.
    void                  slotCopyChatText();
    // The user clicked the "find text" option in the context menu
    void                  slotFindChatText();
    // Open a new url clicked in the khtml widget
    void                  slotOpenURLRequest( const KUrl &url );
    // The user clicked the "select all" option in the context menu.
    void                  slotSelectAllChatText();
    // The user right clicked at the KHTMLPart to show a popup.
    void                  slotShowContextMenu(const QString &url, const QPoint &point);
    // Switch the editor between ink editor and text editor
    void                  slotSwitchEditor();
    // The user clicked the "visit address" or "send email" option in the context menu, or clicked a link in the ChatMessageView
    void                  slotVisitAddress();
    // Update the editor's font to match the account's font
    void                  updateEditorFont();


  protected: // Protected attributes
    // The contact sidebar, which goes in the sidebar
    ContactSidebar       *contactSidebar_;
    // A pointer to the instance of the current account
    CurrentAccount       *currentAccount_;

  private: // Private attributes
    QList<ChatMessage*>   chatMessages_;
    // The embedded khtml part
    ChatMessageView      *chatMessageView_;
    // The XSL transformation handler
    ChatMessageStyle     *chatStyle_;
    // URL address which has been right-clicked
    KUrl                  chatViewClickedUrl_;
    // The contact actions in the invite menu
    QList<ContactAction*> contactActions_;
    // Stores the current level of zoom
    int                   currentZoom_;
    // Whether or not typing messages should be sent (true except the very moment a text message is sent)
    bool                  doSendTypingMessages_;
    // Whether or not the object was initialized
    bool                  initialized_;
    // The ink editor
    InkEdit               *inkEdit_;
    // The list of invitable contacts
    KMenu                 *inviteMenu_;
    // The last messages sent by the same contact. This is used to combine them
    QList<ChatMessage*>   lastContactMessages_;
    // The messages that contain custom emoticons
    QStringList           pendingEmoticonTags_;
    // A timer to time whether a "user is typing" message should be sent.
    QTimer                userTypingTimer_;
    // Index for fast-retype
    int                   indexSentences_;
    // Last sentence in writing
    QString               lastSentence_;

  private slots:
    // The color in the text box changed.
    void                  editorColorChanged(const QColor &color);
    // Update the chat style
    void                  updateChatStyle();
    // Invite a contact to the chat
    virtual void          inviteContact(QString handle) = 0;

  signals: // Public signals
    // Signal that the user wants to add a new custom emoticon copying it from a friend's
    void                 addEmoticon(QString handle, QString shortcut);
    // Signal that there's an application command
    void                  appCommand(QString cookie, QString contact, QString method);
    // Signal that the given text should be sent as a message
    void                  sendMessageToContact(QString text);
    // Signal that the given ink should be sent as a message
    void                  sendInkToContact( QByteArray ink );
    // Signal that the user wants to transfer some files
    void                  sendFiles( QList<QUrl> urls );
    // Signal for ChatMessageView to replace an application's accept/reject/cancel links with another text
    void                  updateApplicationMessage( const QString &messageId, const QString &newMessage );
    // Signal that the user is typing
    void                  userIsTyping();
};

#endif
