/***************************************************************************
                          chatwindow.cpp  -  description
                             -------------------
    begin                : Tue Apr 23 2002
    copyright            : (C) 2002 by Mike K. Bennett
    email                : mkb137b@hotmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "chatwindow.h"

#include "../network/msnswitchboardconnection.h"
#include "../utils/richtextparser.h"
#include "../utils/kmessconfig.h"
#include "../currentaccount.h"
#include "../kmess.h"
#include "../kmessapplication.h"
#include "../kmessdebug.h"
#include "chat.h"
#include "config-kmess.h"
#include "contactsidebar.h"
#include "emoticonsidebar.h"

#include <QShortcut>

#include <KAction>
#include <KActionMenu>
#include <KColorDialog>
#include <KComponentData>
#include <KConfigGroup>
#include <KFontDialog>
#include <KHelpMenu>
#include <KLocale>
#include <KMenu>
#include <KMenuBar>
#include <KMessageBox>
#include <KSqueezedTextLabel>
#include <KStandardAction>
#include <KStatusBar>
#include <KTabBar>
#include <KToolBar>
#include <KToolInvocation>


#ifdef KMESSDEBUG_CHATWINDOW
//   #define KMESSDEBUG_CHATWINDOW_TYPING_MESSAGES
#endif


// The constructor
ChatWindow::ChatWindow( QWidget *parent )
: KMainWindow( parent )
, Ui::ChatWindow()
, currentAccount_( 0 )
, currentInviteMenu_( 0 )
, customEmoticonSidebar_( 0 )
, initialized_( false )
, isSidebarTemporary_( false )
, lastActiveChatIndex_( -1 )
, standardEmoticonSidebar_( 0 )
{
  // Create the interface
  QWidget *mainWidget = new QWidget( this );
  setupUi( mainWidget );
  setCentralWidget( mainWidget );

  // Create the menus
  initMenus();

  // Create the tabs bar
  tabBar_ = new KTabBar( this );
  tabBar_->setElideMode( Qt::ElideRight );
  tabBar_->setDrawBase( false );
  tabBar_->setUsesScrollButtons( false );
#if KDE_IS_VERSION(4,0,70)
  tabBar_->setCloseButtonEnabled( true );
#else
  tabBar_->setHoverCloseButton( true );
  tabBar_->setHoverCloseButtonDelayed( true );
#endif

  tabFrame_->layout()->addWidget( tabBar_ );
  tabBar_->hide();  // Start it hidden when there's only one tab
  connect( tabBar_, SIGNAL(         currentChanged(int) ),
           this,    SLOT  (          slotSwitchTab(int) ) );
  connect( tabBar_, SIGNAL(           closeRequest(int) ),
           this,    SLOT  ( slotSwitchAndDeleteTab(int) ) );
  connect( tabBar_, SIGNAL(       mouseMiddleClick(int) ),
           this,    SLOT  ( slotSwitchAndDeleteTab(int) ) );
  connect( tabBar_, SIGNAL(             wheelDelta(int) ),
           this,    SLOT  (    slotWheelMouseMoved(int) ) );

  // Associate the keyboard shortcuts
  new QShortcut( QKeySequence("Esc"),                       this, SLOT(           queryClose() ) );
  new QShortcut( QKeySequence("Shift+Esc"),                 this, SLOT(        queryCloseAll() ) );
  new QShortcut( QKeySequence(QKeySequence::NextChild),     this, SLOT( slotSwitchToRightTab() ) );
  new QShortcut( QKeySequence(QKeySequence::PreviousChild), this, SLOT(  slotSwitchToLeftTab() ) );
  new QShortcut( QKeySequence(QKeySequence::Forward),       this, SLOT( slotSwitchToRightTab() ) );
  new QShortcut( QKeySequence(QKeySequence::Back),          this, SLOT(  slotSwitchToLeftTab() ) );


  // Create the status bar
  statusLabel_ = new KSqueezedTextLabel( this );
  statusBar()->addWidget( statusLabel_, 1 );
  // NOTE: Workaround - if the next statement is omitted, the first status message ever shown never appears
  showStatusMessage( " " );

  // Install a filter to catch window activation events, to stop title blinking
  installEventFilter( this );

  // Setup the window title blinking timer
  blinkTimer_.setSingleShot( false );
  connect( &blinkTimer_,   SIGNAL(           timeout() ),
           this,           SLOT  (  slotBlinkCaption() ) );

  // Setup the status bar text timer
  statusTimer_.setSingleShot( true );
  connect( &statusTimer_,  SIGNAL(           timeout() ),
           this,           SLOT  ( showStatusMessage() ) );
}



// The destructor
ChatWindow::~ChatWindow()
{
  // Save the window properties
  saveProperties();

#ifdef KMESSDEBUG_CHATWINDOW
  kDebug() << "DESTROYED.";
#endif
}



// Show the "about KDE" message.
void ChatWindow::aboutKDE()
{
  helpMenu_->aboutKDE();
}



// Show the "about KMess" message.
void ChatWindow::aboutKMess()
{
  helpMenu_->aboutApplication();
}



// Add a new chat tab
Chat *ChatWindow::addChatTab( Chat *newChat )
{
  int newIndex = chats_.indexOf( newChat );

  // Add the new tab to the list of chats
  chats_.append( newChat );

  // Also create a tab for the new chat
  tabBar_->insertTab( newIndex, newChat->getCaption() );

  // If there's more than one tab, show the bar and the Close All option
  bool moreThanOneTabs = ( chats_.count() > 1 );
  tabBar_->setVisible( moreThanOneTabs );
  closeAllAction_->setVisible( moreThanOneTabs );

  // Connect its main signals
  connect( newChat, SIGNAL(      chatInfoChanged()                         ),
           this,    SLOT  (  slotChatInfoChanged()                         ) );
  connect( newChat, SIGNAL(             gotNudge()                         ),
           this,    SLOT  (         slotGotNudge()                         ) );
  connect( newChat, SIGNAL( sendMessageToContact(QString)                  ),
           this,    SLOT  (  slotUserSentMessage()                         ) );
  connect( newChat, SIGNAL(       gotChatMessage(const ChatMessage&,Chat*) ),
           this,    SLOT  (   slotGotChatMessage(const ChatMessage&,Chat*) ) );
  connect( newChat, SIGNAL(     gotTypingMessage(Chat*)                    ),
           this,    SLOT  ( slotGotTypingMessage(Chat*)                    ) );

  // Add the chat's view and contact sidebar to the window's frames
  chatFrame_->layout()->addWidget( newChat );
  contactSidebarFrame_->layout()->addWidget( newChat->getContactSidebar() );

  return newChat;
}



// Make the caption blink if the window still doesn't have focus
void ChatWindow::slotBlinkCaption()
{
  // If the window now has focus, stop blinking
  if( isActiveWindow() )
  {
    // Stop the timer
    blinkTimer_.stop();

    // Reset the caption to normal
    setWindowTitle();
    return;
  }

  // Make the caption alternate case
  if ( blinkToUpper_ )
  {
    setWindowTitle( QString( "* " + caption_.toUpper() ) );
  }
  else
  {
    setWindowTitle( QString( "* " + caption_.toLower() ) );
  }

  blinkToUpper_ = ! blinkToUpper_;
}



// Toggle the sidebar based on account preferences
void ChatWindow::checkSidebarPreferences()
{
  if( KMESS_NULL( sidebar_ ) )        return;
  if( KMESS_NULL( currentAccount_ ) ) return;

  if( currentAccount_->getShowSidebar() )
  {
    isSidebarTemporary_ = false;
    sidebar_->show();
  }
  else
  {
    isSidebarTemporary_ = true;
    sidebar_->hide();
  }
}



// put the marked text/object into the clipboard
void ChatWindow::editCopy()
{
  getCurrentChat()->editCopy();
}



// put the marked text/object into the clipboard and remove
//  it from the document
void ChatWindow::editCut()
{
  KTextEdit *typingField = getCurrentChat()->getMessageEditField();

  if( typingField->hasFocus() )
  {
    typingField->cut();
  }
}



// Bring up a dialog to change the message font color.
void ChatWindow::editFont()
{
  QFont    selection;
  QString fontFamily;
  int      leftBracket;

  selection = currentAccount_->getFont();

  int result = KFontDialog::getFont( selection, KFontChooser::NoDisplayFlags, this );
  if ( result == KFontDialog::Accepted )
  {
    currentAccount_->setFont( selection );
    fontFamily = selection.family();
    // KDE3 seems to want to throw some stuff in square brackets at the
    //  end of font names.  Strip that stuff off.
    // I'm assuming no real proper font has "[" in the name.
    leftBracket = fontFamily.indexOf( "[" );
    if ( leftBracket >= 0 )
    {
      fontFamily = fontFamily.left( leftBracket );
    }
    fontFamily = fontFamily.trimmed();
    selection.setFamily( fontFamily );
    // Set it to the stored account
    currentAccount_->setFont( selection );

    // Commit changes to disk, now
    currentAccount_->saveProperties();
  }
}



// Bring up a dialog to change the message font color.
void ChatWindow::editFontColor()
{
  QColor  color = currentAccount_->getFontColor();
  QString fontColor;

  // Show the color dialog
  int result = KColorDialog::getColor( color, this );
  if ( result == KColorDialog::Accepted )
  {
    // Get the string version of the selected color
    fontColor = color.name();
    // Set it to the stored account
    currentAccount_->setFontColor(fontColor);

    // Commit changes to disk, now
    currentAccount_->saveProperties();
  }
}



// paste the clipboard into the document
void ChatWindow::editPaste()
{
  // Regardless of focus, always paste in message edit.
  getCurrentChat()->getMessageEditField()->paste();
}



// enlarge the font size
void ChatWindow::editZoomIn()
{
  getCurrentChat()->changeZoomFactor( true );
}



// decrease the font size
void ChatWindow::editZoomOut()
{
  getCurrentChat()->changeZoomFactor( false );
}




// The emoticon button was pressed.
void ChatWindow::showEmoticonPanel()
{
  // Find out what toolbox does the emoticon sidebar have
  int emoticonSidebarId = sidebar_->indexOf( standardEmoticonSidebar_ );

  // When the side bar is hidden, temporary show it.
  if( ! sidebar_->isVisible() )
  {
    isSidebarTemporary_ = true;
    sidebar_->setCurrentIndex( emoticonSidebarId );
    sidebar_->show();
  }
  else
  {
    // Sidebar is visible

    // Show the emoticon sidebar
    if( sidebar_->currentIndex() != emoticonSidebarId )
    {
      // Show emoticons
      sidebar_->setCurrentIndex( emoticonSidebarId );
    }
    else
    {
      // Emotion sidebar is displayed currently
      // Hide the emoticon sidebar
      if( isSidebarTemporary_ )
      {
        // Hide again
        sidebar_->hide();
      }
      else
      {
        // Show Main tab (contacts) again
        sidebar_->setCurrentIndex( sidebar_->indexOf( getCurrentChat()->getContactSidebar() ) );
      }
    }
  }
}



// Filter to catch window activation events
bool ChatWindow::eventFilter( QObject */*obj*/, QEvent *event )
{

  switch( event->type() )
  {
    case QEvent::WindowActivate:
      // Stop blinking immediately when the window is activated
      if( blinkTimer_.isActive() )
      {
        blinkTimer_.stop();
        setWindowTitle();
      }
      break;

    default:
      break;
  }

  return false;
}



/**
 * @brief Return a reference to a chat
 *
 * When called without parameters, this method returns the active chat.
 * If the given index is not valid, the current index is returned, too.
 *
 * @param index  Number of the chat to retrieve, or nothing to get the currently active one.
 * @return Chat *
 */
Chat *ChatWindow::getChat( int index )
{
#ifdef KMESSTEST
  KMESS_ASSERT( tabBar_->count() > 0 );
#endif

  if( index < 0 || index >= chats_.count() )
  {
    return chats_.at( tabBar_->currentIndex() );
  }

  return chats_.at( index );
}



// Return the chat object for a given switchboard connection, if found.
Chat * ChatWindow::getChatBySwitchboardConnection( const MsnSwitchboardConnection *connection ) const
{
  foreach( Chat *chat, chats_ )
  {
    MsnSwitchboardConnection *chatConnection = chat->getSwitchboardConnection();
    if( chatConnection == connection )
    {
      return chat;
    }
  }

  return 0;
}



// Return a reference to the active chat
Chat *ChatWindow::getCurrentChat()
{
#ifdef KMESSTEST
  KMESS_ASSERT( tabBar_->count() > 0 );
#endif

  // This really shouldn't happen, but instead of getting a qFatal(),
  // let the code crash with a normal null pointer instead.
  if( tabBar_->count() == 0 )
  {
    return 0;
  }

  return chats_.at( tabBar_->currentIndex() );
}



// Show the user manual for KMess.
void ChatWindow::helpContents()
{
  KToolInvocation::invokeHelp("kmess/index");
}




// Create the chat menu
void ChatWindow::initChatMenu()
{
  KAction     *closeAction, *saveAction;

  // Create the menu and the Invite Contact sub-menu
  chatMenu_          = new KMenu( i18n("&Chat"),   this );

  // Create the actions
  sendAction_        = new KAction( KIcon("folder-remote"),                         i18n("Send a &File"),     this );
  meetingAction_     = new KAction( KIcon("gnomemeeting"),                          i18n("Start a &Meeting"), this );
  nudgeAction_       = new KAction( KIcon("preferences-desktop-notification-bell"), i18n("Send a &Nudge!"),   this );
  saveAction         = new KAction( KIcon("document-save"),                         i18n("Save chat"),        this );
  addEmoticonAction_ = new KAction( KIcon("emoticons"),                             i18n("&Emoticons"),       this );
  closeAction        = KStandardAction::close( this, SLOT( queryClose() ), actionCollection_ );
  closeAllAction_    = new KAction( KIcon("dialog-close"),                          i18n("Close &All Tabs"),  this );

  // Initially the chat has one tab only, so the Close All option is not useful
  closeAllAction_->setVisible( false );

  // Add shorter texts for the toolbar
  meetingAction_ ->setIconText( i18n("Meeting") );
  nudgeAction_   ->setIconText( i18n("Nudge") );

  // Add invite button for the toolbar
  inviteButton_      = new KActionMenu( KIcon("list-add-user"), i18n("&Invite"), this );
  inviteButton_->setDelayed( false );

  // Connect them
  connect( sendAction_,        SIGNAL(         triggered(bool) ),
           this,               SLOT  ( startFileTransfer()     ) );
  connect( nudgeAction_,       SIGNAL(         triggered(bool) ),
           this,               SLOT  (         sendNudge()     ) );
  connect( meetingAction_,     SIGNAL(         triggered(bool) ),
           this,               SLOT  (      startMeeting()     ) );
  connect( saveAction,         SIGNAL(         triggered(bool) ),
           this,               SLOT  (          saveChat()     ) );
  connect( addEmoticonAction_, SIGNAL(         triggered()     ),
           this,               SLOT  ( showEmoticonPanel()     ) );
  connect( closeAllAction_,    SIGNAL(         triggered()     ),
           this,               SLOT  (     queryCloseAll()     ) );

#ifdef HAS_KPHONE  // If KPhone is not enabled, don't support voice conversation
  KAction *conversation = new KAction( "kphone", i18n("Start or Stop a &Conversation"), this );
  connect( send, SIGNAL(triggered(bool)), this, SLOT( startConversation() ) );
#endif

  // Plug the items into the Chat menu
  // The Invite menu goes before this separator
  firstChatMenuItem_ = chatMenu_->addSeparator();
  chatMenu_->addAction( sendAction_        );
  chatMenu_->addAction( meetingAction_     );
#ifdef HAS_KPHONE  // If KPhone is not enabled, don't support voice conversation
  chatMenu_->addAction( conversation       );
#endif  
  chatMenu_->addAction( nudgeAction_       );
  chatMenu_->addSeparator();
  chatMenu_->addAction( saveAction         );
  chatMenu_->addAction( closeAction        );
  chatMenu_->addAction( closeAllAction_    );

  // Also plug the icons into the toolbar
  // removed 'close', not a frequently used action (and you'll have Alt+F4 and the window close button for that).
  toolBar()->addAction( inviteButton_      );
  toolBar()->addAction( sendAction_        );
  toolBar()->addAction( nudgeAction_       );
  toolBar()->addAction( addEmoticonAction_ );

  // Plug the menu into the menubar
  menuBar()->  addMenu( chatMenu_          );
}



// Create the edit menu
void ChatWindow::initEditMenu()
{
  KAction         *changeFont, *changeFontColor;
  KAction         *copy, *zoomIn, *zoomOut;
  KMenu      *editMenu;

  // Create the menu
  editMenu = new KMenu( i18n("&Edit"), this );

  // Create the "edit" menus
  changeFont      = new KAction( KIcon("preferences-desktop-font"), i18n("Change &Font"), this );
  changeFontColor = new KAction( KIcon("format-stroke-color"), i18n("Change Font &Color"), this );
  connect( changeFont,      SIGNAL(     triggered(bool) ),
           this,              SLOT(      editFont()     ) );
  connect( changeFontColor, SIGNAL(     triggered(bool) ),
           this,              SLOT( editFontColor()     ) );

  // Add shorter texts for the toolbar
  changeFont ->setIconText( i18n("&Font") );
  changeFontColor->setIconText( i18n("Font &Color") );

  // More standard actions
  cutAction_   = KStandardAction::cut    ( this, SLOT( editCut()     ), actionCollection_ );
  copy         = KStandardAction::copy   ( this, SLOT( editCopy()    ), actionCollection_ );
  pasteAction_ = KStandardAction::paste  ( this, SLOT( editPaste()   ), actionCollection_ );
  zoomIn       = KStandardAction::zoomIn ( this, SLOT( editZoomIn()  ), actionCollection_ );
  zoomOut      = KStandardAction::zoomOut( this, SLOT( editZoomOut() ), actionCollection_ );

  // Plug the items into our new shiny Edit menu
  editMenu->addAction( cutAction_      );
  editMenu->addAction( copy            );
  editMenu->addAction( pasteAction_    );
  editMenu->addSeparator();
  editMenu->addAction( changeFont      );
  editMenu->addAction( changeFontColor );
  editMenu->addSeparator();
  editMenu->addAction( zoomIn          );
  editMenu->addAction( zoomOut         );

  // Plug the icons to the toolbar.
  // Toolbar is only for fast access to frequently used actions.
  // So cut/copy/paste are no longer in the toolbar by default.. It's not an office app you know.
  KToolBar *mainToolBar = toolBar();
  mainToolBar->addSeparator();

  // Plug the menu into the menubar
  menuBar()->addMenu( editMenu );
}



// Create the help menu
void ChatWindow::initHelpMenu()
{
  KAction *aboutKDE, *aboutKMess, *helpContents;
  KMenu   *helpMenu;

  // Create the menu
  helpMenu = new KMenu( i18n("&Help"), this );

  // Create the help menu.
  helpMenu_ = new KHelpMenu( this, KGlobal::mainComponent().aboutData() );

  // Create the help menu items
  aboutKMess   = KStandardAction::aboutApp    ( this, SLOT( aboutKMess()   ), this );
  aboutKDE     = KStandardAction::aboutKDE    ( this, SLOT( aboutKDE()     ), this );
  helpContents = KStandardAction::helpContents( this, SLOT( helpContents() ), actionCollection_ );

  // Plug the items into the new Help menu
  helpMenu->addAction( helpContents );
  helpMenu->addSeparator();
  helpMenu->addAction( aboutKMess );
  helpMenu->addAction( aboutKDE );

  // Plug the menu into the menubar
  menuBar()->addMenu( helpMenu );
}



// Initialize the user interface
bool ChatWindow::initialize()
{
  if( initialized_ )
  {
    kDebug() << "already initialized.";
    return false;
  }

  if( ! initializeCurrentAccount() )
  {
    kDebug() << "Couldn't setup the current account.";
    return false;
  }

  // Create the emoticon sidebars
  standardEmoticonSidebar_ = new EmoticonSidebar( sidebar_, false );
  customEmoticonSidebar_   = new EmoticonSidebar( sidebar_, true  );

  sidebar_->addItem( standardEmoticonSidebar_, i18n( "Emoticons"    ) );
  sidebar_->addItem( customEmoticonSidebar_,   i18n( "My emoticons" ) );

  // Read the window properties
  readProperties();
  checkSidebarPreferences();

  // Update the chat window caption with the contact name
  update();

  initialized_ = true;
  return true;
}



// Set up the current account
bool ChatWindow::initializeCurrentAccount()
{
#ifdef KMESSTEST
  KMESS_ASSERT( currentAccount_ == 0 );
#endif

  currentAccount_ = CurrentAccount::instance();
  if ( currentAccount_ == 0 )
  {
    kDebug() << "Couldn't get an instance of the current account.";
    return false;
  }

  // Set up some widgets based on account settings
  emoticonAction_->setChecked( currentAccount_->getUseEmoticons() );

#ifdef KMESSTEST
  KMESS_ASSERT( currentAccount_ != 0 );
#endif
  return true;
}



// Create the menus.
void ChatWindow::initMenus()
{
  // ActionCollection to collect the actions with a shortcut
  actionCollection_ = new KActionCollection( this, KComponentData() );

  // Initialize the menus with their separate functions
  initChatMenu();
  initEditMenu();
  initSettingsMenu();
  initHelpMenu();

  // Get a fast reference to the main toolbar object
  KToolBar *mainToolbar = toolBar();

  // Add the emoticon button
  mainToolbar->addAction( addEmoticonAction_ );
  // Plug the sidebar button in after the emoticon button
  mainToolbar->addAction( sidebarAction_ );

  // Add collected actions
  addActions( actionCollection_->actions() );
}



// Create the settings menu
void ChatWindow::initSettingsMenu()
{
  KMenu         *settingsMenu;
  KToggleAction *spellCheckAction;

  // Create the menu
  settingsMenu = new KMenu( i18n("&Settings"), this );

  // Create the "settings" menus
  spellCheckAction = new KToggleAction( KIcon("tools-check-spelling"), i18n("Use &Spell Checking"),   this );
  emoticonAction_  = new KToggleAction( KIcon("services"),             i18n("Show &Emoticons"),       this );
  sidebarAction_   = new KAction( KIcon("view-sidetree"), i18n("Hide Side&bar"), this );
  showToolBar_     = new KToggleAction( KIcon("view-sidetree"), i18n("Hide &Toolbar"), actionCollection_ );
  showStatusBar_   = KStandardAction::showStatusbar( this, SLOT( showStatusBar() ), this );
  showMenuBar_     = KStandardAction::showMenubar( this, SLOT( showMenuBar() ), actionCollection_ );

  // Assign to the sidebar action a keyboard shortcut and a tool tip like the standard actions
  sidebarAction_->setShortcut( QKeySequence( "Ctrl+T" ) );
  sidebarAction_->setToolTip( i18n("Hide the contact sidebar") );  // tooltip and title updated by ChatWindow::toggleSidebar()

  // Connect the do... slots!
  connect( emoticonAction_,   SIGNAL(          toggled(bool) ),
           this,              SLOT  (  toggleEmoticons(bool) ) );
  connect( sidebarAction_,    SIGNAL(        activated()     ),
           this,              SLOT  (    toggleSidebar()     ) );
  connect( showToolBar_,      SIGNAL(          toggled(bool) ),
           this,              SLOT  (      showToolBar()     ) );
  connect( spellCheckAction,  SIGNAL(          toggled(bool) ),
           this,              SLOT  ( toggleSpellCheck(bool) ) );

  // Plug the items into the new Settings menu
  settingsMenu->addAction( spellCheckAction );
  settingsMenu->addSeparator();
  settingsMenu->addAction( emoticonAction_  );
  settingsMenu->addAction( showMenuBar_     );
  settingsMenu->addAction( showToolBar_     );
  settingsMenu->addAction( showStatusBar_   );
  settingsMenu->addAction( sidebarAction_   );

  // Plug the menu into the menubar
  menuBar()->addMenu( settingsMenu );
}



// Return whether the side bar is enabled
bool ChatWindow::isSidebarVisible()
{
  return sidebar_->isVisible();
}



// A chat tab is closing
bool ChatWindow::queryClose()
{
#ifdef KMESSDEBUG_CHATWINDOW
  kDebug() << "Closing current chat tab";
#endif

  int currentIndex = tabBar_->currentIndex();
  Chat *currentChat = getCurrentChat();
  if( currentChat != 0 )
  {
    // Save the chat's UI settings before closing it, so the next selected tab will read them
    currentChat->saveProperties();

    currentChat->queryClose();

    if( ! removeChatTab( currentIndex ) )
    {
      kWarning() << "Unable to close the current chat!";
    }

    // If there still are open tabs, don't close
    if( chats_.count() > 0 )
    {
      // Select the leftmost tab
      int newIndex = ( currentIndex == 0 ) ? currentIndex : currentIndex - 1;
      tabBar_->setCurrentIndex( newIndex );
      slotSwitchTab( newIndex );
      return false;
    }
  }

  // Kill this window
  hide();
  deleteLater();

  // If the main window is not visible, make sure only this window closes
  KMessApplication *kmessApp = static_cast<KMessApplication*>( kapp );
  if( ! kmessApp->getContactListWindow()->isVisible() && ! kmessApp->quitSelected() )
  {
#ifdef KMESSDEBUG_KMESSINTERFACE
    kDebug() << "Rejecting quit request, only close this window.";
#endif

    // Reject close / quit request.
    return false;
  }
  else
  {
#ifdef KMESSDEBUG_KMESSINTERFACE
    kDebug() << "Accepting quit request.";
#endif

    // Normal procedure is ok
    return true;
  }
}



// The chat window is closing
bool ChatWindow::queryCloseAll()
{
  if( chats_.count() > 1 )
  {
    int res = KMessageBox::questionYesNoCancel( this,
                                                i18n( "<html>There are multiple tabs open in this chat window. Do you want to close the "
                                                "current tab only, or all tabs?<br /><br />"
                                                "<i>Note:You can also close all tabs at once by pressing Shift+Esc.</i></html>" ),
                                                i18nc( "Caption when closing a single chat window tab", "Closing a Chat Tab" ),
                                                KGuiItem( i18n( "Close all tabs" ), "window-close" ), // Yes option
                                                KGuiItem( i18n( "Close current tab" ), "tab-close" ), // No option
                                                KStandardGuiItem::cancel(),
                                                "closeOneChatTabInfo" );
    switch( res )
    {
      case KMessageBox::Yes   : break;
      case KMessageBox::No    : return queryClose();
      case KMessageBox::Cancel:
      default                 : return false;
    }
  }

  // Just close all tabs one by one, until the last one closes the window too
  while( chats_.count() > 0 )
  {
    queryClose();
  }

  return true;
}



// The application is exiting
bool ChatWindow::queryExit()
{
#ifdef KMESSDEBUG_CHATWINDOW
  kDebug() << "Closing all open chats";
#endif

  // Allow performing last operations to all open chats and delete them
  foreach( Chat *chat, chats_ )
  {
    chat->queryExit();
    delete chat;
  }

  // Clear the list of chat tabs
  chats_.clear();

  // Force a call to the destructor, to ensure correct cleanup by ChatMaster
  deleteLater();

  // Always accept this one, otherwise the chat window won't close.
  // The quit is precented in KMessApplication::slotLastWindowClosed() too.
#ifdef KMESSDEBUG_KMESSINTERFACE
  kDebug() << "Accepting exit request.";
#endif

  return true;
}



// Restore the window properties (called by KMainWindow)
void ChatWindow::readProperties(const KConfigGroup &config )
{
  KConfigGroup group;

  // Choose the config group if it was not given
  if( &config == 0 )
  {
    group = KMessConfig::instance()->getAccountConfig( currentAccount_->getHandle(), "ChatWindow" );
  }
  else
  {
    group = config;
  }

  // First load the general interface settings
  applyMainWindowSettings( group );

  // Set the default widget sizes. The sidebar has a default read size of 1 so it'll be
  //  as small as possible, without becoming invisible
  int minSidebarWidth = sidebar_->minimumSize().width();
  sidebarSplitter_->setSizes( group.readEntry( "SidebarSplitterSizes", QList<int>() << minSidebarWidth << 1 ) );

  // Read the window size and position with a default for common screen dimensions
  QSize size = group.readEntry( "Size", QSize(550,450) );
  if( size.isValid() )
  {
    resize( size );
  }

  QPoint pos = group.readEntry( "Position", QPoint() );
  if( ! pos.isNull() )
  {
    move( pos );
  }

  // bar status settings
  showMenuBar_  -> setChecked( group.readEntry( "showMenuBar",   true ) );
  showToolBar_  -> setChecked( group.readEntry( "showToolBar",   true ) );
  showStatusBar_-> setChecked( group.readEntry( "showStatusBar", true ) );

  showToolBar();
  showStatusBar();

  // Toolbar settings
  toolBar()->applySettings( group );
}



// Delete an existing chat tab
bool ChatWindow::removeChatTab( int index )
{
#ifdef KMESSTEST
  KMESS_ASSERT( chats_.count() > 0 );
#endif

  // Find which chat has the given index
  Chat *chat = getChat( index );

  if( chat == 0 )
  {
    kWarning() << "Asked to remove index" << index << ", but there are" << chats_.count() << "open tabs!";
    return false;
  }
#ifdef KMESSDEBUG_CHATWINDOW
  kDebug() << "Removing Chat:" << chat;
  kDebug() << "All chats:" << chats_;
  kDebug() << "Index of tab:" << index;
  kDebug() << "Index in the list:" << chats_.indexOf( chat );
#endif

  // Remove the tab
  tabBar_->removeTab( chats_.indexOf( chat ) );

  // Remove the chat from the list of open chats
  chats_.removeAll( chat );

  // If there's more than one tab, show the bar and the Close All option
  bool moreThanOneTabs = ( chats_.count() > 1 );
  tabBar_->setVisible( moreThanOneTabs );
  closeAllAction_->setVisible( moreThanOneTabs );

  // Remove the chat's view and contact sidebar from the window's frames
  ContactSidebar *contactSideBar = chat->getContactSidebar();
  chatFrame_->layout()->removeWidget( chat );
  contactSidebarFrame_->layout()->removeWidget( contactSideBar  );
  delete contactSideBar;
  delete chat;

  return true;
}



// Save the chat according to the user's request
void ChatWindow::saveChat()
{
  getCurrentChat()->showSaveChatDialog();
}



// Save the window properties (called by KMainWindow)
void ChatWindow::saveProperties( KConfigGroup &config )
{
  KConfigGroup group;

  // Choose the config group if it was not given
  if( &config == 0 )
  {
    group = KMessConfig::instance()->getAccountConfig( currentAccount_->getHandle(), "ChatWindow" );
  }
  else
  {
    group = config;
  }

  // First save the general interface settings
  saveMainWindowSettings( group );

  // Save widgets sizes
  group.writeEntry( "SidebarSplitterSizes", sidebarSplitter_->sizes()   );

  // Save the window geometry
  group.writeEntry( "Size",                 size()                      );
  group.writeEntry( "Position",             pos()                       );
  group.writeEntry( "showMenuBar",          showMenuBar_  ->isChecked() );
  group.writeEntry( "showToolBar",          showToolBar_  ->isChecked() );
  group.writeEntry( "showStatusBar",        showStatusBar_->isChecked() );

  // Toolbar settings
  toolBar()->saveSettings( group );

  // Only save the chats properties when there are some
  if( chats_.count() > 0 )
  {
    // Save the last active chat's settings
    getCurrentChat()->saveProperties( group ) ;
  }
}



// Enable or disable the parts of the window which allow user interaction
void ChatWindow::setEnabled( bool isEnabled )
{
  // Set the state of the main UI elements
  customEmoticonSidebar_   ->setEnabled( isEnabled );
  standardEmoticonSidebar_ ->setEnabled( isEnabled );

  // Set the state of the extra contact actions
  meetingAction_           ->setEnabled( isEnabled );
  nudgeAction_             ->setEnabled( isEnabled );
  sendAction_              ->setEnabled( isEnabled );

  // Set the state of cut and paste
  cutAction_               ->setEnabled( isEnabled );
  pasteAction_             ->setEnabled( isEnabled );

  // The window is being enabled back, replace some stuff which changes
  // on every connection
  if( isEnabled )
  {
    // Re-read the current account pointer
    // because it likely is a different one
    currentAccount_ = CurrentAccount::instance();
  }
}



// "Show menu bar" was toggled.
void ChatWindow::showMenuBar()
{
  if(!showMenuBar_->isChecked())
  {
    menuBar()->hide();
  }
  else
  {
    menuBar()->show();
  }
}



// "Show status bar" was toggled.
void ChatWindow::showStatusBar()
{
  statusBar()->setVisible( showStatusBar_->isChecked() );
}



// Show a status message for some seconds in the status dialog
void ChatWindow::showStatusMessage( QString message, int duration )
{
  // Avoid newlines which cause problems to the UI layout.
  message = Qt::escape( message.replace( "\n", " " ) );

#ifdef KMESSDEBUG_CHATWINDOW
  kDebug() << "Showing status bar text" << message << "with duration" << duration;
#endif

  statusLabel_->setText( message );

  // Each message will have an expiration time of its own or none at all
  statusTimer_.stop();

  // Non-expiring messages do not trigger the timer
  if( duration <= 0 || message.isEmpty() )
  {
    return;
  }

#ifdef KMESSDEBUG_CHATWINDOW
  kDebug() << "Setting status bar update timer to" << duration << "sec.";
#endif
  statusTimer_.start( duration * 1000 );
}



// "Show toolbar" was toggled.
void ChatWindow::showToolBar()
{
  toolBar( "mainToolBar" )->setVisible( showToolBar_->isChecked() );
}



// A message has been received
void ChatWindow::slotGotChatMessage( const ChatMessage &message, Chat *chat )
{
  // Update the sidebar
  checkSidebarPreferences();

  // If is the user the sender of the chat message, switch to the contacts sidebar
  if( ! message.isIncoming() )
  {
    // If the emoticon sidebar is still visible, show contact bar again.
    if( sidebar_->currentIndex() == sidebar_->indexOf( standardEmoticonSidebar_ ) )
    {
      sidebar_->setCurrentIndex( sidebar_->indexOf( getCurrentChat()->getContactSidebar() ) );
    }

    // When the sidebar was only temporary visible, hide it again
    if( isSidebarTemporary_ && ! currentAccount_->getShowSidebar() )
    {
      sidebar_->hide();
    }
  }

  // Show the alert icon
  if( chat != getCurrentChat() )
  {
    tabBar_->setTabIcon( chats_.indexOf( chat ), KIcon( "mail-mark-important" ) );
  }

  // If the window does not have focus, blink the caption as visual cue of an event
  if( isActiveWindow() )
  {
    return;
  }

  // Start blinking the caption
  blinkToUpper_ = true;
  slotBlinkCaption();
  blinkTimer_.start( 2000 );
}



// A nudge has been received: shake the window.
void ChatWindow::slotGotNudge()
{
  if( ! currentAccount_->getShakeNudge() )
  {
    return;
  }


  // avoid shaking again in a short period.
  if( lastShake_.secsTo( QTime::currentTime() ) < 15 )
  {
#ifdef KMESSDEBUG_CHATWINDOW
    kDebug() << "Not shaking the window until"
             << lastShake_.secsTo( QTime::currentTime() ) << "more seconds.";
#endif
    return;
  }
  lastShake_ = QTime::currentTime();

  QTime t;
  int xp = x();
  int yp = y();

  t.start();

  // Code example from: http://ariya.blogspot.com/2005/12/buzz-or-shake-my-window.html

  // Shake the window.
  for ( int i = 16; i > 0; )
  {
    kapp->processEvents();

    if ( t.elapsed() >= 50 )
    {
      int delta = i >> 2;
      int dir   = i & 3;
      int dx    = (dir == 1 || dir == 2) ? delta : -delta;
      int dy    = (dir < 2) ? delta : -delta;

      move( xp+dx, yp+dy );
      t.restart();
      i--;
    }
  }

  // Move to start again
  move( xp, yp );
}



// Wheel mouse moved: change the tab.
void ChatWindow::slotWheelMouseMoved( int delta )
{
  // If delta < 0 moves right trought tabs
  if( delta < 0 )
  {
    slotSwitchToRightTab();
  }
  // If delta > 0 moves left trought tabs
  else
  {
    slotSwitchToLeftTab();
  }
}



// The 'Send nudge' button has been clicked
void ChatWindow::sendNudge()
{
  getCurrentChat()->slotSendNudge();
}



// Change the active chat
void ChatWindow::setCurrentChat( Chat *chat )
{
  tabBar_->setCurrentIndex( chats_.indexOf( chat ) );
}



// Start a file transfer
void ChatWindow::startFileTransfer()
{
  getCurrentChat()->startFileTransfer();
}



// Switch the window to the clicked tab
void ChatWindow::slotChatInfoChanged()
{
  // Grep the chat has emitted the signal
  Chat *chat = static_cast< Chat *>( sender() );

  // Search it into the list
  int index = chats_.indexOf( chat );
  if( index != -1 )
  {
    // Change the tab caption
    tabBar_->setTabText( index, chat->getCaption() );

    // If chat is current chat then calls update caption method
    if ( chat == getCurrentChat() )
    {
      setWindowTitle();
    }
  }
}



// Switch the window to the clicked tab
void ChatWindow::slotSwitchTab( int tabIndex )
{
  // Avoid connecting to non-existent sidebar tabs
  if( ! initialized_ )
  {
#ifdef KMESSDEBUG_CHATWINDOW
    kDebug() << "Not switching tab while not initialized.";
#endif
    return;
  }

  if( tabIndex < 0 || tabIndex >= chats_.count() )
  {
    kWarning() << "Chat tab index" << tabIndex << "doesn't exist in the tabs list!";
    return;
  }

  // Don't reactivate already selected chats
  // Don't return if current tab and lastActiveChat are 0 beacuse if we close chat 0
  // and lastActiveChat is 0 the chat doesn't refresh!
  if( lastActiveChatIndex_ == tabIndex && ( lastActiveChatIndex_ != 0 && tabIndex != 0 ) )
  {
    return;
  }

  Chat *chat = chats_.at( tabIndex );
  ContactSidebar *contactSidebar = chat->getContactSidebar();

  // Remove the widgets of the various open chats from the chat and contact sidebar frame
  Chat *oldChat = getChat( lastActiveChatIndex_ );
  if( oldChat )
  {
    // Save the chat's UI settings
    oldChat->saveProperties();

    ContactSidebar *oldContactSidebar = oldChat->getContactSidebar();

    // Disconnect the slots from the previously selected chat
    disconnect( oldContactSidebar,        0, this,    0 );
    disconnect( standardEmoticonSidebar_, 0, oldChat, 0 );
    disconnect( customEmoticonSidebar_,   0, oldChat, 0 );

    oldChat->getInviteMenu()->hide();
    oldContactSidebar->hide();
    oldChat->hide();
  }

  // Load the old chat's UI settings into this chat
  chat->readProperties();

  // Reset the icon
  tabBar_->setTabIcon( tabIndex, KIcon() );

  // Display the new chat
  chat->show();
  contactSidebar->show();

  // connect some signals so that the sidebars can send commands back
  connect( standardEmoticonSidebar_, SIGNAL(    insertEmoticon(QString)      ),
           chat,                     SLOT  (    insertEmoticon(QString)      ) );
  connect( customEmoticonSidebar_,   SIGNAL(    insertEmoticon(QString)      ),
           chat,                     SLOT  (    insertEmoticon(QString)      ) );

  // Add the chat's invite menu in the toolbar
  inviteButton_->setMenu( chat->getInviteMenu() );

  // Replace the old chat's invite menu with the new one's
  chatMenu_->removeAction( currentInviteMenu_ );
  currentInviteMenu_ = chatMenu_->insertMenu( firstChatMenuItem_, chat->getInviteMenu() );

  // If the new chat is an offline one, disable the unavailable features
  MsnSwitchboardConnection *connection = chat->getSwitchboardConnection();
  bool enableFeatures = ( connection && ! connection->isInactive() );
  sendAction_       ->setEnabled( enableFeatures );
  nudgeAction_      ->setEnabled( enableFeatures );
  meetingAction_    ->setEnabled( enableFeatures );
  currentInviteMenu_->setEnabled( enableFeatures );

  // Save the new chat as last active one
  lastActiveChatIndex_ = tabIndex;

  // Update the UI with this chat's details
  setWindowTitle();
  slotGotTypingMessage( chat );
}



// Switch the window to the tab at the left of the current one
void ChatWindow::slotSwitchToLeftTab()
{
  // Find out at what tab are we and how many are there
  int current = tabBar_->currentIndex();
  int count = tabBar_->count();

  // Do nothing if there's only one tab
  if( count <= 1 )
  {
    return;
  }

  // If the first tab was selected, then restart from the last tab
  if( current == 0 )
  {
    tabBar_->setCurrentIndex( count - 1 );
    return;
  }

  tabBar_->setCurrentIndex( current - 1 );
}



// Switch the window to the tab at the right of the current one
void ChatWindow::slotSwitchToRightTab()
{
  // Find out at what tab are we and how many are there
  int current = tabBar_->currentIndex();
  int count = tabBar_->count();

  // Do nothing if there's only one tab
  if( count <= 1 )
  {
    return;
  }

  // If the last tab was selected, then restart from the first tab
  if( current >= ( count - 1 ) )
  {
    tabBar_->setCurrentIndex( 0 );
    return;
  }

  tabBar_->setCurrentIndex( current + 1 );
}



// Switch the window to the clicked tab and remove it
void ChatWindow::slotSwitchAndDeleteTab( int tabIndex )
{
  tabBar_->setCurrentIndex( tabIndex );
  queryClose();
}



// The user sent a new message
void ChatWindow::slotUserSentMessage()
{
  // If the emoticon sidebar is still visible, show contact bar again.
  if( sidebar_->currentIndex() == sidebar_->indexOf( standardEmoticonSidebar_ ) )
  {
    sidebar_->setCurrentIndex( sidebar_->indexOf( getCurrentChat()->getContactSidebar() ) );
  }

  // When the sidebar was only temporary visible, hide it again
  if( isSidebarTemporary_ && ! currentAccount_->getShowSidebar() )
  {
    sidebar_->hide();
  }
}



// Start GnomeMeeting with a contact.
void ChatWindow::startMeeting()
{
  getCurrentChat()->startMeeting();
}



// Start a voice conversation
void ChatWindow::startConversation()
{
  getCurrentChat()->startConversation();
}



// Called when the "use emoticons" action is called.
void ChatWindow::toggleEmoticons( bool useEmoticons )
{
  // Set the new emoticon use value to the account
  currentAccount_->setUseEmoticons( useEmoticons );
}



// Called when the "show sidebar" action is called.
void ChatWindow::toggleSidebar()
{
  if( sidebar_->isVisible() )
  {
    sidebar_->hide();
    sidebarAction_->setText( i18n("Show Side&bar") );
    sidebarAction_->setToolTip( i18n("Show the contact sidebar") );
  }
  else
  {
    sidebar_->show();
    sidebarAction_->setText( i18n("Hide Side&bar") );
    sidebarAction_->setToolTip( i18n("Hide the contact sidebar") );
  }

  // Toggle the current account's setting.
  if ( currentAccount_ != 0 )
  {
    currentAccount_->setShowSidebar( sidebar_->isVisible() );

#ifdef KMESSDEBUG_CHATWINDOW
    kDebug() << "Set 'show sidebar' to " << currentAccount_->getShowSidebar() << ".";
#endif
  }
}



// Called when the "use spell checking" action is called.
void ChatWindow::toggleSpellCheck( bool useSpellCheck )
{
  // Don't crash if no tabs are available
  if( KMESS_NULL(getCurrentChat()) ) return;

#ifdef KMESSDEBUG_CHATWINDOW
  kDebug() << "Spell checking is now " << useSpellCheck;
#endif

  // Change the spell checking feature status
  getCurrentChat()->getMessageEditField()->setCheckSpellingEnabled( useSpellCheck );
}



// Update the contact typing messages in the status bar
void ChatWindow::slotGotTypingMessage( Chat *chat )
{
  // Get the active chat if called with no or invalid parameters
  if( chat == 0 )
  {
    chat = getCurrentChat();
#ifdef KMESSDEBUG_CHATWINDOW_TYPING_MESSAGES
    kDebug() << "Using current chat" << chat;
#endif
  }
#ifdef KMESSDEBUG_CHATWINDOW_TYPING_MESSAGES
  else
  {
    kDebug() << "Using chat:" << chat;
  }
#endif

  QPalette *palette = new QPalette();
  int tabIndex = chats_.indexOf( chat );
  QStringList typingContacts = chat->getTypingContacts();

#ifdef KMESSDEBUG_CHATWINDOW_TYPING_MESSAGES
  kDebug() << "These contacts are typing:" << typingContacts;
#endif

  // Update the color of this chat's tab
  if( typingContacts.isEmpty() )
  {
    tabBar_->setTabTextColor( tabIndex, palette->color( QPalette::WindowText ) );
  }
  else
  {
    tabBar_->setTabTextColor( tabIndex, palette->color( QPalette::Highlight ) );
  }

  // If the typing message was not intended for the currently active tab, skip
  // the status bar processing altogether
  if( chat != getCurrentChat() )
  {
#ifdef KMESSDEBUG_CHATWINDOW_TYPING_MESSAGES
    kDebug() << "Typing message was not for the active chat, skipping status bar update.";
#endif
    return;
  }

  // If no one is typing anymore, clean up the status bar
  if( typingContacts.isEmpty() )
  {
    showStatusMessage( QString() );
    return;
  }

  QString statusString;

  // Decide what string will be used depending on how many contacts are typing
  if( typingContacts.count() == 1 )
  {
    QString first = typingContacts.first();
    statusString = i18n("%1 is typing.", first );
  }
  else
  {
    int     count = typingContacts.count() - 2;
    QString first = typingContacts[0], second = typingContacts[1];

    if( count == 0 )
    {
      statusString = i18n("%1 and %2 are typing.", first, second );
    }
    else
    {
      statusString = i18n("%1, %2 and %3 others are typing.", first, second, count );
    }
  }

  // Set the message to the status bar without a time limit: the chat itself will send another
  // signal when the current message expires.
  showStatusMessage( statusString, 0 );
}



// Change the caption of the chat window
void ChatWindow::setWindowTitle( const QString &caption )
{
  // Remember the window caption when it is fetched from the current chat
  if( caption.isEmpty() )
  {
    Chat *currentChat = getCurrentChat();
    if( currentChat != 0 )
    {
      // Avoid newlines which would cause problems to the UI layout
      caption_ = currentChat->getCaption().replace( "\n", " " );
    }
    else
    {
      caption_ = QString();
    }
  }

  // Always append the "Chat" prefix to the chat
  if( caption.isEmpty() && caption_.isEmpty() )
  {
    setCaption( i18nc( "Chat window caption, without contact name", "Chat" ) );
  }
  else
  {
    setCaption( i18nc( "Chat window caption, with contact name", "%1 - Chat", ( caption.isEmpty() ? caption_ : caption ) ) );
  }
}



#include "chatwindow.moc"
