/***************************************************************************
                          chatwindow.h  -  description
                             -------------------
    begin                : Tue Apr 23 2002
    copyright            : (C) 2002 by Mike K. Bennett
    email                : mkb137b@hotmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef CHATWINDOW_H
#define CHATWINDOW_H

#include "ui_chatwindow.h"

#include <QList>
#include <QTimer>
#include <QWidget>

#include <KAction>
#include <KActionCollection>
#include <KMainWindow>
#include <KToggleAction>


// forward declarations
class Chat;
class ChatMessage;
class ChatView;
class CurrentAccount;
class EmoticonSidebar;
class MsnSwitchboardConnection;

class QLabel;

class KActionMenu;
class KConfigGroup;
class KHelpMenu;
class KMenu;
class KTabBar;
class KToolBarButton;
class KToolBarPopupAction;



/**
 * The interface for the chat window, defining the menus.
 *
 * @author Mike K. Bennett
 * @ingroup Chat
 */
class ChatWindow : public KMainWindow, private Ui::ChatWindow
{
  Q_OBJECT

  friend class KMessTest;

  public:
    // The constructor
                     ChatWindow(QWidget *parent=0);
    // The destructor
    virtual         ~ChatWindow();
    // Add a new chat tab
    Chat            *addChatTab( Chat *newChat );
    // Return a reference to a chat
    Chat            *getChat( int index = -1 );
    // Return a reference to the active chat
    Chat            *getCurrentChat();
    // Change the active chat
    void             setCurrentChat( Chat *chat );
    // Initialize the object
    bool             initialize();
    // Return whether the side bar is enabled
    bool             isSidebarVisible();
    // Return the chat object for a given switchboard connection, if found.
    Chat            *getChatBySwitchboardConnection( const MsnSwitchboardConnection *connection ) const;
    // Enable or disable the parts of the window which allow user interaction
    void             setEnabled( bool isEnabled );

  private: // Private methods
    // Toggle the sidebar based on account preferences
    void             checkSidebarPreferences();
    // Filter to catch window activation and keyboard shortcut events
    bool             eventFilter( QObject *obj, QEvent *event );
    // Initialize the connect menu.
    void             initChatMenu();
    // Initialize the settings menu.
    void             initEditMenu();
    // Initialize the help menu.
    void             initHelpMenu();
    // Set up the preferences which depend on the current account
    bool             initializeCurrentAccount();
    // Create the menus.
    void             initMenus();
    // Initialize the settings menu.
    void             initSettingsMenu();
    // Restore the window properties (called by KMainWindow)
    void             readProperties( const KConfigGroup &config = *((const KConfigGroup *)0) );
    // Delete an existing chat tab
    bool             removeChatTab( int index );
    // The application is exiting
    bool             queryExit();
    // Save the window properties (called by KMainWindow)
    void             saveProperties( KConfigGroup &config = *((KConfigGroup *)0) );

  public slots:
    // Show a status message temporary in the status dialog
    void             showStatusMessage( QString message = QString(), int duration = 5000 );

  private slots: // Private slots
    // Show the "about KDE" message.
    void             aboutKDE();
    // Show the "about KMess" message.
    void             aboutKMess();
    // put the marked text/object into the clipboard
    void             editCopy();
    // put the marked text/object into the clipboard and remove it from the document
    void             editCut();
    // Bring up a dialog to change the message font color.
    void             editFont();
    // Bring up a dialog to change the message font color.
    void             editFontColor();
    // paste the clipboard into the document
    void             editPaste();
    // enlarge the text size
    void             editZoomIn();
    // decrease the text size
    void             editZoomOut();
     // Show the user manual for KMess.
    void             helpContents();
    // A chat tab is closing
    bool             queryClose();
    // The chat window is closing
    bool             queryCloseAll();
    // Save the chat according to the user's request
    void             saveChat();
    // Change the caption of the chat window
    void             setWindowTitle( const QString &caption = QString() );
    // The emoticon button was pressed.
    void             showEmoticonPanel();
    // "Show status bar" was toggled.
    void             showStatusBar();
    // "Show toolbar" was toggled.
    void             showToolBar();
    // "Show menu bar" was toggled.
    void             showMenuBar();
    // Make the caption blink if the window still doesn't have focus
    void             slotBlinkCaption();
    // A message has been received
    void             slotGotChatMessage( const ChatMessage &message, Chat *chat );
    // A nudge has been received: shake the window.
    void             slotGotNudge();
    // Update the status label text
    void             slotGotTypingMessage( Chat *chat = 0 );
    // Wheel mouse moved: change the tab.
    void             slotWheelMouseMoved( int delta );
    // The 'Send nudge' button has been clicked
    void             sendNudge();
    // Change the info for tab
    void             slotChatInfoChanged();
    // Switch the window to the clicked tab
    void             slotSwitchTab( int tabIndex );
    // Switch the window to the clicked tab and remove it
    void             slotSwitchAndDeleteTab( int tabIndex );
    // Switch the window to the tab at the left of the current one
    void             slotSwitchToLeftTab();
    // Switch the window to the tab at the right of the current one
    void             slotSwitchToRightTab();
    // The user sent a new message
    void             slotUserSentMessage();
    // Start a voice conversation
    void             startConversation();
    // Start a file transfer
    void             startFileTransfer();
    // Start GnomeMeeting with a contact.
    void             startMeeting();
    // Called when the "use emoticons" action is called.
    void             toggleEmoticons( bool useEmoticons );
    // Called when the "show sidebar" action is called.
    void             toggleSidebar();
    // Called when the "use spell checking" action is called.
    void             toggleSpellCheck( bool useSpellCheck );

  private: // private attributes
    // The add emoticon action
    KAction         *addEmoticonAction_;
    // A timer used to make the caption blink on new messages.
    QTimer           blinkTimer_;
    // Whether to blink to upper (true) or lower (false) case
    bool             blinkToUpper_;
    // The stored window caption
    QString          caption_;
    // The Chat menu
    KMenu           *chatMenu_;
    // The list of active chat tabs
    QList<Chat*>     chats_;
    // Action to close the window along with all of its tabs
    KAction         *closeAllAction_;
    // A pointer to the instance of the current account
    CurrentAccount  *currentAccount_;
    // The currently shown Invite menu
    QAction         *currentInviteMenu_;
    // The custom emoticons sidebar
    EmoticonSidebar *customEmoticonSidebar_;
    // Action to cut text to the clipboard
    KAction         *cutAction_;
    // The emoticon sidebar toggling action
    KAction         *emoticonAction_;
    // The first item in the Chat Menu (used to add the Invite menu before it)
    QAction         *firstChatMenuItem_;
    // The help menu launcher
    KHelpMenu       *helpMenu_;
    // Whether or not the object was initialized
    bool             initialized_;
    // The contact invite action
    KActionMenu     *inviteButton_;
    // Whether the sidebar should only be temporary visible
    bool             isSidebarTemporary_;
    // The index of the chat tab which was activated last
    int              lastActiveChatIndex_;
    // The last time the window was shaked
    QTime            lastShake_;
    // Action to open a meeting
    KAction         *meetingAction_;
    // Action to send a nudge
    KAction         *nudgeAction_;
    // Action to paste the clipboard contents
    KAction         *pasteAction_;
    // Toggle actions for showing and hiding the statusbar
    KToggleAction   *showStatusBar_;
    // Toggle actions for showing and hiding the toolbar
    KToggleAction   *showToolBar_;
    // Toggle actions for showing and hiding the menu bar
    KToggleAction   *showMenuBar_;
    // The sidebar toggle
    KAction         *sidebarAction_;
    // Action to start a file transfer
    KAction         *sendAction_;
    // The standard emoticons sidebar
    EmoticonSidebar *standardEmoticonSidebar_;
    // The status label
    QLabel          *statusLabel_;
    // A timer to time the duration that a status message is shown
    QTimer           statusTimer_;
    // The bar of chat tabs
    KTabBar         *tabBar_;
    // ActionCollection
    KActionCollection *actionCollection_;

  signals: // Public signals
    // Signal that the window is about to close.
    void               closing(QObject *object);
};
#endif
