/***************************************************************************
                          contactframe.cpp  -  description
                             -------------------
    begin                : Thu Jan 16 2003
    copyright            : (C) 2003 by Mike K. Bennett
    email                : mkb137b@hotmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "contactframe.h"

#include "../contact/contact.h"
#include "../contact/contactlist.h"
#include "../contact/msnstatus.h"
#include "../dialogs/contactpropertiesdialog.h"
#include "../utils/kmessshared.h"
#include "../utils/richtextparser.h"
#include "../currentaccount.h"
#include "../kmessdebug.h"

#include <QClipboard>
#include <QDateTime>
#include <QFileInfo>
#include <QLabel>
#include <QLayout>
#include <QMouseEvent>
#include <QPixmap>
#include <QPushButton>
#include <QTextDocument>
#include <QToolButton>
#include <QToolTip>

#include <KApplication>
#include <KActionMenu>
#include <KGlobal>
#include <KIconEffect>
#include <KIconLoader>
#include <KLocale>
#include <KMenu>
#include <KRun>
#include <KSqueezedTextLabel>
#include <KStandardDirs>
#include <KUrl>


// Time in milliseconds before the glowing done for a typing user fades away
#define GLOWING_PICTURE_TIMEOUT 5000



// The constructor
ContactFrame::ContactFrame(QWidget *parent )
 : QWidget(parent)
 , Ui::ContactFrame()
 , activated_(false)
 , contact_(0)
 , contactPropertiesDialog_(0)
 , detailedContact_(0)
 , handle_(QString::null)
 , currentMode_(ModeNormal)
{
  KStandardDirs *dirs   = KGlobal::dirs();
  QPixmap        pixmap;
  QString        path;

  // Setup the UI as first thing
  setupUi( this );

  // Just for show
  path = dirs->findResource( "data", "kmess/pics/unknown.png" );
  pixmap.load( path );
  contactPixmapLabel_->setPixmap( pixmap );
  contactPixmapLabel_->setMinimumSize( pixmap.width(), pixmap.height() );
  nowListeningPixmapLabel_->setPixmap( KIconLoader::global()->loadIcon("knotify", KIconLoader::NoGroup, KIconLoader::SizeSmall ) );
  nowListeningPixmapLabel_->hide();

  contactPixmapLabel_   ->installEventFilter( this );
  statusLabel_          ->installEventFilter( this );
  personalMessageLabel_ ->installEventFilter( this );

  installEventFilter( this );

  // Set up and connect the typing timer
  typingTimer_.setSingleShot( true );
  connect( &typingTimer_,  SIGNAL(     timeout() ),
           this,           SLOT  (  stopTyping() ) );
}



ContactFrame::~ContactFrame()
{
}



// Activate the frame by giving it a contact
void ContactFrame::activate( ContactBase *contact )
{
  if( contact == 0 )
  {
#ifdef KMESSDEBUG_CONTACTFRAME
    kDebug() << "Deactivating frame for contact:" << handle_;
#endif

    disconnect( contact_, 0 );
    disconnect( detailedContact_, 0 );

    contact_         = 0;
    detailedContact_ = 0;
    activated_       = false;

    return;
  }

  if ( contact_ != 0 )
  {
    kWarning() << "Already got a contact (" << contact_->getHandle() << "), refusing new one (" << contact->getHandle() << ").";
    return;
  }

  contact_ = contact;
  handle_  = contact_->getHandle();
  detailedContact_ = CurrentAccount::instance()->getContactList()->getContactByHandle( handle_ );

  // Connect up the contact to update the frame on name, status, and picture changes
  connect( contact_, SIGNAL(    changedFriendlyName()         ),
           this,     SLOT  (    updateStatusWidgets()         ) );
  connect( contact_, SIGNAL(          changedStatus()         ),
           this,     SLOT  (    updateStatusWidgets()         ) );
  connect( contact_, SIGNAL(         changedPicture()         ),
           this,     SLOT  (          updatePicture()         ) );
  connect( contact_, SIGNAL(              destroyed(QObject*) ),
           this,     SLOT  (       contactDestroyed(QObject*) ) );

  if( detailedContact_ != 0 )
  {
    connect( detailedContact_, SIGNAL( changedPersonalMessage(Contact*) ),
             this,             SLOT  (    updateStatusWidgets()         ) );
    connect( detailedContact_, SIGNAL(            changedList(Contact*) ),
             this,             SLOT  (    updateStatusWidgets()         ) );
  }

  activated_ = true;
  show();

  // Activate the right-click context menu
  initContactPopup();

  // And prepare the widgets
  updatePicture();
  updateStatusWidgets();
}



// Allow this contact to see our MSN status
void ContactFrame::allowContact()
{
  if( detailedContact_ == 0 )
  {
    return;
  }

  emit setContactAllowed( handle_ );

  updateStatusWidgets();
}



// The contact was destroyed
void ContactFrame::contactDestroyed( QObject *object )
{
  // This method is relevant for invited contacts, the following happens:
  // - an InvitedContact leaves the chat, SB notifies
  //   - ContactFrame::setDisplayMode() will be called, and updates the widgets.
  //   - CurrentAccount deletes the InvitedContact object.
  // - somehow a code paths leads to this class (Chat::setSwitchboardConnection(0) is already patched).
  //   - crash.

#ifdef KMESSDEBUG_CONTACTFRAME
  if( contact_ != object )
  {
    kWarning() << "Received unexpected contact reference to" << object->metaObject()->className();
  }
  else
  {
    kDebug() << "Resetting contact reference.";
  }
#endif

  activated_ = false;
  contact_ = 0;
  detailedContact_ = 0;
}



// Copy some details of the contact to the clipboard.
// Only one method is used to copy all different details, and that's to avoid having three identical methods which only copy different
// text bits on the clipboard.
void ContactFrame::copyText()
{
  // Get the signal sender to find out who called us
  KAction *action = static_cast<KAction*>( const_cast<QObject*>( sender() ) );
  if(KMESS_NULL(action)) return;

  if( action == popupCopyFriendlyName_ )
  {
    kapp->clipboard()->setText( contact_->getFriendlyName() );
  }
  else if( action == popupCopyHandle_ )
  {
    kapp->clipboard()->setText( handle_ );
  }
  else if( action == popupCopyPersonalMessage_ && detailedContact_ != 0 )
  {
    kapp->clipboard()->setText( detailedContact_->getPersonalMessage() );
  }
  else if( detailedContact_ != 0 )
  {
    // Copy the link from PM or Friendly Name
    kapp->clipboard()->setText( action->toolTip() );
  }
}



void ContactFrame::editNotes()
{
  // Check if there are details for contact
  if( detailedContact_ != 0 && detailedContact_->getExtension() != 0 )
  {
    // Create the contact properties dialog
    if( contactPropertiesDialog_ == 0 )
    {
      contactPropertiesDialog_ = new ContactPropertiesDialog( this );
    }

    contactPropertiesDialog_->launch( detailedContact_, true );
  }
}



// The personal status message received an event.
bool ContactFrame::eventFilter( QObject *obj, QEvent *event )
{
  if( event->type() != QEvent::MouseButtonRelease )
  {
    return false;  // don't stop processing.
  }

  QMouseEvent *mouseEvent = static_cast<QMouseEvent*>( event );

  if( mouseEvent->button() == Qt::RightButton )
  {
    showContactPopup( mouseEvent->globalPos() );
  }
  else if( obj == contactPixmapLabel_ )
  {
    showProfile();
  }

  return false;  // don't stop processing.
}



// Return the handle of this frame's contact
QString ContactFrame::getHandle() const
{
  return handle_;
}



// Initialize the contact popup
bool ContactFrame::initContactPopup()
{
  if( contact_ == 0 )
  {
    return false;
  }

  // Initialize context popup actions
  popupEmailContact_        = new KAction( KIcon("mail-message-new"),         i18n("&Send email"),       this );
  popupMsnProfile_          = new KAction( KIcon("preferences-desktop-user"), i18n("&View Profile"),     this );
  popupEditNotes_           = new KAction( KIcon("user-properties"),          i18n("Ed&it Notes"),       this );

  popupAddContact_          = new KAction( KIcon("list-add"),         i18n("&Add Contact"),      this );
  popupAllowContact_        = new KAction( KIcon("dialog-ok-apply"),  i18n("A&llow Contact"),    this );
  popupRemoveContact_       = new KAction( KIcon("list-remove-user"), i18n("&Delete Contact"),   this );

  popupBlockContact_        = new KAction( KIcon("dialog-cancel"),    i18n("&Block Contact"),    this );
  popupUnblockContact_      = new KAction( KIcon("dialog-ok"),        i18n("&Unblock Contact"),  this );

  popupCopyFriendlyName_    = new KAction(                            i18n("&Friendly Name"),    this );
  popupCopyPersonalMessage_ = new KAction(                            i18n("&Personal Message"), this );
  popupCopyHandle_          = new KAction(                            i18n("&Email Address"),    this );


  // Connect the actions
  connect( popupEmailContact_,        SIGNAL(triggered(bool)),   this,  SLOT( sendEmail()             ) );
  connect( popupMsnProfile_,          SIGNAL(triggered(bool)),   this,  SLOT( showProfile()           ) );
  connect( popupEditNotes_,           SIGNAL(triggered(bool)),   this,  SLOT( editNotes()             ) );

  connect( popupAddContact_,          SIGNAL(triggered(bool)),   this,  SLOT( toggleContactAdded()    ) );
  connect( popupAllowContact_,        SIGNAL(triggered(bool)),   this,  SLOT( allowContact()          ) );
  connect( popupRemoveContact_,       SIGNAL(triggered(bool)),   this,  SLOT( toggleContactAdded()    ) );

  connect( popupBlockContact_,        SIGNAL(triggered(bool)),   this,  SLOT( toggleContactBlocked()  ) );
  connect( popupUnblockContact_,      SIGNAL(triggered(bool)),   this,  SLOT( toggleContactBlocked()  ) );

  connect( popupCopyFriendlyName_,    SIGNAL(triggered(bool)),   this,  SLOT( copyText()              ) );
  connect( popupCopyPersonalMessage_, SIGNAL(triggered(bool)),   this,  SLOT( copyText()              ) );
  connect( popupCopyHandle_,          SIGNAL(triggered(bool)),   this,  SLOT( copyText()              ) );


  // Initialize sub popups
  popupCopyMenu_ = new KActionMenu( i18n("&Copy..."), this );
  popupCopyMenu_ ->addAction( popupCopyFriendlyName_ );
  popupCopyMenu_ ->addAction( popupCopyPersonalMessage_ );
  popupCopyMenu_ ->addAction( popupCopyHandle_ );
  popupCopyMenu_->addSeparator();

  // If detailed contact information is available, search for links
  if( detailedContact_ != 0 )
  {
    // Append friendly name and personal message so cycle only one time to search links
    QString nameAndPm = detailedContact_->getFriendlyName() + " " + detailedContact_->getPersonalMessage();

    //Initialize index, link RegExp and found boolean for insert separator only at first found link
    int pos = 0;
    QRegExp linkRegExp( "(http://|https://|ftp://|sftp://|www\\..)\\S+" );

    while( ( pos = linkRegExp.indexIn( nameAndPm, pos ) ) != -1 )
    {
      // Grep the found link and update the position for the next cycle
      QString foundLink = linkRegExp.cap( 0 );
      pos += linkRegExp.matchedLength();

      // Put the found link into KAction, replace & with && to avoid shortcut problem and use tooltip
      // for future grep of link ( to avoid shortcut problem too )
      popupCopyLink_ = new KAction( foundLink.replace( "&", "&&" ), this );
      popupCopyLink_->setToolTip( foundLink );
      connect( popupCopyLink_, SIGNAL( triggered( bool ) ), this, SLOT( copyText() ) );

      // Add action to copy menu
      popupCopyMenu_->addAction( popupCopyLink_ );
    }
  }

  // Initialize the popup popup
  contactActionPopup_ = new KMenu( handle_, this );

  // Attach the actions to the menu
  contactActionPopup_->addTitle( contact_->getHandle() );
  contactActionPopup_->addAction( popupEmailContact_ );
  contactActionPopup_->addAction( popupMsnProfile_ );
  contactActionPopup_->addAction( popupCopyMenu_ );
  contactActionPopup_->addAction( popupEditNotes_ );
  contactActionPopup_->addSeparator();
  contactActionPopup_->addAction( popupAddContact_ );
  contactActionPopup_->addAction( popupAllowContact_ );
  contactActionPopup_->addAction( popupRemoveContact_ );
  contactActionPopup_->addSeparator();
  contactActionPopup_->addAction( popupBlockContact_ );
  contactActionPopup_->addAction( popupUnblockContact_ );

  return true;
}



// Whether or not the frame has been activated
bool ContactFrame::isActivated() const
{
  return activated_;
}



// The user received a message from this contact
void ContactFrame::messageReceived()
{
  // Stop the "user is typing" glow
  typingTimer_.stop();
  stopTyping();
}



// Email the contact
void ContactFrame::sendEmail()
{
  if ( contact_ != 0 )
  {
    CurrentAccount::instance()->openMailAtCompose( handle_ );
  }
}



// Change the display mode of the frame when the contact leaves the chat or there are many active contacts
void ContactFrame::setDisplayMode( DisplayMode mode )
{
  bool isEnabled = ( mode & ModeDisabled ) ? false : true;

  // Reset the widgets only if enabling or disabling have been requested
  if( ( mode & ModeEnabled ) || ( mode & ModeDisabled ) )
  {
    // Enable or disable the frame contents
    contactPixmapLabel_      ->setEnabled( isEnabled );
  }

#ifdef KMESSDEBUG_CONTACTFRAME
  kDebug() << "Frame" << this << "- Enabled:" << isEnabled << "- Mode:" << mode;
#endif

  // Change the displayed widgets
  if( ( mode & ModeNormal ) || ( mode & ModeSingle ) )
  {
    contactPixmapLabel_      ->setVisible( ( detailedContact_ != 0 ) ); // Hide the picture for invited contacts
    statusPixmapLabel_       ->setVisible( true  );
    statusLabel_             ->setVisible( true  );
    personalMessageLabel_    ->setVisible( true  );
    nowListeningPixmapLabel_ ->setVisible( true  );
  }
  else if( mode & ModeSmall )
  {
    contactPixmapLabel_      ->setVisible( false );
    statusPixmapLabel_       ->setVisible( true  );
    statusLabel_             ->setVisible( true  );
    personalMessageLabel_    ->setVisible( true  );
    nowListeningPixmapLabel_ ->setVisible( true  );
  }
  else if( mode & ModeTiny )
  {
    contactPixmapLabel_      ->setVisible( false );
    statusPixmapLabel_       ->setVisible( true  );
    statusLabel_             ->setVisible( true  );
    personalMessageLabel_    ->setVisible( false );
    nowListeningPixmapLabel_ ->setVisible( false );
  }

  // Set the current mode internally, so contact property updates don't change the way we display the content
  currentMode_ = mode;

  // Update the status widgets
  updateStatusWidgets();
}



// Update and show the contact popup menu
void ContactFrame::showContactPopup( const QPoint &point )
{
  bool isAdded     = false;
  bool isAddable   = false;
  bool isAllowed   = true;
  bool isBlocked   = false;
  bool isBlockable = false;
  bool hasPersonalMessage = false;

  if( detailedContact_ != 0 )
  {
    isAddable   = true;
    isBlockable = true;

    isAdded   = detailedContact_->isFriend();
    isAllowed = detailedContact_->isAllowed();
    isBlocked = detailedContact_->isBlocked();
    hasPersonalMessage = ! detailedContact_->getPersonalMessage().isEmpty();
  }

  popupAllowContact_       ->setVisible( ! isAllowed );
  popupCopyPersonalMessage_->setVisible( hasPersonalMessage );

  popupAddContact_         ->setVisible( isAddable   && ! isAdded   );
  popupRemoveContact_      ->setVisible( isAddable   &&   isAdded   );

  popupBlockContact_       ->setVisible( isBlockable && ! isBlocked );
  popupUnblockContact_     ->setVisible( isBlockable &&   isBlocked );

  contactActionPopup_->popup( point );
}



// Show the contact's profile
void ContactFrame::showProfile()
{
  if( contact_ == 0 )
  {
    return;
  }

  // Create a URL to the msn profile page, localized with our system's locale.
  KUrl url = KUrl( "http://members.msn.com/default.msnw?mem=" + handle_ + "&mkt=" + KGlobal::locale()->country() );

  // Launch the browser for the given URL
  KMessShared::openBrowser( url );
}



// Receive notice that the contact is typing
void ContactFrame::startTyping()
{
  contactPixmapLabel_->setPixmap( contactTypingPicture_ );

  // Make the contact picture glow
  // Start the timer to disable the label
  typingTimer_.start( GLOWING_PICTURE_TIMEOUT );
}



// Disable the typing label when the timer has timed out
void ContactFrame::stopTyping()
{
  contactPixmapLabel_->setPixmap( contactPicture_ );
}



// Add or remove the contact from the contact list
void ContactFrame::toggleContactAdded()
{
  if( detailedContact_ == 0 )
  {
    return;
  }

  emit setContactAdded( handle_, ! detailedContact_->isFriend() );

  updateStatusWidgets();
}



// Change the contact's blocked/unblocked status
void ContactFrame::toggleContactBlocked()
{
  if( detailedContact_ == 0 )
  {
    return;
  }

  emit setContactBlocked( handle_, ! detailedContact_->isBlocked() );

  updateStatusWidgets();
}



// Update the contact picture
void ContactFrame::updatePicture()
{
  if ( contact_ == 0 )
  {
    return;
  }

  // See if the contact picture is not updated. Avoids GUI flashing on older boxes
  QString contactPicturePath = contact_->getContactPicturePath();
  QFileInfo contactPictureInfo( contactPicturePath );
  QDateTime contactPictureDate = contactPictureInfo.lastModified();

  if( contactPicturePath == contactPicturePath_
  &&  contactPictureDate == contactPictureDate_ )
  {
#ifdef KMESSDEBUG_CONTACTFRAME
    kDebug() << "Picture was already loaded.";
#endif
    return;
  }

#ifdef KMESSDEBUG_CONTACTFRAME
  kDebug() << "New contact picture:" << contact_->getContactPicturePath();
#endif

  QImage contactImage;
  bool loaded = contactImage.load( contactPicturePath );
  if( loaded )
  {
    // Loaded
    contactPicturePath_ = contactPicturePath;
    contactPictureDate_ = contactPictureDate;
  }
  else
  {
    // Not laoded (PNG error?). load default if custom image could not be loaded
    loaded = contactImage.load( contact_->getContactDefaultPicturePath() );

    if( ! loaded )
    {
      return;
    }
  }

  // Setup the normal and typing pixmaps
  contactPicture_ = QPixmap::fromImage( contactImage );
  KIconEffect::toGamma( contactImage, 0.8f );
  contactTypingPicture_ = QPixmap::fromImage( contactImage );

  // Display the image
  contactPixmapLabel_->setPixmap( contactPicture_ );
  contactPixmapLabel_->setMinimumSize( contactImage.width(), contactImage.height() );
}



// Update the status widgets
void ContactFrame::updateStatusWidgets()
{
  QString label, statusIdentifier;
  Status  status;
  Flags   flags = FlagNone;

  if( contact_ == 0 )
  {
    return;
  }

  status = contact_->getStatus();

  if( detailedContact_ != 0 && detailedContact_->isBlocked())
  {
    // Show blocked regardless of status
    statusIdentifier = i18n( "Blocked" );
    flags = FlagBlocked;
  }
  else
  {
    statusIdentifier = MsnStatus::getName( status );
  }

  bool showFormatting = CurrentAccount::instance()->getUseChatFormatting();

  if( showFormatting )
  {
    statusLabel_->setText( contact_->getFriendlyName( STRING_FORMATTED  ) );
  }
  else
  {
    statusLabel_->setText( Qt::escape( contact_->getFriendlyName( STRING_CLEANED ) ) );
  }

  statusLabel_       ->setToolTip( contact_->getFriendlyName() );
  statusPixmapLabel_ ->setToolTip( i18n( "The contact is %1", statusIdentifier ) );
  statusPixmapLabel_ ->setPixmap( MsnStatus::getIcon( status, flags ) );

  // Update the detailed info widgets, if available, and if we're not in Tiny display mode, which means that no
  // personal message/song is displayed at all
  if( detailedContact_ == 0 || currentMode_ == ModeTiny )
  {
#ifdef KMESSDEBUG_CONTACTFRAME
    kDebug() << "Not updating widgets. Mode:" << currentMode_ << " Detailed info:" << detailedContact_;
#endif

    personalMessageLabel_   ->hide();
    nowListeningPixmapLabel_->hide();
    return;
  }

  // Update the label with currently playing media
  label = detailedContact_->getCurrentMediaString();

  // If the contact has no media playing, show the personal message
  if( label.isEmpty() )
  {
    nowListeningPixmapLabel_->hide();
    if( showFormatting )
    {
      label = detailedContact_->getPersonalMessage( STRING_FORMATTED );
    }
    else
    {
      label = Qt::escape( detailedContact_->getPersonalMessage( STRING_CLEANED ) );
    }
  }
  else
  {
    nowListeningPixmapLabel_->show();
  }

  // If neither the media nor the message are available, hide the label
  if( ! label.isEmpty() )
  {

    personalMessageLabel_->show();
    personalMessageLabel_->setText( label );
    personalMessageLabel_->setToolTip( label );
  }
  else
  {
    personalMessageLabel_->hide();
  }
}



#include "contactframe.moc"
