/***************************************************************************
                          contactsidebar.cpp  -  description
                             -------------------
    begin                : Thu Jan 16 2003
    copyright            : (C) 2003 by Mike K. Bennett
    email                : mkb137b@hotmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "contactsidebar.h"

#include "../contact/contactbase.h"
#include "../currentaccount.h"
#include "../kmessdebug.h"

#include <QDragEnterEvent>
#include <QDragMoveEvent>
#include <QDropEvent>
#include <QLayout>
#include <QObject>
#include <QVBoxLayout>



// The constructor
ContactSidebar::ContactSidebar( QWidget *parent )
 : QWidget( parent )
{
  // Stretch widget
  setSizePolicy( QSizePolicy::MinimumExpanding, QSizePolicy::MinimumExpanding );

  // Create a layout within the scrollview so that added widgets are laid out nicely
  viewBox_ = new QVBoxLayout( this );

  // Allow drag&drop
  setAcceptDrops( true );

  // Add some more space at the bottom
  viewBox_->addStretch();
}



// The destructor
ContactSidebar::~ContactSidebar()
{
  // Remove all contact frames
  qDeleteAll( contactFrames_ );
  contactFrames_.clear();
}



// A contact joined the chat
void ContactSidebar::contactJoined( QString handle, QString )
{
kDebug()<< handle;
  ContactBase *contact;
  ContactFrame *contactFrame;
  ContactFrame::DisplayMode frameMode;

  // See if the contact already has a frame
  contactFrame = getContactFrameByHandle( handle );
  if( contactFrame != 0 )
  {
    // Reactivate the contact's frame
    contactFrame->setDisplayMode( ContactFrame::ModeEnabled );
    return;
  }

  // Get the first available contact frame
  contactFrame = getFirstInactiveContactFrame();
  if( KMESS_NULL(contactFrame) ) return;

  // Get the given contact
  contact = CurrentAccount::instance()->getContactByHandle( handle );

  if( KMESS_NULL(contact) ) return;

  // Activate the frame
  contactFrame->activate( contact );

  // Find the most suitable frame display mode
  frameMode = getBestFrameMode();

  // Set it to all the frames, including the new one
  foreach( ContactFrame *frame, contactFrames_ )
  {
    frame->setDisplayMode( frameMode );
  }
}



// A contact left the chat
void ContactSidebar::contactLeft( QString handle, bool /*isChatIdle*/ )
{
kDebug()<< handle;
  ContactFrame::DisplayMode frameMode;
  ContactFrame *contactFrame = getContactFrameByHandle( handle );

  if ( contactFrame == 0 )
  {
    return;
  }

  // Find the most suitable frame mode
  frameMode = getBestFrameMode();

  // Set it to all the frames
  foreach( ContactFrame *frame, contactFrames_ )
  {
    frame->setDisplayMode( frameMode );
  }

  // Delete the frames whenever there is more than one; when only one is left, grey it out instead.
  if( contactFrames_.count() > 1 )
  {
    contactFrames_.removeAll( contactFrame );
    delete contactFrame;
    return;
  }

  // Deactivate the contact frame
  contactFrame->setDisplayMode( ContactFrame::ModeDisabled );
}



// A contact is typing
void ContactSidebar::contactTyping(QString handle, QString)
{
  ContactFrame *contactFrame = getContactFrameByHandle( handle );
  if ( contactFrame != 0 )
  {
    contactFrame->startTyping();
  }
}



// Get the most suitable frame size, depending on how many contacts are currently in chat with us
ContactFrame::DisplayMode ContactSidebar::getBestFrameMode()
{
  // Find out how many are chatting at the moment
  int count = contactFrames_.count();


  if( count > 5 )
  {
    // More than 5 contacts in chat with us: it's a big group chat and we need to be able to display as many frames as possible
    return ContactFrame::ModeTiny;
  }
  else if( count > 2 )
  {
    // A normal group chat: reduce the size of the frames to be able to see all of our contacts at once
    return ContactFrame::ModeSmall;
  }
  else if( count == 2 )
  {
    // A small group chat: Show all details, there's still a lot of space
    return ContactFrame::ModeNormal;
  }

  // A simple 1-on-1 chat: we can show as much detail as possible
  return ContactFrame::ModeSingle;
}



// Find the contact frame with the given handle
ContactFrame* ContactSidebar::getContactFrameByHandle(const QString& handle)
{
  foreach( ContactFrame *contactFrame, contactFrames_ )
  {
    if( contactFrame->getHandle() == handle )
    {
      return contactFrame;
    }
  }

  return 0;
}



// Return the first inactive contact frame
ContactFrame* ContactSidebar::getFirstInactiveContactFrame()
{
  ContactFrame *contactFrame;
  
  // If there's an inactive frame, use it
  foreach( contactFrame, contactFrames_ )
  {
    if( ! contactFrame->isActivated() )
    {
      return contactFrame;
    }
  }

  // Well we don't have any, so make another.
  contactFrame = new ContactFrame( this );

  connect( contactFrame,   SIGNAL( setContactAllowed( QString )  ),
           this,           SIGNAL( setContactAllowed( QString )  ) );
  connect( contactFrame,   SIGNAL(   setContactAdded( QString, bool )  ),
           this,           SIGNAL(   setContactAdded( QString, bool )  ) );
  connect( contactFrame,   SIGNAL( setContactBlocked( QString, bool )  ),
           this,           SIGNAL( setContactBlocked( QString, bool )  ) );

  // Resize it, and add it to the viewBox so it appears in the sidebar
  contactFrame->setSizePolicy( QSizePolicy::Ignored, QSizePolicy::Preferred );
  viewBox_->insertWidget( children().count() - 2, contactFrame );

  // put it in the list of frames so we can find it again.
  contactFrames_.append( contactFrame );

  return contactFrame;
}



// Initialize the object
bool ContactSidebar::initialize()
{
  // The important frame-creating code has to go in the constructor, so there's less to initialize, here.

  return true;
}



// A message was received from one of the contacts... notify its frame
void ContactSidebar::messageReceived(const QString& handle)
{
  ContactFrame *frame;
  frame = getContactFrameByHandle( handle );
  if ( frame != 0 )
  {
    frame->messageReceived();
  }
}



// Enable/disable the frames
void ContactSidebar::setEnabled( bool isEnabled )
{
  ContactBase *contact;
  CurrentAccount *account = CurrentAccount::instance();

  if( isEnabled )
  {
    // Loop through all frames
    foreach( ContactFrame *frame, contactFrames_ )
    {
      contact = account->getContactByHandle( frame->getHandle() );

      // Only reactivate frames for contacts which exist in the account
      if( contact )
      {
        // Reactivate the frame
        frame->activate( contact );
      }
    }
  }
  else
  {
    // Loop through all frames
    foreach( ContactFrame *frame, contactFrames_ )
    {
      // Remove the activation status
      frame->activate( 0 );
    }
  }
}


#include "contactsidebar.moc"
