/***************************************************************************
                          emoticonsidebar.cpp -  description
                             -------------------
    begin                : Mon 12 09 2005
    copyright            : (C) 2005 by Diederik van der Boor
    email                : "vdboor" --at-- "codingdomain.com"
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "emoticonsidebar.h"

#include "../emoticonmanager.h"
#include "../emoticontheme.h"
#include "../kmessdebug.h"
#include "../dialogs/addemoticondialog.h"
#include "../accountsmanager.h"
#include "../currentaccount.h"

#include <QVBoxLayout>
#include <QLabel>
#include <QListWidgetItem>
#include <QMouseEvent>
#include <QPoint>

#include <KAction>
#include <KMenu>
#include <KLocale>


// The constructor
EmoticonSidebar::EmoticonSidebar( QWidget *parent, bool useCustomTheme )
 : QWidget(parent)
 , emoticonDialog_(0)
 , isCustom_(useCustomTheme)
 , list_(0)
{
  // Get and connect to the emoticon manager. We'll need to update the sidebar when the emoticon
  // theme changes
  manager_ = EmoticonManager::instance();
  connect( manager_, SIGNAL(          updated() ),
           this,     SLOT  ( slotThemeUpdated() ) );

  // Create and set the properties for list widget
  list_ = new EmoticonListWidget( this );

  // Connect to click of each button
  connect( list_ , SIGNAL( leftClicked    ( const QListWidgetItem* ) ),
           this,   SLOT  ( slotItemClicked( const QListWidgetItem* ) ) );

  // Enable the right-click context menu
  connect( list_, SIGNAL( rightClicked   (const QListWidgetItem*,const QPoint& ) ),
           this,  SLOT  ( showContextMenu(const QListWidgetItem*,const QPoint& ) ) );

  // Create and set the properties for the warning message
  noEmoticonsWarning_ = new QLabel( this );
  noEmoticonsWarning_->setScaledContents( true );
  noEmoticonsWarning_->setWordWrap( true );
  noEmoticonsWarning_->setText( i18nc( "Informative label on the chat's Sidebar",
                                       "<html><b>No custom emoticons have been added yet!</b><br/>"
                                       "They are like normal emoticons, but <i>you</i> choose the ones you want.<br/>"
                                       "To open the custom emoticons settings, <a href='#'>click here</a>.</html>" ) );
  // Enable the link click action
  connect( noEmoticonsWarning_, SIGNAL(              linkActivated(const QString&) ),
           this,                SLOT  ( slotNoEmoticonsLinkClicked(const QString&) ) );

  // Create the layout
  layout_ = new QVBoxLayout( this );
  layout_->addWidget( noEmoticonsWarning_ );
  layout_->addWidget( list_ );

  // Force adding the theme's emoticons
  slotThemeUpdated();
}



// Constructor for emoticon list widget class
EmoticonListWidget::EmoticonListWidget( QWidget *parent )
  : QListWidget(parent)
{
  // Set the list properties
  setViewMode( QListView::IconMode );
  setResizeMode( QListView::Adjust );
  setIconSize( QSize( 16,16 ) );
  setMovement( QListView::Static );
  setSelectionMode( QAbstractItemView::SingleSelection );
  setHorizontalScrollBarPolicy( Qt::ScrollBarAlwaysOff );
  setFrameShape(QFrame::NoFrame );
  setStyleSheet( "QListWidget { background-color: transparent; }" );
}



// Reimplement the mouse press event to grep left and right click
void EmoticonListWidget::mousePressEvent( QMouseEvent *e )
{
  QListWidget::mousePressEvent( e );
  QListWidgetItem *item = itemAt( e->pos() );
  if( item == 0 )
  {
    return;
  }
  if( e->button() == Qt::LeftButton )
  {
    emit leftClicked( item );
  }
  else if( e->button() == Qt::RightButton )
  {
    emit rightClicked( item, e->pos() );
  }
}



// The destructor
EmoticonSidebar::~EmoticonSidebar()
{
  delete emoticonDialog_;
  delete list_;
  delete layout_;
}



// One item was clicked
void EmoticonSidebar::slotItemClicked( const QListWidgetItem * item )
{
  // Deselect current item so it has the click effect
  list_->setCurrentItem( 0 );

  emit insertEmoticon( item->data( Qt::UserRole ).toString() );
}



// The "Click to open the custom emoticons settings" link has been clicked
void EmoticonSidebar::slotNoEmoticonsLinkClicked( const QString &url )
{
  Q_UNUSED( url );

  AccountsManager::instance()->showAccountSettings( CurrentAccount::instance(), window(), AccountSettingsDialog::PageEmoticonsTabCustom );
}



// The theme has changed, we need to reload it
void EmoticonSidebar::slotThemeUpdated()
{
#ifdef KMESSDEBUG_EMOTICONSIDEBAR
  kDebug() << "Removing old theme from the sidebar.";
#endif

  // Clear list
  list_->clear();

#ifdef KMESSDEBUG_EMOTICONSIDEBAR
  kDebug() << "Updating sidebar with current emoticons.";
#endif

  // Add all theme emoticons to the list
  EmoticonTheme *theme = manager_->getTheme( isCustom_ );
  const QList<Emoticon*> &emoticons = theme->getEmoticons();

  list_              ->setVisible( ! emoticons.isEmpty() );
  noEmoticonsWarning_->setVisible(   emoticons.isEmpty() && isCustom_ );

  foreach( Emoticon *emoticon, emoticons )
  {
    if( ! emoticon->isValid() )
    {
      continue;
    }

    QListWidgetItem *item = new QListWidgetItem( QIcon( emoticon->getPicturePath() ), "", list_ );

    // Set the shortcut into data of current item widget
    item->setData( ShortcutRole, emoticon->getShortcut()    );
    item->setData( PictureRole,  emoticon->getPicturePath() );

    // Add the emoticon tooltip
    QString tooltip( emoticon->getTooltip() );
    Emoticon::htmlEscape( tooltip );
    item->setToolTip( QString( emoticon->getHtml(false) + " " + tooltip ) );
  }

#ifdef KMESSDEBUG_EMOTICONSIDEBAR
  kDebug() << "Updated total" << emoticons.count() << "emoticons";
#endif
}




// Right mouse button was clicked
void EmoticonSidebar::showContextMenu( const QListWidgetItem *item, const QPoint &point )
{
  // Deselect current item so it has the click effect
  list_->setCurrentItem( 0 );

  // The popup menu is only for custom emoticon
  if( isCustom_ == false )
  {
    return;
  }

  // Create the popup menu for custom emoticon
  KMenu menu;
  KAction addAction(    KIcon( "insert-image" ),  i18n( "Add to Chat" ), this );
  KAction insertAction( KIcon( "list-add" ),      i18n( "Insert new" ),  this );
  KAction editAction(   KIcon( "draw-freehand" ), i18n( "Edit"   ),      this );
  KAction removeAction( KIcon( "list-remove" ),   i18n( "Remove" ),      this );

  menu.addAction( &addAction    );
  menu.addAction( &insertAction );
  menu.addAction( &editAction   );
  menu.addAction( &removeAction );

  // Exec the menu popup and grep the action selected
  QAction *clickedAction = menu.exec( mapToGlobal( point ) );

  if( clickedAction == 0 )
  {
    // No action clicked
    return;
  }

  // Check which action was clicked
  if( clickedAction == &addAction )
  {
    slotItemClicked( item );
    return;
  }
  else if( clickedAction == &removeAction )
  {
    manager_->getTheme( isCustom_ )->removeEmoticon( item->data( ShortcutRole ).toString() );
    return;
  }

  if( emoticonDialog_ == 0 )
  {
    emoticonDialog_ = new AddEmoticonDialog( manager_->getTheme( true ), this );
  }

  if( clickedAction == &insertAction )
  {
    emoticonDialog_->preSelect( "", "" );
  }
  else if( clickedAction == &editAction )
  {
    emoticonDialog_->editEmoticon( item->data( PictureRole ).toString(), item->data( ShortcutRole ).toString() );
  }

  emoticonDialog_->show();
}


#include "emoticonsidebar.moc"
