/***************************************************************************
                          contactbase.cpp  -  description
                             -------------------
    begin                : Thu Jan 16 2003
    copyright            : (C) 2003 by Mike K. Bennett
    email                : mkb137b@hotmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "contactbase.h"

#include "../contact/msnobject.h"   // only used for detectClientAppName()!
#include "../network/applications/applicationlist.h"
#include "../emoticon.h"
#include "../kmessdebug.h"

#include <QFile>
#include <QImage>
#include <QList>

#include <KLocale>
#include <KStandardDirs>



// The constructor
ContactBase::ContactBase(QString handle, QString friendlyName, uint capabilities)
 : capabilities_(capabilities)
 , friendlyName_(friendlyName)
 , handle_(handle)
 , applicationList_(0)
{
}



// The destructor
ContactBase::~ContactBase()
{
  delete applicationList_;
}



// Add a custom emoticon definition
void ContactBase::addEmoticonDefinition(const QString &emoticonCode, const QString &msnObjectDataHash )
{
#ifdef KMESSDEBUG_CONTACTBASE
  kDebug() << "adding '" << emoticonCode << "' to pending replacements";
#endif

  // Escape the code so the HTML parser won't be fooled.
  QString shortcut = emoticonCode;
  Emoticon::htmlEscape( shortcut );

  // Add emoticon to pending list, until Msn object is received.
  pendingEmoticons_.insert( msnObjectDataHash, shortcut );

  // Escape the code to insert it safely into a regular expression
  QString escapedCode = QRegExp::escape( shortcut );

  // Update regexp pattern
  if( pendingRegExp_.isEmpty() )
  {
    pendingRegExp_ = QRegExp( escapedCode );
  }
  else
  {
    pendingRegExp_ = QRegExp( pendingRegExp_.pattern() + "|" + escapedCode );
  }

#ifdef KMESSTEST
  KMESS_ASSERT( pendingRegExp_.isValid() );
#endif
}



// Add a custom emoticon file
void ContactBase::addEmoticonFile(const QString &msnObjectDataHash, const QString &filename)
{
  // Check whether the mapping is available.
  if( ! pendingEmoticons_.contains(msnObjectDataHash) )
  {
    kWarning() << "Received an custom emoticon, but no emoticon code found!";
    return;
  }

  // Get the emoticon code from pending list, remove it
  QString emoticonCode = pendingEmoticons_[msnObjectDataHash];    // no QString& because record gets removed.
  pendingEmoticons_.remove(msnObjectDataHash);

#ifdef KMESSDEBUG_CONTACTBASE
  kDebug() << "linking '" << emoticonCode << "' to received file.";
#endif

  // Check if filename exists, attempt to read it.
  if( ! QFile::exists(filename) )
  {
    kWarning() << "Could not read emoticon file: file not found!";
    return;
  }
  QImage emoticonData(filename);
  if( emoticonData.isNull() )
  {
    kWarning() << "Could not read emoticon file!";
    return;
  }

#ifdef KMESSDEBUG_CONTACT_GENERAL
  kDebug() << "Setting image for '" << emoticonCode << " to " << filename;
#endif

  // Escape the code to insert it safely into a regular expression
  QString escapedCode = QRegExp::escape( emoticonCode );

  // Update regexp pattern
  if( emoticonRegExp_.isEmpty() )
  {
    emoticonRegExp_ = QRegExp( escapedCode );
  }
  else
  {
    emoticonRegExp_ = QRegExp( emoticonRegExp_.pattern() + "|" + escapedCode );
  }


  // Update pending regexp pattern.
  pendingRegExp_ = QRegExp(pendingRegExp_.pattern().remove( QRegExp("(^|\\|)" + escapedCode) ));

#ifdef KMESSTEST
  KMESS_ASSERT( emoticonRegExp_.isValid() );
  KMESS_ASSERT( pendingRegExp_.isValid() );
#endif


  // Get the image dimensions from the file
  int originalWidth = emoticonData.width();
  int originalHeight = emoticonData.height();

  // Resize the displayed image to match KMess' maximum emoticon size
  int width  = originalWidth; // qMin( EMOTICONS_MAX_SIZE, originalWidth ); // Require only a max height
  int height = qMin( EMOTICONS_MAX_SIZE, originalHeight );

  if( originalWidth > originalHeight )
  {
      height = ( width * originalHeight ) / originalWidth;
  }
  else
  {
      width = ( height * originalWidth ) / originalHeight;
  }

  // Add replacement.
  QString replacement = "<img src='" + filename
                      + "' alt='" + emoticonCode
                      + "' width='" + QString::number( width )
                      + "' height='" + QString::number( height )
                      + "' valign='middle'"
                      +  " class='customEmoticon' />";
  customEmoticons_.insert(emoticonCode, replacement);
}



// Register the contact as participant in the chat session
void ContactBase::addSwitchboardConnection( const MsnSwitchboardConnection *connection )
{
#ifdef KMESSDEBUG_CONTACTBASE
  kDebug() << "adding connection to list.";
#endif

  if( chatSessions_.contains(connection) )
  {
    kWarning() << "chat session is already added.";
    return;
  }

  chatSessions_.append( connection );
}



// Initialize an application list for the contact.
ApplicationList * ContactBase::createApplicationList()
{
  if( applicationList_ != 0 )
  {
    kWarning() << "application list already created for contact '" << handle_ << "'.";
  }
  else
  {
    applicationList_ = new ApplicationList( handle_ );

    connect( applicationList_, SIGNAL(        applicationsAborted(const QString &) ),
             this,             SLOT  ( slotApplicationListAborted()                ) );
  }

  return applicationList_;
}



// Set the application name based on the capabilities
void ContactBase::detectClientAppName( uint capabilities, const MsnObject *msnObject )
{
  // Use capabilities to give an indication what client we're dealing with.
  uint clientVersion = capabilities & 0xF0000000;

  // Find msn version
  float msnVersion = 0;
  switch( clientVersion )
  {
    case 0: break;
    case MSN_CAP_MSN60: msnVersion = 6.0f; break;
    case MSN_CAP_MSN61: msnVersion = 6.1f; break;
    case MSN_CAP_MSN62: msnVersion = 6.2f; break;
    case MSN_CAP_MSN70: msnVersion = 7.0f; break;  // default caps 0x4000C024: ink-gif, multi-packet, direct-im, winks
    case MSN_CAP_MSN75: msnVersion = 7.5f; break;
    case MSN_CAP_MSN80: msnVersion = 8.0f; break;
    case MSN_CAP_MSN81: msnVersion = 8.1f; break;  // default caps 0x765DC02C: ink-gif, (ink-isf), multi-packet, direct-im, winks, shared-search, voice-clips, secure-channel, sip, folder-sharing
    case MSN_CAP_MSN85: msnVersion = 8.5f; break;  // default caps 0x865DC02C: ink-gif, ink-isf, multi-packet, direct-im, winks, shared-search, voice-clips, secure-channel, sip, folder-sharing
    default: msnVersion = 999;
  }


  // Third party clients also include their name in the MsnObject,
  // since that's already available at the notification connection.
  QString pictureClientName;
  if( msnObject != 0 )
  {
    const QString &pictureLocation = msnObject->getLocation();

    if( ! pictureLocation.isEmpty()
    &&  pictureLocation != "0"   // WLM 8.1
    &&  ! pictureLocation.startsWith("TFR") )  // MSN / WLM
    {
      // Known variations:
      if( pictureLocation == "KMess.tmp" )
      {
        pictureClientName = "KMess";
      }
      else if( pictureLocation == "Mercury.tmp" )
      {
        pictureClientName = "Mercury";
      }
      else if( pictureLocation == "kopete.tmp" )
      {
        pictureClientName = "Kopete";
      }
      else if( pictureLocation == "EncartaIcon.png" )
      {
        // Encarta Instant Answers.
        // ignore, no special string for this one.
      }
      else
      {
#ifdef KMESSDEBUG_CONTACTBASE
        kDebug() << "NOTICE: Unknown client name in MsnObject location:" << pictureLocation;
#endif
      }
    }
  }

  if( ! pictureClientName.isNull() )
  {
    clientAppName_ = pictureClientName;
  }
  // Check capabilities for special software
  else if( capabilities & MSN_CAP_WIN_MOBILE )
  {
    clientAppName_ = i18n("Windows Mobile");
  }
  else if( capabilities & MSN_CAP_WEB_CLIENT )
  {
    clientAppName_ = i18n("Web Messenger");
  }
  else if( capabilities & MSN_CAP_MSO_CLIENT )
  {
    clientAppName_ = i18n("Office Communicator");
  }
  else if( capabilities & MSN_CAP_BOT )
  {
    clientAppName_ = i18n("Messenger Bot");
  }
  else
  {
    if( msnVersion == 0 )
    {
      // pre p2p clients. msn 4.x is locked out after the switch to SSL logins.
      clientAppName_ = i18n("MSN Messenger %1 compatible", QString("5.0"));
    }
    else if( msnVersion < 8.0 )
    {
      // pre wlm client
      clientAppName_ = i18n("MSN Messenger %1 compatible", QString::number( msnVersion ) );
    }
    else if( msnVersion < 999 )
    {
      // wlm client
      if( ( capabilities & 0x065DC024 ) == 0x065DC024 )   // is 0x765DC02C without ink-isf.
      {
        // Supports all default caps of 8.1, remove the "compatible" string. This truely is WLM.
        clientAppName_ = i18n("Windows Live Messenger %1", QString::number( msnVersion ) );
      }
      else
      {
        clientAppName_ = i18n("Windows Live Messenger %1 compatible", QString::number( msnVersion ) );
      }
    }
    else
    {
      // newer version of wlm 
      clientAppName_ = i18n("Windows Live Messenger");
    } 
  }

#ifdef KMESSDEBUG_CONTACTBASE
  if( isOnline() )
  {
    QStringList caps;
    if( capabilities & MSN_CAP_WIN_MOBILE     ) caps << "mobile-client";
    if( capabilities & MSN_CAP_MSO_CLIENT     ) caps << "office-client";
    if( capabilities & MSN_CAP_WEB_CLIENT     ) caps << "web-client";
    if( capabilities & MSN_CAP_BOT            ) caps << "bot-client";

    if( capabilities & MSN_CAP_INK_GIF        ) caps << "ink-gif";
    if( capabilities & MSN_CAP_INK_ISF        ) caps << "ink-isf";
    if( capabilities & MSN_CAP_MULTI_PACKET   ) caps << "multi-packet";
    if( capabilities & MSN_CAP_DIRECT_IM      ) caps << "direct-im";
    if( capabilities & MSN_CAP_WINKS          ) caps << "winks";
    if( capabilities & MSN_CAP_SEARCH         ) caps << "shared-search";
    if( capabilities & MSN_CAP_VOICE_IM       ) caps << "voice-clips";
    if( capabilities & MSN_CAP_SECURE_IM      ) caps << "secure-channel";
    if( capabilities & MSN_CAP_SIP_INVITE     ) caps << "sip";
    if( capabilities & MSN_CAP_FOLDER_SHARING ) caps << "folder-sharing";

    if( capabilities & MSN_CAP_VIDEO_CHAT     ) caps << "has-webcam";
    if( capabilities & MSN_CAP_MSN_MOBILE     ) caps << "has-mobile";
    if( capabilities & MSN_CAP_MSN_DIRECT     ) caps << "has-direct";
    if( capabilities & MSN_CAP_LIVE_SPACE     ) caps << "has-space";

    kDebug() << "contact" << getHandle() << "uses client:" << clientAppName_ << "msn version:" << msnVersion << "capabilities:" << capabilities << "(" << caps << ")";
  }
#endif
}



// Return the application list of the contact
ApplicationList * ContactBase::getApplicationList() const
{
  return applicationList_;
}



// The capabilities of the client
uint ContactBase::getCapabilities() const
{
  return capabilities_;
}



// Return a user readable client name string.
const QString & ContactBase::getClientAppName() const
{
  return clientAppName_;
}



// Return the third party client nane field
const QString & ContactBase::getClientName() const
{
  return clientName_;
}



// Return the default contact picture path
QString ContactBase::getContactDefaultPicturePath() const
{
  KStandardDirs *dirs   = KGlobal::dirs();
  return dirs->findResource( "data", "kmess/pics/unknown.png" );
}



// Return the custom emoticon code for an msn object
QString ContactBase::getEmoticonCode(const QString &msnObjectDataHash) const
{
  return pendingEmoticons_[ msnObjectDataHash ];
}



// Return the custom emoticon search pattern.
const QRegExp & ContactBase::getEmoticonPattern() const
{
  return emoticonRegExp_;
}



// Return the custom emoticon replacements.
const QMap<QString,QString>& ContactBase::getEmoticonReplacements() const
{
  return customEmoticons_;
}



// Return the contact's friendly name
QString ContactBase::getFriendlyName( FormattingMode mode ) const
{
  return friendlyName_.getString( mode );
}



// Return the contact's handle
QString ContactBase::getHandle() const
{
  return handle_;
}



// Return the search pattern for pending custom emoticons.
const QRegExp & ContactBase::getPendingEmoticonPattern() const
{
  return pendingRegExp_;
}



// Return the list of switchboard connections
const QList<const MsnSwitchboardConnection*> ContactBase::getSwitchboardConnections() const
{
  return chatSessions_;
}



// Returns whether the contact has an application list initialized.
bool ContactBase::hasApplicationList() const
{
  return (applicationList_ != 0);
}



// Returns whether the contact has capability
bool  ContactBase::hasCapability( MsnClientCapabilities flag ) const
{
  return ( capabilities_ & flag ) == flag;
}



// Returns whether the contact's client supports the given MSNP2P version
bool ContactBase::hasP2PSupport( MsnClientCapabilities minimalVersion ) const
{
#ifdef KMESSTEST
  KMESS_ASSERT( minimalVersion != 0 && (minimalVersion & 0x0fffffff) == 0 );  // don't pass some other capability.
#endif

  return capabilities_ >= (uint) minimalVersion;
}



// Return true if the contact is offline
bool ContactBase::isOffline() const
{
  return getStatus() == STATUS_OFFLINE;
}



// Return true if the contact is online
bool ContactBase::isOnline() const
{
  return getStatus() != STATUS_OFFLINE;
}



// Unregister the contact as participant, contact has left the chat.
void ContactBase::removeSwitchboardConnection( const MsnSwitchboardConnection *connection, bool userInitiated )
{
#ifdef KMESSDEBUG_CONTACTBASE
  kDebug() << "removing connection, userInitiated=" << userInitiated;
#endif

  bool found = chatSessions_.removeAll( connection ) > 0;
  if( ! found )
  {
    kWarning() << "chat session not found.";
  }

  if( chatSessions_.isEmpty() )
  {
#ifdef KMESSDEBUG_CONTACTBASE
    kDebug() << handle_ << " is no longer in any other chat conversation, cleaning up.";
#endif

    // Inform the ApplicationList if we have one.
    // It's possible some sessions need to be terminated now.
    if( applicationList_ != 0 )
    {
      bool hasAbortingApps = applicationList_->contactLeftChat( userInitiated );

      // If there aren't any applications aborting, the object can be deleted now.
      // Also avoid deleting the object if there is still a direct connection.
      // It's possible a switchboard is replaced with another one (with WLM -> background chat -> normal chat).
      // Note the object could be destroyed from slotApplicationListAborted() too.
      if( ! hasAbortingApps && applicationList_ != 0 && ! applicationList_->hasAuthorizedDirectConnection() )
      {
#ifdef KMESSDEBUG_CONTACTBASE
        kDebug() << "no applications to abort, deleting ApplicationList object.";
#endif

        delete applicationList_;
        applicationList_ = 0;
      }
    }

    // This will be used by CurrentAccount to clean up temporary InvitedContact objects.
    emit leftAllChats( this );
  }
}



// Set the client capabilities
void ContactBase::setCapabilities( uint capabilities )
{
  // See if the client software could have changed.
  // Enable flags which could be toggled by the client.
  bool clientChanged = ( capabilities_ | MSN_CAP_VIDEO_CHAT | MSN_CAP_MSN_MOBILE | MSN_CAP_MSN_DIRECT | MSN_CAP_LIVE_SPACE ) !=
                       ( capabilities  | MSN_CAP_VIDEO_CHAT | MSN_CAP_MSN_MOBILE | MSN_CAP_MSN_DIRECT | MSN_CAP_LIVE_SPACE );

  capabilities_ = capabilities;

  // Reset the client name information if the capabilities changed.
  if( clientChanged )
  {
    clientAppName_.clear();
    clientName_.clear();
  }
}



// Restore the user readable client application name
void ContactBase::setClientAppName( const QString &clientAppName )
{
  clientAppName_ = clientAppName;
}



// Set the third part client name field
void ContactBase::setClientName( const QString& clientName )
{
  clientName_ = clientName;

  // Also give a more precise real name
  clientAppName_ = clientName_.replace("/", " ");

  // Weird 'Purple 2.3.1' string means the user actually has Pidgin or Adium.
  if( clientAppName_.startsWith("Purple ") )
  {
    clientAppName_ = "Pidgin/Adium (" + clientAppName_ + ")";
  }
}



// The application list completed aborting it's applications.
void ContactBase::slotApplicationListAborted()
{
  // Avoid killing an object if it was filled with new entries again
  if( ! applicationList_->isEmpty() || applicationList_->hasAuthorizedDirectConnection() )
  {
#ifdef KMESSDEBUG_CONTACTBASE
    kDebug() << "requested abortion completed, but ApplicationList object has object references: not deleting object.";
#endif
    return;
  }

#ifdef KMESSDEBUG_CONTACTBASE
  kDebug() << "requested abortion completed, deleting ApplicationList object.";
#endif


  // Use deleteLater() because this is called from the signal loop.
  applicationList_->deleteLater();
  applicationList_ = 0;
}



#include "contactbase.moc"
