/***************************************************************************
                          contactextension.cpp  -  description
                             -------------------
    begin                : Fri 13 Dec 2002
    copyright            : (C) 2003 by Michael Curtis.
    email                : mdcurtis@users.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "contactextension.h"

#include "../currentaccount.h"
#include "../kmessdebug.h"
#include "contactlist.h"

#include <QDir>
#include <QFile>

#include <KLocale>
#include <KGlobal>
#include <KConfigGroup>



// The constructor with specified contact handle
ContactExtension::ContactExtension( Contact *contact )
 : QObject(),
   alternativeName_(),
   alternativePictureFile_(""),
   notifyOnline_(true),
   pictureFile_(""),
   soundFile_(""),
   useAlternativeName_(false)
{
  // Set the current contact and connect the contactOffline event for lastSeen entry
  handle_ = contact->getHandle();
  contact_ = contact;
  connect( contact_ , SIGNAL( contactOffline( Contact*, bool ) ), this, SLOT( setLastSeen() ) );
}


// The destructor
ContactExtension::~ContactExtension()
{

}



// Clear the cache
void ContactExtension::clearCache()
{
  QString fileName;
  QFile   file;

  // Grep all pictures and remove these
  foreach( fileName,  pictureFileList_ )
  {
    // Remove the picture file
    file.setFileName( fileName );
    if( file.exists() )
    {
      file.remove();
      file.close();
    }

    // Remove the dat file
    file.setFileName( fileName  + ".dat" );
    if( file.exists() )
    {
      file.remove();
      file.close();
    }
  }
  pictureFileList_.clear();
}



// Read property of QString  alternativeName_.
const QString & ContactExtension::getAlternativeName( FormattingMode mode ) const
{
  return alternativeName_.getString( mode );
}



// Read property of QString alternativePictureFile_
const QString& ContactExtension::getContactAlternativePicturePath() const
{
  return alternativePictureFile_;
}



// Read property of QString  PictureFile_.
const QString ContactExtension::getContactPicturePath() const
{
  if(QFile::exists(pictureFile_))
  {
    return pictureFile_;
  }
  else
  {
    return QString::null;
  }
}



// Read property of QString  SoundFile_.
const QString ContactExtension::getContactSoundPath() const
{
  if(QFile::exists(soundFile_))
  {
    return soundFile_;
  }
  else
  {
    return QString::null;
  }
}



// Read property of QString LastSeen
const QString& ContactExtension::getLastSeen()
{
  // If the contact is connects
  if( contact_->isOnline() )
  {
    lastSeen_ = i18n( "Connected" );
  }

  if( lastSeen_.isEmpty() )
  {
    lastSeen_ = i18n( "Never seen" );
  }

  return lastSeen_;
}



// Read property of QString notes_
const QString& ContactExtension::getNotes() const
{
  return notes_;
}



// Read property of bool notifyOnline_
bool ContactExtension::getNotifyOnline() const
{
  return notifyOnline_;
}



// Read property of bool useAlternativeName_.
bool ContactExtension::getUseAlternativeName() const
{
  return useAlternativeName_;
}



// Read property of pictures list
QStringList ContactExtension::getPictureList() const
{
  return pictureFileList_;
}



// Load extension settings for a contact
void ContactExtension::readProperties( const KConfigGroup &config )
{
  // Get the configuration group where this contact's data is stored
  KConfigGroup extensionGroup = config.group( "Extension" );

  // When changing any of those default values, also change the saving condition
  // in saveProperties() !
  alternativeName_       = extensionGroup.readEntry("alternativeName",          ""            );
  alternativePictureFile_= extensionGroup.readEntry("alternativePictureFile",   ""            );
  lastSeen_              = extensionGroup.readEntry("lastSeen",                 ""            );
  notifyOnline_          = extensionGroup.readEntry("notifyOnline",             true          );
  notes_                 = extensionGroup.readEntry("notes",                    ""            );
  pictureFile_           = extensionGroup.readEntry("pictureFile",              ""            );
  pictureFileList_       = extensionGroup.readEntry("pictureFileList",          QStringList() );
  soundFile_             = extensionGroup.readEntry("soundFile",                ""            );

  // We use the accessor function here because it handles the update, if required
  setUseAlternativeName  ( extensionGroup.readEntry("useAlternativeName",       false       ) );
}



// Save extension settings for a contact
void ContactExtension::saveProperties( KConfigGroup &config )
{
  // Get the configuration group where this contact's data is stored
  KConfigGroup extensionGroup = config.group( "Extension" );

  // If no info have been changed, don't waste space writing it, and if it's present already, remove the group.
  if(  ! useAlternativeName_
      && notifyOnline_
      && alternativeName_.getOriginal().isEmpty()
      && alternativePictureFile_       .isEmpty()
      && lastSeen_                     .isEmpty()
      && pictureFile_                  .isEmpty()
      && soundFile_                    .isEmpty() )
  {
#ifdef KMESSDEBUG_CONTACTEXTENSION
    if( extensionGroup.hasGroup( "Extension" ) )
    {
      kDebug() << "Deleting unused contact info for" << handle_;
    }
    else
    {
      kDebug() << "Not saving unused contact info for" << handle_;
    }
#endif

    config.deleteGroup( "Extension" );
    return;
  }

  extensionGroup.writeEntry( "useAlternativeName",     useAlternativeName_            );
  extensionGroup.writeEntry( "alternativePictureFile", alternativePictureFile_        );
  extensionGroup.writeEntry( "alternativeName",        alternativeName_.getOriginal() );
  extensionGroup.writeEntry( "notifyOnline",           notifyOnline_                  );
  extensionGroup.writeEntry( "notes",                  notes_                         );

  // If the contact is connects, refresh the lastSeen before saves it
  if( contact_->isOnline() )
  {
    setLastSeen();
  }

  extensionGroup.writeEntry( "lastSeen",           lastSeen_           );
  extensionGroup.writeEntry( "pictureFile",        pictureFile_        );
  extensionGroup.writeEntry( "pictureFileList",    pictureFileList_    );
  extensionGroup.writeEntry( "soundFile",          soundFile_          );
}



// Write property of QString alternativeName_.
void ContactExtension::setAlternativeName( const QString& newAlternativeName )
{
  // Ignore invalid requests
  if( newAlternativeName.isEmpty() || newAlternativeName == alternativeName_.getOriginal() )
  {
    return;
  }

  // Copy the new name to the three versions
  alternativeName_.setString( newAlternativeName );

  if( getUseAlternativeName() )
  {
    emit changedFriendlyName();
  }
}



// Write property of QString alternativePictureFile_
void ContactExtension::setContactAlternativePicturePath( const QString& _newVal )
{
  alternativePictureFile_ = _newVal;
}



// Write property of QString  pictureFile_.
void ContactExtension::setContactPicturePath( const QString & _newVal )
{
  pictureFile_ = _newVal;

  if( ! _newVal.isEmpty()  )
  {
    // If the current picture isn't in the list of all used pictures,
    // then add it.
    if( ! pictureFileList_.contains( _newVal ) )
    {
      pictureFileList_.append( _newVal );
    }

    // If the picture file hasn't a dat file, then create it
    // and put into the current date
    QFile fileDat( _newVal + ".dat" );
    if( ! fileDat.exists () )
    {
      if ( fileDat.open( QIODevice::WriteOnly | QIODevice::Text ) )
      {
        QTextStream out( &fileDat );
        out << QDate::currentDate ().toString( Qt::SystemLocaleDate );
        fileDat.close();
      }
    }
  }

  emit changedPicture();
}



// Write property of QString  soundFile_.
void ContactExtension::setContactSoundPath( const QString & _newVal)
{
  soundFile_ = _newVal;
}



// Write property of QString LastSeen
void ContactExtension::setLastSeen()
{
 QString time = QDateTime::currentDateTime().toString();
 lastSeen_ = time;
}



// Write property of QString notes
void ContactExtension::setNotes( const QString& _newVal )
{
  notes_ = _newVal;
}



// Write propety of bool notifyOnline_
void ContactExtension::setNotifyOnline( bool _newVal )
{
  notifyOnline_ = _newVal;
}



// Write property of bool useAlternativeName_
void ContactExtension::setUseAlternativeName( bool _newVal)
{
  bool needUpdate;

  needUpdate = ( useAlternativeName_ != _newVal);

  useAlternativeName_ = _newVal;

  if ( needUpdate == true ) emit changedFriendlyName();
}



#include "contactextension.moc"
