/***************************************************************************
                          contactlist.cpp  -  description
                             -------------------
    begin                : Sun Jan 5 2003
    copyright            : (C) 2003 by Mike K. Bennett
                           (C) 2005 by Diederik van der Boor
    email                : mkb137b@hotmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "contactlist.h"

#include "../contact/msnobject.h"
#include "../network/specialgroups.h"
#include "../utils/kmessconfig.h"
#include "../currentaccount.h"
#include "../kmessdebug.h"
#include "contact.h"
#include "contactextension.h"
#include "group.h"

#include <QFile>

#include <KLocale>


#ifdef KMESSDEBUG_CONTACTLIST
  #define KMESSDEBUG_CONTACTLIST_GENERAL
  #define KMESSDEBUG_CONTACTLIST_TOOLTIP
  #define KMESSDEBUG_CONTACTLIST_DISCONNECTION
  #define KMESSDEBUG_CONTACTLIST_MODEL
#endif



// The constructor
ContactList::ContactList()
: rootNode_(0)
{
}



// The destructor
ContactList::~ContactList()
{
  // Destroy the list's contents, without regenerating the special groups and the model
  reset( false );

#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
  kDebug() << "DESTROYED.";
#endif
}



// Add a contact to the contact list
Contact * ContactList::addContact( QString handle, QString friendlyName, int lists, QStringList groupIds, QString guid )
{
  // Ignore call if we are closing...
  // TODO: prevent this issue from happening in the first place.
  if( KMESS_NULL(rootNode_) ) return 0;

#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
  kDebug() << "Adding to list contact" << handle << " in groups" << groupIds;
#endif
#ifdef KMESSTEST
  KMESS_ASSERT( ! handle.isEmpty() );
#endif

  // More consistency checks
  if(getContactByHandle(handle) != 0)
  {
    kWarning() << "attempted to add contact " << handle << " twice.";
    return 0;
  }

  if(handle.isEmpty())
  {
    return 0;
  }


  // Create the contact
  Contact *contact = new Contact( handle, friendlyName, lists, groupIds, guid );
  contacts_[ handle ] = contact;

  // Add the contact to the special groups
  if( contact->isFriend() )
  {
    // Initially, all friends are Offline (we receive the status for connected ones later)
    groupIds << SpecialGroups::OFFLINE;
  }
  else
  {
    if( contact->isAllowed() )
    {
      groupIds << SpecialGroups::ALLOWED;
    }
    else
    {
      groupIds << SpecialGroups::REMOVED;
    }
  }


  // Update the model
  // Add it to the nodes it belongs to
  foreach( QString id, groupIds )
  {
    ContactListModelItem *group = rootNode_->childGroup( id );
    if( ! group )
    {
      kWarning() << "Contact's group" << id << "doesn't exist!";
      continue;
    }

    int row = group->childCount();

    beginInsertRows( index( group->row(), 0, QModelIndex() ), row, row );
    new ContactListModelItem( contact, group );
    endInsertRows();
  }

  // If the contact hasn't been added to any node - other than the Offline or Online one - add it to the root node
  if( contact->isFriend() && groupIds.count() == 1 )
  {
    int row = rootNode_->childCount();

    beginInsertRows( QModelIndex(), row, row );
    new ContactListModelItem( contact, rootNode_ );
    endInsertRows();
  }


  // Attach the signals
  connect( contact, SIGNAL(                   changedMsnObject(Contact*)      ),   // Contact changed msnobject
           this,    SIGNAL(            contactChangedMsnObject(Contact*)      ));
  connect( contact, SIGNAL(                     contactOffline(Contact*,bool) ),   // Contact went offline
           this,    SLOT  (          slotForwardContactOffline(Contact*,bool) ));
  connect( contact, SIGNAL(                      contactOnline(Contact*,bool) ),   // Contact went online
           this,    SLOT  (           slotForwardContactOnline(Contact*,bool) ));
  connect( contact, SIGNAL(                       changedGroup(Contact*)      ),   // Contact moved
           this,    SLOT  (                   slotContactMoved(Contact*)      ));
  connect( contact, SIGNAL(                        changedList(Contact*)      ),   // Contact was added/removed from certain lists
           this,    SLOT  (                   slotContactMoved(Contact*)      ));

  return contact;
}



// Add a group to the contact list
Group * ContactList::addGroup( QString groupId, QString groupName )
{
  // Ignore call if we are closing...
  // TODO: prevent this issue from happening in the first place.
  if( KMESS_NULL(rootNode_) ) return 0;

#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
  kDebug() << groupName << "(" << groupId << ")";
#endif

  // Check if the group already exists
  if ( groupExists( groupId ) )
  {
    kDebug() << "Group " << groupId << " already exists.";
    return 0;
  }

  // Create the group
  Group *group = new Group( groupId, groupName );
  groups_.append( group );

  if( group->isSpecialGroup() )
  {
    group->setSortPosition( 9999 );
  }
  else
  {
    group->setSortPosition( groups_.count() - 1 );
    connect( group, SIGNAL( changedPosition( Group*, int ) ),
             this,    SLOT(  slotGroupMoved( Group*, int ) ) );
  }

  // Update the model.
  int row = rootNode_->childCount();

  beginInsertRows( QModelIndex(), row, row );
  new ContactListModelItem( group, rootNode_ );
  endInsertRows();


  // Signal the UI
  emit groupAdded(group);

  return group;
}



// Change a contact's status (and update the friendly name)
void ContactList::changeContactStatus( QString handle, Status status, QString friendlyName,
                                       uint capabilities, QString msnObjectString, bool showBaloon )
{
  // Ignore call if we are closing...
  // TODO: prevent this issue from happening in the first place.
  if( KMESS_NULL(rootNode_) ) return;

  // Get the contact
  Contact *contact;
  bool     wasOnline;
  Status   oldStatus;

  contact = getContactByHandle( handle );
  if ( contact == 0 )
  {
    kWarning() << "changeContactStatus - Couldn't find contact " << handle << " in the contact list.";
    return;
  }

  // Store the old online state to detect a change
  wasOnline = contact->isOnline();
  oldStatus = contact->getStatus();

  // Update the settings
  if( ! friendlyName.isEmpty() )
  {
    contact->setFriendlyName( friendlyName );
  }
  contact->setStatus( status, showBaloon );
  contact->loadMsnObject( msnObjectString );
  contact->setCapabilities( capabilities );   // also resets clientName if caps changed.

  // Try to detect the client name when the contact was offline before (e.g. connected/reconnected with a new client).
  if( ! wasOnline )
  {
    if( contact->getClientName().isEmpty() )
    {
      const MsnObject *msnObject = contact->getMsnObject();  // only in derived Contact class, not in ContactBase
      contact->detectClientAppName( capabilities, msnObject );
    }
  }

  // Signal that the contact has changed status
  if( oldStatus != status )
  {
    emit contactChangedStatus( contact, showBaloon );
    forwardDataChanged( rootNode_->childContacts( handle ) );
  }

  // Switch the contact between the Offline and Online groups as needed
  if( wasOnline != contact->isOnline() )
  {
    // Get a reference to both groups
    ContactListModelItem *onlineGroup  = rootNode_->childGroup( SpecialGroups::ONLINE  );
    ContactListModelItem *offlineGroup = rootNode_->childGroup( SpecialGroups::OFFLINE );
    if( ! onlineGroup || ! offlineGroup )
    {
      kWarning() << "The Online or Offline groups are not available!";
      return;
    }

    // Then find the contact in both groups and move it
    // NOTE The childContacts() calls in the two foreach loops should return only one item,
    //      because we're searching for a contact into a specific group

    if( contact->isOnline() )
    {
      ModelItemList offlineList( offlineGroup->childContacts( handle ) );
      foreach( ContactListModelItem *item, offlineList )
      {
        item->setParent( onlineGroup );
        forwardDataChanged( item );
      }
    }
    else
    {
      ModelItemList onlineList( onlineGroup->childContacts( handle ) );
      foreach( ContactListModelItem *item, onlineList )
      {
        item->setParent( offlineGroup );
        forwardDataChanged( item );
      }
    }
  }

}



/**
 * Get the number of columns present in an index
 *
 * @param parent The index to count columns of
 * @return int
 */
int ContactList::columnCount( const QModelIndex &parent ) const
{
     if (! parent.isValid())
     {
         return rootNode_->columnCount();
     }

  return static_cast<ContactListModelItem*>( parent.internalPointer() )->columnCount();
}



// Return whether or not the contact with the given handle exists
bool ContactList::contactExists( QString handle )
{
  return getContactByHandle( handle ) != 0;
}



/**
 * Return the data contained by an index in the model
 *
 * @param index  Points to the actual data to be returned
 * @param role   Indicates what data from the index needs to be obtained.
 * @see http://doc.trolltech.com/4.3/qt.html#ItemDataRole-enum for available roles
 */
QVariant ContactList::data( const QModelIndex &index, int role ) const
{
  // If the index isn't valid or the caller wants a non-displayable role, do nothing
  if( ! index.isValid() || role != Qt::DisplayRole )
    return QVariant();

  ContactListModelItem *item = static_cast<ContactListModelItem*>( index.internalPointer() );

  if( ! item )
  {
    return QVariant();
  }

  return item->data( index.column() );
}



/**
 * Dumps the contact list contents to the debug output
 *
 * This method is temporary and for debugging purposes exclusively.
 * This method is recursive. Do not call with any arguments, they are for internal use only.
 *
 * @param start Item in the model where to start from
 * @param depth Current search depth
 */
void ContactList::dump( ContactListModelItem *start, int depth )
{
#ifdef KMESSDEBUG_CONTACTLIST_MODEL
  bool started = false;
  if( ! start )
  {
    start = rootNode_;
    kDebug() << "Contact List model stats";
    kDebug() << "--------------------------------------";
    kDebug() << "Columns:"            << columnCount( createIndex( 0, 0, rootNode_ ) );
    kDebug() << "Rows:"               << rowCount( createIndex( 0, 0, rootNode_ ) );
    kDebug() << "Root group Columns:" << rootNode_->columnCount();
    kDebug() << "Root group Rows:"    << rootNode_->childCount();
    kDebug();
    kDebug() << "Contact List model dump";
    kDebug() << "--------------------------------------";
    started = true;
  }

  QString spacer = QString().fill( '-', 2 * depth ) + ">";

  // Print the current node's contents, and manage children recursively
  for( int i = 0; i < start->childCount(); i++ )
  {
    ContactListModelItem* item = start->child( i );
    int type = item->getType();

    switch( type )
    {
      case ContactListModelItem::ItemContact:
        kDebug() << spacer << item->data( 1 ).toInt() << ") Contact:" << item->data( 0 ).toMap()["handle"].toString();
        break;

      case ContactListModelItem::ItemGroup:
        kDebug() << spacer << item->data( 1 ).toInt() << ") Group:" << item->data( 0 ).toMap()["id"].toString();
        dump( item, depth + 1 );
        break;

      default:
        kDebug() << spacer << "Unknown type of item: " + type;
        break;
    }
  }

  if( started )
  {
    kDebug() << "--------------------------------------";
  }
#else
  Q_UNUSED( start );
  Q_UNUSED( depth );
#endif
}



/**
 * Return the flags which characterize the index
 *
 * @param index  Points to the data to get flags of
 * @return The index's flags
 */
Qt::ItemFlags ContactList::flags( const QModelIndex &index ) const
{
  if( ! index.isValid() )
    return 0;

  return Qt::ItemIsEnabled | Qt::ItemIsSelectable;
}



/**
 * Emits the Data Changed signal on one item
 *
 * This is only an overloaded version for the list forwardDataChanged() method
 *
 * @param item The item which changed
 */
void ContactList::forwardDataChanged( ContactListModelItem *item )
{
  forwardDataChanged( ( ModelItemList() << item ) );
}



/**
 * Emits the Data Changed signal on some items
 *
 * @param itemList  The list of items which changed
 */
void ContactList::forwardDataChanged( ModelItemList itemList )
{
  if( itemList.isEmpty() )
  {
#ifdef KMESSDEBUG_CONTACTLIST_MODEL
    kDebug() << "Called with an empty list!";
#endif
    return;
  }

  foreach( ContactListModelItem* item, itemList )
  {
    // Get an index for the parent of this item
    QModelIndex parent( createIndex( item->parent()->row(), 0, item->parent() ) );

    // Get an index for the changed item itself and signal that it's changed
    QModelIndex updatedIndex( index( item->row(), 0, parent ) );
    emit dataChanged( updatedIndex, updatedIndex );
  }
}



// Return the contact with the given guid
Contact* ContactList::getContactByGuid( QString guid ) const
{
  // No need to optimize, only called for ADC en REM now.
  foreach( Contact *contact, contacts_ )
  {
    if(contact->getGuid() == guid)
    {
      return contact;
    }
  }

  return 0;
}



// Return the contact with the given handle
Contact* ContactList::getContactByHandle( QString handle ) const
{
  // Optimized with hashing because it's called a lot
  return contacts_.value( handle, 0 );
}



// Return the list of contacts
const QHash<QString,Contact*>& ContactList::getContactList() const
{
  return contacts_;
}



// Return the group with the given id
Group* ContactList::getGroupById( QString groupId ) const
{
  foreach( Group *group, groups_ )
  {
    if( group->getId() == groupId )
    {
      return group;
    }
  }

  return 0;
}



// Return the group with the given sort position
Group* ContactList::getGroupBySortPosition( int sortPosition, bool next ) const
{
  if( next )
  {
    foreach( Group *group, groups_ )
    {
      if ( group->getSortPosition() >= sortPosition && ! group->isSpecialGroup() )
      {
        return group;
      }
    }
  }
  else
  {
    Group *group;
    for( int i = groups_.count() - 1; i >= 0; i-- )
    {
      group = groups_.at( i );
      if ( group->getSortPosition() <= sortPosition && ! group->isSpecialGroup() )
      {
        return group;
      }
    }
  }

  return 0;
}



// Return the list of groups
const QList<Group*>& ContactList::getGroupList() const
{
  return groups_;
}



// Return whether or not a group with the given id exists
bool ContactList::groupExists( QString groupId )
{
  return getGroupById( groupId ) != 0;
}



/**
 * Return the text for a list or tree column header.
 *
 * @see http://doc.trolltech.com/4.3/qt.html#ItemDataRole-enum for available roles
 *
 * @param section Row or column of the header text to return
 * @param orientation Whether the headers are horizontal or vertical
 * @param role   Indicates what type of text needs to be obtained.
 * @return A QVariant version of the text
 */
QVariant ContactList::headerData( int section, Qt::Orientation orientation, int role ) const
{
  if( rootNode_ && orientation == Qt::Horizontal && role == Qt::DisplayRole )
    return rootNode_->data( section );

  return QVariant();
}



/**
 * Obtain a model index for a specifically located item
 *
 * @param row     Row of the item
 * @param column  Column of the item
 * @param parent  Parent of the model which contains the row and column. If invalid, the root group is used.
 * @return        An index for the specified item, or an invalid index if it hasn't been found
 */
QModelIndex ContactList::index( int row, int column, const QModelIndex &parent ) const
{
  if( ! hasIndex( row, column, parent ) )
  {
    return QModelIndex();
  }

  ContactListModelItem *parentItem;

  if( ! parent.isValid() )
  {
    parentItem = rootNode_;
  }
  else
  {
    parentItem = static_cast<ContactListModelItem*>( parent.internalPointer() );
  }

  ContactListModelItem *childItem = parentItem->child( row );
  if( childItem )
  {
    return createIndex( row, column, childItem );
  }
  else
  {
    return QModelIndex();
  }
}



/**
 * Obtain a model index's parent index
 *
 * @param index  Points to the item to find the parent of
 * @return       An index for the parent of the specified item, or an invalid index if it hasn't been found
 */
QModelIndex ContactList::parent( const QModelIndex &index ) const
{
  if( ! index.isValid() )
  {
    return QModelIndex();
  }

  ContactListModelItem *item = static_cast<ContactListModelItem*>( index.internalPointer() );

  if( ! item )
  {
    return QModelIndex();
  }

  ContactListModelItem *parentItem = item->parent();

  if( ! parentItem || parentItem == rootNode_ )
  {
    return QModelIndex();
  }

  return createIndex( parentItem->row(), 0, parentItem );
}



/**
 * Apply to the model a contact's movement through lists and groups
 *
 * This method is called when a contact changes between two lists, for example from blocked to allowed;
 * and when changes group, for example when it is moved or copied from a group to another.
 * First it identifies lists and groups where the contact was and will be; then the changes are applied.
 * The contact's list items are then iteratively moved from the old to the new lists/groups, and any
 * remaining addition and removal is taken care of.
 *
 * @param contact  The contact to update in the model
 */
void ContactList::slotContactMoved( Contact *contact )
{
  // The oldItems will be copied into the newNodes
  ModelItemList oldItems;
  ModelItemList newNodes;
  QStringList groupIds = contact->getGroupIds();

  // Add the groups to the list of new groups of which the contact is member
  if( groupIds.isEmpty() )
  {
    newNodes << rootNode_;
  }
  foreach( QString groupId, groupIds )
  {
    ContactListModelItem *item = rootNode_->childGroup( groupId );
    if( item )
    {
      newNodes << item;
    }
  }

  // Add the lists to the list of new groups of which the contact is member
  if( contact->isFriend() )
  {
    if( contact->isOnline() )
    {
      newNodes << rootNode_->childGroup( SpecialGroups::ONLINE );
    }
    else
    {
      newNodes << rootNode_->childGroup( SpecialGroups::OFFLINE );
    }
  }
  else
  {
    if( contact->isAllowed() )
    {
      newNodes << rootNode_->childGroup( SpecialGroups::ALLOWED );
    }
    else
    {
      newNodes << rootNode_->childGroup( SpecialGroups::REMOVED );
    }
  }

  // Loop through the nodes which contain this contact, avoid moving the contact if it's still in the same group/list
  ModelItemList itemsList( rootNode_->childContacts( contact->getHandle() ) );
  foreach( ContactListModelItem *oldNode, itemsList )
  {
    if( newNodes.contains( oldNode->parent() ) )
    {
      // It was and is still in this node, don't move it
      newNodes.removeAll( oldNode->parent() );
    }
    else
    {
      // It's not in this node anymore, we can move this node elsewhere
      oldItems << oldNode;
    }
  }

  // The contact is still in the same nodes, avoid any moving around
  if( oldItems.isEmpty() && newNodes.isEmpty() )
  {
#ifdef KMESSDEBUG_CONTACTLIST_MODEL
    kDebug() << "Not moving anything, the contact's nodes are unchanged.";
#endif
    // Just update the contact's nodes
    forwardDataChanged( rootNode_->childContacts( contact->getHandle() ) );
    return;
  }

#ifdef KMESSDEBUG_CONTACTLIST_MODEL
  kDebug() << "Moving" << oldItems.count() << "items of contact" << contact->getHandle() << "to" << newNodes.count() << "nodes.";
#endif

  // Move the contact's items between changed nodes
  for( int idx = 0; idx < qMax( oldItems.count(), newNodes.count() ); ++idx )
  {
    bool hasOldNode = ( idx < oldItems.count() );
    bool hasNewNode = ( idx < newNodes.count() );

    int insertIntoPos, removeFromPos;
    QModelIndex oldParentIndex, newParentIndex;
    ContactListModelItem *item = 0, *oldNode = 0, *newNode = 0;

    // If there's an item to remove from a node, start a removal process
    if( hasOldNode )
    {
      item           = oldItems.at( idx );
      oldNode        = item->parent();
      removeFromPos  = item->row();
      oldParentIndex = ( oldNode == rootNode_ ) ? QModelIndex() : index( oldNode->row(), 0, QModelIndex() );

      beginRemoveRows( oldParentIndex, removeFromPos, removeFromPos );
    }
    // If there's an item to add to a node, start an addition process
    if( hasNewNode )
    {
      newNode        = newNodes.at( idx );
      insertIntoPos  = newNode->childCount();
      newParentIndex = ( newNode == rootNode_ ) ? QModelIndex() : index( newNode->row(), 0, QModelIndex() );

      beginInsertRows( newParentIndex, insertIntoPos, insertIntoPos );
    }

    // Move the contact between two nodes
    if( hasOldNode && hasNewNode )
    {
      item->setParent( newNode );
    }
    // Remove an item now unneeded
    else if( hasOldNode )
    {
      delete item;
    }
    // Add a new item
    else if( hasNewNode )
    {
      new ContactListModelItem( contact, newNode );
    }

    // Complete the addition and/or deletion process
    if( hasNewNode )
    {
      endInsertRows();
    }
    if( hasOldNode )
    {
      endRemoveRows();
    }
  }

#ifdef KMESSDEBUG_CONTACTLIST_MODEL
  kDebug() << "Moving complete. sorting the list.";
#endif

  // A lot of items have possibly changed, sort the list again
  sort();
}



// Forward from a group that is has move
void ContactList::slotGroupMoved( Group *group, int newPosition )
{
  int oldPosition = groups_.indexOf( group );
  if( newPosition < groups_.count() && oldPosition != -1 )
  {
    groups_.swap( oldPosition, newPosition );
  }
}



// Forward from a contact that it is offline
void ContactList::slotForwardContactOffline(Contact *contact, bool showBaloon)
{
#ifdef KMESSTEST
  KMESS_ASSERT( contact != 0 );
#endif

#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
  if ( contact != 0 )
  {
    kDebug() << "Forwarding that " << contact->getHandle() << " went offline.";
  }
#endif

  emit contactOffline( contact, showBaloon );
}



// Forward from a contact that it is online
void ContactList::slotForwardContactOnline(Contact *contact, bool showBaloon)
{
#ifdef KMESSTEST
  KMESS_ASSERT( contact != 0 );
#endif

#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
  if ( contact != 0 )
  {
    kDebug() << "Forwarding that " << contact->getHandle() << " went online.";
  }
#endif

  emit contactOnline( contact, showBaloon );
}



// Remove a group
void ContactList::removeGroup( QString groupId )
{
  Group *group = getGroupById( groupId );
  ContactListModelItem *nodeToRemove = rootNode_->childGroup( groupId );

  if ( group == 0 || nodeToRemove == 0 )
  {
    kDebug() << "Group " << groupId << " doesn't exist in the contact list.";
    return;
  }

  if( nodeToRemove->childCount() )
  {
    kWarning() << "Group" << groupId << "was not empty!";
  }

  int row = nodeToRemove->row();

  // Remove the group
  beginRemoveRows( QModelIndex(), row, row );
  groups_.removeAll( group );
  endRemoveRows();

  emit groupRemoved( group ); // Signal the UI

  delete nodeToRemove;
  delete group;
}



// Change a contact's name
void ContactList::renameContact( QString handle, QString newName )
{
  Contact *contact;

  contact = getContactByHandle( handle );
  if ( contact == 0 )
  {
    kDebug() << "Couldn't find contact " << handle << " in the contact list.";
    return;
  }

  contact->setFriendlyName( newName );

  // If the contact exists in the list, signal that it's changed
  forwardDataChanged( rootNode_->childContacts( handle ) );
}



// Change a group's name
void ContactList::renameGroup( QString groupId, QString groupName )
{
  Group *group = getGroupById( groupId );
  if( group == 0 )
  {
    kWarning() << "Group '" << groupId << "' "
                << "not found in the contact list." << endl;
    return;
  }

  group->setName( groupName );

  // If the group exists in the list, signal that it's changed
  forwardDataChanged( rootNode_->childGroup( groupId ) );
}


/**
 * Empty the contact list
 *
 * Deletes the contact list's contents and regenerates the special groups and the model.
 * When exiting, the regeneration part can be skipped.
 */
void ContactList::reset( bool restore )
{
  if( contacts_.count() || groups_.count() )
  {
#ifdef KMESSDEBUG_CONTACTLIST_DISCONNECTION
    kDebug() << "Resetting contacts...";
#endif

    // Delete all contacts
    qDeleteAll( contacts_ );
    contacts_.clear();

#ifdef KMESSDEBUG_CONTACTLIST_DISCONNECTION
    kDebug() << "Resetting groups...";
#endif

    // Delete all groups
    qDeleteAll( groups_ );
    groups_.clear();
  }

  // Be quiet if we're not doing anything.
  if( rootNode_ == 0 && ! restore )
  {
    return;
  }

#ifdef KMESSDEBUG_CONTACTLIST_MODEL
  kDebug() << "Resetting model tree...";
#endif

  // Delete the root node
  delete rootNode_;
  rootNode_ = 0;

  // Notify views that the old model is gone
  QAbstractItemModel::reset();

  // Do not recreate the groups
  if( ! restore )
  {
    return;
  }

  // Regenerate the model's root node, the only one which contains groups and contacts
  rootNode_ = new ContactListModelItem();

  // Create the special groups
  addGroup( SpecialGroups::ONLINE,  i18n("Online")  );
  addGroup( SpecialGroups::OFFLINE, i18n("Offline") );
  addGroup( SpecialGroups::ALLOWED, i18n("Allowed") );
  addGroup( SpecialGroups::REMOVED, i18n("Removed") );

#ifdef KMESSTEST
  KMESS_ASSERT( groups_.count() == 4 );
#endif
}



// Read groups and contacts properties
void ContactList::readProperties()
{
#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
  kDebug() << "Loading properties for the contact list";
#endif

  KConfigGroup config;
  QString handle = CurrentAccount::instance()->getHandle();

  config = KMessConfig::instance()->getContactListConfig( handle, "Groups" );

  // Load all groups
  QList<Group *> sortedList( groups_ );
  Group *group;
  int newPosition;

  for( int i = 0; i < groups_.count(); i++ )
  {
    group = groups_.value( i );
    group->readProperties( config );

    // Signal that this group has changed, so it will be correctly repainted
    forwardDataChanged( rootNode_->childGroup( group->getId() ) );

    if( group->isSpecialGroup() )
    {
      continue;
    }

    newPosition = group->getSortPosition();

    if( newPosition >= groups_.count() )
    {
      newPosition = groups_.count() - 1;
    }

    sortedList.swap( i, newPosition );
  }

  groups_.clear();
  groups_ = sortedList;

  config.sync();
  config = KMessConfig::instance()->getContactListConfig( handle, "Contacts" );

  // Load all contact extensions
  foreach( Contact *contact, contacts_ )
  {
    contact->readProperties( config );
  }

  config.sync();
}



/**
 * Get the number of rows present in an index
 *
 * @param parent The index to count rows of
 * @return int
 */
int ContactList::rowCount( const QModelIndex &parent ) const
{
  ContactListModelItem *parentItem = 0;

  if( ! parent.isValid() )
  {
    parentItem = rootNode_;
  }
  else if( parent.column() > 0 )
  {
    return 0;
  }
  else
  {
    parentItem = static_cast<ContactListModelItem*>( parent.internalPointer() );
  }

  if( ! parentItem )
  {
    return 0;
  }

  return parentItem->childCount();
}



// Sort the model's items
void ContactList::sort( int column, Qt::SortOrder order )
{
  Q_UNUSED( column );

  // Signal the views that we're about to change the model someway
  emit layoutAboutToBeChanged();

  rootNode_->sort( order );

  // Complete the model update
  emit layoutChanged();
}



// Save groups and contacts properties
void ContactList::saveProperties()
{
#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
  kDebug() << "Saving properties for the contact list";
#endif

  // Do not save account properties when it's a temporary account
  if( CurrentAccount::instance()->isGuestAccount() )
  {
    kWarning() << "Requested to save properties, but this is a guest account!";
    return;
  }

  // Get the position of the contacts settings of this account
  KConfigGroup config;
  QString handle = CurrentAccount::instance()->getHandle();

  // Skip saving groups if there aren't any
  if( groups_.count() > 0 )
  {
    config = KMessConfig::instance()->getContactListConfig( handle, "Groups" );

    // Save all groups
    foreach( Group *group, groups_ )
    {
      // The position for contact is the same of the group in qlist
      group->setSortPosition( groups_.indexOf( group ) );
      group->saveProperties( config );
    }

    // Write the data now!
    config.sync();
  }
#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
  else
  {
    kDebug() << "No groups to save";
  }
#endif

  // Skip saving contacts if there aren't any, as above
  if( ! contacts_.isEmpty() )
  {
    config = KMessConfig::instance()->getContactListConfig( handle, "Contacts" );

    // We want to delete unused contact info; so here we get a list of all contacts and remove
    // the used ones. Then we remove what's left in this list
    QStringList pendingRemoval = config.groupList();

    // Save all contact extensions
    foreach( Contact *contact, contacts_ )
    {
      pendingRemoval.removeAll( contact->getHandle() );
      contact->saveProperties( config );
    }

    // Delete the remaining contacts info, which at this point are only old contacts not in list anymore
    QFile picture;
    KConfigGroup contactConfig;
    foreach( QString unusedContact, pendingRemoval )
    {
  #ifdef KMESSDEBUG_CONTACTLIST_GENERAL
      kDebug() << "Deleting unused contact info for " << unusedContact;
  #endif

      // Remove the custom picture associated with this contact, if there is any
      contactConfig = config.group( unusedContact );
      picture.setFileName( contactConfig.readEntry( "pictureFile", "" ) );
      if( picture.exists() )
      {
        picture.remove();
      }
      config.deleteGroup( unusedContact );
    }

    // Write the data now!
    config.sync();
  }
#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
  else
  {
  kDebug() << "No contacts to save";
  }
#endif
}



#include "contactlist.moc"
