/***************************************************************************
                          contactlist.h  -  description
                             -------------------
    begin                : Sun Jan 5 2003
    copyright            : (C) 2003 by Mike K. Bennett
                           (C) 2005 by Diederik van der Boor
    email                : mkb137b@hotmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef CONTACTLIST_H
#define CONTACTLIST_H

#include <QAbstractItemModel>
#include <QHash>
#include <QObject>

#include "contact/msnstatus.h"
#include "contactlistmodelitem.h"


// Forward declarations
class Contact;
class CurrentAccount;
class Group;
class KMessTest;



/**
 * @brief Data class for the contact list.
 *
 * This class stores the state of the current contact list, it does not update the server.
 * The MsnNotificationConnection class sends update requests to the server.
 * When the server acknowledges the change, it will be updated here.
 * The user interface classes repond to the signals of this class to update their view.
 *
 * The instance of this class is shared between the MsnNotificationConnection class,
 * and re-used when the user connects with a different account.
 * After the MsnNotificationConnection class is initialized, a pointer
 * to the ContactList is set in the CurrentAccount class too.
 * The CurrentAccount class is only used to provide access to the current contact list,
 * the administration of the contacts is still done through the MsnNotificationConnection class,
 *
 * @author Mike K. Bennett
 * @ingroup Contact
 */
class ContactList : public QAbstractItemModel
{
  Q_OBJECT

  friend class KMessTest;

  public: // Public methods
    // The constructor
                         ContactList();
    // The destructor
    virtual             ~ContactList();
    // Add a contact to the contact list
    Contact *            addContact( QString handle, QString friendlyName, int lists, QStringList groupIds, QString guid );
    // Add a group to the contact list
    Group *              addGroup( QString groupId, QString groupName );
    // Change a contact's status (and update the friendly name)
    void                 changeContactStatus( QString handle, Status status, QString friendlyName = QString(),
                                              uint capabilities = 0, QString msnObjectString = QString(), bool showBaloon = true );
    // Return the contact with the given guid
    Contact*             getContactByGuid( QString guid ) const;
    // Return the contact with the given handle
    Contact*             getContactByHandle( QString handle ) const;
    // Return the list of contacts
    const QHash<QString,Contact*>& getContactList() const;
    // Return the group with the given id
    Group*               getGroupById( QString groupId ) const;
    // Return the group with the given sort position
    Group*               getGroupBySortPosition( int sortPosition, bool next ) const;
    // Return the list of groups
    const QList<Group*>& getGroupList() const;
    // Read groups and contacts properties
    void                 readProperties();
    // Remove a group
    void                 removeGroup( QString groupId );
    // Change a contact's name
    void                 renameContact( QString handle, QString newName );
    // Change a group's name
    void                 renameGroup( QString groupId, QString newName );
    // Restore the orginal state of the contact list (empty it)
    void                 reset( bool restore = true );
    // Save groups and contacts properties
    void                 saveProperties();

  public:  // Public model methods
    // Get the number of columns present in an index
    int                  columnCount( const QModelIndex &parent = QModelIndex() ) const;
    // Return the data contained by an index in the model
    QVariant             data( const QModelIndex &index, int role ) const;
    // Dumps the contact list contents to the debug output
    void                 dump( ContactListModelItem *start = 0, int depth = 1 );
    // Return the flags which characterize the index
    Qt::ItemFlags        flags( const QModelIndex &index ) const;
    // Return the text for a list or tree column header.
    QVariant             headerData( int section, Qt::Orientation orientation, int role = Qt::DisplayRole ) const;
    // Obtain a model index for a specificly located item
    QModelIndex          index( int row, int column, const QModelIndex &parent = QModelIndex() ) const;
    // Obtain a model index's parent index
    QModelIndex          parent( const QModelIndex &index ) const;
    // Get the number of rows present in an index
    int                  rowCount( const QModelIndex &parent = QModelIndex() ) const;
    // Sort the model's items
    void                 sort( int column = 0, Qt::SortOrder order = Qt::AscendingOrder );

  private: // Private methods
    // Return whether or not the contact with the given handle exists
    bool                 contactExists( QString handle );
    // Emits the Data Changed signal on one item
    void                 forwardDataChanged( ContactListModelItem *item );
    // Emits the Data Changed signal on some items
    void                 forwardDataChanged( ModelItemList itemList );
    // Return whether or not a group with the given id exists
    bool                 groupExists( QString groupId );

  private slots: // Private slots
    // Forward from a contact that is has moved
    void                 slotContactMoved( Contact *contact );
    // Forward from a group that is has move
    void                 slotGroupMoved( Group *group, int newPosition );
    // Forward from a contact that it is offline
    void                 slotForwardContactOffline( Contact *contact, bool showBaloon );
    // Forward from a contact that it is contact is online
    void                 slotForwardContactOnline( Contact *contact, bool showBaloon );


  private: // Private attributes
    // The list of contacts
    QHash<QString,Contact*> contacts_;
    // The list of groups
    QList<Group*>         groups_;
    // The root node which makes up the contact list tree
    ContactListModelItem *rootNode_;

  signals: // Signals
    // Signal that a contact changed it's msnobject.
    void                 contactChangedMsnObject(Contact *contact);
    // Signal that a contact changed it's status.
    void                 contactChangedStatus( Contact *contact, bool showBaloon );
    // Signal that a contact is offline
    void                 contactOffline( Contact *contact, bool showBaloon );
    // Signal that a contact is online
    void                 contactOnline( Contact *contact, bool showBaloon );
    // Signal that a group has been added
    void                 groupAdded( const Group *group );
    // Signal that a group has been added
    void                 groupRemoved(const Group *group);
};

#endif
