/***************************************************************************
                          contactlistmodelitem.cpp  -  description
                             -------------------
    begin                : Sat Feb 16 2008
    copyright            : (C) 2008 by Valerio Pilo
    email                : valerio@kmess.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "contactlistmodelitem.h"

#include "../network/specialgroups.h"
#include "contact.h"
#include "group.h"
#include "kmessdebug.h"

#include <QtAlgorithms>



/**
 * Root constructor
 *
 * Creates a root node
 */
ContactListModelItem::ContactListModelItem()
 : parentNode_(0)
 , object_(0)
 , itemType_(ItemRoot)
{
}



/**
 * Contact constructor
 *
 * Creates a Contact node
 */
ContactListModelItem::ContactListModelItem( Contact *contact, ContactListModelItem *parent )
 : parentNode_(parent)
 , object_(contact)
 , itemType_(ItemContact)
{
  // Add this node as a child of the parent
  parent->appendChild( this );
}



/**
 * Group constructor
 *
 * Creates a Group node
 */
ContactListModelItem::ContactListModelItem( Group *group, ContactListModelItem *parent )
 : parentNode_(parent)
 , object_(group)
 , itemType_(ItemGroup)
{
  // Add this node as a child of the parent
  parent->appendChild( this );
}



/**
 * Destructor
 *
 * Since all the items contained by the model are directly managed by ContactList, their deletion
 * is performed by the Contact List itself.
 */
ContactListModelItem::~ContactListModelItem()
{
  qDeleteAll( childItems_ );
  childItems_.clear();

  // Delete this node from the parent's children
  if( parentNode_ )
  {
    parentNode_->removeChild( this );
  }
}



/**
 * Add a new child node
 *
 * Insert the given item as a child of this node, but only if it doesn't already exist
 *
 * @param item  Item to add as child item
 */
void ContactListModelItem::appendChild( ContactListModelItem *item )
{
  // Don't add groups as children of other groups which are not the root one
  if( itemType_ != ItemRoot && item->getType() == ItemGroup )
  {
    kWarning() << "Cannot insert a group as child of another group!";
    return;
  }

  // Don't add items which are already present
  if( ! childItems_.contains( item ) )
  {
    childItems_.append( item );
  }
}



/**
 * Return the row-th child of this item
 *
 * @param row  Requested row
 * @return Node at the requested position
 */
ContactListModelItem *ContactListModelItem::child( int row )
{
  if( row >= childItems_.count() )
  {
    return 0;
  }

  return childItems_.at( row );
}



/**
 * Return all Contact children of this node which are identified by an handle
 *
 * This method is made so that Contacts added in different Groups are all found (for example,
 * a Contact may be in the Online group and in an user-defined group too).
 * If called from the root node, also search for children contacts in all groups.
 *
 * @param handle  Handle of the contact to find
 * @return List of items for the requested handle
 */
ModelItemList ContactListModelItem::childContacts( QString handle )
{
  ModelItemList list;
  foreach( ContactListModelItem *item, childItems_ )
  {
    if(    item->getType() == ItemContact
        && static_cast<Contact*>( item->object() )->getHandle() == handle )
    {
        list << item;
    }
    else if( itemType_ == ItemRoot && item->getType() == ItemGroup )
    {
      ModelItemList subItems = item->childContacts( handle );
      if( ! subItems.isEmpty() )
      {
        list << subItems;
      }
    }
  }

  return list;
}



/**
 * Return a child Group identified by its ID
 *
 * @param id  Id of the requested group
 * @return Node of the group
 */
ContactListModelItem *ContactListModelItem::childGroup( QString id )
{
  foreach( ContactListModelItem *item, childItems_ )
  {
    if(    item->getType() == ItemGroup
        && static_cast<Group*>( item->object() )->getId() == id )
    {
      return item;
    }
  }

  return 0;
}



/**
 * Return the number of children of this item
 *
 * @return int
 */
int ContactListModelItem::childCount() const
{
  return childItems_.count();
}



/**
 * Return the number of data columns
 *
 * The first column (index 0) contains the displayed data.
 * The second column contains sort indexing information.
 * The third column contains a string to be used for quick comparison based by the displayed text.
 *
 * @return int
 */
int ContactListModelItem::columnCount() const
{
  return 3;
}



/**
 * Find if one item is lesser than another
 *
 * @param firstItem   First item to compare
 * @param secondItem  Other item to compare
 * @return bool
 */
bool ContactListModelItem::compareItems( const ContactListModelItem *firstItem, const ContactListModelItem *secondItem )
{
  if( ! firstItem || ! secondItem )
  {
    kWarning() << "Attempted comparison between invalid items!";
    return false;
  }

  // Save the items' values for a faster comparison
  int first  = firstItem ->data( 1 ).toInt();
  int second = secondItem->data( 1 ).toInt();

  // Sort alphabetically contacts having the same status
  if( first == second )
  {
    int comparison = QString::localeAwareCompare( firstItem->data( 2 ).toString(), secondItem->data( 2 ).toString() );
    return ( comparison < 0 );
  }

  // The contacts are in different states
  return ( first < second );
}



/**
 * Return a collection of data from this item
 *
 * Data on the first column is actual item object's contents; data on the second
 * column is a numeric index, used to sort the list contents; data on the third
 * one is a string used to quickly compare two contacts for sorting.
 *
 * @param column  Specify the kind of item data to obtain
 * @return QVariant containing int, QMap or QString for the three columns respectively.
 */
QVariant ContactListModelItem::data( int column ) const
{
  // Output the sorting information
  if( column == 1 )
  {
    int sortIndex = 9999;

    if( itemType_ == ItemContact )
    {
      Contact *contact = static_cast<Contact*>( object_ );

      // Contacts are sorted by their status
      switch( contact->getStatus() )
      {
        case STATUS_ONLINE:        sortIndex = 1; break;
        case STATUS_BE_RIGHT_BACK: sortIndex = 2; break;
        case STATUS_AWAY:          sortIndex = 3; break;
        case STATUS_IDLE:          sortIndex = 3; break;
        case STATUS_OUT_TO_LUNCH:  sortIndex = 4; break;
        case STATUS_ON_THE_PHONE:  sortIndex = 5; break;
        case STATUS_BUSY:          sortIndex = 6; break;

        case STATUS_INVISIBLE:
        case STATUS_OFFLINE:
        default:                   sortIndex = 8; break;
      }
    }
    else if( itemType_ == ItemGroup )
    {
      Group *group = static_cast<Group*>( object_ );

      // This number is set so that the groups always come after the contacts
      sortIndex = group->getSortPosition() + 9;
    }

    return sortIndex;
  }

  // Return a string to use for quick item comparison
  if( column == 2 )
  {
    if( itemType_ == ItemContact )
    {
      Contact *contact = static_cast<Contact*>( object_ );
      return contact->getFriendlyName();
    }
    else if( itemType_ == ItemGroup )
    {
      Group *group = static_cast<Group*>( object_ );

      // Sort groups how we want them
      QString groupPrefix;

      // The Online group must be the first
      if( group->getId() == SpecialGroups::ONLINE )
      {
        groupPrefix = "0";
      }
      // The Offline group must be the second
      else if( group->getId() == SpecialGroups::OFFLINE )
      {
        groupPrefix = "1";
      }
      // Then non-special groups
      else if( ! group->isSpecialGroup() )
      {
        groupPrefix = "2";
      }
      // Then the remaining special groups
      else
      {
        groupPrefix = "3";
      }

      return ( groupPrefix + "-" + group->getName() );
    }
  }

  // Column 1 and 2 are done, and we have 3 columns only
  if( column != 0 )
  {
    return QVariant();
  }

  ModelDataList itemData;

  // Fill the returned data map with a selection of useful contact information
  if( itemType_ == ItemContact )
  {
    Contact *contact = static_cast<Contact*>( object_ );

    itemData[ "type"                     ] = ItemContact;
    itemData[ "handle"                   ] = contact->getHandle();
    itemData[ "friendly"                 ] = contact->getFriendlyName( STRING_CLEANED );
    itemData[ "friendlyFormatted"        ] = contact->getFriendlyName( STRING_FORMATTED );
    itemData[ "trueFriendly"             ] = contact->getTrueFriendlyName( STRING_CLEANED );
    itemData[ "trueFriendlyFormatted"    ] = contact->getTrueFriendlyName( STRING_FORMATTED );
    itemData[ "personalMessage"          ] = contact->getPersonalMessage( STRING_CLEANED );
    itemData[ "personalMessageFormatted" ] = contact->getPersonalMessage( STRING_FORMATTED );
    itemData[ "displayPicture"           ] = contact->getContactPicturePath();
    itemData[ "groups"                   ] = contact->getGroupIds();
    itemData[ "mediaType"                ] = contact->getCurrentMediaType();
    itemData[ "mediaString"              ] = contact->getCurrentMediaString();
    itemData[ "status"                   ] = (int) contact->getStatus();
    itemData[ "isBlocked"                ] = contact->isBlocked();
    itemData[ "isReverse"                ] = contact->isReverse();
    itemData[ "clientName"               ] = contact->getClientAppName();

    if( parentNode_->object() )
    {
      Group *group = static_cast<Group*>( parentNode_->object() );
      itemData[ "group"            ] = group->getId();
      itemData[ "isInSpecialGroup" ] = group->isSpecialGroup();
    }
    else
    {
      itemData[ "group"            ] = QString();
      itemData[ "isInSpecialGroup" ] = false;
    }
  }
  else if( itemType_ == ItemGroup )
  {
    // Get the item's group to retrieve its data
    Group *group = static_cast<Group*>( object_ );

    // Find out how many of the group's contacts are online
    uint numOnlineContacts = 0;
    foreach( ContactListModelItem *item, childItems_ )
    {
      if( item->getType() != ItemContact )
      {
        continue;
      }

      Contact *contact = static_cast<Contact*>( item->object() );

      if( contact && contact->isOnline() )
      {
        numOnlineContacts++;
      }
    }

    itemData[ "type"           ] = ItemGroup;
    itemData[ "id"             ] = group->getId();
    itemData[ "name"           ] = group->getName();
    itemData[ "onlineContacts" ] = QString::number( numOnlineContacts );
    itemData[ "totalContacts"  ] = QString::number( childCount() );
    itemData[ "isExpanded"     ] = group->isExpanded();
  }
  else
  {
    itemData[ "type"           ] = ItemRoot;
  }

  return itemData;
}



/**
 * Return the type of the internal stored object
 *
 * @return ItemType
 */
ContactListModelItem::ItemType ContactListModelItem::getType() const
{
  return itemType_;
}



/**
 * Move a child item up or down the internal list
 *
 * @param movement  Use a negative number to move it up one position, a positive one to
 *                  move it down. Zero is positive.
 * @param item      The item to move
 */
void ContactListModelItem::moveChild( int movement, ContactListModelItem *item )
{
  if( ! item || childItems_.indexOf( item ) < 0 )
  {
    return;
  }

  int row = childItems_.indexOf( item );
  movement = (movement >= 0) ? (row + 1) : (row - 1);

  // Verify if the destination row is within the list's dimensions
  if( movement > childItems_.count() || movement < 0 )
  {
    return;
  }

  childItems_.move( row, movement );
}



/**
 * Return the stored object
 *
 * The object should be casted accordingly to the result of a getType() call.
 *
 * @return void*
 */
void *ContactListModelItem::object() const
{
  return object_;
}



/**
 * Get the parent node of this item
 *
 * @return Parent node
 */
ContactListModelItem *ContactListModelItem::parent() const
{
  return parentNode_;
}



/**
 * Remove a child item
 *
 * The child's object itself is not deleted, but is removed from the parent node only.
 * You have to manually call delete on it after removal.
 *
 * @param child  The item to delete
 */
void ContactListModelItem::removeChild( ContactListModelItem *child )
{
  if( ! childItems_.contains( child ) )
    return;

  childItems_.removeAll( child );
}



/**
 * Get this item's position within the parent node
 *
 * @return int
 */
int ContactListModelItem::row() const
{
  if( parentNode_ && ! parentNode_->childItems_.isEmpty() )
  {
    return parentNode_->childItems_.indexOf( const_cast<ContactListModelItem*>( this ) );
  }

  return 0;
}



/**
 * Set another node to be the parent of this item
 *
 * This method can be used to move items between nodes
 *
 * @param newParent  The node where this item will be moved to
 */
void ContactListModelItem::setParent( ContactListModelItem *newParent )
{
  if( ! parentNode_ || ! newParent )
  {
    return;
  }

  parentNode_->removeChild( this );
  newParent->appendChild( this );

  parentNode_ = newParent;
}



/**
 * Sort this item's children
 *
 * Currently sorting is performed by status only.
 * Adding sort ordering would require an UI option and another compareItems method.
 *
 * @param order         Sorting order. Currently ignored.
 * @param sortChildren  False to avoid sorting child nodes after this node's items.
 */
void ContactListModelItem::sort( Qt::SortOrder order, bool sortChildren )
{
  Q_UNUSED( order );

  // Contacts have no children
  if( childItems_.count() == 0 )
  {
    return;
  }

  // Stable sort the child items: already ordered items are not changed
  qStableSort( childItems_.begin(), childItems_.end(), ContactListModelItem::compareItems );

  // This option can be used to sort a single group, or the main group alone
  if( ! sortChildren )
  {
    return;
  }

  // Only sort child groups, to sort faster
  foreach( ContactListModelItem *item, childItems_ )
  {
    if( item->getType() == ItemGroup )
    {
      item->sort( order );
    }
  }
}



/**
 * Custom operator to display nodes and items in the debug output
 *
 * @param out   Output debugging stream
 * @param item  Item to print
 * @return QDebug  The debug stream itself
 */
QDebug operator<<( QDebug out, const ContactListModelItem *item )
{
  QString type;
  Contact *contact;
  Group *group;

  switch( item->getType() )
  {
    case ContactListModelItem::ItemRoot:
      type = "Root Node";
      break;

    case ContactListModelItem::ItemGroup:
      group = static_cast<Group *>( item->object() );
      type = "Group Node, id " + group->getId();
      break;

    case ContactListModelItem::ItemContact:
      contact = static_cast<Contact *>( item->object() );

      type = "Contact Node, handle " + contact->getHandle();

      if( item->parent()->getType() == ContactListModelItem::ItemRoot )
      {
        type += " in root node";
      }
      // Only the root node should have no parent, this is symptom of an error
      else if( ! item->parent() )
      {
        type += " without parent (!)";
      }
      else
      {
        group = static_cast<Group *>( item->parent()->object() );
        type += " in node " + group->getId();
      }
      break;

    default:
      type = "Unknown node type";
      break;
  }

  out << type << "(at address" << QString::number( (qlonglong)item, 16 ) << ")";

  return out;
}


