/***************************************************************************
                          contactlistmodelitem.h  -  description
                             -------------------
    begin                : Sat Feb 16 2008
    copyright            : (C) 2008 by Valerio Pilo
    email                : valerio@kmess.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef CONTACTLISTMODELITEM_H
#define CONTACTLISTMODELITEM_H

#include <QHash>
#include <QList>
#include <QString>


// Forward declarations
class Contact;
class Group;
class ContactListModelItem;


// Custom types
typedef QMap<QString, QVariant> ModelDataList;
typedef QList<ContactListModelItem *> ModelItemList;



// Custom operator to display nodes and items in the debug output
QDebug operator<<( QDebug s, const ContactListModelItem *item );



/**
 * This class represent a Contact List node or item in the Model/View paradigm.
 *
 * The Model is a way to universally represent data which could be reused later by different views:
 * for example, the contact list, or a list of contacts available to be invited in a chat.
 * The data output by the model can be easily expanded for future uses.
 * A model's item doesn't contain anything. A node is an element which contains other nodes or items.
 *
 * @author Valerio Pilo <valerio@kmess.org>
 * @ingroup Contact
 */
class ContactListModelItem
{
  public:  // Public enumerations
    enum ItemType
    {
      ItemRoot     = -1      // This element should be unique, it contains no object
    , ItemGroup    =  1      // Groups have this item type
    , ItemContact  =  2      // Contacts have this type
    };


  public:  // Public methods
    // Contact constructor, creates a root node
                                  ContactListModelItem();
    // Contact constructor, creates a Contact node
                                  ContactListModelItem( Contact *contact, ContactListModelItem *parent = 0 );
    // Group constructor, creates a Group node
                                  ContactListModelItem( Group *group, ContactListModelItem *parent = 0 );
    // Destructor
                                 ~ContactListModelItem();
    // Add a new child node
    void                          appendChild( ContactListModelItem *child );
    // Return a child identified by its row
    ContactListModelItem         *child( int row );
    // Return all Contact children identified by an handle
    ModelItemList                 childContacts( QString handle );
    // Return the number of children of this item
    int                           childCount() const;
    // Return a child Group identified by its ID
    ContactListModelItem         *childGroup( QString id );
    // Return the number of data columns
    int                           columnCount() const;
    // Return a collection of data from this item
    QVariant                      data( int column ) const;
    // Return the type of the internal stored object
    ItemType                      getType() const;
    // Move a child item up or down the internal list
    void                          moveChild( int movement, ContactListModelItem *item );
    // Return the stored object
    void                         *object() const;
    // Get the parent node of this item
    ContactListModelItem         *parent() const;
    // Remove a child item
    void                          removeChild( ContactListModelItem *child );
    // Get this item's position within the parent node
    int                           row() const;
    // Set another node to be the parent of this item
    void                          setParent( ContactListModelItem *newParent );
    // Sort this item's children
    void                          sort( Qt::SortOrder order, bool sortChildren = true );

  public: // Public static methods
    // Compare this item to another, checking if this is less than the other
    static bool                   compareItems( const ContactListModelItem *firstItem, const ContactListModelItem *secondItem );

  private:  // Private properties
    // The internal list of the item's children
    ModelItemList                 childItems_;
    // Pointer to the parent node
    ContactListModelItem         *parentNode_;
    // Pointer to this item's data
    void                         *object_;
    // Type of the contained data
    ItemType                      itemType_;
};

#endif
