/***************************************************************************
                          contactlistviewdelegate.cpp - description
                             -------------------
    begin                : Sat Feb 16 2008
    copyright            : (C) 2008 by Valerio Pilo
    email                : valerio@kmess.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "contactlistviewdelegate.h"

#include "currentaccount.h"
#include "emoticonmanager.h"
#include "contact/contactlistmodelitem.h"
#include "contact/msnstatus.h"
#include "utils/richtextparser.h"
#include "kmessdebug.h"

#include <QModelIndex>
#include <QPainter>
#include <QPalette>
#include <QTextDocument>

#include <KIconLoader>
#include <KLocale>
#include <KStandardDirs>





/**
 * Constructor
 *
 * Creates two labels which will be used to paint the contact list elements at the location of every item
 *
 * @param parent  The parent object, usually a QTreeView
 */
ContactListViewDelegate::ContactListViewDelegate( QWidget *parent )
    : QAbstractItemDelegate( parent )
    , iconLoader_( KIconLoader::global() )
{
  // Create test document
  textDocument_ = new QTextDocument();

  // Save the current account's reference
  currentAccount_ = CurrentAccount::instance();
}



/**
 * Destructor
 */
ContactListViewDelegate::~ContactListViewDelegate()
{
  delete textDocument_;
}



/**
 * Paint an item of the contact list
 *
 * This method is called whenever a contact list item needs do be drawn.
 * The 'index' parameter is used to extract the data from the list item, and depending on the list item type
 * (contact or group mainly) it decides what needs to be drawn and how.
 *
 * @param painter Painter item, we'll use it to draw our item's contents
 * @param option  Contains some details about the item we have to draw
 * @param index   Points to the actual data to be represented
 */
void ContactListViewDelegate::paint( QPainter *painter, const QStyleOptionViewItem &option, const QModelIndex &index ) const
{
  Flags statusFlags;
  QString groupIconName;
  ContactListModelItem *item = static_cast<ContactListModelItem*>( index.internalPointer() );
  QString text;

  // Only render column zero, and avoid invalid tree items which can't be rendered
  if( index.column() != 0 || item->getType() == ContactListModelItem::ItemRoot )
  {
    return;
  }

  // Save the painter state to avoid messing up further painting operations with it
  painter->save();

  // Reset coordinate translation in the painter (defaults are relative to the parent's window) - and make rendering prettier possibly
  painter->translate( QPoint( 0, 0 ) );
  painter->setRenderHint( QPainter::Antialiasing, true );

  // Paint the item as selected, that is with a colored background
  if( option.state & QStyle::State_Selected )
  {
    painter->fillRect( option.rect, option.palette.highlight() );
  }

  // Read the actual data off the model index
  QMap<QString,QVariant> itemData = item->data( index.column() ).toMap();

  // Set the initial position where to paint.
  // Leave one pixel above, and one on the left, to help spacing.
  QPoint position( option.rect.topLeft() );
  position.rx()++;
  position.ry()++;

  // Paint the different types of list items
  switch( item->getType() )
  {
    case ContactListModelItem::ItemGroup:
    {
      // Only show the expansion arrow if there are contacts to show
      if( (   currentAccount_->getShowOfflineContacts() && itemData[ "totalContacts"  ].toInt() > 0 )
      ||  ( ! currentAccount_->getShowOfflineContacts() && itemData[ "onlineContacts" ].toInt() > 0 ) )
      {
        groupIconName = ( itemData[ "isExpanded" ].toBool() ? "arrow-down" : "arrow-right" );
      }
      else
      {
        // Show only arrow down if there aren't contacts to show in group
        groupIconName = "arrow-down";
      }

      // Get the contact's picture
      QPixmap pixmap( KIconLoader::global()->loadIcon( groupIconName
                      , KIconLoader::NoGroup
                      , KIconLoader::SizeSmall ) );

      painter->drawPixmap( position, pixmap );

      // Leave a small margin next to the arrow (even if it's not drawn, so we align the group names anyway)
      position.rx() += pixmap.width() + 2;
      position.ry() += pixmap.height() -3;

      // Prepare and paint the group name
      text =  i18nc( "Group name in the contact list with online/total contacts of that group"
                     , "%1 (%2/%3)"
                     , Qt::escape( itemData[ "name" ].toString() )
                     , itemData[ "onlineContacts" ].toString()
                     , itemData[ "totalContacts"  ].toString() );


      // Set bold font 
      QFont font;
      font.setBold( true );
      painter->setFont( font );
      painter->drawText( position, text );
      break;
    }


    case ContactListModelItem::ItemContact:
    {
      int displayPictureSize = currentAccount_->getListPictureViewMode();

      // Use this variable to store the position when we draw the icon status on bottom edge.
      QPoint oldPosition;

      // Only show the picture if it's been enabled by the user
      if( displayPictureSize != 0 )
      {
        QPixmap pixmap( itemData[ "displayPicture" ].toString() );
        pixmap = pixmap.scaled( displayPictureSize,  displayPictureSize,
                                Qt::KeepAspectRatio, Qt::FastTransformation );

        painter->drawPixmap( position, pixmap );

        // If the display picture is enough, show small status icon over the picture
        if( displayPictureSize >= 48 )
        {
          oldPosition = position;
          position.ry() += displayPictureSize - KIconLoader::SizeSmall;
        }
        else
        {
          // Leave a small margin next to the display picture
          position.rx() += displayPictureSize + 2;
        }
      }

      // Determine which status image to load
      statusFlags = FlagNone;
      if( itemData[ "isBlocked" ].toBool() )
      {
        statusFlags = FlagBlocked;
      }

      painter->drawPixmap( position, MsnStatus::getIcon( (Status) itemData[ "status" ].toInt(), statusFlags ) );

      if( displayPictureSize >= 48 )
      {
        position = oldPosition;
        position.rx() += displayPictureSize + 4;
      }
      else
      {
        // Leave a small margin next to the status picture
        position.rx() += KIconLoader::SizeSmall + 4;
      }

      // Get the contact information.
      QString friendlyName, personalMessage;
      QString mediaType       = itemData[ "mediaType"       ].toString();
      QString mediaString     = itemData[ "mediaString"     ].toString();

      // Parse the name and message of the contact, adding emoticons and formatting
      if( currentAccount_->getUseListFormatting() )
      {
        friendlyName    = itemData[ "friendlyFormatted"        ].toString();
        personalMessage = itemData[ "personalMessageFormatted" ].toString();
      }
      else
      {
        friendlyName    = Qt::escape( itemData[ "friendly"        ].toString() );
        personalMessage = Qt::escape( itemData[ "personalMessage" ].toString() );
      }

      QString messageString;

      if( ! mediaString.isEmpty() )
      {
        // Determine icon for the various media types
        QString mediaEmoticon;
        if( mediaType == "Music" )
        {
          mediaEmoticon = "(8)";
        }
        else if( mediaType == "Gaming" )
        {
          mediaEmoticon = "(xx)";
        }

        // Set the icon if it's found in the emoticon theme.
        if( ! mediaEmoticon.isEmpty() )
        {
          mediaEmoticon = EmoticonManager::instance()->getReplacement( mediaEmoticon, true );
        }

        messageString = mediaEmoticon + Qt::escape( mediaString );
      }
      else if( ! personalMessage.isEmpty() ) // When there's no media, show the PM
      {
        messageString = "<i>" + personalMessage + "</i>";
      }

      // Fix HTML entities in the friendly name
      if( ! itemData[ "isReverse" ].toBool() ) // The contact does not have us in its list
      {
        friendlyName = "<i>" + friendlyName + "</i>";
      }

      // Draw the text widget
      if( messageString.isEmpty() )
      {
        // Only person name
        text = friendlyName;
      }
      else if( ! currentAccount_->getListPictureViewMode() )
      {
        // Person with message after it
        text = friendlyName + "<small><font style='color:palette(window-text)'> - " + messageString + "</font></small>";
      }
      else
      {
        // Message below, 
        text = friendlyName + "<br /><small><font style='color:palette(window-text)'>" + messageString + "</font></small>";
      }

      textDocument_->setHtml( text );
      painter->translate( position.x(), option.rect.topLeft().y() - 2 );
      textDocument_->drawContents( painter );
      break;
    }

    default:
      kWarning() << "Unknown type of item:" << item->getType();
      break;
  }

  // Restore the painter to the previous state
  painter->restore();
}



/**
 * Return an hint about the size of an item
 *
 * @param option  Contains some details about the item we have to determine size of
 * @param index   Points to the actual list item
 * @return The ize for the given index
 */
QSize ContactListViewDelegate::sizeHint( const QStyleOptionViewItem &option, const QModelIndex &index ) const
{
  Q_UNUSED( option );

  ContactListModelItem *item = static_cast<ContactListModelItem*>( index.internalPointer() );

  QSize size( KIconLoader::SizeSmall + 2, KIconLoader::SizeSmall + 4 );

  if( ! item )
  {
    return size;
  }

  int picturesDimension = currentAccount_->getListPictureViewMode();
  switch( item->getType() )
  {
    case ContactListModelItem::ItemContact:
      if( picturesDimension > 0 )
      {
        QSize pictureSize( size );
        if( picturesDimension <= 32 )
        {
          // Fix the margin
          pictureSize.rheight() = 36;
        }
        else
        {
          pictureSize.rheight() = picturesDimension + 2;
        }

        return pictureSize;
      }
      else
      {
        return size;
      }

      break;

    case ContactListModelItem::ItemGroup:
    default:
      return size;
      break;
  }
}

