/***************************************************************************
                          contactpropertiesdialog.cpp  -  description
                             -------------------
    begin                : Sun Dec 15 2002
    copyright            : (C) 2002 by Michael Curtis
    email                : magnalium@hotmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "contactpropertiesdialog.h"

#include "../contact/contact.h"
#include "../contact/contactextension.h"
#include "../contact/contactlist.h"
#include "../contact/group.h"
#include "../utils/kmessconfig.h"
#include "../utils/richtextparser.h"
#include "../currentaccount.h"
#include "../kmessdebug.h"

#include <QTextDocument>

#include <KFileDialog>
#include <KIconEffect>
#include <KIconLoader>
#include <KStandardDirs>



ContactPropertiesDialog::ContactPropertiesDialog( QWidget *parent )
 : KDialog( parent )
 , Ui::ContactProperties()
 , ok_( false )
 , mediaSound_(0)
{
  // Set up the interface and the dialog
  QWidget *mainWidget = new QWidget( this );
  setupUi( mainWidget );
  setMainWidget( mainWidget );

  // Configure the dialog
  setWindowModality( Qt::WindowModal );

  KIconLoader *loader = KIconLoader::global();

  // Set the dialog size
  setButtons( Ok | Cancel );
  setDefaultButton( Ok );
  restoreDialogSize( KMessConfig::instance()->getGlobalConfig( "ContactPropertiesDialog" ) );

  // Set the icons for the buttons
  soundButton_->setIcon(      loader->loadIcon( "document-open",        KIconLoader::Small ) );
  playButton_->setIcon(       loader->loadIcon( "media-playback-start", KIconLoader::Small ) );

  // Set properties for clear cache button
  clearCacheButton_->setIcon( loader->loadIcon( "draw-eraser",          KIconLoader::Small ) );

  connect( clearCacheButton_, SIGNAL( clicked() ), this, SLOT( slotClearCache()         ) );
  connect( restoreButton_,    SIGNAL( clicked() ), this, SLOT( restoreContactPicture()  ) );
  connect( soundButton_,      SIGNAL( clicked() ), this, SLOT( chooseSound()            ) );
  connect( playButton_,       SIGNAL( clicked() ), this, SLOT( slotPlaySound()          ) );

  imageContact_->installEventFilter( this );
}



ContactPropertiesDialog::~ContactPropertiesDialog()
{
  KConfigGroup group = KMessConfig::instance()->getGlobalConfig( "ContactPropertiesDialog" );
  saveDialogSize( group );

  delete mediaSound_;
#ifdef KMESSDEBUG_CONTACTPROPERTIES
  kDebug() << "DESTROYED.";
#endif
}



// The alt name checkbox was toggled.
void ContactPropertiesDialog::alternativeNameCheckBoxToggled(bool checked)
{
  nameEdit_->setEnabled( checked );
}



// Apply all values in the widgets to the contact extension.
void ContactPropertiesDialog::applyChanges()
{
  ContactExtension *extension;

  extension = contact_->getExtension();

  // if the user clears the name edit, be user-friendly and decide that what
  // they really wanted to do was turn off the alternative name override.
  if ( nameEdit_->text().isEmpty() )
  {
    extension->setUseAlternativeName( false );
  }
  else
  {
    extension->setUseAlternativeName( alternativeNameCheckBox_->isChecked() );
  }

  extension->setAlternativeName   ( nameEdit_->text()                   );
  extension->setNotes             ( notesEdit_->toPlainText()           );
  extension->setNotifyOnline      ( notifyOnlineCheckBox_->isChecked()  );
  extension->setContactSoundPath  ( soundEdit_->text()                  );
  extension->setContactAlternativePicturePath(  alternativePicturePath_ );


  const QStringList groupIdsList = contact_->getGroupIds();
  QListWidgetItem *item;
  QString currentId;

  // Search for added/removed contact group
  for( int i = 0; i < groupList_->count(); i++ )
  {
    item = groupList_->item( i );
    currentId = item->data( 6 ).toString();

    if( item->checkState() == Qt::Checked )
    {
      // If item is checked and the groupIdsList doesn't contain
      // the currentId, it means the user want to add the contact in other group
      if( ! groupIdsList.contains( currentId ) )
      {
        emit addContactToGroup( contact_->getHandle(), currentId );
      }
      // else the user is already in the group
      continue;
    }
    else
    {
      // If item is unchecked and the groupIdsList contains
      // the currentId, it means the user want to remove the contact from the group
      if( groupIdsList.contains( currentId ) )
      {
        emit removeContactFromGroup( contact_->getHandle(), currentId );
      }
      // else the user isn't already in the group
    }
  }
}



// Get the location of a file from the user, using a KFile dialog
//  Uses the contents of the KLineEdit as a default, and replaces the contents
//   of said widget with the result of the file selection dialog.
void ContactPropertiesDialog::chooseFile(KLineEdit *edit, QString wildcard)
{
  QString file;

  file = edit->text();

  file = KFileDialog::getOpenFileName( file, wildcard, this );

  if ( !file.isEmpty() )
  {
    edit->setText( file );
  }
}



// Choose a picture file (KFile dialog)
void ContactPropertiesDialog::choosePicture()
{
  KUrl file;
  file = KFileDialog::getImageOpenUrl( alternativePicturePath_ , this );
  if ( ! file.isEmpty() )
  {
    QPixmap pixmap( file.toLocalFile() );

    // If it's a valid image
    if( ! pixmap.isNull() )
    {
      imageContact_->setPixmap( pixmap );
      alternativePicturePath_ =  file.toLocalFile();
    }
  }
}



// Choose a sound file (KFile dialog)
void ContactPropertiesDialog::chooseSound()
{
  // Grep the standard sound directory
  QStringList soundDirectoryList( KGlobal::dirs()->findDirs( "sound", "" ) );
  if( ! soundDirectoryList.isEmpty () )
  {
    KUrl soundDirectory( soundDirectoryList.first() );
    soundEdit_->setUrl( soundDirectory );
  }

  chooseFile( soundEdit_, "*.ogg *.wav" );
}



bool ContactPropertiesDialog::eventFilter(QObject *obj, QEvent *event )
{
  if( obj == imageContact_ )
  {
    QString image;
    if( alternativePicturePath_.isEmpty() )
    {
      image = contact_->getExtension()->getContactPicturePath();

      if( image.isEmpty() )
      {
        image = contact_->getContactDefaultPicturePath();
      }
    }
    else
    {
      image = alternativePicturePath_;
    }

    int eventType = event->type();

    if( eventType == QEvent::Enter )
    {
      QImage glowingImage( image );
      KIconEffect::toGamma( glowingImage, 0.8f );
      imageContact_->setPixmap( QPixmap::fromImage( glowingImage ) );
    }
    else if( eventType == QEvent::Leave )
    {
      imageContact_->setPixmap( image );
    }
    else if( eventType == QEvent::MouseButtonRelease )
    {
      choosePicture();
    }
  }

  return false;
}



// Show the dialog and obtain the contact handle.
bool ContactPropertiesDialog::launch( Contact *contact, bool openNotesTab )
{
#ifdef KMESSDEBUG_CONTACTPROPERTIES
  kDebug() << "Launch with contact " << contact->getHandle();
#endif
  QString dialogCaption;

  // Copy the pointer to the contact class
  contact_ = contact;

  // Set the dialog title
  dialogCaption = i18n( "Contact Properties for %1", contact->getHandle() );
  setCaption( dialogCaption );

  // Setup the dialog
  setupWidgets();

  // If openNotesTab is true, set the current index to notes tab
  if( openNotesTab )
  {
    tabs_->setCurrentIndex( 2 );
    notesEdit_->setFocus();
  }

  // Show the dialog modally
  exec();

  // If the ok button was pressed..
  if ( ok_ )
  {
    applyChanges();
  }

  return ok_;
}



// The online notification checkbox was toggled
void ContactPropertiesDialog::notifyOnlineCheckBoxToggled(bool checked)
{
  soundEdit_->setEnabled( checked );
  soundButton_->setEnabled( checked );
  playButton_->setEnabled( checked );
}



// Restore the original contact picture
void ContactPropertiesDialog::restoreContactPicture()
{
  // Restore the contact's picture
  QString originalPicture = contact_->getExtension()->getContactPicturePath();

  if( originalPicture.isEmpty() )
  {
    originalPicture = contact_->getContactDefaultPicturePath();
  }

  QPixmap pixmap( originalPicture );

  if( ! pixmap.isNull() )
  {
    imageContact_->setPixmap( pixmap );
    alternativePicturePath_ = "";
  }

  restoreButton_->hide();
}



// Setup the dialog's widgets
void ContactPropertiesDialog::setupWidgets()
{
  // Use bold font for the friendly name of the contact
  QFont font;
  font.setBold(true);
  ContactExtension *extension = contact_->getExtension();

  // Set the button default view to hide
  restoreButton_->hide();

  trueNameLabel_->setFont( font );
  trueNameLabel_->setText( Qt::escape( contact_->getTrueFriendlyName() ) );

  // Set the informations about contact
  QString pm = Qt::escape( contact_->getPersonalMessage() );
  if( pm.isEmpty() )
  {
    // Hide to avoid blank space
    personalMessageLabel_->hide();
  }
  else
  {
    personalMessageLabel_->setText( pm );
  }

  lastSeenLabel_       ->setText( i18n( "Last seen: " ) + extension->getLastSeen()        );
  emailLabel_          ->setText( i18n( "Email: "     ) + contact_->getHandle()           );
  clientLabel_         ->setText( i18n( "Client: "    ) + contact_->getClientAppName()    );
  nameEdit_            ->setText(                         extension->getAlternativeName() );
  notesEdit_      ->setPlainText(                         extension->getNotes()           );

  alternativeNameCheckBox_->setChecked( extension->getUseAlternativeName() );
  alternativeNameCheckBoxToggled(       extension->getUseAlternativeName() );
  notifyOnlineCheckBox_->setChecked(    extension->getNotifyOnline()       );
  notifyOnlineCheckBoxToggled(          extension->getNotifyOnline()       );

  // Set the contact picture
  QString picturePath = contact_->getContactPicturePath();

  QPixmap pixmap( picturePath );

  // If pixmap is null set the default contact picture
  if( pixmap.isNull() )
  {
    pixmap.load( contact_->getContactDefaultPicturePath() );
  }

  imageContact_->setPixmap( pixmap );

  alternativePicturePath_ = extension->getContactAlternativePicturePath();

  // If the alternative picture is used, show the restore button
  if( ! alternativePicturePath_.isEmpty() )
  {
    restoreButton_->show();
  }

  soundEdit_->setText( extension->getContactSoundPath()   );

  // Set the group list
  QStringList groupIdsList = contact_->getGroupIds();
  QList<Group *> groupContactList = CurrentAccount::instance()->getContactList()->getGroupList();
  QListWidgetItem *item;

  foreach( Group *group, groupContactList )
  {
    if( ! group->isSpecialGroup() )
    {
      item = new QListWidgetItem( group->getName(), groupList_ );
      item->setFlags( Qt::ItemIsUserCheckable | Qt::ItemIsEnabled );
      item->setData( 6, group->getId() );

      if( groupIdsList.contains( group->getId() ) )
      {
        item->setCheckState( Qt::Checked );
      }
      else
      {
        item->setCheckState( Qt::Unchecked );
      }
    }
  }

  // Set the pictures list
  QStringList pictureList( extension->getPictureList() );
  QListWidgetItem *pictureItem;

  foreach( QString pictureFile, pictureList )
  {
    // Load into QPixmap and check if the pixmap is valid
    if( ! pixmap.load( pictureFile ) )
    {
      continue;
    }

    // Set the new item and put it into pictures list
    pictureItem = new QListWidgetItem();
    pictureItem->setIcon( QIcon( pixmap ) );

    QFile fileDat( pictureFile + ".dat" );
    if ( fileDat.exists() && fileDat.open( QIODevice::ReadOnly | QIODevice::Text ) )
    {
      QTextStream in( &fileDat );
      pictureItem->setText( in.readLine() );
      fileDat.close();
    }

    pictureList_->addItem( pictureItem );
  }
}



// The Ok or Cancel button has been pressed
void ContactPropertiesDialog::slotButtonClicked( int button )
{
  // If the Cancel button has been pressed, close without saving
  if( button != Ok )
  {
    KDialog::slotButtonClicked( button );
    return;
  }

  ok_ = true;
  accept();
}



// The "Clear cache" button has been pressed
void ContactPropertiesDialog::slotClearCache()
{
  if( pictureList_->count() != 0 )
  {
    ContactExtension *extension = contact_->getExtension();
    extension->clearCache();
    pictureList_->clear();
  }
}



// The play sound button has been pressed
void ContactPropertiesDialog::slotPlaySound()
{
  QString currentFile = soundEdit_->text();
  if( currentFile.isEmpty() )
  {
    return;
  }

  // Initialize the media sound object and connect it with slot sound finished
  if( mediaSound_ == 0 )
  {
    mediaSound_ = Phonon::createPlayer( Phonon::MusicCategory );
    connect( mediaSound_, SIGNAL( finished() ), this, SLOT( slotSoundFinished() ) );
  }

  // If the state is playing, stop the sound media and reset the icon
  if( mediaSound_->state() == Phonon::PlayingState )
  {
    mediaSound_->stop();
    playButton_->setIcon( KIcon( "media-playback-start" ) );
    return;
  }

  // Play the sound and change the icon
  mediaSound_->setCurrentSource( currentFile );
  mediaSound_->play();
  playButton_->setIcon( KIcon( "media-playback-stop" ) );
}



// Called when media sound finished
void ContactPropertiesDialog::slotSoundFinished()
{
  // When sound finished, change reset the icon
  playButton_->setIcon( KIcon( "media-playback-start" ) );
}


#include "contactpropertiesdialog.moc"
