/***************************************************************************
                          kmess.cpp  -  description
                             -------------------
    begin                : Sun Jan  5 15:18:36 CST 2003
    copyright            : (C) 2003 by Mike K. Bennett
    email                : mkb137b@hotmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "kmess.h"

#include "chat/chat.h"
#include "chat/chatmaster.h"
#include "contact/contact.h"
#include "contact/contactlist.h"
#include "contact/group.h"
#include "contact/msnstatus.h"
#include "dialogs/addcontactdialog.h"
#include "dialogs/awaymessagedialog.h"
#include "dialogs/contactaddeduserdialog.h"
#include "dialogs/networkwindow.h"
#include "dialogs/transferwindow.h"
#include "network/msnnotificationconnection.h"
#include "notification/notificationmanager.h"
#include "notification/chatnotification.h"
#include "notification/contactstatusnotification.h"
#include "notification/newemailnotification.h"
#include "notification/passivepopup.h"
#include "settings/accountsettingsdialog.h"
#include "utils/idletimer.h"
#include "utils/kmessconfig.h"
#include "utils/kmessdbus.h"
#include "utils/kmessshared.h"
#include "utils/nowlisteningclient.h"
#include "utils/richtextparser.h"
#include "account.h"
#include "accountsmanager.h"
#include "accountaction.h"
#include "currentaccount.h"
#include "emoticonmanager.h"
#include "initialview.h"
#include "kmessdebug.h"
#include "kmessinterface.h"
#include "kmessview.h"
#include "systemtraywidget.h"

#include <QDir>

#include <KAction>
#include <KActionMenu>
#include <KConfig>
#include <KGlobal>
#include <KInputDialog>
#include <KLocale>
#include <KMenu>
#include <KMessageBox>
#include <KSelectAction>
#include <KStandardDirs>
#include <KStatusBar>
#include <KToggleAction>
#include <KUrl>



// The constructor
KMess::KMess( QWidget *parent )
 : KMessInterface( parent ),
   chatMaster_(0),
   chatNotification_(0),
   contactStatusNotification_(0),
   newEmailNotification_(0),
   currentStatus_(STATUS_OFFLINE),
   idleTimer_(0),
   initialized_(false),
   initialView_(0),
   nowListeningClient_(0),
   msnNotificationConnection_(0),
   view_(0)
{
}



// The destructor
KMess::~KMess()
{
#ifdef KMESSDEBUG_KMESS
  kDebug() << "shutting down KMess...";
#endif

  /*
   * The destructor is called after closeEvent(),
   * when KMess is quit by the user.
   *
   * If the KDE session ends, the destructor won't be called.
   * only saveProperties() will be called, and the execution ends.
   */


  // Saving properties manually.
  saveProperties();

  // Delete all chat windows
  if(chatMaster_ != 0)
  {
    chatMaster_->disconnected();
    chatMaster_->deleteLater();
    chatMaster_ = 0;
  }

  // Disconnect main connection so views can clean up/save properties.
  if(msnNotificationConnection_->isConnected())
  {
    // Make sure KMess::disconnected() does attempt to update
    // the user interface when we're trying to kill it
    disconnect(msnNotificationConnection_, SIGNAL(disconnected()), this, SLOT(disconnected()) );

    // This also destroys all contacts
    msnNotificationConnection_->closeConnection();
  }

  // Delete other created objects
  delete idleTimer_;
  delete chatNotification_;
  delete contactStatusNotification_;
  delete notificationManager_;
  delete newEmailNotification_;
  delete nowListeningClient_;
  delete msnNotificationConnection_;

  AccountsManager::destroy();
  CurrentAccount::destroy();
  TransferWindow::destroy();
  KMessConfig::destroy();
#ifdef KMESS_NETWORK_WINDOW
  NetworkWindow::destroy();
#endif

#ifdef KMESSDEBUG_KMESS
  kDebug() << "DESTROYED.";
#endif
}



// Add the account to the GUI
void KMess::slotAccountAdded(Account *account)
{
  AccountAction *accountAction;

  // Make the action for the "Connect..." menu
  accountAction = new AccountAction( account, this );
  connectActionMenu_->addAction( accountAction );
  connectMenuItems_[ account ] = accountAction;

  connect( accountAction, SIGNAL(          activated(Account*) ),
           this,          SLOT  ( connectWithAccount(Account*) ) );


  // Make the action for the "Settings..." menu
  accountAction = new AccountAction( account, this );
  settingsActionMenu_->addAction( accountAction );
  settingsMenuItems_[ account ] = accountAction;

  connect( accountAction, SIGNAL(              activated(Account*) ),
           this,          SLOT  ( showSettingsForAccount(Account*) ) );

  // If the initial login window is visible, add the new account there, too
  if( initialView_ )
  {
    initialView_->addAccount( account );
  }
}



// "Add a new contact" was selected from the menu.
void KMess::addNewContact()
{
#ifdef KMESSTEST
  KMESS_ASSERT( msnNotificationConnection_ != 0 );
#endif
#ifdef KMESSDEBUG_KMESS
  kDebug();
#endif

  QString groupId;
  QString newHandle;

  // Launch a modal dialog to get the new contact handle and, optionally, group
  new AddContactDialog( newHandle, groupId, this );

  // Do nothing if the user has clicked cancel
  if( newHandle.isEmpty() )
  {
    return;
  }

  // Ask the server to add the contact
  msnNotificationConnection_->addNewContact( newHandle, groupId );
}



// "Add a new group" was selected from the menu.
void KMess::addNewGroup()
{
#ifdef KMESSTEST
  KMESS_ASSERT( msnNotificationConnection_ != 0 );
#endif
#ifdef KMESSDEBUG_KMESS
  kDebug();
#endif
  QString newGroupName;
  bool    okPressed = false;

  // Set a default group name
  newGroupName = i18n( "New Group" );

  // Launch a dialog to get a new group name
  newGroupName = KInputDialog::getText(i18n("Add a group"),
                                       i18n("Enter the new name for this group:"),
                                       newGroupName, &okPressed, this);

  if(okPressed)
  {
    // Ask the server to add the group
    msnNotificationConnection_->addGroup( newGroupName );
  }
}



// The application is closing, after queryClose() was approved
void KMess::applicationClosing()
{
  // Prepare closing, KMessInterface::queryClose() accepted the close request.
  // After this method returns, KApplication::aboutToQuit() also kicks in, handled by KMessApplication.

  if( chatMaster_ != 0 )
  {
#ifdef KMESSDEBUG_KMESS
    kDebug() << "destroying chat master";
#endif

    // Close the chat windows earlier, so logs are saved.
    chatMaster_->deleteLater();
    chatMaster_ = 0;
  }
}



// An account's settings have been changed
void KMess::slotAccountChanged( Account *account, QString oldHandle, QString oldFriendlyName )
{
  QString newHandle = account->getHandle();

#ifdef KMESSDEBUG_KMESS
  kDebug() << "Saving UI settings for" << newHandle;
#endif

  if ( msnNotificationConnection_ != 0 )
  {
    // While we're saving stuff, also save the contact list settings
    msnNotificationConnection_->saveProperties();

    bool isCurrentAccount          = ( oldHandle == currentAccount_->getHandle() );
    bool isConnectedCurrentAccount = ( isCurrentAccount && msnNotificationConnection_->isConnected() );

    // If the account matches the current account and the friendly name was changed, ask the server to change it
    if( isConnectedCurrentAccount && account->isVerified() && account->getFriendlyName( STRING_ORIGINAL ) != oldFriendlyName )
    {
      msnNotificationConnection_->changeFriendlyName( account->getFriendlyName( STRING_ORIGINAL ) );
    }
  }

  // Change the account name in the UI when it's been changed
  if( oldHandle == newHandle )
  {
    return;
  }

  // Update it in the initial view's list of accounts
  if( initialView_ )
  {
#ifdef KMESSDEBUG_KMESS
  kDebug() << "Updating initial view";
#endif
    // Notify the initial view when the account handle has changed.
    initialView_->changedAccount( oldHandle, newHandle );
  }

  // Update the menu items
  connectMenuItems_ [ account ]->updateText();
  settingsMenuItems_[ account ]->updateText();
}



// A view pictures mode has been selected from the menu.
void KMess::changedListPictureViewMode( int mode )
{
  int newMode = 0;
  switch( mode )
  {
    case 0:
      break;
    case 1:
      newMode = 22;
      break;
    case 2:
      newMode = 32;
      break;
    case 3:
      newMode = 40;
      break;
    case 4:
      newMode = 48;
      break;
    case 5:
      newMode = 64;
      break;

    default:
      break;
  }

  currentAccount_->setListPictureViewMode( newMode );
}



// A status was selected from the menu.
void KMess::changeStatus( int index )
{
  Status             newStatus;
  bool               useAutoreply;
  AwayMessageDialog *awayMessageDialog;
  QString            awayMessage;

  isIdleForced_ = false;
  currentAccount_->setAutoreply( false );

  // Find out what status has been chosen
  switch( index )
  {
    // Away with auto-reply
    case 2:
      newStatus = STATUS_AWAY;

      // Create the away message dialog
      awayMessageDialog = new AwayMessageDialog( this );

      // Confirm the auto-away message with the user
      awayMessage = CurrentAccount::instance()->getAutoreplyMessage();
      useAutoreply = false;
      useAutoreply = awayMessageDialog->useMessage( awayMessage );
      if( useAutoreply )
      {
        // Tell the account to respond with an autoreply message
        CurrentAccount::instance()->setAutoreplyMessage( awayMessage );
        CurrentAccount::instance()->setAutoreply( true );
      }
      delete awayMessageDialog;
      break;
    // Idle
    case 3:
      newStatus = STATUS_IDLE;
      isIdleForced_ = true;
      break;

    case 1:  newStatus = STATUS_AWAY;          break;
    case 4:  newStatus = STATUS_BE_RIGHT_BACK; break;
    case 5:  newStatus = STATUS_BUSY;          break;
    case 6:  newStatus = STATUS_OUT_TO_LUNCH;  break;
    case 7:  newStatus = STATUS_ON_THE_PHONE;  break;
    case 8:  newStatus = STATUS_INVISIBLE;     break;
    case 0:
    default: newStatus = STATUS_ONLINE;        break;
  }

#ifdef KMESSDEBUG_KMESS
  kDebug() << "Selected status: " << MsnStatus::getCode( newStatus );
#endif

  msnNotificationConnection_->changeStatus( newStatus );
}



// Tray notifications settings were changed
void KMess::changedNotificationsSettings()
{
  chatNotification_         ->setEnabled( currentAccount_->getNotifyContactsChatStart(),
                                          currentAccount_->getNotifyContactsChatMessages() );
  contactStatusNotification_->setEnabled( currentAccount_->getNotifyContactsOnline(),
                                          currentAccount_->getNotifyContactsOffline(),
                                          currentAccount_->getNotifyContactsStatus() );
  newEmailNotification_     ->setEnabled( currentAccount_->getNotifyInboxEmails(),
                                          currentAccount_->getNotifyOtherEmails() );
}



// The current now listening settings have changed.
void KMess::changedNowListeningSettings()
{
  // Enable now listening if current account is online and likes to enable now listening.
  Status status = currentAccount_->getStatus();

  // Enable only if not offline or invisible
  bool isOnline = ( status != STATUS_OFFLINE && status != STATUS_INVISIBLE );
  nowListeningClient_->setEnabled( currentAccount_->getShowNowListening() && isOnline );

#ifdef KMESSDEBUG_KMESS
  kDebug() << "Now listening setting is now"
           << ( currentAccount_->getShowNowListening() ? "enabled" : "disabled") << endl;
#endif
}



// The currently playing song was changed.
void KMess::changedSong( const QString &artist, const QString &album, const QString &track, bool playing )
{
  if( msnNotificationConnection_->isConnected() )
  {
    if( ! playing )
    {
      // Update notification connection and KMessView.
      msnNotificationConnection_->removeCurrentMedia();
    }
    else
    {
      // Update notification connection.
      QStringList arguments;
      arguments << track << artist << album << "" << "";
      msnNotificationConnection_->changeCurrentMedia( "", "Music", true, "{0} - {1}", arguments );
    }

    // Update KMessView
    view_->changedSong( artist, album, track, playing );
  }
}



// The status was changed
void KMess::changedStatus( Account *account )
{
  int item = -1;
  Status newStatus;

  if( account != 0 )
  {
    newStatus = account->getInitialStatus();
  }
  else if ( currentAccount_ != 0 )
  {
    newStatus = currentAccount_->getStatus();
  }
  else
  {
    kWarning() << "Changed status dropdown without an account set!";
    return;
  }

#ifdef KMESSDEBUG_KMESS
  kDebug() << "Changed status to" << MsnStatus::getCode( newStatus )
           << "from" << MsnStatus::getCode( currentStatus_ ) << ". "
           << "Autoreply is" <<  currentAccount_->getAutoreply() << endl;
#endif

  // Make sure the drop down list matches the user's status
  switch( newStatus )
  {
    case STATUS_AWAY:
      if ( !currentAccount_->getAutoreply() )
      {
        // Change status to Away
        item = 1;
      }
      else
      {
        // Change status to Away with autoreply
        item = 2;
      }
      break;
    case STATUS_IDLE:          item = 3; break;
    case STATUS_BE_RIGHT_BACK: item = 4; break;
    case STATUS_BUSY:          item = 5; break;
    case STATUS_OUT_TO_LUNCH:  item = 6; break;
    case STATUS_ON_THE_PHONE:  item = 7; break;
    case STATUS_INVISIBLE:     item = 8; break;
    case STATUS_ONLINE:
    default:                   item = 0; break;
  }

#ifdef KMESSDEBUG_KMESS
  kDebug() << "Changed the select status to item " << item;
#endif

  if ( item >= 0 )
  {
    status_->setCurrentItem( item );
  }
}



// A view mode has been selected from the menu.
void KMess::changeViewMode(int mode)
{
#ifdef KMESSTEST
  KMESS_ASSERT( ( mode == 0 ) || ( mode == 1 ) );
#endif
  if ( mode == 0 )
  {
    CurrentAccount::instance()->setShowContactsByGroup( true );
  }
  else if ( mode == 1 )
  {
    CurrentAccount::instance()->setShowContactsByGroup( false );
  }
  else
  {
    kDebug() << "changeViewMode() - WARNING - Invalid view mode = " << mode;
  }
}



// Show a "Contact added you" dialog
void KMess::showContactAddedUserDialog(const Contact *contact)
{
  const QString &handle = contact->getHandle();

#ifdef KMESSDEBUG_KMESS
  kDebug() << "Contact added user";
#endif

  // Show a dialog and let the user choose whether to add, allow, or block the contact.
  ContactAddedUserDialog dialog;
  ContactAddedUserDialog::ReturnCode returnValue = dialog.launch( handle, contact->getFriendlyName() );

  switch( returnValue )
  {
    case ContactAddedUserDialog::ADD:
      // Add the contact to the allowed and friends lists
      msnNotificationConnection_->addNewContact( handle );
      break;

    case ContactAddedUserDialog::ALLOW:
      // Add the contact to the allowed list
      msnNotificationConnection_->allowContact( handle );
      break;

    case ContactAddedUserDialog::BLOCK:
      // Add the contact to the blocked list.
      msnNotificationConnection_->blockContact(handle);
      break;
  }
}



#ifdef KMESS_NETWORK_WINDOW
// Opens the network window
void KMess::showNetworkWindow()
{
  NetworkWindow::instance()->show();
}
#endif


// Show a dialog before removing the contact
void KMess::showRemoveContactDialog(QString handle)
{
  QString message = i18n( "<qt>Are you sure you want to remove the contact <b>%1</b> from your contact list?</qt>", handle );

  int result = KMessageBox::warningYesNoCancel( this, message, i18n( "Remove contact" )
                                              , KGuiItem(i18n("Remove"), "edit-delete")             // Yes
                                              , KGuiItem(i18n("Remove and block"), "list-remove")       // No
                                              , KStandardGuiItem::cancel()
                                              , QString()
                                              , KMessageBox::Dangerous
                                              );

  switch( result )
  {
    // User has pressed "Remove and block"
    case KMessageBox::No :
      msnNotificationConnection_->removeContact( handle, true );
      break;

    // User has pressed "Remove only"
    case KMessageBox::Yes :
      msnNotificationConnection_->removeContact( handle, false );
      break;

    // User has canceled the action, no nothing
    case KMessageBox::Cancel :
    default:
      break;
  }

}


// Show a dialog before removing the group
void KMess::showRemoveGroupDialog(QString groupId)
{
  Group *group = currentAccount_->getContactList()->getGroupById(groupId);
 
  // If the group was found
  if(group == 0)
  {
    kDebug() << "Couldn't find a matching group.";
    return;
  }

  QString message = i18n( "<qt>Are you sure you want to remove the group <b>%1</b> from your contact list?</qt>", group->getName());

  int result = KMessageBox::warningContinueCancel( this, message, i18n( "Remove group" ),
                                                   KGuiItem(i18n("Remove"), "edit-delete"),
                                                   KStandardGuiItem::cancel(),
                                                   QString(),
                                                   KMessageBox::Dangerous
                                                  );

  if(result == KMessageBox::Continue)
  {
    msnNotificationConnection_->removeGroup(groupId);
  }
}


// Show a "Rename group" dialog
void KMess::showRenameGroupDialog(QString groupId)
{
  Group  *group = currentAccount_->getContactList()->getGroupById(groupId);
  QString newGroupName;
  bool    okPressed = false;

  // If the group was found
  if ( group == 0 )
  {
    kDebug() << "Couldn't find a matching group.";
    return;
  }

  // If the group is not a special group...
  if(group->isSpecialGroup())
  {
    // Show a message that it can't be deleted
    KMessageBox::error( 0, i18n("This is a special group and can't be changed.") );
    return;
  }
  else
  {
    // Otherwise, get the new group name..
    newGroupName = group->getName();

    // Launch a dialog to get a new group name
  newGroupName = KInputDialog::getText(i18n("Rename group"),
                                       i18n("Enter the new name for this group:"),
                                       newGroupName, &okPressed, this);

    if(okPressed)
    {
      // Request that the group be renamed
      msnNotificationConnection_->renameGroup(groupId, newGroupName);
    }
  }
}


// Opens the transfer manager
void KMess::showTransferWindow()
{
  TransferWindow *transferWindow = TransferWindow::getInstance();
  transferWindow->show();
}


// Autologin with the first account that has autologin enabled
void KMess::checkAutologin(QString handle)
{
  Account *loginAccount = 0;

  // If no handle was given...
  if ( handle.isEmpty() )
  {
#ifdef KMESSDEBUG_KMESS
    kDebug() << "No handle was given, find an account that uses autologin.";
#endif

    const QList<Account*> accountsList = AccountsManager::instance()->getAccounts();
    foreach( Account *account, accountsList )
    {
      // For completeness, also test for guest accounts.
      if( account->isGuestAccount() )
      {
        continue;
      }

#ifdef KMESSDEBUG_KMESS
      kDebug() << "Account " << account->getHandle()
                << " useAutologin = " << account->getUseAutologin() << "." << endl;
#endif
      if ( account->getUseAutologin() )
      {
        loginAccount = account;
        break;
      }
    }
  }
  else
  {
#ifdef KMESSDEBUG_KMESS
  kDebug() << "finding account: " << handle << ".";
#endif
    loginAccount = AccountsManager::instance()->getAccountByHandle(handle);
  }

  bool doAutoLogin = ( loginAccount != 0 );

  // Wait for the user password if needed
  AccountsManager::instance()->initializePasswordManager( doAutoLogin );

  if( doAutoLogin )
  {
    // Set the info about login
    initialView_->startConnecting( loginAccount->getHandle() );
    connectWithAccount( loginAccount );
  }
}



// A connection has been made with the notification server.
void KMess::connected()
{
  // Enable/disable the menus
  enableMenus( true );

  // Show the status bar now that we're connected
  if( showStatusBar_->isChecked() )
  {
    statusBar()->show();
  }

  // Set up the toggle and view mode tools to match the account settings
  showAllowedAction_->setChecked( currentAccount_->getShowAllowedContacts() );
  showOfflineAction_->setChecked( currentAccount_->getShowOfflineContacts() );
  showRemovedAction_->setChecked( currentAccount_->getShowRemovedContacts() );

  // Set the pictures mode
  int newMode = 0;
  switch( currentAccount_->getListPictureViewMode() )
  {
    case 0:
      break;
    case 22:
      newMode = 1;
      break;
    case 32:
      newMode = 2;
      break;
    case 40:
      newMode = 3;
      break;
    case 48:
      newMode = 4;
      break;
    case 64:
      newMode = 5;
      break;

    default:
      break;
  }

  listPictureViewMode_->setCurrentItem(  newMode  );

  if ( currentAccount_->getShowContactsByGroup() )
  {
    viewMode_->setCurrentItem( 0 );
  }
  else
  {
    viewMode_->setCurrentItem( 1 );
  }
  // Show the connected message
  statusMessage( i18n("Connected"), false );

  // The connection was successful, this means the supplied
  // email and password were correct: save the settings now
  saveProperties();

  // Notify ChatMaster we're connected
  chatMaster_->connected();

  // Also read the account settings
//   msnNotificationConnection_->readProperties( config );

  // Load and show the contact list
  switchViewToContactList();

  // Set the caption
  setCaptionToUser();
}



// Connect to the server with the given account
void KMess::connectWithAccount(Account *account)
{
  if(KMESS_NULL(account)) return;

#ifdef KMESSDEBUG_KMESS
  kDebug() << account->getHandle() << ".";
#endif

  // First disconnect, to unregister all contacts
  if(msnNotificationConnection_->isConnected())
  {
    msnNotificationConnection_->closeConnection();
  }

  // Update the main view's controls to match this account's details.
  changedStatus( account );

  // Copy the account to the current account
#ifdef KMESSDEBUG_KMESS
  kDebug() << "copying account data for '" << account->getHandle() << "' to the CurrentAccount object.";
#endif
  currentAccount_->copyAccount( account );

  // Allow users to cancel the login attempt
  disconnect_->setEnabled( true );

  // Connect to the server.
  msnNotificationConnection_->openConnection();
}



// Connect to the server with the given account, possibly temporary or new.
void KMess::connectWithAccount(QString handle, QString password, bool rememberAccount, bool rememberPassword, Status initialStatus )
{
  // Try to find an existing account
  Account *account = AccountsManager::instance()->getAccountByHandle(handle);
  if( account != 0 )
  {
    // Found an existing account.
    // The user may enter a new password at the login dialog.
    // When the login succeeds, save the new temporary password as new account password.
    account->setTemporaryPassword( password );
    account->setSavePassword( rememberPassword );
    account->setInitialStatus( initialStatus );
    connectWithAccount( account );
    return;
  }

  // No account found, create a new account.
  account = new Account();
  account->setLoginInformation( handle, handle, password );
  account->setGuestAccount( ! rememberAccount );
  account->setSavePassword( rememberPassword );
  account->setInitialStatus( initialStatus );
  AccountsManager::instance()->addAccount( account );

  // Connect
  connectWithAccount( account );
}



// Create the program's default directories in .kde/share/apps/
bool KMess::createDirectories()
{
#ifdef KMESSDEBUG_KMESS
  kDebug() << "Creating default directories.";
#endif
  KStandardDirs *dirs   = KGlobal::dirs();
  QString        localKdeDir;
  QDir           appsDir, kmessDir;

  localKdeDir = dirs->localkdedir();
#ifdef KMESSDEBUG_KMESS
  kDebug() << "Local KDE dir is " << localKdeDir << ".";
#endif
  appsDir.setPath( localKdeDir + "/share/apps" );

  if ( appsDir.exists() )
  {
    kmessDir.setPath( appsDir.absolutePath() + "/kmess" );
#ifdef KMESSDEBUG_KMESS
    kDebug() << "kmess dir should be at " << kmessDir.absolutePath() << ".";
#endif
    if ( !kmessDir.exists() )
    {
#ifdef KMESSDEBUG_KMESS
      kDebug() << "Creating kmess dir.";
#endif
      appsDir.mkdir( kmessDir.absolutePath() );
    }
  }
#ifdef KMESSDEBUG_KMESS
  else
  {
    kDebug() << "" << appsDir.absolutePath() << " doesn't exist!";
  }
#endif
#ifdef KMESSTEST
  KMESS_ASSERT( kmessDir.exists() );
#endif

  return true;
}



// "Add new account" has been selected from the menu.
void KMess::createNewAccount()
{
  AccountsManager::instance()->showAccountSettings();
}



// The current account changed its name, so set the corresponding account's name
void KMess::currentAccountChangedName()
{
  Account *account = AccountsManager::instance()->getAccountByHandle( currentAccount_->getHandle() );
  if( account != 0 )
  {
    // Set the account's name to the current account's name
    account->setFriendlyName( currentAccount_->getFriendlyName( STRING_ORIGINAL ) );
  }

  // Set the caption to the account's new name
  setCaptionToUser();
}



// Delete the given account from the UI
void KMess::slotAccountDeleted( Account *account )
{
  AccountAction *accountAction;
  if( KMESS_NULL(account) ) return;

  // Remove it from the initial screen's list
  if( initialView_ )
  {
    initialView_->deleteAccount( account );
  }

  // Remove from connect menu.
  accountAction = connectMenuItems_[ account ];
  if( ! KMESS_NULL(accountAction) )
  {
    connectActionMenu_->removeAction( accountAction );
    delete connectMenuItems_.take( account );
  }

  // Remove from settings menu.
  // This can't be done from AccountAction, because the menu item should
  // remain accessable when you're logged in.
  accountAction = settingsMenuItems_[ account ];
  if( ! KMESS_NULL(accountAction) )
  {
    settingsActionMenu_->removeAction(accountAction);
    delete settingsMenuItems_.take( account );
  }
}



// Disconnect was selected from the menu.
void KMess::disconnectClicked()
{
#ifdef KMESSTEST
  KMESS_ASSERT( msnNotificationConnection_ != 0 );
#endif
#ifdef KMESSDEBUG_KMESSINTERFACE
  kDebug() << "Disconnecting.";
#endif
  msnNotificationConnection_->closeConnection();
}



// The program is not connected (initially) or no longer connected (after a disconnect) to the notification server.
void KMess::disconnected()
{
#ifdef KMESSTEST
  KMESS_ASSERT( chatMaster_ != 0 );
#endif
#ifdef KMESSDEBUG_KMESS
  kDebug() << "Disconnected.";
#endif

  // this method is also called when initializing the GUI, avoid mangling the settings at that point.
  if( initialized_ )
  {
    // ContactList is saved by the MsnNotificationConnection already.
    // A nice alternative would be currentAccount_->disconnected() saving the list,
    // however CurrentAccount has a const reference to the list
    // because MsnNotificationConnection maintains the list instead.

    // Save the account. If it is not found, it's the initial disconnect call.
    Account *account = AccountsManager::instance()->getAccountByHandle( currentAccount_->getHandle() );
    if( account )
    {
      if( account->isDeleted() )
      {
#ifdef KMESSDEBUG_KMESS
        kDebug() << "Account" << account->getHandle() << "was deleted, removing menu items.";
#endif
        slotAccountDeleted( account );
      }
      else
      {
#ifdef KMESSDEBUG_KMESS
        kDebug() << "Saving the CurrentAccount.";
#endif

        // Transfer account settings that are made in the GUI.
        account->copyAccountUISettings( currentAccount_ );
      }
    }
  }

  // Update the GUI
  disconnect_->setEnabled( false );
  enableMenus(false);
  statusMessage( i18n("Disconnected"), true );

  // Hide the status bar now that we're not connected
  if( showStatusBar_->isChecked() )
  {
    statusBar()->hide();
  }

  // Inform the chat master that we're disconnected
  chatMaster_->disconnected();

  // Update the current account
  currentAccount_->setStatus( STATUS_OFFLINE );
  currentAccount_->setAutoreply( false );
#ifdef KMESSDEBUG_KMESS
  kDebug() << "Swap views.";
#endif

  // Show the screen where to insert login details to log back in
  switchViewToInitialScreen();

#ifdef KMESSDEBUG_KMESS
  kDebug() << "Blank the caption.";
#endif

  // Set the caption
  setCaption( QString::null );
}



// Initialize the class
bool KMess::initialize()
{
  if ( initialized_ )
  {
    kDebug() << "Already initialized.";
    return false;
  }

#ifdef KMESSDEBUG_KMESS
  kDebug() << "initializing main class.";
#endif

  // Initialize the Rich Text parser engine
  RichTextParser::initialize();

  // Create current account instance
  currentAccount_ = CurrentAccount::instance();
  if ( currentAccount_ == 0 )
  {
    kDebug() << "Couldn't get a pointer to the instance of the current account.";
    return false;
  }
  if ( !KMessInterface::initialize() )
  {
    kDebug() << "Couldn't initialize ancestor.";
    return false;
  }
  if ( !initMsnNotificationConnection() )
  {
    kDebug() << "Couldn't initialize the MsnNotificationConnection class.";
    return false;
  }
  if ( !initIdleTimer() )
  {
    kDebug() << "Couldn't initialize the IdleTimer class.";
    return false;
  }
  if ( !initChatMaster() )
  {
    kDebug() << "Couldn't initialize the Chat master.";
    return false;
  }
  if ( !initNotifications() )
  {
    kDebug() << "Couldn't initialize the notifications.";
    return false;
  }
  if ( !initNowListening() )
  {
    kDebug() << "Couldn't initialize the now listening support.";
    return false;
  }
  if ( !initEmoticonManager() )
  {
    kDebug() << "Couldn't initialize the emoticon manager.";
    return false;
  }
  if ( !createDirectories() )
  {
    kDebug() << "Couldn't create local KDE directories.";
    return false;
  }

  // Connect current account signals for Now Listening
  connect( currentAccount_, SIGNAL(            accountInvisible() ),
           this,            SLOT  ( changedNowListeningSettings() ));
  connect( currentAccount_, SIGNAL(              accountOffline() ),
           this,            SLOT  ( changedNowListeningSettings() ));
  connect( currentAccount_, SIGNAL(               accountOnline() ),
           this,            SLOT  ( changedNowListeningSettings() ));

  // Connect current account
  connect( currentAccount_, SIGNAL(         changedFriendlyName() ),
           this,            SLOT  (   currentAccountChangedName() ));
  connect( currentAccount_, SIGNAL( changedNowListeningSettings() ),
           this,            SLOT  ( changedNowListeningSettings() ));
  connect( currentAccount_, SIGNAL(changedNotificationsSettings() ),
           this,            SLOT  (changedNotificationsSettings() ));
  connect( currentAccount_, SIGNAL(               changedStatus() ),
           this,            SLOT  (               changedStatus() ));
  connect( currentAccount_, SIGNAL(            changedMsnObject() ),
           msnNotificationConnection_, SLOT(   changedMsnObject() ));
  connect( currentAccount_, SIGNAL(     changedEmoticonSettings() ),
           EmoticonManager::instance(),SLOT(slotChangedEmoticonSettings() ) );

  // Connect the Account Changed signal to update the GUI
  AccountsManager *accountsManager = AccountsManager::instance();
  connect( accountsManager, SIGNAL(       accountAdded(Account*)                 ),
           this,            SLOT  (   slotAccountAdded(Account*)                 ) );
  connect( accountsManager, SIGNAL(     accountChanged(Account*,QString,QString) ),
           this,            SLOT  ( slotAccountChanged(Account*,QString,QString) ) );
  connect( accountsManager, SIGNAL(     accountDeleted(Account*)                 ),
           this,            SLOT  ( slotAccountDeleted(Account*)                 ) );

  // Read the accounts and properties
  readProperties();

  // Start disconnected.
  // Added debug here because this has a potential to cause problems..
  // It would be better if each class got things right from it's construtor or initialize method.
#ifdef KMESSDEBUG_KMESS
  kDebug() << "faking disconnect signal.";
#endif
  disconnected();

  // Setting up DBus
  dbus_ = new KMessDBus( this );

  // All done, mark as initialized
#ifdef KMESSDEBUG_KMESS
  kDebug() << "main class is initialized.";
#endif

  initialized_ = true;
  return true;
}



// Initialize the chat master
bool KMess::initChatMaster()
{
#ifdef KMESSTEST
  KMESS_ASSERT( chatMaster_ == 0 );
#endif
  if ( msnNotificationConnection_ == 0 )
  {
    kDebug() << "The notification connection must be initialized before the chat master.";
    return false;
  }
  chatMaster_ = new ChatMaster( this );
  if ( chatMaster_ == 0 )
  {
    kDebug() << "Couldn't create the chat master.";
    return false;
  }
  if ( !chatMaster_->initialize() )
  {
    kDebug() << "Couldn't initialize the chat master.";
    return false;
  }

  // Make the chat master's connections
  connect( chatMaster_,                SIGNAL( requestSwitchboard(QString,ChatInformation::ConnectionType) ),
           msnNotificationConnection_, SLOT  ( requestSwitchboard(QString,ChatInformation::ConnectionType) ));
  connect( chatMaster_,                SIGNAL(       blockContact(QString) ),
           msnNotificationConnection_, SLOT  (       blockContact(QString) ) );
  connect( chatMaster_,                SIGNAL(     unblockContact(QString) ),
           msnNotificationConnection_, SLOT  (     unblockContact(QString) ) );
  connect( chatMaster_,                SIGNAL(         addContact(QString) ),
           msnNotificationConnection_, SLOT  (      addNewContact(QString) ) );
  connect( chatMaster_,                SIGNAL(      removeContact(QString,bool) ),
           msnNotificationConnection_, SLOT  (      removeContact(QString,bool) ) );
  connect( chatMaster_,                SIGNAL(       allowContact(QString) ),
           msnNotificationConnection_, SLOT  (       allowContact(QString) ) );

  connect( msnNotificationConnection_, SIGNAL(   startSwitchboard(const ChatInformation&) ),
           chatMaster_,                SLOT  (   startSwitchboard(const ChatInformation&) ) );
  connect( msnNotificationConnection_, SIGNAL(     offlineMessage(const ChatMessage&) ),
           chatMaster_,                SLOT  ( showSpecialMessage(const ChatMessage&) ));
  connect( msnNotificationConnection_, SIGNAL(           pingSent() ),
           chatMaster_,                SLOT  (        timedUpdate() ));

#ifdef KMESSTEST
  KMESS_ASSERT( chatMaster_ != 0 );
#endif
  return true;
}



// Initialize the emoticon manager
bool KMess::initEmoticonManager()
{
  if ( msnNotificationConnection_ == 0 )
  {
    kDebug() << "MsnNotificationConnection must be initialized before EmoticonManager";
    return false;
  }

  EmoticonManager *manager = EmoticonManager::instance();

  connect( msnNotificationConnection_, SIGNAL(     connected() ),
           manager,                    SLOT  (     connected() ) );
  connect( msnNotificationConnection_, SIGNAL(  disconnected() ),
           manager,                    SLOT  (  disconnected() ) );

  return true;
}



// Initialize the idle timer
bool KMess::initIdleTimer()
{
#ifdef KMESSTEST
  KMESS_ASSERT( idleTimer_ == 0 );
#endif
  idleTimer_ = new IdleTimer();
  if ( idleTimer_ == 0 )
  {
    kDebug() << "Couldn't create the idle timer.";
    return false;
  }
  // Connect the timer to signal when the user is away.
  connect( idleTimer_, SIGNAL( timeout()       ),
           this,       SLOT  ( userIsIdle() )    );
  connect( idleTimer_, SIGNAL( activity()      ),
           this,       SLOT  ( userIsNotIdle() ) );
#ifdef KMESSTEST
  KMESS_ASSERT( idleTimer_ != 0 );
#endif
  return true;
}



// Initialize the MSN notification connection
bool KMess::initMsnNotificationConnection()
{
  bool initialized;
  
  msnNotificationConnection_ = new MsnNotificationConnection();

  initialized = msnNotificationConnection_->initialize();
  if ( !initialized )
  {
    kDebug() << "Couldn't initialize MsnNotificationConnection.";
    return false;
  }

  // Connect the signals
  connect( msnNotificationConnection_, SIGNAL(                  connected()               ),   // Connected to msn
           this,                       SLOT  (                  connected()               ) );
  connect( msnNotificationConnection_, SIGNAL(               disconnected()               ),   // Disconnected from msn
           this,                       SLOT  (               disconnected()               ) );
  connect( msnNotificationConnection_, SIGNAL(           contactAddedUser(const Contact*) ),   // Contact added user
           this,                       SLOT  ( showContactAddedUserDialog(const Contact*) ) );
  connect( msnNotificationConnection_, SIGNAL(              statusMessage(QString,bool)   ),   // Display a status message
           this,                       SLOT  (              statusMessage(QString,bool)   ) );

  return true;
}



// Initialize notification objects
bool KMess::initNotifications()
{
#ifdef KMESSTEST
  KMESS_ASSERT( currentAccount_             != 0 );
  KMESS_ASSERT( systemTrayWidget_           != 0 );
  KMESS_ASSERT( msnNotificationConnection_  != 0 );

  KMESS_ASSERT( chatNotification_           == 0 );
  KMESS_ASSERT( contactStatusNotification_  == 0 );
  KMESS_ASSERT( newEmailNotification_       == 0 );
#endif

  const ContactList *contactList;

  if ( systemTrayWidget_ == 0 )
  {
    kDebug() << "The system tray widget must be initialized before the notifications!";
    return false;
  }
  if ( msnNotificationConnection_ == 0 )
  {
    kDebug() << "The notification connection must be initialized before the notifications!";
    return false;
  }


  // Add check whether the 'eventsrc' file can be found.
  // The warning is a bit obtruisive, but should help to user to fix the problem.
  if( KGlobal::dirs()->findResource( "data", "kmess/kmess.notifyrc" ).isNull() )
  {
    QString dirs = KGlobal::dirs()->findDirs( "data", QString::null ).join("/kmess<br>- ");
    if( dirs.isEmpty() )
    {
      dirs = "(none)";
    }
    else
    {
      dirs = "<br>- " + dirs + "/kmess";
      dirs = dirs.replace( "//kmess", "/kmess" );  // possibly added by /kmess suffix.
    }

    // Always show warning at the console.
    kWarning() << "Sounds and notifications will not be available, the file 'kmess/kmess.notifyrc' could not be found.";

    // Show the message in the GUI as well.
    // Allow to user to choose "don't show this message again", in case he/she doesn't know how to fix it.
    KMessageBox::information( this, i18n( "<qt><p>KMess will not be able to play sounds and notifications.</p>"
                                          "<p>The KMess file 'kmess.notifyrc' could not be found in one of the following folders: %1</p>"
                                          "<p>Please verify your installation, and copy the missing file to one of the listed folders.</p>"
                                          "</qt>", dirs ),
                                    i18n( "Error with notifications" ),
                                    "eventsrcNotFound" );
  }


  contactList = currentAccount_->getContactList();
  notificationManager_ = NotificationManager::instance();
  notificationManager_->setTrayObject( systemTrayWidget_ );

  // Create the notifiers
  chatNotification_          = new ChatNotification         ( notificationManager_ );
  contactStatusNotification_ = new ContactStatusNotification( notificationManager_ );
  newEmailNotification_      = new NewEmailNotification     ( notificationManager_ );

    // Connect the chat messages notification signals
  connect( chatMaster_,                 SIGNAL( newChatMessage(const ChatMessage&, Chat*)                   ),
           chatNotification_,           SLOT  (         notify(const ChatMessage&, Chat*)                   ) );
  connect( chatNotification_,           SIGNAL(      raiseChat(Chat*,bool)                                  ),
           chatMaster_,                 SLOT  (      raiseChat(Chat*,bool)                                  ) );

    // Connect the status change notification signals
  connect( contactList,                 SIGNAL(        contactOnline(Contact*,bool)                          ),
           contactStatusNotification_,  SLOT  (               notify(Contact*,bool)                          ) );
  connect( contactList,                 SIGNAL(       contactOffline(Contact*,bool)                          ),
           contactStatusNotification_,  SLOT  (               notify(Contact*,bool)                          ) );
  connect( contactList,                 SIGNAL( contactChangedStatus(Contact*,bool)                          ),
           contactStatusNotification_,  SLOT  (               notify(Contact*,bool)                          ) );

    // Connect the "Email notification" signals
  connect( msnNotificationConnection_,  SIGNAL( newEmail(QString, QString, bool, QString, QString, QString) ),
           newEmailNotification_,       SLOT  (   notify(QString, QString, bool, QString, QString, QString) ) );

  // Connect the "Chat request" signals
  connect( contactStatusNotification_,  SIGNAL(          startChat(QString,ChatInformation::ConnectionType) ),
           msnNotificationConnection_,  SLOT  ( requestSwitchboard(QString,ChatInformation::ConnectionType) ) );


  // If specified in the settings, connect the relative signals
  changedNotificationsSettings();


#ifdef KMESSTEST
  KMESS_ASSERT( chatNotification_          != 0 );
  KMESS_ASSERT( contactStatusNotification_ != 0 );
  KMESS_ASSERT( newEmailNotification_      != 0 );
#endif

  return true;
}



// Initialize the now listening support.
bool KMess::initNowListening()
{
  if ( msnNotificationConnection_ == 0 )
  {
    kDebug() << "The notification connection must be initialized before the now listening support!";
    return false;
  }

  // Create and connect client.
  nowListeningClient_ = new NowListeningClient();
  connect( nowListeningClient_, SIGNAL( changedSong( const QString&, const QString&, const QString&, bool ) ),
           this,                SLOT  ( changedSong( const QString&, const QString&, const QString&, bool ) ));
  
  return true;
}



// Check if the connection to the server is active
bool KMess::isConnected()
{
  if( KMESS_NULL(msnNotificationConnection_) )
  {
    return false;
  }

  return msnNotificationConnection_->isConnected();
}



// Read in account and other properties
void KMess::readProperties( const KConfigGroup &config )
{
  KConfigGroup group;

  // Choose the config group if it was not given
  if( &config == 0 )
  {
    group = KMessConfig::instance()->getGlobalConfig( "General" );
  }
  else
  {
    group = config;
  }

  KMessInterface::readProperties( group );

#ifdef KMESSDEBUG_KMESS
  kDebug() << "reading properties";
#endif

  AccountsManager *accountsManager = AccountsManager::instance();

  // Load the saved accounts from disk and create them
  accountsManager->readProperties();
}



// Save account and other properties
void KMess::saveProperties( KConfigGroup &config )
{
  KConfigGroup group;

  // Save the accounts' passwords
    AccountsManager::instance()->storeAccountsPasswords();

  // Choose the config group if it was not given
  if( &config == 0 )
  {
    group = KMessConfig::instance()->getGlobalConfig( "General" );
  }
  else
  {
    group = config;
  }

  // Save the UI configuration
  KMessInterface::saveProperties( group );

#ifdef KMESSDEBUG_KMESS
  kDebug() << "saving properties";
#endif

  // Have all the accounts save their properties
  const QList<Account*> accountsList = AccountsManager::instance()->getAccounts();
  foreach( Account *account, accountsList )
  {
    // Avoid saving guest accounts
    if( account->isGuestAccount() || account->isDeleted() ) 
    {
      continue;
    }

    // If the account is the current account...
    if ( account->getHandle() == currentAccount_->getHandle() )
    {
      // Move some settings from the current account to the account that are set by the user interface
#ifdef KMESSDEBUG_KMESS
      kDebug() << "copying global user-interface settings from CurrentAccount to the Account object.";
#endif
      account->copyAccountUISettings( currentAccount_ );
    }

    // Save properties of accounts
    account->saveProperties();
  }

  // Write data now!
  group.sync();
}



// Set the caption
void KMess::setCaptionToUser()
{
  if ( currentAccount_ != 0 )
  {
    // Get the friendly name, but avoid newlines which cause problems to the titlebar layout
    QString friendlyName( currentAccount_->getFriendlyName().replace( "\n", " " ) );

    // Set the caption
    setPlainCaption( i18nc( "Main window caption", "KMess - %1", friendlyName ) );
  }
}



// Show the settings dialog for a given account
void KMess::showSettingsForAccount( Account *account )
{
  AccountsManager::instance()->showAccountSettings( account );
}



// Show the settings dialog for the current account.
void KMess::showSettingsForCurrentAccount()
{
  // Use the Account object so it's consistent with the other AccountAction events.
  Account *accountData = AccountsManager::instance()->getAccountByHandle( CurrentAccount::instance()->getHandle() );
  if( KMESS_NULL(accountData) ) return;
  showSettingsForAccount( accountData );
}



// Show the user's MSN profile.
void KMess::showUserProfile()
{
  if( currentAccount_ == 0 )
  {
    return;
  }

  KMessShared::openProfile( currentAccount_->getHandle() );
}



// Switch the view to the contact list
void KMess::switchViewToContactList()
{
  // If the view was already set, reload it
  if( view_ )
  {
    delete view_;
  }

#ifdef KMESSTEST
  KMESS_ASSERT( msnNotificationConnection_ != 0 );
#endif

  // Create and initialize the view's widgets
  view_ = new KMessView( this );
  view_->initialize( msnNotificationConnection_->getContactListModel() );
  view_->show();

  // Connect other signals to msnNotificationConnection
  connect( view_,                      SIGNAL(    addContact(QString)                   ),   // Add a contact
           msnNotificationConnection_, SLOT  (addExistingContact(QString)               ) );
  connect( view_,                      SIGNAL(  allowContact(QString)                   ),   // Allow the contact
           msnNotificationConnection_, SLOT  (  allowContact(QString)                   ) );
  connect( view_,                      SIGNAL(  blockContact(QString)                   ),   // Block the contact
           msnNotificationConnection_, SLOT  (  blockContact(QString)                   ) );
  connect( view_,                      SIGNAL( changePersonalMessage(QString)           ),   // Change the personal message.
           msnNotificationConnection_, SLOT  ( changePersonalMessage(QString)           ) );
  connect( view_,                      SIGNAL(                   changeStatus(int)      ),
           this,                       SLOT  (                  changeStatus(int )      ) ); 
  connect( view_,                      SIGNAL(       copyContact(QString, QString)      ),
           msnNotificationConnection_, SLOT  ( addContactToGroup(QString, QString)      ) );
  connect( view_,                      SIGNAL( moveContact(QString, QString, QString)   ),   // Move the contact
           msnNotificationConnection_, SLOT  ( moveContact(QString, QString, QString)   ) );
  connect( view_,                      SIGNAL( showSettings()                           ),   // Show settings
           this,                       SLOT  ( showSettingsForCurrentAccount()          ) );
  connect( view_,                      SIGNAL(requestChat( QString )                    ),   // Start a chat
           chatMaster_,                SLOT  (requestChat( QString )                    ) );
  connect( view_,                      SIGNAL( removeContact(QString)                   ),   // Remove the contact
           this,                       SLOT  ( showRemoveContactDialog(QString)         ) );
  connect( view_,                      SIGNAL( removeContactFromGroup(QString, QString) ),
           msnNotificationConnection_, SLOT  ( removeContactFromGroup(QString, QString) ) );
  connect( view_,                      SIGNAL(   removeGroup(QString)                   ),   // Remove the group
           this,                       SLOT  ( showRemoveGroupDialog(QString)           ) );
  connect( view_,                      SIGNAL(   renameGroup(QString)                   ),   // Rename the group
           this,                       SLOT  ( showRenameGroupDialog(QString)           ) );
  connect( view_,                      SIGNAL( unblockContact(QString)                  ),   // Unblock the contact
           msnNotificationConnection_, SLOT  ( unblockContact(QString)                  ) );
  // Set it as the application's main widget
  setCentralWidget( view_ );

  // Make sure the other views' pointers are zero: setCentralWidget deletes the old central widget
  initialView_ = 0;
}



// Switch the view to the initial login screen
void KMess::switchViewToInitialScreen()
{
  // If the view was already set, reload it
  if( initialView_ )
  {
    delete initialView_;
  }

  // Create and set up the new widget
  initialView_ = new InitialView( this );
  initialView_->updateView();
  initialView_->show();

  // Connect it so it triggers the login process
  connect( initialView_, SIGNAL( connectWithAccount(QString,QString,bool,bool,Status) ),
           this,         SLOT  ( connectWithAccount(QString,QString,bool,bool,Status) ) );
  connect( initialView_, SIGNAL(                                  disconnectClicked() ),
           this,         SLOT  (                                  disconnectClicked() ) );

  connect( msnNotificationConnection_, SIGNAL( statusMessage(QString,bool) ),
           initialView_,               SLOT  ( statusMessage(QString)      ) );


  // Set it as the window's main widget
  setCentralWidget( initialView_ );

  // Make sure the other views' pointers are zero: setCentralWidget deletes the old central widget
  view_ = 0;
}



// The "show allowed contacts" menu item has been toggled.
void KMess::toggleShowAllowed(bool show)
{
  currentAccount_->setShowAllowedContacts( show );
}



// The "show search in contact list" menu item has been toggled.
void KMess::toggleShowSearchFrame(bool show)
{
  if( view_ != 0 )
  {
    view_->toggleShowSearchFrame( show );
  }
}


// The "show offline contacts" menu item has been toggled.
void KMess::toggleShowOffline(bool show)
{
  currentAccount_->setShowOfflineContacts( show );
}



// The "show removed contacts" menu item has been toggled.
void KMess::toggleShowRemoved(bool show)
{
  currentAccount_->setShowRemovedContacts( show );
}



// The user has gone idle
void KMess::userIsIdle()
{
#ifdef KMESSDEBUG_KMESS
    kDebug();
#endif
  if ( ( currentAccount_ != 0 ) && ( msnNotificationConnection_ != 0 ) )
  {
#ifdef KMESSDEBUG_KMESS
    kDebug() << "Current status is " << currentAccount_->getStatus() << ".";
#endif
    // Only change the state if the user is currently set as online
    if ( currentAccount_->getStatus() == STATUS_ONLINE )
    {
#ifdef KMESSDEBUG_KMESS
      kDebug() << "Change status to Idle.";
#endif
      // Request a status change to Idle
      msnNotificationConnection_->changeStatus( STATUS_IDLE );
      isIdleForced_ = false;
    }
  }
}



// The user is no longer idle
void KMess::userIsNotIdle()
{
#ifdef KMESSDEBUG_KMESS
    kDebug();
#endif
  if ( ( currentAccount_ != 0 ) && ( msnNotificationConnection_ != 0 ) )
  {
#ifdef KMESSDEBUG_KMESS
    kDebug() << "User is no longer idle, current status is " << currentAccount_->getStatus();
#endif
    // Only change the state if the user is currently set as idle AND the idle status was not forced by the user
    if ( currentAccount_->getStatus() == STATUS_IDLE && ! isIdleForced_ )
    {
#ifdef KMESSDEBUG_KMESS
      kDebug() << "Change status to Online";
#endif
      // Request a status change to Online
      msnNotificationConnection_->changeStatus( STATUS_ONLINE );
    }
  }
#ifdef KMESSDEBUG_KMESS
    kDebug() << "Done userIsNotIdle().";
#endif
}


#include "kmess.moc"
