/***************************************************************************
                          kmessapplication.cpp -  description
                             -------------------
    begin                : Wed 7 13 2005
    copyright            : (C) 2005 by Diederik van der Boor
    email                : vdboor --at-- codingdomain.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "kmessapplication.h"

#include "kmess.h"
#include "kmesstest.h"
#include "kmessdebug.h"
#include "utils/crashhandler.h"

#include "config-kmess.h"

#include <QStringList>
#include <QTimer>

#include <KCmdLineArgs>
#include <KIconLoader>
#include <KLocalizedString>
#include <KStandardDirs>


#if ( ENABLE_BINRELOC != 0 )
  // Required for autopackage support.
  // Handles detection of paths at run time.
  #include "binreloc.h"
#endif



/**
 * The application constructor
 */
KMessApplication::KMessApplication()
  : KApplication()
  , contactListWindow_(0)
  , quitSelected_(false)
{
  // KMessApplication is created from main.cpp.
  // It continues the initialisation of the application.

  // Install a message handler, so KMESS_ASSERT won't do a exit(1) or abort()
  qInstallMsgHandler( kmessDebugPrinter );

  // Install our crash handler.
  CrashHandler::activate();

  // Start remaining initialisation
  initializePaths();
  initializeMainWindow();

//   // Delay further initialisation until the event loop is active.
//   // This makes it easier to show messages boxes, etc..
//   QTimer::singleShot( 0, this, SLOT( slotEventLoopActive() ) );
}



/**
 * Destructor
 */
KMessApplication::~KMessApplication()
{
  // automatically destroyed
  contactListWindow_ = 0;
}



/**
 * Return the contact list window
 */
KMess* KMessApplication::getContactListWindow() const
{
  return contactListWindow_;
}


/**
 * Initialisation of the main window.
 */
void KMessApplication::initializeMainWindow()
{
  // Fetch the command line arguments
  KCmdLineArgs *args = KCmdLineArgs::parsedArgs();
  QString autologinHandle = args->getOption("autologin");

  // Create the main window and initialize it
  contactListWindow_ = new KMess( 0 );
  if( ! contactListWindow_->initialize() )
  {
    exit(1);
    return;
  }

  // Initialize KApplication
  setTopWidget( contactListWindow_ );

  // Apply our own quit policy.
  // Only quit when the user wanted to, or KDE is logging out.
  setQuitOnLastWindowClosed( false );
  connect( this, SIGNAL(lastWindowClosed()), this, SLOT(slotLastWindowClosed()) );
  connect( this, SIGNAL(aboutToQuit()),      this, SLOT(slotAboutToQuit()) );

#ifdef KMESSTEST
  // The debug build allows to run tests from the command line.
  QString testName = args->getOption("runtest");
  if( ! testName.isEmpty() )
  {
    KMessTest *kmessTest = new KMessTest( contactListWindow_ );
    kmessTest->runTest( testName );
    delete kmessTest;
    return;
  }
#endif


  // Avoid showing the main window is KMess is started with KDE.
  if( ! kapp->isSessionRestored() )
  {
    contactListWindow_->show();
  }

  // Start auto login
  contactListWindow_->checkAutologin( autologinHandle );
}



/**
 * Initialize additional paths
 */
void KMessApplication::initializePaths()
{
#if ( ENABLE_BINRELOC != 0 )
  // Register run-time detected paths for Autopackage support.
  // This also prevents the "square boxes images" problem from happening
  // if the --prefix of ./configure was incorrect.
  BrInitError brError;
  if( br_init( &brError ) == 0 && brError != BR_INIT_ERROR_DISABLED)
  {
    kWarning() << "BinReloc failed to initialize (error code " << brError << ")";
    kDebug()   << "Will fallback to hardcoded compile-time path (" << KMESS_PREFIX << ").";
  }

  // Detect run-time paths, fallback to KMESS_PREFIX of configure
  KGlobal::dirs()       -> addPrefix( br_find_prefix( KMESS_PREFIX ) );
  KIconLoader::global() -> addAppDir( br_find_data_dir( KMESS_PREFIX "/share" ) );
#else
  // Add compile time paths as fallback
  KGlobal::dirs()       -> addPrefix( KMESS_PREFIX );
  KIconLoader::global() -> addAppDir( KMESS_PREFIX "/share" );
#endif

  // Test whether the prefix is correct.
  if( KGlobal::dirs()->findResource( "appdata", "pics/kmesslogo.png" ).isNull() )
  {
    kWarning() << "KMess could not find resources in the search paths: "
               << KGlobal::dirs()->findDirs( "appdata", QString::null ).join(", ") << endl;
  }
}



/**
 * Called when QCoreApplication notifies the app is really quitting now.
 */
void KMessApplication::slotAboutToQuit()
{
  // No user-interaction is possible here anymore..!

#ifdef KMESSDEBUG_KMESSAPPLICATION
  kDebug() << "KMess is about to quit...";
#endif
}



/**
 * Start the main window.
 */
void KMessApplication::slotEventLoopActive()
{
}



/**
 * Called when the last window is closed.
 *
 * Also see: KMessInterface::queryClose().
 */
void KMessApplication::slotLastWindowClosed()
{
#ifdef KMESSTEST
  KMESS_ASSERT( quitOnLastWindowClosed() == false );
#endif

  if( quitSelected_ || sessionSaving() )
  {
#ifdef KMESSDEBUG_KMESSAPPLICATION
    kDebug() << "Last window closed, user really wants to quit now. Initiate quit() call.";
#endif

    quit();
  }
  else
  {
#ifdef KMESSDEBUG_KMESSAPPLICATION
    kDebug() << "User closed last window, but keep KMess running with all windows closed.";
#endif
  }
}



/**
 * Return true if quit was selected
 */
bool KMessApplication::quitSelected() const
{
  return quitSelected_;
}



/**
 * Tell the application that quit was selected
 */
void KMessApplication::setQuitSelected(bool quitSelected)
{
  quitSelected_ = quitSelected;
}


#include "kmessapplication.moc"
