/***************************************************************************
                          kmessview.cpp  -  description
                             -------------------
    begin                : Thu Jan 9 2003
    copyright            : (C) 2003 by Mike K. Bennett
                           (C) 2005 by Diederik van der Boor
    email                : mkb137b@hotmail.com
                           vdboor --at-- codingdomain.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "kmessview.h"

#include "contact/contact.h"
#include "contact/contactlist.h"
#include "contact/group.h"
#include "contact/msnstatus.h"
#include "dialogs/contactpropertiesdialog.h"
#include "network/specialgroups.h"
#include "utils/kmessshared.h"
#include "utils/richtextparser.h"
#include "contactlistviewdelegate.h"
#include "currentaccount.h"
#include "emoticonmanager.h"
#include "kmessdebug.h"

#include <QClipboard>
#include <QEvent>
#include <QHeaderView>
#include <QImage>
#include <QItemSelectionModel>
#include <QLabel>
#include <QLayout>
#include <QPalette>
#include <QPixmap>
#include <QPushButton>
#include <QSignalMapper>
#include <QStringList>
#include <QTextDocument>
#include <QToolTip>

#include <KAction>
#include <KActionMenu>
#include <KApplication>
#include <KGlobal>
#include <KIconLoader>
#include <KIconEffect>
#include <KLineEdit>
#include <KLocale>
#include <KMenu>
#include <KRun>
#include <KSqueezedTextLabel>
#include <KUrl>
#include <KUrlLabel>


#ifdef KMESSDEBUG_KMESSVIEW
  #define KMESSDEBUG_KMESSVIEW_GENERAL
//   #define KMESSDEBUG_KMESSVIEW_ITEM_VISIBILITY
#endif



// The constructor
KMessView::KMessView( QWidget *parent )
  : QWidget( parent ),
    Ui::KMessView(),
    contactCopyMapper_(0),
    contactMoveMapper_(0),
    currentAccount_(0),
    initialized_(false)
{
  setupUi( this );
  searchFrame_->hide();

  // Capture focus events without subclassing the widget from the .UI file
  personalMessageInput_->installEventFilter(this);
  currentAccountDisplayPic_->installEventFilter(this);
  contactListView_->installEventFilter(this);

  // Fake an event to configure the personal message input
  QEvent ev(QEvent::FocusOut);
  eventFilter(personalMessageInput_, &ev);
}



// The destructor
KMessView::~KMessView()
{
  // Delete the contact actions for copy/move to group
  qDeleteAll( groupCopyActionsList_ );
  qDeleteAll( groupMoveActionsList_ );
  qDeleteAll( groupCopyLinkActionsList_ );
  delete contactCopyMapper_;
  delete contactMoveMapper_;

#ifdef KMESSDEBUG_KMESSVIEW_GENERAL
  kDebug() << "DESTROYED.";
#endif
}



// The currently playing song was changed.
void KMessView::changedSong( const QString &artist, const QString &/*album*/, const QString &track, bool playing )
{
  if( playing )
  {
    // Show status line
    if( ! nowListeningFrame_->isVisible() )
    {
      nowListeningFrame_->setHidden(false);
      nowListeningFrame_->setMaximumHeight(40);
    }

    // Update label
    QString longTitle = artist + (artist.isEmpty() ? "" : " - ") + track;
    nowListeningLabel_->setText( i18n("Now listening to %1", longTitle) );
  }
  else
  {
    // Hide status line
    nowListeningFrame_->setHidden(true);
    nowListeningFrame_->setMaximumHeight(0);
  }
}



// Get the specified item's or the current item's data
ModelDataList KMessView::getItemData( const QModelIndex &index )
{
  QModelIndex     itemIndex = index;
  QModelIndexList selection = selectionModel_->selectedIndexes();

  // If the index is not valid, get the current selection
  if( ! itemIndex.isValid() && ! selection.isEmpty() )
  {
    // Only one index can be returned at once by the contact list (has single selection mode)
    itemIndex = selection.first();
  }

  // Do nothing if the index is not valid
  if( ! itemIndex.isValid() )
  {
    return ModelDataList();
  }

  // Access the index's data
  ContactListModelItem *item = static_cast<ContactListModelItem*>( itemIndex.internalPointer() );

  // Do nothing if the item is not of a valid type
  if(    item->getType() != ContactListModelItem::ItemGroup
      && item->getType() != ContactListModelItem::ItemContact )
  {
    return ModelDataList();
  }

  return item->data( itemIndex.column() ).toMap();
}



// Initialize the class
bool KMessView::initialize( QAbstractItemModel *viewModel )
{
  bool         success;
  KIconLoader *loader = KIconLoader::global();

  // Get a ref to the current account
  currentAccount_ = CurrentAccount::instance();

  if ( initialized_ )
  {
    kDebug() << "already initialized.";
    return false;
  }

  success = initContactListView( viewModel );
  if(! success)
  {
    kDebug() << "Couldn't initialize contact list view.";
    return false;
  }

  success = initContactPopup();
  if(! success)
  {
    kDebug() << "Couldn't initialize contact popup.";
    return false;
  }

  success = initGroupPopup();
  if(! success)
  {
    kDebug() << "Couldn't initialize group popup.";
    return false;
  }

  // Set an icon for the email label
  // Using "knotify" icon because it's available everywhere. "multimedia" is SUSE-specific.
  emailPixmapLabel_->setPixmap( loader->loadIcon( "mail-mark-unread", KIconLoader::NoGroup ) );
  nowListeningPixmapLabel_->setPixmap( loader->loadIcon( "speaker", KIconLoader::NoGroup ) );
  searchIcon_->setPixmap( loader->loadIcon( "edit-find-user", KIconLoader::Small ) );

  // Run the update functions with the current user data
  slotUpdateEmailDisplay();
  slotUpdateUserStatus();
  slotUpdateNoEmails();
  slotUpdateDisplayPicture();

  // Create the menu for status button
  QMenu *menu = new QMenu();

  QStringList states;
  states << i18n("Online")
         << i18n("Away")
         << i18n("Away with auto-reply")
         << i18n("Idle")
         << i18n("Be Right Back")
         << i18n("Busy")
         << i18n("Out to Lunch")
         << i18n("On the Phone")
         << i18n("Invisible");

  int i = 0;
  QAction *action;
  foreach( QString actionText, states )
  {
    // Set the informations for actions
    action = new QAction( this );
    action->setText( actionText );
    action->setData( i++ );
    menu  ->addAction( action );
  }

  // Set the menu into button and connect the signals
  statusButton_->setMenu( menu );
  connect( statusButton_, SIGNAL(                         clicked() ),
           statusButton_, SLOT(                          showMenu() ) );

#if QT_VERSION >= 0x040400
  connect( statusButton_, SIGNAL(               triggered(QAction*) ),
           this,          SLOT(   slotForwardChangeStatus(QAction*) ) );
#else
  connect( menu,          SIGNAL(               triggered(QAction*) ),
           this,          SLOT(   slotForwardChangeStatus(QAction*) ) );
#endif

  // Enable the labels
  onlineStatusLabel_->setEnabled(true);
  emailLabel_->setEnabled(true);

  // Hide the now listening until something is playing
  changedSong( QString::null, QString::null, QString::null, false );

  // Refresh the personal message text
  slotPersonalMessageLostFocus( true );


  // Connect signals for the current account
  connect( currentAccount_, SIGNAL(            changedMsnObject() ),
           this,            SLOT  (    slotUpdateDisplayPicture() ) );  // Changed users display pic
  connect( currentAccount_, SIGNAL(             changedNoEmails() ),    // Changed e-mail count
           this,            SLOT  (          slotUpdateNoEmails() ) );
  connect( currentAccount_, SIGNAL(               changedStatus() ),    // Changed status
           this,            SLOT  (        slotUpdateUserStatus() ) );
  connect( currentAccount_, SIGNAL(         changedFriendlyName() ),    // Changed name
           this,            SLOT  (        slotUpdateUserStatus() ) );
  connect( currentAccount_, SIGNAL( changedEmailDisplaySettings() ),    // Changed e-mail display settings
           this,            SLOT  (      slotUpdateEmailDisplay() ) );
  connect( currentAccount_, SIGNAL(      changedDisplaySettings() ),    // Changed group-display settings
           this,            SLOT  (   slotUpdateDisplaySettings() ) );
  connect( currentAccount_, SIGNAL(             changedViewMode() ),    // Changed the requested view mode
           this,            SLOT  (          slotUpdateViewMode() ) );

  // Connect signals for the contactlist
  const ContactList *contactList = currentAccount_->getContactList();
  connect( contactList, SIGNAL(            groupAdded(const Group*)  ),   // A group was added
           this,          SLOT( rebuildContactActions()              ) );
  connect( contactList, SIGNAL(          groupRemoved(const Group*)  ),   // A group was removed
           this,          SLOT( rebuildContactActions()              ) );

  // Connect signals for the history combobox and put the first string into it
  connect( contactList, SIGNAL(            contactOnline( Contact*,  bool ) ),
           this,          SLOT( slotContactChangedStatus( Contact *, bool ) ) );
  connect( contactList, SIGNAL(           contactOffline( Contact*,  bool ) ),
           this,          SLOT( slotContactChangedStatus( Contact *, bool ) ) );

  historyBox_->insertItem( 0, i18n( "[%1] Logged with %2", QTime::currentTime ().toString( "HH:mm:ss" )
                                                         , currentAccount_->getHandle() ) );

  // Connect the search edit line with slot for searching
  connect( searchEdit_, SIGNAL( textChanged ( QString ) ), this, SLOT( slotSearchContact( QString ) ) );
  return true;
}



// Called when a contact is now online/offline
void KMessView::slotContactChangedStatus( Contact *contact, bool showNotify )
{
  // One contact is now online/offline, put it in history combobox
  if( showNotify )
  {
    QString item;
    QString friendlyname = contact->getFriendlyName();

    if( contact->isOnline() )
    {
      item = i18n( "[%1] %2 goes online", QTime::currentTime ().toString( "HH:mm:ss" )
                                        , Qt::escape( friendlyname )  );
    }
    else
    {
      item = i18n( "[%1] %2 goes offline", QTime::currentTime ().toString( "HH:mm:ss" )
                                         ,Qt::escape( friendlyname  ) );
    }

    // Remove oldest item if the count is >= 500 elements
    if( historyBox_->count() >= 500 )
    {
      historyBox_->removeItem( historyBox_->count() - 1 );
    }

    historyBox_->insertItem( 0, item );
    historyBox_->setCurrentIndex ( 0 );
  }
}



// Check whether an item will be hidden in the list
bool KMessView::isItemHidden( ModelDataList itemData )
{
  bool isHidden = false;

  // Read the settings
  bool showAllowed = currentAccount_->getShowAllowedContacts();
  bool showRemoved = currentAccount_->getShowRemovedContacts();
  bool showByGroup = currentAccount_->getShowContactsByGroup();
  bool showOffline = currentAccount_->getShowOfflineContacts();

  if( itemData.isEmpty() )
  {
#ifdef KMESSDEBUG_KMESSVIEW_ITEM_VISIBILITY
    kDebug() << "Skipping empty item" << itemData;
#endif
    return true;
  }

  switch( itemData[ "type" ].toInt() )
  {
    case ContactListModelItem::ItemContact:

      // If the searchContact string is not empty show the contacts that match to string
      if( ! searchContact_.isEmpty() )
      {
        if( ! itemData[ "friendly" ].toString().contains( searchContact_, Qt::CaseInsensitive ) &&
            ! itemData[ "handle"   ].toString().contains( searchContact_, Qt::CaseInsensitive ) )
        {
          isHidden = true;
        }
        break;
      }

      // Contacts in special groups are always visible. Their group may be not, though
      if( ! itemData[ "isInSpecialGroup" ].toBool() )
      {
        // Hide contacts without group when showing by online/offline
        if( ( ! showByGroup ) && itemData[ "group" ].toString().isEmpty() )
        {
          isHidden = true;
        }
        else
        {
          isHidden = ( ( (Status) itemData[ "status" ].toInt() ) == STATUS_OFFLINE && ! showOffline );
        }
      }

#ifdef KMESSDEBUG_KMESSVIEW_ITEM_VISIBILITY
      kDebug() << "Visibility check: contact" << itemData[ "handle" ].toString() << "is"
                                              << ( isHidden ? "hidden" : "visible" );
#endif
      break;

    case ContactListModelItem::ItemGroup:

      if(      itemData[ "id" ] == SpecialGroups::ALLOWED )
      {
        isHidden = ( ! showAllowed );
      }
      else if( itemData[ "id" ] == SpecialGroups::REMOVED )
      {
        isHidden = ( ! showRemoved );
      }
      else if( itemData[ "id" ] == SpecialGroups::ONLINE )
      {
        isHidden = ( showByGroup );
      }
      else if( itemData[ "id" ] == SpecialGroups::OFFLINE )
      {
        isHidden = ( showByGroup || ! showOffline );
      }
      else
      {
        isHidden = ( ! showByGroup );
      }

#ifdef KMESSDEBUG_KMESSVIEW_ITEM_VISIBILITY
      kDebug() << "Visibility check: group" << itemData[ "name" ].toString() << "is"
                                            << ( isHidden ? "hidden" : "visible" );
#endif
      break;

    default:
      kWarning() << "Unknown item:" << itemData;
      break;
  }

  return isHidden;
}



// The "show search in contact list" menu item has been toggled.
void KMessView::toggleShowSearchFrame(bool show)
{
  searchFrame_->setShown( show );
  searchEdit_->setFocus();
}



// The personal status message received an event.
bool KMessView::eventFilter(QObject *obj, QEvent *event)
{
  if( obj != personalMessageInput_
  &&  obj != currentAccountDisplayPic_
  &&  obj != contactListView_ )
  {
#ifdef KMESSDEBUG_KMESSVIEW_GENERAL
    kWarning() << "Received event '" << event->type() << "' from object '" << obj->objectName() << "'.";
#endif
    return false;
  }

  QEvent::Type eventType = event->type();

  if( obj == currentAccountDisplayPic_ )
  {
    if( eventType == QEvent::Enter )
    {
      QImage glowingImage( currentAccount_->getPicturePath() );
      KIconEffect::toGamma( glowingImage, 0.8f );
      currentAccountDisplayPic_->setPixmap( QPixmap::fromImage( glowingImage ) );
    }
    else if( eventType == QEvent::Leave )
    {
      currentAccountDisplayPic_->setPixmap( QPixmap( currentAccount_->getPicturePath() ) );
    }
    else if( eventType == QEvent::MouseButtonRelease )
    {
      emit showSettings();
    }
  }
  else if( obj == contactListView_ )
  {
    if( eventType == QEvent::ToolTip )
    {
      showToolTip( QCursor::pos() );
    }
  }
  else if( obj == personalMessageInput_ )
  {
    if( eventType == QEvent::Enter )
    {
      // Mouse enters widget's boundaries
      // Just a bit different from the original client, already change the background line web-forms also do.
      personalMessageInput_->setStyleSheet( "background-color: base;" );
    }
    else if( eventType == QEvent::Leave )
    {
      // Mouse leaves the widget.
      if( ! personalMessageInput_->hasFocus() )
      {
      personalMessageInput_->setStyleSheet( "background-color: transparent;" );
      }
    }
    else if( eventType == QEvent::FocusIn )
    {
      // Personal status mesage got the focus
      personalMessageInput_->setStyleSheet( "background-color: base;" );

      // Clear if there is no status message, but a placeholder label
      if( currentAccount_->getPersonalMessage().isEmpty() )
      {
        personalMessageInput_->clear();
      }
      else
      {
        // Select all text for editing
        personalMessageInput_->selectAll();
      }
    }
    else if( event->type() == QEvent::FocusOut )
    {
      // Lost the focus
      personalMessageInput_->setStyleSheet( "background-color: transparent;" );
    }
  }

  return false;  // don't stop processing.
}



// Initialize the contact list view
bool KMessView::initContactListView( QAbstractItemModel *viewModel )
{
  KIconLoader *loader = KIconLoader::global();

  // Create a selection model to get data about the list selections
  selectionModel_ = new QItemSelectionModel( viewModel );

  // Set contact list view's properties
#if QT_VERSION >= 0x040400
  contactListView_->setHeaderHidden( true );
#else
  contactListView_->header()->hide();
#endif

  contactListView_->setItemDelegate( new ContactListViewDelegate( contactListView_ ) );
  contactListView_->setSelectionMode( QAbstractItemView::SingleSelection );

  // Connect the list to the model and selection model
  contactListView_->setModel( viewModel );
  contactListView_->setSelectionModel( selectionModel_ );

  // Set a background pixmap
  QString backgroundPicture = loader->iconPath( "background", KIconLoader::User );
  contactListView_->setStyleSheet( "QTreeView { background-image: url(" + backgroundPicture + "); "
                                   "            background-color: palette(alternate-base); }"
                                   "QTreeView::branch:has-siblings:adjoins-item { border-image: url(empty.png) 0; }"
                                   "QTreeView::branch:!has-children:!has-siblings:adjoins-item { border-image: url(empty.png) 0; }" );

  // Connect the model update signals
  connect( viewModel, SIGNAL(               dataChanged(const QModelIndex&,const QModelIndex&) ),
           this,        SLOT(            slotUpdateItem(const QModelIndex&,const QModelIndex&) ) );
  connect( viewModel, SIGNAL(             layoutChanged()                                      ),
           this,        SLOT( slotUpdateDisplaySettings()                                      ) );

  // Connect the list's signals
  connect( contactListView_, SIGNAL(               clicked(const QModelIndex&) ),
           this,               SLOT( slotItemSingleClicked(const QModelIndex&) ) );
  connect( contactListView_, SIGNAL(         doubleClicked(const QModelIndex&) ),
           this,               SLOT( slotItemDoubleClicked(const QModelIndex&) ) );
  connect( contactListView_, SIGNAL(             collapsed(const QModelIndex&) ),
           this,               SLOT(      slotGroupChanged(const QModelIndex&) ) );
  connect( contactListView_, SIGNAL(              expanded(const QModelIndex&) ),
           this,               SLOT(      slotGroupChanged(const QModelIndex&) ) );

  // Enable the right-click context menu
  contactListView_->setContextMenuPolicy( Qt::CustomContextMenu );
  connect( contactListView_, SIGNAL( customContextMenuRequested(const QPoint&) ),
           this,               SLOT(            showContextMenu(const QPoint&) ) );

  // Enable list sorting
  contactListView_->setColumnHidden( 1, true );
  contactListView_->setColumnHidden( 2, true );
  contactListView_->setSortingEnabled( true );
  contactListView_->sortByColumn( 1, Qt::AscendingOrder );

  // Update the list's visible groups
  slotUpdateDisplaySettings();

  return true;
}



// Initialize the contact popup
bool KMessView::initContactPopup()
{
  // Initialize context menu actions
  chatWithContact_   = new KAction( KIcon("user-group-new"),   i18n("Cha&t"), this );
  emailContact_      = new KAction( KIcon("mail-message-new"), i18n("&Send email"), this );
  msnProfile_        = new KAction( KIcon("preferences-desktop-user"),    i18n("&View Profile"), this );
  contactProperties_ = new KAction( KIcon("user-properties"),  i18n("&Properties"), this );

  addContact_        = new KAction( KIcon("list-add"),         i18n("&Add Contact"), this );
  allowContact_      = new KAction( KIcon("dialog-ok-apply"),  i18n("A&llow Contact"), this );
  blockContact_      = new KAction( KIcon("dialog-cancel"),    i18n("&Block Contact"), this );
  unblockContact_    = new KAction( KIcon("dialog-ok"),        i18n("&Unblock Contact"), this );
  removeContact_     = new KAction( KIcon("list-remove-user"), i18n("&Delete Contact"), this );
  removeFromGroup_   = new KAction( KIcon("list-remove"),      i18n("&Remove from group"), this );

  popupCopyFriendlyName_    = new KAction(                     i18n("&Friendly Name"),    this );
  popupCopyPersonalMessage_ = new KAction(                     i18n("&Personal Message"), this );
  popupCopyHandle_          = new KAction(                     i18n("&Email Address"),    this );
  popupCopyMusic_           = new KAction(                     i18n("&Listening Music"),  this );

  // Connect the actions
  connect( chatWithContact_,   SIGNAL(activated()),   this,  SLOT(slotForwardStartChat())      );
  connect( emailContact_,      SIGNAL(activated()),   this,  SLOT(slotEmailContact())          );
  connect( msnProfile_,        SIGNAL(activated()),   this,  SLOT(slotShowContactProfile())    );
  connect( contactProperties_, SIGNAL(activated()),   this,  SLOT(slotShowContactProperties()) );

  connect( addContact_,        SIGNAL(activated()),   this,  SLOT(slotForwardAddContact())     );
  connect( allowContact_,      SIGNAL(activated()),   this,  SLOT(slotForwardAllowContact())   );
  connect( blockContact_,      SIGNAL(activated()),   this,  SLOT(slotForwardBlockContact())   );
  connect( unblockContact_,    SIGNAL(activated()),   this,  SLOT(slotForwardUnblockContact()) );
  connect( removeContact_,     SIGNAL(activated()),   this,  SLOT(slotForwardRemoveContact())  );
  connect( removeFromGroup_,   SIGNAL(activated()),   this,  SLOT(slotForwardRemoveFromGroup()));

  connect( popupCopyFriendlyName_,    SIGNAL(triggered(bool)),   this,  SLOT( copyText()     ) );
  connect( popupCopyPersonalMessage_, SIGNAL(triggered(bool)),   this,  SLOT( copyText()     ) );
  connect( popupCopyHandle_,          SIGNAL(triggered(bool)),   this,  SLOT( copyText()     ) );
  connect( popupCopyMusic_,           SIGNAL(triggered(bool)),   this,  SLOT( copyText()     ) );

  // Create the signal mappers needed to assign all actions to the move/copy contact slots
  contactCopyMapper_ = new QSignalMapper( this );
  contactMoveMapper_ = new QSignalMapper( this );
  connect( contactCopyMapper_, SIGNAL(mapped(const QString&)),   this,  SLOT(slotForwardCopyContact(const QString&)) );
  connect( contactMoveMapper_, SIGNAL(mapped(const QString&)),   this,  SLOT(slotForwardMoveContact(const QString&)) );

  // Initialize the sub menus to copy/move contacts
  copyContactToGroup_ = new KActionMenu( i18n("&Copy to Group"), this );
  moveContactToGroup_ = new KActionMenu( i18n("&Move to Group"), this );

  // Fill them with the current groups names
  rebuildContactActions();

  // Initialize Copy sub popups
  popupCopyMenu_ = new KActionMenu( i18n("&Copy..."), this );
  popupCopyMenu_ ->addAction( popupCopyFriendlyName_ );
  popupCopyMenu_ ->addAction( popupCopyPersonalMessage_ );
  popupCopyMenu_ ->addAction( popupCopyHandle_ );
  popupCopyMenu_ ->addAction( popupCopyMusic_ );
  popupCopyMenu_->addSeparator();

  // Initialize the popup menu
  contactActionPopup_ = new KMenu( "KMess", this );

  contactActionPopup_->addAction( chatWithContact_ );
  contactActionPopup_->addAction( emailContact_ );
  contactActionPopup_->addAction( msnProfile_ );
  contactActionPopup_->addAction( popupCopyMenu_ );
  contactActionPopup_->addAction( contactProperties_ );
  contactActionPopup_->addSeparator();

  contactActionPopup_->addAction( addContact_ );
  contactActionPopup_->addAction( allowContact_ );

  contactActionPopup_->addAction( blockContact_ );
  contactActionPopup_->addAction( unblockContact_ );

  contactActionPopup_->addAction( moveContactToGroup_ );
  contactActionPopup_->addAction( copyContactToGroup_ );
  contactActionPopup_->addAction( removeFromGroup_ );
  contactActionPopup_->addAction( removeContact_ );

  return true;
}



// Initialize the group popup
bool KMessView::initGroupPopup()
{
  // Initialize context menu actions
  moveGroupDown_ = new KAction( KIcon("arrow-down"),  i18n("Move Group &Down"), this );
  moveGroupUp_   = new KAction( KIcon("arrow-up"),    i18n("Move Group &Up"),   this );
  removeGroup_   = new KAction( KIcon("edit-delete"), i18n("Re&move Group"),    this );
  renameGroup_   = new KAction( KIcon("edit"),        i18n("Re&name Group"),    this );

  // Connect the actions
  connect( moveGroupDown_, SIGNAL(activated()),   this,  SLOT(     slotMoveGroupDown()) );
  connect( moveGroupUp_,   SIGNAL(activated()),   this,  SLOT(       slotMoveGroupUp()) );
  connect( removeGroup_,   SIGNAL(activated()),   this,  SLOT(slotForwardRemoveGroup()) );
  connect( renameGroup_,   SIGNAL(activated()),   this,  SLOT(slotForwardRenameGroup()) );

  // Initialize the popup menu
  groupActionPopup_ = new KMenu( "KMess", this );

  groupActionPopup_->addAction( moveGroupUp_ );
  groupActionPopup_->addAction( moveGroupDown_ );
  groupActionPopup_->addSeparator();

  groupActionPopup_->addAction( renameGroup_ );
  groupActionPopup_->addAction( removeGroup_ );

  return true;
}



// Generate the lists of actions for the "copy/move contact to group" menus
void KMessView::rebuildContactActions()
{
  // Remove all copy and move actions
  foreach( KAction *action, groupCopyActionsList_ )
  {
    copyContactToGroup_->removeAction( action );
    groupCopyActionsList_.removeAll( action );
    delete action;
  }
  foreach( KAction *action, groupMoveActionsList_ )
  {
    moveContactToGroup_->removeAction( action );
    groupMoveActionsList_.removeAll( action );
    delete action;
  }

  // Create the list of groups between which contacts can be moved & copied
  QList<Group*>groupsList( currentAccount_->getContactList()->getGroupList() );
  foreach( Group *group, groupsList )
  {
    // Disallow copying and moving to special groups
    if( group->isSpecialGroup() )
    {
      continue;
    }

    // Since we need both a name to display and a group ID to distinguish groups,
    // but KActions only support the name, we'll use the Qt objectName to store
    // their internal id
    QString id = group->getId();
    KAction *copyAction = new KAction( group->getName(), copyContactToGroup_ );
    KAction *moveAction = new KAction( group->getName(), moveContactToGroup_ );
    copyAction->setObjectName( id );
    moveAction->setObjectName( id );

    copyContactToGroup_->addAction( copyAction );
    moveContactToGroup_->addAction( moveAction );

    connect( copyAction, SIGNAL(triggered(bool)), contactCopyMapper_, SLOT(map()) );
    connect( moveAction, SIGNAL(triggered(bool)), contactMoveMapper_, SLOT(map()) );
    contactCopyMapper_->setMapping( copyAction, id );
    contactMoveMapper_->setMapping( moveAction, id );

    groupCopyActionsList_.append( copyAction );
    groupMoveActionsList_.append( moveAction );
  }
}



// The email label was clicked so open the user's preferred email url
void KMessView::slotEmailLabelClicked()
{
#ifdef KMESSDEBUG_KMESSVIEW_GENERAL
  kDebug();
#endif

  currentAccount_->openMailAtInbox();
}



// Personal Status changed
void KMessView::slotForwardChangeStatus( QAction *action )
{
  int index = action->data().toInt();
  emit changeStatus( index );
}


// A group was expanded or collapsed
void KMessView::slotGroupChanged( const QModelIndex &index )
{
  // Get the data from the selected index
  ModelDataList itemData = getItemData( index );

  // Do nothing if the input index was not valid
  if( itemData.isEmpty() || itemData[ "type" ] != ContactListModelItem::ItemGroup )
  {
    return;
  }

  Group *group = currentAccount_->getContactList()->getGroupById( itemData[ "id" ].toString() );

  // Bail out if the group had not been found
  if( ! group )
  {
    return;
  }

  group->setExpanded( contactListView_->isExpanded( index ) );
}



// Show the context menu
void KMessView::showContextMenu( const QPoint &point )
{
  const ContactList *contactList = currentAccount_->getContactList();

  // Get the data from the selected index
  ModelDataList itemData = getItemData( contactListView_->indexAt( point ) );

  // The index was empty
  if( itemData.isEmpty() )
  {
    return;
  }

  // Depending on the item type, show the appropriate menu
  if( itemData[ "type" ] == ContactListModelItem::ItemContact )
  {
    Contact *contact = contactList->getContactByHandle( itemData[ "handle" ].toString() );

    bool isAllowed = contact->isAllowed();
    bool isBlocked = contact->isBlocked();
    bool isReverse = contact->isReverse();
    bool isFriend  = contact->isFriend();
    bool isOnline  = contact->isOnline();
    QStringList contactGroups = contact->getGroupIds();

    // Only chat if contact is friends list, and not blocked.
    chatWithContact_      ->setVisible( isFriend && ! isBlocked && isOnline );

    moveContactToGroup_   ->setVisible(   isFriend );

    // Only if the contact is not in the 'Implicit' group (group 0) or there is only one group;
    // (this means it will be the fifth ('Implicit', Online, Offline, Allowed, Removed);
    if( contactGroups.isEmpty() )
    {
      copyContactToGroup_ ->setVisible( false );
    }
    else if( contactList->getGroupList().count() == 5 )
    {
      copyContactToGroup_ ->setVisible( false );
      moveContactToGroup_ ->setVisible( false );
    }
    else
    {
      copyContactToGroup_ ->setVisible( isFriend );
    }

    // Only if not added yet
    addContact_           ->setVisible( ! isFriend );

    // Only allow if contact is on the "removed" group.
    allowContact_         ->setVisible( ! isFriend && ! isAllowed && ! isBlocked );

    // Only one of these is enabled:
    blockContact_         ->setVisible( ! isBlocked );
    unblockContact_       ->setVisible(   isBlocked );

    // Only if the contact is on the friends list, or no longer
    // added to reverse list (otherwise an "contact has added you" message appears)
    removeContact_        ->setVisible(   isFriend || ! isReverse );

    // Only if the contact is in the 'Implicit' group
    if( contactGroups.isEmpty() )
    {
      removeFromGroup_  ->setVisible(false);
    }
    else
    {
      removeFromGroup_  ->setVisible(true);
    }

    // Show all copy and move actions
    foreach( KAction *action, groupCopyActionsList_ )
      action->setVisible( true );
    foreach( KAction *action, groupMoveActionsList_ )
      action->setVisible( true );

    // Then hide only the invalid choices
    int enabledCopyActions = groupCopyActionsList_.count();
    int enabledMoveActions = groupMoveActionsList_.count();
    foreach( QString groupId, contactGroups )
    {
      foreach( KAction *action, groupCopyActionsList_ )
      {
        // See initContactPopup() for the choice of using objectName to store the group ID
        if( action->objectName() == groupId )
        {
          enabledCopyActions--;
          action->setVisible( false );
          break;
        }
      }
      foreach( KAction *action, groupMoveActionsList_ )
      {
        // See initContactPopup() for the choice of using objectName to store the group ID
        if( action->objectName() == groupId )
        {
          enabledMoveActions--;
          action->setVisible( false );
          break;
        }
      }
    }

    // If no actions have been enabled in a menu, disable it too
    if( enabledCopyActions <= 0 )
    {
      copyContactToGroup_ ->setVisible( false );
    }
    if( enabledMoveActions <= 0 )
    {
      moveContactToGroup_ ->setVisible( false );
    }

    // Clear the previous actions to avoid duplication of links
    qDeleteAll( groupCopyLinkActionsList_ );
    groupCopyLinkActionsList_.clear();

     // If detailed contact information is available, search for links
    if( contact != 0 )
    {
      // Append friendly name and personal message so cycle only one time to search links
      QString nameAndPm = contact->getFriendlyName() + " " + contact->getPersonalMessage();

      //Initialize index, link RegExp and found boolean for insert separator only at first found link
      int pos = 0;
      QRegExp linkRegExp( "(http://|https://|ftp://|sftp://|www\\..)\\S+" );

      while( ( pos = linkRegExp.indexIn( nameAndPm, pos ) ) != -1 )
      {
        // Grep the found link and update the position for the next cycle
        QString foundLink = linkRegExp.cap( 0 );
        pos += linkRegExp.matchedLength();

        // Put the found link into KAction, replace & with && to avoid shortcut problem and use tooltip
        // for future grep of link ( to avoid shortcut problem too )
        popupCopyLink_ = new KAction( foundLink.replace( "&", "&&" ), this );
        popupCopyLink_->setToolTip( foundLink );
        connect( popupCopyLink_, SIGNAL( triggered( bool ) ), this, SLOT( copyText() ) );

        // Append current KAction to the copy link list
        groupCopyLinkActionsList_.append( popupCopyLink_ );

        // Add action to copy menu
        popupCopyMenu_->addAction( popupCopyLink_ );
      }
    }

    // Set tooltip ( informations ) of other copy actions
    popupCopyFriendlyName_   ->setToolTip( contact->getFriendlyName() );
    popupCopyPersonalMessage_->setToolTip( contact->getPersonalMessage() );
    popupCopyHandle_         ->setToolTip( contact->getHandle() );
    popupCopyMusic_          ->setToolTip( contact->getCurrentMediaString() );

    // Show the popup
    contactActionPopup_->popup( QCursor::pos() );
  }
  else if( itemData[ "type" ] == ContactListModelItem::ItemGroup )
  {
    Group *group = contactList->getGroupById( itemData[ "id" ].toString() );

    // No context menu for special groups
    if( ! group || group->isSpecialGroup() )
    {
      return;
    }

    int pos = group->getSortPosition();

    moveGroupUp_   ->setEnabled( contactList->getGroupBySortPosition( pos - 1, false ) != 0 );
    moveGroupDown_ ->setEnabled( contactList->getGroupBySortPosition( pos + 1, true  ) != 0 );

    // Show the popup
    groupActionPopup_->popup( QCursor::pos() );
  }
  else
  {
    kWarning() << "Invalid type of list item" << itemData[ "type" ];
  }
}



// Copy some details of the contact to the clipboard.
void KMessView::copyText()
{
  // Grep the action sender for copy the tooltip that contains the information
  KAction *action = static_cast<KAction*>( const_cast<QObject*>( sender() ) );
  kapp->clipboard()->setText( action->toolTip() );
}



// Show the contact & group tool tip
void KMessView::showToolTip( const QPoint &point )
{
  QString tipText;
  QModelIndex index = contactListView_->indexAt( contactListView_->mapFromGlobal( point ) );

  if( !index.isValid() )
  {
    return;
  }

  // Access the index's data
  ContactListModelItem *item = static_cast<ContactListModelItem*>( index.internalPointer() );

  // Read the data off the model's index
  ModelDataList itemData = item->data( index.column() ).toMap();

  // Fields used for the Contact tooltip
  QString clientName, mediaType, mediaString, friendlyName, personalMessage;

  // Paint the different types of list items
  switch( item->getType() )
  {
    case ContactListModelItem::ItemContact:
      tipText = "<qt><table border='0'>"
                  "<tr>"
                    "<td width='64' valign='top'><img src='%1' /></td>"
                    "<td align='left' valign='top'>"
                      "<b>%2</b> (%3)<br />"
                      "%4 %5"
                    "</td>"
                  "</tr>"
                  "<tr>"
                    "<td colspan='2'>"
                      "%6%7"
                    "</td>"
                  "</tr>"
                "</table></qt>";

      // Add the display picture (if there is any), or the default KMess logo
      tipText = tipText.arg( itemData[ "displayPicture" ].toString() );

      // Add the contact's real name - the alias is not shown:
      // you can read it in the contact list, after all
      friendlyName = Qt::escape( itemData[ "trueFriendly" ].toString() );
      tipText = tipText.arg( friendlyName );

      // Add status
      tipText = tipText.arg( MsnStatus::getName( (Status) itemData[ "status" ].toInt() ) );

      // Add the personal message
      personalMessage = Qt::escape( itemData[ "personalMessage" ].toString() );
      if( ! personalMessage.isEmpty() )
      {
        tipText = tipText.arg( "<i>" + personalMessage + "</i>" );
      }
      else
      {
        tipText = tipText.arg( "" );
      }

      // If the contact is listening to music, show it
      mediaType   = itemData[ "mediaType"   ].toString();
      mediaString = itemData[ "mediaString" ].toString();
      if( ! mediaString.isEmpty() )
      {
        QString mediaEmoticon;
        QString newlineBefore( personalMessage.isEmpty() ? "" : "<br />" );

        // Determine icon for special media types
        if( mediaType == "Music" )
        {
          mediaEmoticon = "(8)";
        }
        else if( mediaType == "Gaming" )
        {
          mediaEmoticon = "(xx)";
        }

        // Set the icon if it's found in the emoticon theme.
        if( ! mediaEmoticon.isEmpty() )
        {
          mediaEmoticon = EmoticonManager::instance()->getReplacement( mediaEmoticon, true );
        }

        tipText = tipText.arg( newlineBefore + "<small><font color=\"#83838F\">" + mediaEmoticon + Qt::escape( mediaString ) + "</font></small>" );
      }
      else
      {
        tipText = tipText.arg( "" );
      }

      // Add handle
      tipText = tipText.arg( i18nc( "Contact email in list tooltip", "Email: %1", Qt::escape( itemData[ "handle" ].toString() ) ) );

      // Add client identifier
      // NOTE If you want to add new client identifiers here's the right place
      clientName = itemData[ "clientName" ].toString();
      if( ! clientName.isEmpty() )
      {
        tipText = tipText.arg( "<br>" + i18nc( "Contact MSN client in list tooltip", "Client: %1", Qt::escape( clientName ) ) );
      }
      else
      {
        tipText = tipText.arg( "" );
      }
      break;


    case ContactListModelItem::ItemGroup:
      tipText = "<qt>%1<br /><i>(%2)</i></qt>";

      // Add the group name
      tipText = tipText.arg( i18nc( "Group name in group tooltip", "Group <b>%1</b>", itemData[ "name" ].toString() ) );

      // Add the number of contacts
      tipText = tipText.arg( i18nc( "Contacts count in group tooltip", "%1 contacts", itemData[ "totalContacts" ].toString() ) );
      break;


    default:
      // Nothing to do!
      return;
  }

  // Show the tooltip
  QToolTip::showText( point, tipText, contactListView_, contactListView_->visualRect( index ) );
}



// E-mail the current contact
void KMessView::slotEmailContact()
{
  // Get the data from the currently selected index
  ModelDataList itemData = getItemData();

  // The index was not valid, do nothing
  if( itemData.isEmpty() || itemData[ "type" ] != ContactListModelItem::ItemContact )
  {
    return;
  }

  // Send mail to that account
  currentAccount_->openMailAtCompose( itemData[ "handle" ].toString() );
}


// An item was double clicked
void KMessView::slotItemDoubleClicked( const QModelIndex &index )
{
  // Detect accidental clicks
  if( KGlobalSettings::singleClick() )
  {
#ifdef KMESSDEBUG_KMESSVIEW_GENERAL
    kDebug() << "Accidental double click detected, skipping click event.";
#endif
    return;
  }

  slotItemClicked( index );
}



void KMessView::slotItemSingleClicked( const QModelIndex &index )
{
  bool accidental = false;
  if( ! KGlobalSettings::singleClick() )
  {
#ifdef KMESSDEBUG_KMESSVIEW_GENERAL
    kDebug() << "Accidental single click detected, skipping click event.";
#endif

    accidental = true;
  }

  slotItemClicked( index, accidental );
}



// An item was clicked
void KMessView::slotItemClicked( const QModelIndex &index, const bool accidental )
{
  // Get the data from the selected index
  ModelDataList itemData = getItemData( index );

  // The index was empty
  if( itemData.isEmpty() )
  {
    kWarning() << "Invalid list index given!";
    return;
  }

  switch( itemData[ "type" ].toInt() )
  {
    // Expand/collapse the groups when clicked
    case ContactListModelItem::ItemGroup:
      contactListView_->setExpanded( index, ! itemData[ "isExpanded" ].toBool() );
      break;


    //  Request a chat with that contact
    case ContactListModelItem::ItemContact:
      if( ! accidental )
      {
        // Tell the Chat Master we'd like to start a chat.
        emit requestChat( itemData[ "handle" ].toString() );
      }
      break;


    default:
#ifdef KMESSDEBUG_KMESSVIEW_GENERAL
      kWarning() << "Nothing to do!";
#endif
      break;
  }
}



// Move down the selected group
void KMessView::slotMoveGroupDown()
{
  const ContactList *contactList = currentAccount_->getContactList();

  // Get the group data from the selected index
  ModelDataList itemData = getItemData();
  Group *group = contactList->getGroupById( itemData[ "id" ].toString() );
  if( ! group || itemData.isEmpty() )
  {
    return;
  }

  // Obtain the next listed group
  Group *nextGroup = contactList->getGroupBySortPosition( group->getSortPosition() + 1, true );
  if( ! nextGroup )
  {
    return;
  }

  // Get the group sorting position for both groups
  int oldGroupPos = group    ->getSortPosition();
  int newGroupPos = nextGroup->getSortPosition();

  // Swap them
  group    ->setSortPosition( newGroupPos );
  nextGroup->setSortPosition( oldGroupPos );

  // Re-sort the list
  contactListView_->sortByColumn( 1, Qt::AscendingOrder );
}



// Move up the selected group
void KMessView::slotMoveGroupUp()
{
  const ContactList *contactList = currentAccount_->getContactList();

  // Get the group data from the selected index
  ModelDataList itemData = getItemData();
  Group *group = contactList->getGroupById( itemData[ "id" ].toString() );
  if( ! group || itemData.isEmpty() )
  {
    return;
  }

  // Obtain the previous listed group
  Group *prevGroup = contactList->getGroupBySortPosition( group->getSortPosition() - 1, false );
  if( ! prevGroup )
  {
    return;
  }

  // Get the group sorting position for both groups
  int oldGroupPos = group    ->getSortPosition();
  int newGroupPos = prevGroup->getSortPosition();

  // Swap them
  group    ->setSortPosition( newGroupPos );
  prevGroup->setSortPosition( oldGroupPos );

  // Re-sort the list
  contactListView_->sortByColumn( 1, Qt::AscendingOrder );
}



// Forward the "add contact" menu action
void KMessView::slotForwardAddContact()
{
  // Get the data from the currently selected index
  ModelDataList itemData = getItemData();

  // The index was not empty
  if( itemData.isEmpty() || itemData[ "type" ] != ContactListModelItem::ItemContact )
  {
    return;
  }

  // Add that contact
  emit addContact( itemData[ "handle" ].toString() );
}



// Forward the "allow contact" menu action
void KMessView::slotForwardAllowContact()
{
  // Get the data from the currently selected index
  ModelDataList itemData = getItemData();

  // The index was not empty
  if( itemData.isEmpty() || itemData[ "type" ] != ContactListModelItem::ItemContact )
  {
    return;
  }

  // Allow that contact
  emit allowContact( itemData[ "handle" ].toString() );
}



// Forward the "block contact" menu action
void KMessView::slotForwardBlockContact()
{
  // Get the data from the currently selected index
  ModelDataList itemData = getItemData();

  // The index was not empty
  if( itemData.isEmpty() || itemData[ "type" ] != ContactListModelItem::ItemContact )
  {
    return;
  }

  // Block that contact
  emit blockContact( itemData[ "handle" ].toString() );
}



// Called when a contact should be copied
void KMessView::slotForwardCopyContact( const QString &newGroupId )
{
  // Get the data from the currently selected index
  ModelDataList itemData = getItemData();

  // The index was not empty
  if( itemData.isEmpty() || itemData[ "type" ] != ContactListModelItem::ItemContact )
  {
    return;
  }

  // Copy that contact
  emit copyContact( itemData[ "handle" ].toString(), newGroupId );
}



// Called when a contact should be moved
void KMessView::slotForwardMoveContact( const QString &newGroupId )
{
  // Get the data from the currently selected index
  ModelDataList itemData = getItemData();

  // The index was not empty
  if( itemData.isEmpty() || itemData[ "type" ] != ContactListModelItem::ItemContact )
  {
    return;
  }

  // Get the old group ID, and if it's not valid, set it to zero
  QString oldGroupId;
  ModelDataList parentItemData = getItemData( selectionModel_->selectedIndexes().first().parent() );
  if( ! parentItemData.isEmpty() && parentItemData[ "type" ] == ContactListModelItem::ItemGroup )
  {
    oldGroupId = parentItemData[ "id" ].toString();
  }
  else
  {
    oldGroupId = "0";
  }

  // If the old and new groups are the same one, do nothing
  if( oldGroupId == newGroupId )
  {
    return;
  }

  // Move that contact
  emit moveContact( itemData[ "handle" ].toString(), oldGroupId, newGroupId );
}



// Forward the "remove contact" menu action
void KMessView::slotForwardRemoveContact()
{
  // Get the data from the currently selected index
  ModelDataList itemData = getItemData();

  // The index was not empty
  if( itemData.isEmpty() || itemData[ "type" ] != ContactListModelItem::ItemContact )
  {
    return;
  }

  // Remove that contact
  emit removeContact( itemData[ "handle" ].toString() );
}


void KMessView::slotForwardRemoveFromGroup()
{
  // Get the data from the currently selected index
  ModelDataList itemData = getItemData();

  // The index was invalid
  if( itemData.isEmpty() )
  {
    return;
  }

  // Get the ID of the group which contained this instance of the contact
  ModelDataList parentItemData = getItemData( selectionModel_->selectedIndexes().first().parent() );
  if( parentItemData.isEmpty() || parentItemData[ "type" ] != ContactListModelItem::ItemGroup )
  {
    return;
  }

  // Remove that contact
  emit removeContactFromGroup( itemData[ "handle" ].toString(), parentItemData[ "id" ].toString() );
}


// Forward the "remove group" menu action
void KMessView::slotForwardRemoveGroup()
{
  // Get the data from the currently selected index
  ModelDataList itemData = getItemData();

  // The index was invalid
  if( itemData.isEmpty() || itemData[ "type" ] != ContactListModelItem::ItemGroup )
  {
    return;
  }

  // Remove that group
  emit removeGroup( itemData[ "id" ].toString() );
}




// Forward the "rename group" menu action
void KMessView::slotForwardRenameGroup()
{
  // Get the data from the currently selected index
  ModelDataList itemData = getItemData();

  // The index was invalid
  if( itemData.isEmpty() || itemData[ "type" ] != ContactListModelItem::ItemGroup )
  {
    return;
  }

  // Rename that group
  emit renameGroup( itemData[ "id" ].toString() );
}



// Forward the "start chat" menu action
void KMessView::slotForwardStartChat()
{
  // Only one index can be returned at once by the contact list (has single selection mode)
  slotItemClicked( selectionModel_->selectedIndexes().first() );
}



// Forward the "unblock contact" menu action
void KMessView::slotForwardUnblockContact()
{
  // Get the data from the currently selected index
  ModelDataList itemData = getItemData();

  // The index was not empty
  if( itemData.isEmpty() || itemData[ "type" ] != ContactListModelItem::ItemContact )
  {
    return;
  }

  // Unblock that contact
  emit unblockContact( itemData[ "handle" ].toString() );
}



// Called when the enter key is pressed on the personal message input
void KMessView::slotPersonalMessageEnterPressed()
{
  personalMessageInput_->clearFocus();
}



/**
 * @brief Called when the personal message input looses focus
 *
 * Sends to the server the new personal message selected by the user.
 *
 * @param refreshOnly  When true, the PM widget is only updated with the
 *                     current account's message, or with the default
 *                     message, and nothing is sent to the server.
 */
void KMessView::slotPersonalMessageLostFocus( bool refreshOnly )
{
  const QString defaultMessage( i18nc( "Default personal message shown in the contact list", "&lt;Enter your personal message here&gt;" ) );

  personalMessageInput_->clearFocus();
  contactListView_->setFocus();

  // When refreshing the message or when the focus losing isn't voluntary, don't update it on the server
  if( refreshOnly || ! isActiveWindow() || ! topLevelWidget()->isActiveWindow() )
  {
#ifdef KMESSDEBUG_KMESSVIEW_GENERAL
    kDebug() << "Refreshing message only: initial update or non-intentional focus lose.";
#endif

    // Refresh the message
    const QString &message = currentAccount_->getPersonalMessage();
    if( message.isEmpty() )
    {
      personalMessageInput_->setText( defaultMessage );
    }
    else
    {
      personalMessageInput_->setText( message );
    }

    return;
  }

  // Get message
  QString message = personalMessageInput_->text();

#ifdef KMESSDEBUG_KMESSVIEW_GENERAL
  kDebug() << "Updating personal message to:" << message;
#endif

  // Protect against weird situations where the label is seen as text.
  // The lost focus event is also sent when the while window looses focus
  if( message == defaultMessage )
  {
    kWarning() << "Personal message label was about to be used as personal message!";
    return;
  }

  // Only update the message if it's been changed
  if( currentAccount_->getPersonalMessage() != message )
  {
    // Update message and save the changes
    currentAccount_->setPersonalMessage( message );
    // Commit changes to disk, now
    currentAccount_->saveProperties();

    // Notify changes to the MSN server
    emit changePersonalMessage( message );
  }

  // Restore the placeholder.
  if( message.isEmpty() )
  {
    personalMessageInput_->setText( defaultMessage );
  }
}



// Slot for searching contact
void KMessView::slotSearchContact( const QString& searchFor )
{
  searchContact_ = searchFor;

  // Update and show only the contacts that match the search
  slotUpdateDisplaySettings();
}



// Display the profile of the current contact
void KMessView::slotShowContactProfile()
{
  // Get the data from the currently selected index
  ModelDataList itemData = getItemData();

  // The index was not empty
  if( itemData.isEmpty() || itemData[ "type" ] != ContactListModelItem::ItemContact )
  {
    return;
  }

  KMessShared::openProfile( itemData[ "handle" ].toString() );
}



// Display the properties of the current contact
void KMessView::slotShowContactProperties()
{
  // Get the data from the currently selected index
  ModelDataList itemData = getItemData();

  // The index was not empty
  if( itemData.isEmpty() || itemData[ "type" ] != ContactListModelItem::ItemContact )
  {
    return;
  }

  // Get the contact data to pass to the properties dialog
  Contact *contact = currentAccount_->getContactList()->getContactByHandle( itemData[ "handle" ].toString() );

  // If no corresponding contact has been found, do nothing
  if( ! contact )
  {
    return;
  }

  ContactPropertiesDialog *dialog = new ContactPropertiesDialog( this );
  connect( dialog, SIGNAL( removeContactFromGroup(QString,QString) ),
           this,   SIGNAL( removeContactFromGroup(QString,QString) ) );
  connect( dialog, SIGNAL(      addContactToGroup(QString,QString) ),
           this,   SIGNAL(            copyContact(QString,QString) ) );
  dialog->launch( contact );
  delete dialog;
}


// Update the users display pic
void KMessView::slotUpdateDisplayPicture()
{
  QString image = "";

  if( currentAccount_->getShowPicture() )
  {
    displayPicFrame_->show();
    image = currentAccount_->getPicturePath();
    currentAccountDisplayPic_->setPixmap( QPixmap( image ) );
  }
  else // The user does not want to show a display pic, so just show an icon for their sign in status
  {
    displayPicFrame_->hide();
  }
}


// Change whether the allowed,removed groups are displayed.
void KMessView::slotUpdateDisplaySettings( QModelIndex rootIndex )
{
#ifdef KMESSDEBUG_KMESSVIEW_ITEM_VISIBILITY
  // Only show the debug message when the update is starting
  if( ! rootIndex.isValid() )
  {
    kDebug() << "Starting list items visibility update";
  }
#endif

  QAbstractItemModel *model = contactListView_->model();
  int rows = model->rowCount( rootIndex );

  for( int row = 0; row < rows; row++ )
  {
    QModelIndex index( model->index( row, 0, rootIndex ) );

    if( ! index.isValid() )
    {
#ifdef KMESSDEBUG_KMESSVIEW_ITEM_VISIBILITY
      kDebug() << "Skipping an invalid item at row" << row;
#endif
      continue;
    }

    // Access the index's data
    ModelDataList itemData = getItemData( index );

    contactListView_->setRowHidden( row, rootIndex, isItemHidden( itemData ) );

    // If the item has children (groups) also update them
    if( index.child( 0, 0 ).isValid() )
    {
      slotUpdateDisplaySettings( index );
      contactListView_->setExpanded( index, itemData[ "isExpanded" ].toBool() );
    }
  }
}



// Change whether or not the email label is displayed based on account settings.
void KMessView::slotUpdateEmailDisplay()
{
  if ( currentAccount_->getShowEmail() && currentAccount_->getEmailSupported() )
  {
    emailFrame_->setHidden(false);
    //emailFrame_->setMargin(statusLayout_->margin());
    emailFrame_->setMaximumHeight(40);
  }
  else
  {
    emailFrame_->setHidden(true);
    //emailFrame_->setMargin(0);
    emailFrame_->setMaximumHeight(0);
  }
}



// Update a contact's visibility on the list
void KMessView::slotUpdateItem( const QModelIndex &topLeft,const QModelIndex &bottomRight )
{
  if( ! topLeft.isValid() || ! bottomRight.isValid() )
  {
    return;
  }

  QModelIndex index( topLeft );
  int  itemsToUpdate = bottomRight.row() - topLeft.row();

#ifdef KMESSDEBUG_KMESSVIEW_ITEM_VISIBILITY
  kDebug() << "Updating list items: rows" << topLeft.row() << "-" << bottomRight.row()
                                          << " of parent" << topLeft.parent();
#endif

  do
  {
    // Access the index's data
    ModelDataList itemData = getItemData( index );

    contactListView_->setRowHidden( index.row(), index.parent(), isItemHidden( itemData ) );

    index = contactListView_->indexBelow( index );
  }
  while( index.isValid() && itemsToUpdate-- );

  // Re-sort the list
  contactListView_->sortByColumn( 1, Qt::AscendingOrder );
}



// Update the email count.
void KMessView::slotUpdateNoEmails()
{
  QString message = i18np( "%1 new email message", "%1 new email messages",
                            currentAccount_->getNoEmails() );

  // Update the label
  emailLabel_->setText( message );

  // Set the user's preferred email website.
  emailLabel_->setUrl( currentAccount_->getEmailUrl() );
}



// Update the user's status.
void KMessView::slotUpdateUserStatus()
{
  QString friendlyName;
  QString image;
  Status  status;

  FormattingMode mode = currentAccount_->getUseListFormatting()
                        ? STRING_FORMATTED : STRING_CLEANED;

  // Avoid newlines, which cause problems to the layout.
  friendlyName = currentAccount_->getFriendlyName( mode ).replace( "\n", " " );

  // Escape the friendly name if the mode is cleaned.
  if( mode == STRING_CLEANED )
  {
    friendlyName = Qt::escape( friendlyName );
  }
  status = currentAccount_->getStatus();

#ifdef KMESSDEBUG_KMESSVIEW_GENERAL
  kDebug() << "Updating user status to " << MsnStatus::getName( status );
#endif

  // Update the widgets
  statusButton_->setText( "(" + MsnStatus::getName( status ) + ")" );
  onlineStatusLabel_->setText( friendlyName );
  onlineStatusPixmapLabel_->setPixmap( MsnStatus::getIcon( status ) );
}



// Update the way contact groups are displayed
void KMessView::slotUpdateViewMode()
{
  // Make the groups visible, contacts are already added to the
  // online/offline groups to make this viewmode update
  // easier and faster. (no need to move the contacts)
  slotUpdateDisplaySettings();
}


#include "kmessview.moc"
