/***************************************************************************
                          msnconnection.cpp  -  description
                             -------------------
    begin                : Thu Jan 23 2003
    copyright            : (C) 2003 by Mike K. Bennett
    email                : mkb137b@hotmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "msnconnection.h"

#include "mimemessage.h"
#include "msnsocketbase.h"
#include "msnsockettcp.h"
#include "msnsockethttp.h"
#include "multipacketmessage.h"
#include "soap/httpsoapconnection.h"

#include <KLocale>
#include <KMessageBox>


#ifdef KMESSDEBUG_CONNECTION // Area-specific debug statements
  #define KMESSDEBUG_CONNECTION_GENERAL
  #define KMESSDEBUG_CONNECTION_SOCKET
#endif



/**
 * @brief The constructor
 *
 * Initializes the sockets, buffers and ping timer.
 *
 * @param  serverType  The type of server connection: TCP or HTTP.
 */
MsnConnection::MsnConnection( MsnSocketBase::ServerType serverType )
: QObject(0)
, ack_(0)
, initialized_(false)
{
  // Create the socket needed for this connection: default to the TCP one
  socket_ = new MsnSocketTcp( serverType );

#ifdef KMESSTEST
  KMESS_ASSERT( socket_ != 0 );
#endif
}



/**
 * @brief The destructor
 *
 * Closes the connection, and deletes the sockets.
 */
MsnConnection::~MsnConnection()
{
  // Disconnect from the server on exit
  if(isConnected())
  {
    disconnectFromServer();
  }

  // Clean up collections
  qDeleteAll( multiPacketBuffer_ );
  multiPacketBuffer_.clear();

  // Off course, delete the SOAP clients now here too.
  deleteSoapClients();

  // Delete the socket
  delete socket_;

#ifdef KMESSDEBUG_CONNECTION
  kDebug() << objectName() << "- DESTROYED";
#endif
}



/**
 * @brief Internal function to track SOAP clients.
 *
 * This method connects the HttpSoapConnection::requestFailed() signal,
 * and appends the client to the internal list.
 *
 * @param  client  The SOAP client class.
 */
void MsnConnection::addSoapClient( HttpSoapConnection *client )
{
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
  kDebug() << "Adding SOAP client to list.";
#endif

  // Connect signals
  connect(client, SIGNAL(     requestFailed(HttpSoapConnection*, const QString&) ),
          this,   SLOT  (     slotSoapError(HttpSoapConnection*, const QString&) ));
  connect(client, SIGNAL(     requestFailed(HttpSoapConnection*, const QString&, const QString&, QDomElement) ),
          this,   SLOT  (     slotSoapFault(HttpSoapConnection*, const QString&, const QString&, QDomElement) ));

  // Add to list so it cleans up when the connection closes
  soapClients_.append(client);
}



/**
 * @brief Connect a socket's signals to our slots
 *
 * Connects the signals of a new socket to this class' slots.
 * This is needed to have one place only for socket connection;
 * since socket connections are made in initialize() and in
 * switchToHttpSocket().
 */
void MsnConnection::attachToSocketSignals()
{
#ifdef KMESSDEBUG_CONNECTION
  kDebug() << objectName() << "- Connecting socket signals.";
#endif
#ifdef KMESSTEST
  KMESS_ASSERT( socket_ != 0 );
#endif

  // Connect our integrative methods to the raw socket's events signals
  connect( socket_, SIGNAL(        connected()                                     ),
           this,    SLOT  (    slotConnected()                                     ) );
  connect( socket_, SIGNAL(     disconnected()                                     ),
           this,    SLOT  ( slotDisconnected()                                     ) );
  connect( socket_, SIGNAL(     dataReceived(const QStringList&,const QByteArray&) ),
           this,    SLOT  ( slotDataReceived(const QStringList&,const QByteArray&) ) );
  connect( socket_, SIGNAL(            error(QString,MsnSocketBase::ErrorType)     ),
           this,    SLOT  (        slotError(QString,MsnSocketBase::ErrorType)     ) );
  // Forward the socket's notification signals
  connect( socket_, SIGNAL(    statusMessage(QString,bool)                         ),
           this,    SIGNAL(    statusMessage(QString,bool)                         ) );
  connect( socket_, SIGNAL(         pingSent()                                     ),
           this,    SIGNAL(         pingSent()                                     ) );
}



/**
 * @brief Connect to the given server via the socket
 *
 * The connection attempt is asynchronous.
 * When the connection is made, connectionSuccess() is called,
 * on error connectionFailed() is called.
 * Connection timeouts are not detected here yet.
 *
 * @param  server  The server hostname or IP address.
 * @param  port    The port to connect to.
 */
void MsnConnection::connectToServer( const QString& server, const quint16 port )
{
#ifdef KMESSDEBUG_CONNECTION_SOCKET
  if( server.isEmpty() && port == 0 )
  {
    kDebug() << objectName() << "- Connecting to the default server.";
  }
  else
  {
    kDebug() << objectName() << "- Connecting to server at" << server << ":" << port << ".";
  }
#endif

  // Start connecting now
  socket_->connectToServer( server, port );
}



/**
 * @brief Disconnect and delete a SOAP clients.
 *
 * @param  client  The client to delete.
 * @see deleteSoapClients()
 */
void MsnConnection::deleteSoapClient( HttpSoapConnection *client )
{
  soapClients_.removeAll( client );
  QTimer::singleShot( 10, client, SLOT(deleteLater()) );  // timeout needed for deleting passport login service.
}



/**
 * @brief Disconnect and delete all SOAP clients.
 *
 * This method is called by the derived class when all SOAP clients need to be deleted.
 * This is not done automatically in disconnectFromServer() so the derived class
 * is both responsable for creating and deleting the SOAP client objects.
 */
void MsnConnection::deleteSoapClients()
{
  // Disconnect other possibly active SOAP sessions.
  // Use deleteLater() since this method may be called from loginIncorrect()
  if( ! soapClients_.isEmpty() )
  {
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
    kDebug() << "Closing SOAP sessions.";
#endif

    foreach( HttpSoapConnection *soapClient, soapClients_ )
    {
      soapClient->closeConnection();
      soapClient->deleteLater();
    }

    soapClients_.clear();
  }
}



/**
 * @brief Disconnect from the server
 * 
 * This function is called after closeConnection()
 * It empties all buffers, and resets the sockets.
 *
 * @param  isTransfer  When set, a disconnected() signal won't be fired.
 *                     This is used to handle transfers to another server
 *                     without noticing a disconnect.
 */
void MsnConnection::disconnectFromServer( bool isTransfer )
{
#ifdef KMESSDEBUG_CONNECTION_SOCKET
  kDebug() << objectName() << "- Closing the connection.";
#endif

  // Gracefully disconnect
  if( isConnected() )
  {
    sendCommand( "OUT", "\r\n" );
  }

  // Stop pinging
  setSendPings( false );

  // Clean up collections
  qDeleteAll( multiPacketBuffer_ );
  multiPacketBuffer_.clear();

  socket_->disconnectFromServer( isTransfer );
}



/**
 * @brief Return the local IP address of the socket.
 *
 * This is typically a LAN address, unless the system
 * is directly connected to the Internet.
 *
 * @returns The IP address the socket uses at the system.
 */
QString MsnConnection::getLocalIp() const
{
#ifdef KMESSDEBUG_SWITCHBOARD_FILETRANSFER
  kDebug() << objectName() << "- Local IP is " << socket_->getLocalIp();
#endif

  return socket_->getLocalIp();
}



/**
 * @brief Initialize the object
 *
 * This method should be called before the class is used.
 *
 * It initializes the socket
 * Subclasses may re-implement this method to connect
 * to signals which are not available at startup in the constructor.
 */
bool MsnConnection::initialize()
{
  if( initialized_ )
  {
    kDebug() << objectName() << "- Already initialized!";
    return false;
  }

  // Attach the socket's signals to this class
  attachToSocketSignals();

  initialized_ = true;
  return true;
}



/**
 * @brief Return whether or the socket is connected.
 * @returns True when connected, false otherwise.
 */
bool MsnConnection::isConnected() const
{
  return socket_->isConnected();
}



/**
 * @brief Return whether the given command is a payload command.
 *
 * This method just passes the question to the socket class
 *
 * @returns True when it's a payload command, false otherwise.
 */
bool MsnConnection::isPayloadCommand( const QString &command ) const
{
  return socket_->isPayloadCommand( command );
}



/**
 * Internal processing of multi-packet messages.
 * @param  command      The received command.
 * @param  mimeMessage  The received mime message.
 */
void MsnConnection::parseMultiPacketMimeMessage( const QStringList &command, const MimeMessage &mimeMessage )
{
#ifdef KMESSDEBUG_CONNECTION_SOCKET
  kDebug() << objectName() << "- Message is a multi-packet message, buffering.";
#endif

  // Get associated multi-packet message buffer, create one if it doesn't exist yet.
  QString messageId = mimeMessage.getValue("Message-ID");
  MultiPacketMessage *multiPacketMessage;
  if( ! multiPacketBuffer_.contains( messageId ) )
  {
    multiPacketMessage = new MultiPacketMessage;
    multiPacketBuffer_.insert( messageId, multiPacketMessage );
  }
  else
  {
    multiPacketMessage = multiPacketBuffer_[ messageId ];
  }

  // Add the message to the handler.
  multiPacketMessage->addChunk( mimeMessage );
  if( ! multiPacketMessage->isComplete() )
  {
    return;
  }

  // Parse message if complete multi-packet message is received.
  const MimeMessage &fullMimeMessage = multiPacketMessage->getMessage();
  parseMimeMessage( command, fullMimeMessage );

  // Delete handler class
  delete multiPacketBuffer_.take( messageId );
}



/**
 * @brief Send a command to the server
 *
 * This function automatically inserts the ACK-number
 * between the prefix and text.
 * This can be used later to track error responses.
 *
 * @param  prefix  The three-letter command, which parseCommand() receives in <code>command[0]</code>.
 * @param  text    The command arguments, joined with spaces.
 * @returns The ack-number used by the command.
 */
int MsnConnection::sendCommand(const QString& prefix, const QString &text)
{
#ifdef KMESSTEST
  KMESS_ASSERT( prefix.length() == 3 );
  if ( ( prefix != "QRY" ) && ( prefix != "MSG" ) )
  {
    KMESS_ASSERT( text.right( 2 ) == "\r\n" );
  }
#endif

  int ack = ack_++;

  // Send the data to the server
  QString command = prefix + " " + QString::number( ack ) + " " + text;
  writeData( command );

  // Return the ack used
  return ack;
}



/**
 * @brief Send a MIME message command to the server.
 * @returns, returning the ack
 */
int MsnConnection::sendMimeMessage(AckType ackType, const MimeMessage &message)
{
  // Determine which type to send.
  // Normally a char[4] lookup array would be sufficient,
  // but this is easier to use in the string concatenation with Qt
  const char *ackTypeStr = "?";
  switch(ackType)
  {
    case ACK_NONE:        ackTypeStr = "U"; break;
    case ACK_NAK_ONLY:    ackTypeStr = "N"; break;
    case ACK_ALWAYS:      ackTypeStr = "A"; break;
    case ACK_ALWAYS_P2P:  ackTypeStr = "D"; break;
  }

  // Create the "MSG id type length" header
  QByteArray rawMessage = message.getMessage();
  QString    command;

  // Build the command
  int ack = ack_++;
  command.sprintf("MSG %u %s %u\r\n", ack, ackTypeStr, rawMessage.size());

  // Write the data
  writeBinaryData( command.toUtf8() + rawMessage );

  // Return the ack used
  return ack;
}



// Send a payload command to the server, convenience function
int MsnConnection::sendPayloadMessage(const QString &prefix, const QString &arguments, const QString &payload)
{
  // Send payload, as UTF-8 encoded raw string.
  QByteArray binPayload = payload.toUtf8();
  return sendPayloadMessage( prefix, arguments, binPayload );
}



// Send a payload command to the server
int MsnConnection::sendPayloadMessage(const QString &prefix, const QString &arguments, const QByteArray &payload)
{
  QString command;
  int ack = ack_++;

  // Build the command
  command = prefix + " " + QString::number(ack);
  if( ! arguments.isEmpty() )
  {
    command += " " + arguments;
  }
  command += " " + QString::number(payload.size()) + "\r\n";

  // Write the data
  if( payload.size() > 0 )
  {
    writeBinaryData( command.toUtf8() + payload );
  }
  else
  {
    writeData( command );
  }

  // Return the ack used
  return ack;
}



// Specify which accepted commands carry payload data for this connection
void MsnConnection::setAcceptedPayloadCommands( QStringList commandList )
{
  socket_->setAcceptedPayloadCommands( commandList );
}



// Set whether we're sending pings or not
void MsnConnection::setSendPings( bool sendPings )
{
  socket_->setSendPings( sendPings );
}



/**
 * @brief Called when data is received from the server.
 *
 * This method parses the incoming data for the syntax used by the MSN protocol.
 *
 * A parsed message to handled to one of the following methods:
 * - one-line commands are delivered to parseCommand()
 * - <code>MSG</code> commands with MIME payloads are delivered to parseMimeMessage()
 * - other commands with payloads are delivered to parsePayloadMessage().
 *   This uses isPayloadCommand() to detect if a command has a payload.
 */
void MsnConnection::slotDataReceived( const QStringList &commandLine, const QByteArray &payloadData )
{
  // See if it's a payload command
  bool gotPayloadCommand = ( isPayloadCommand( commandLine[0] ) );

#ifdef KMESSDEBUG_SERVERMESSAGES
  // Display the received command line
  kDebug() << objectName() << "- <<< " << commandLine;
#endif

#ifdef KMESS_NETWORK_WINDOW
  KMESS_NET_RECEIVED( this, commandLine.join(" ").toUtf8() );
  // Also display the payload if present
  if( ! payloadData.isEmpty() )
  {
    KMESS_NET_RECEIVED( this, payloadData );
  }
#endif

  // Parse the messages.
  if( gotPayloadCommand )
  {
    // See if it's a Mime message.
    if( commandLine[0] != "MSG" )
    {
      // Other payload message types are parsed in the subclasses.
      parsePayloadMessage( commandLine, payloadData );
    }
    else
    {
      // It's a mime message
      // Note we can't parse strings here, the message could contain binary p2p data.
      // UTF-8 characters are translated in the MimeMessage constructor itself.
      MimeMessage mimeMessage( payloadData );

      // Detect Multi-packet messages
      if( ! mimeMessage.hasField("Message-ID") )
      {
        // It's a normal message, parse it.
        parseMimeMessage( commandLine, mimeMessage );
      }
      else
      {
        // Handle multi-packet message parts.
        parseMultiPacketMimeMessage( commandLine, mimeMessage );
      }
    }
  }
  else
  {
    // It's a normal command, handle it in the subclass.
    parseCommand( commandLine );
  }
}



/**
 * @brief Called when the socket is disconnected.
 */
void MsnConnection::slotDisconnected()
{
#ifdef KMESS_NETWORK_WINDOW
  KMESS_NET_CLOSE(this);
#endif

  emit disconnected();
}



/**
 * @brief A SOAP request failed with an error message.
 *
 * This method can be overwritten in the derived class to produce more specific error messages.
 * This default implemetentation is a fallback, it simply calls slotError() with a generic error message.
 *
 * @param  client    The SOAP client object.
 * @param  reason    The reported error by the SOAP client object.
 */
void MsnConnection::slotSoapError(HttpSoapConnection *client, const QString &reason)
{
  if(KMESS_NULL(client)) return;

  // Default implementation just gives a standard warning.
  slotError( i18n( "KMess could not access the remote webservice.<br>Details: %1", reason ),
             MsnSocketBase::ERROR_INTERNAL );
}



/**
 * @brief A SOAP request failed with an SOAP fault.
 *
 * A SOAP fault is reported by the server in response to the request.
 * For MSN SOAP services, some SOAP errors need to be acted on.
 *
 * This method could not be named slotSoapFault() as it causes a gcc "<method> was hidden" warning.
 * if the other slotSoapError() method is overwritten only.
 *
 * @param  client       The SOAP client object.
 * @param  faultCode    The SOAP fault code reported by the SOAP server.
 * @param  faultString  The SOAP fault string reported by the SOAP server.
 * @param  faultNode    The complete fault node, to parse any custom tags.
 */
void MsnConnection::slotSoapFault(HttpSoapConnection *client, const QString &faultCode,
                                  const QString &faultString, QDomElement /*faultNode*/)
{
  if(KMESS_NULL(client)) return;

  // Most of the times, the SOAP class itself also handles some errors,
  // and translates it into a specific signal.
  // If an error still gets through it's likely some server issue and it should be reported.

  slotSoapError(client, faultCode + ": " + faultString);
}



/**
 * @brief If it is possible, switch to the HTTP connection
 *
 * This method checks what connection is being established, and if is a TCP
 * connection, will delete it and re-start connecting using the HTTP method.
 *
 * This is used when the TCP connection process fails.
 */
bool MsnConnection::switchToHttpSocket()
{
  // Check if we're already using HTTP
  if( qobject_cast<MsnSocketHttp*>( socket_ ) )
  {
#ifdef KMESSDEBUG_CONNECTION_SOCKET
    kDebug() << "Cannot switch to HTTP connection. Class is" << socket_->metaObject()->className();
#endif
    return false;
  }

#ifdef KMESSDEBUG_CONNECTION_SOCKET
  kDebug() << "Attempting switch to HTTP connection.";
#endif

  // Save the socket details to use the same ones for the new socket
  MsnSocketBase::ServerType serverType = socket_->getServerType();
  QStringList acceptedPayloadCommands = socket_->getAcceptedPayloadCommands();

  // Close and disable the old server
  socket_->blockSignals( true );
  socket_->disconnectFromServer();
  socket_->deleteLater();

  // Replace it with the new one
  socket_ = new MsnSocketHttp( serverType );
  socket_->setAcceptedPayloadCommands( acceptedPayloadCommands );

  // Attach its signals to this class
  attachToSocketSignals();

  // Start connecting
  connectToServer();

  return true;
}



/**
 * @brief Switch back to the TCP connection
 *
 * This method checks what connection is being established, and if is an HTTP
 * connection, will delete it and start using again the TCP method.
 *
 * This is used after disconnecting from an HTTP session, to try again with TCP the next time.
 */
bool MsnConnection::switchToTcpSocket()
{
  // Check if we're already using TCP
  if( qobject_cast<MsnSocketTcp*>( socket_ ) )
  {
#ifdef KMESSDEBUG_CONNECTION_SOCKET
    kDebug() << "Cannot switch to TCP connection. Class is" << socket_->metaObject()->className();
#endif
    return false;
  }

#ifdef KMESSDEBUG_CONNECTION_SOCKET
  kDebug() << "Attempting to switch back to TCP connection.";
#endif

  // Save the socket details to use the same ones for the new socket
  MsnSocketBase::ServerType serverType = socket_->getServerType();
  QStringList acceptedPayloadCommands = socket_->getAcceptedPayloadCommands();

  // Close and disable the old server
  socket_->blockSignals( true );
  socket_->disconnectFromServer();
  socket_->deleteLater();

  // Replace it with the new one
  socket_ = new MsnSocketTcp( serverType );
  socket_->setAcceptedPayloadCommands( acceptedPayloadCommands );

  // Attach its signals to this class
  attachToSocketSignals();

  return true;
}



// Write data to the socket without conversions
void MsnConnection::writeBinaryData( const QByteArray& data )
{
  if( ! isConnected() )
  {
    kWarning() << objectName() << "- Attempting to write binary data while disconnected, aborting.";
    return;
  }

#ifdef KMESSDEBUG_SERVERMESSAGES
// To display the entire contents:
/*
  QByteArray debugData( data );
  for( int i = 0; i < debugData.size(); i++ ) { if( debugData[i] < 10 ) debugData[ i ] = '?'; }
  kDebug() << objectName() << "- >>> " << QString::fromUtf8( debugData.data(), debugData.size() );
*/
// To display the first readable part: (removing the final \r\n)
  kDebug() << objectName() << "- >>> " << QString( data ).trimmed();
#endif

  // Write to the socket
  socket_->writeBinaryData( data );

  // Append the message to the network window
#ifdef KMESS_NETWORK_WINDOW
  KMESS_NET_SENT( this, data );
#endif
}



// Write data to the socket
void MsnConnection::writeData( const QString& data )
{
#ifdef KMESSTEST
  if ( ( data.left(3) != "QRY" ) && ( data.left(3) != "MSG" ) )
  {
    KMESS_ASSERT( ( data.right( 2 ) ) == "\r\n" );
  }
#endif

  if( ! isConnected() )
  {
    kWarning() << objectName() << "- Attempting to write data while disconnected, aborting.";
    return;
  }


#ifdef KMESSDEBUG_SERVERMESSAGES
  kDebug() << objectName() << "- >>> " << data.trimmed();
#endif

  QByteArray binData = data.toUtf8();
  qint64 noBytesWritten = socket_->writeData( binData );

  if( noBytesWritten != binData.length() ) // Note that the previous code used data.length instead of unicode.length(), but I
  {                                         // changed this since that is the string we send to the socket... (could be wrong here)
    kWarning() << objectName() << "- Wanted to write " << data.length() << " bytes to the socket, wrote " << noBytesWritten << ".";
  }

  // Append message to the network window
#ifdef KMESS_NETWORK_WINDOW
  KMESS_NET_SENT( this, data.toUtf8() );
#endif
}


#include "msnconnection.moc"
