/***************************************************************************
                          msnsockethttp.cpp  -  description
                             -------------------
    begin                : Thu Apr 4 2008
    copyright            : (C) 2008 by Valerio Pilo
    email                : valerio@kmess.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "msnsockethttp.h"

#include <QEventLoop>
#include <QHostAddress>
#include <QHttp>
#include <QSslError>
#include <QTimer>

#include <KLocale>
#include <KMessageBox>

#include <config-kmess.h>
#include "../kmessdebug.h"

#ifdef KMESSDEBUG_CONNECTION // Area-specific debug statements
  #define KMESSDEBUG_CONNECTION_SOCKET_HTTP
  #define KMESSDEBUG_CONNECTION_SOCKET_HTTP_DUMPS
#endif


/**
 *  Default Gateway address/port
 *
 * This define holds the default address which KMess uses to send HTTP
 * messages. The port is usually 80.
 */
#define DEFAULT_GATEWAY_ADDRESS "gateway.messenger.hotmail.com"
#define DEFAULT_GATEWAY_PORT    80



/**
 * @brief The constructor
 *
 * Initializes the HTTP interface, prepares the default HTTP header we'll use
 * throughout the connection, and sets up the request/ping timer.
 */
MsnSocketHttp::MsnSocketHttp( ServerType serverType )
: MsnSocketBase( serverType )
, canSendRequests_( true )
, sendPings_( false )
{
  setObjectName( "MsnSocketHttp" );

  // Create the http interface
  http_ = new QHttp( this );

  // Connect the socket's signals for parsing
  connect( http_,          SIGNAL(  requestFinished(int,bool)                ),
           this,           SLOT  ( slotDataReceived(int,bool)                ) );
  connect( http_,          SIGNAL(        sslErrors(const QList<QSslError>&) ),
           this,           SLOT  (        slotError(const QList<QSslError>&) ) );
  // Connect the server request timer
  connect( &pollingTimer_, SIGNAL(          timeout()                        ),
           this,           SLOT  (  slotQueryServer()                        ) );

  // We need the timer to only fire once
  pollingTimer_.setSingleShot( true );

  // Set the default header values
  requestHeader_.setValue      ( "Accept",          "*" "/" "*"            );
  requestHeader_.setValue      ( "User-Agent",      "KMess/" KMESS_VERSION );
  requestHeader_.setValue      ( "Connection",      "Keep-Alive"           );
  requestHeader_.setValue      ( "Cache-Control",   "no-cache"             );
  requestHeader_.setContentType( "text/xml; charset=utf-8"                 );
  baseServerPath_ = QString    ( "/gateway/gateway.dll?"                   );

#ifdef KMESSTEST
  KMESS_ASSERT( http_ != 0 );
#endif
}



/**
 * @brief The destructor
 *
 * Closes the connection, cleans up request queues, and deletes the interface.
 */
MsnSocketHttp::~MsnSocketHttp()
{
  // Disconnect from the server on exit
  if( connected_ )
  {
    disconnectFromServer();
  }

  // Stop the polling timer
  pollingTimer_.stop();

  // Clean up and delete the http interface
  cleanUp();
  delete http_;

#ifdef KMESSDEBUG_CONNECTION_SOCKET_HTTP
  kDebug() << "DESTROYED";
#endif
}



/**
 * @brief Clean up the connection
 *
 * This internal method empties all queues and removes any pending connection -
 * including any in-progress ones.
 */
void MsnSocketHttp::cleanUp()
{
  // Clear the request queues
  qDeleteAll( queuedRequests_     );
  qDeleteAll( unansweredRequests_ );
  queuedRequests_    .clear();
  unansweredRequests_.clear();

  // Abort any pending requests
  if( connected_ )
  {
    http_->abort();
  }
}



/**
 * @brief Connect to the given server
 *
 * The initial connection is always made through the main Gateway.
 * The first server response usually carries the new gateway IP
 * which will be used for the rest of the connection.
 *
 * @param  server  The server hostname or IP address.
 * @param  port    The port to connect to. Not used - all of our
 *                 traffic goes through the default port.
 */
void MsnSocketHttp::connectToServer( const QString& server, const quint16 port )
{
  // We use our default HTTP port only
  Q_UNUSED( port );

  // When no server is specified, use the default msn one
  QString desiredAddress;
  if( server.isEmpty() )
  {
    desiredAddress = "messenger.hotmail.com";
  }
  else
  {
    desiredAddress = server;
  }

#ifdef KMESSDEBUG_CONNECTION_SOCKET_HTTP
  kDebug() << "Connecting to server at " << desiredAddress << ":" << DEFAULT_GATEWAY_PORT << ".";
#endif

  // Set up the gateway where to send our messages
  setGateway( DEFAULT_GATEWAY_ADDRESS );

  // Translate the server type to a 2-letter string to send in our requests
  QString serverType;
  switch( serverType_ )
  {
    case SERVER_NOTIFICATION: serverType = "NS"; break;
    case SERVER_SWITCHBOARD:  serverType = "SB"; break;
    default:
      kFatal() << "Unknown type of server" << serverType_ << "!";
      return;
  }

  // Set the parameters for the next request
  nextRequestParameters_ = "Action=open&Server=" + serverType + "&IP=" + desiredAddress;

  // Flush any active or pending request if the interface is closing a previous connection
  cleanUp();

  // Set our internal state variables
  connected_       = true;
  canSendRequests_ = true;

  // Signal we're ready
  emit connected();
  emit statusMessage( i18n( "Connected" ), false );
}



/**
 * @brief Disconnect from the server
 * 
 * This method closes the connection to the server and does some cleaning up.
 * It empties all buffers, and resets the internal state.
 * To avoid losing data, all pending requests are merged into a single huge
 * request and sent altogether, before closing the connection. We won't
 * receive responses for them; but at least we lose no outgoing data.
 *
 * @param  isTransfer  When set, a disconnected() signal won't be fired.
 *                     This is used to handle transfers to another server
 *                     without noticing a disconnect.
 */
void MsnSocketHttp::disconnectFromServer( bool isTransfer )
{
  // Stop pinging
  setSendPings( false );
  pollingTimer_.stop();

  // Just clean up if we are disconnected already
  if( ! connected_ )
  {
    return;
  }

  connected_ = false;

  // Make sure all remaining requests get sent, so that we don't lose messages.
  // Even if we're disconnecting because of an error, send them anyway: we'll ignore the
  // responses anyway.
  if( ! queuedRequests_.isEmpty() )
  {
#ifdef KMESSDEBUG_CONNECTION_SOCKET_HTTP
    kDebug() << "Sending" << queuedRequests_.count() << "remaining requests before closing.";
#endif

    // Get the oldest message
    Request *request = queuedRequests_.takeFirst();
    QByteArray data( request->data );

    // Fill up the header details
    QHttpRequestHeader header( requestHeader_ );
    header.setRequest( "POST", baseServerPath_ + request->requestParameters );

    // Force merging all remaining messages into one (not deleting anything, that will
    // be done automatically below)
    foreach( Request *req, queuedRequests_ )
    {
      // Avoid sending polling and special requests, which are not useful at this time
      if( req->isPollingRequest || req->isSpecialRequest )
      {
        continue;
      }

      // Merge the data
      data.append( req->data );
    }

    // Set up the size of the data in the header
    header.setContentLength( data.size() );
    if( data.size() == 0 )
    {
      header.removeValue( "Content-Type" );
    }

    // Send it
    http_->request( header, data );

    // Wait until this request has been sent, using a local loop.
    QEventLoop loop( this );
    pollingTimer_.start( 5000 );
    while( http_->currentId() != 0 && pollingTimer_.isActive() )
    {
      loop.processEvents();
    }
  }

  // Close the connection
  http_->close();

  // Clear the queues, so pending messages will be ignored
  cleanUp();

  // Do not signal a disconnection when we are transferring to another gateway
  if( isTransfer )
  {
    return;
  }

#ifdef KMESSDEBUG_CONNECTION_SOCKET_HTTP
  kDebug() << "Emitting disconnection signal.";
#endif

  emit disconnected();
}



/**
 * @brief Return the local IP address
 *
 * Using HTTP, we don't really care about our address. Not just that,
 * we're probably behind a firewall, so finding out our real IP wouldn't
 * help either.
 * This method just returns "127.0.0.1".
 *
 * @returns The IP address the socket uses at the system.
 */
QString MsnSocketHttp::getLocalIp() const
{
  // We don't care about our own IP, return a generic one
  return QHostAddress( QHostAddress::LocalHost ).toString();
}



/**
 * @brief Sends the next queued request
 *
 * Data to send is stored in a queue of requests. This method just takes the
 * oldest message and sends it.
 */
void MsnSocketHttp::sendNextRequest()
{
  // Verify if we can send another message, if there is any
  if( ! connected_ || ! canSendRequests_ || queuedRequests_.isEmpty() )
  {
    return;
  }

  QTime now( QTime::currentTime() );

  // Do not send messages too fast, or the server will ignore them
  int elapsedMsecs = lastRequestTime_.msecsTo( now );
  if( elapsedMsecs <= 1000 )
  {
#ifdef KMESSDEBUG_CONNECTION_SOCKET_HTTP
    kDebug() << "Not sending another message in" << elapsedMsecs << "more milliseconds.";
#endif

    // The polling method will call this method back
    pollingTimer_.start( 1000 - elapsedMsecs );
    return;
  }

  // Lock the sending process: only one message at most can be 'on air' in any given moment
  canSendRequests_ = false;

  // Get the oldest message
  Request *request = queuedRequests_.takeFirst();
  QByteArray data( request->data );

  // Fill up the header details
  QHttpRequestHeader header( requestHeader_ );
  header.setRequest( "POST", baseServerPath_ + request->requestParameters );

  // Polling requests are small and are sent all by theirselves
  if( request->isPollingRequest )
  {
    header.setRequest( "POST", baseServerPath_ + "Action=poll&SessionID=" + nextRequestId_ );
    header.setContentLength( 0 );
    header.removeValue( "Content-Type" );
  }
  else // Normal requests can be packed up with multiple messages
  {
#ifdef KMESSDEBUG_CONNECTION_SOCKET_HTTP
    int mergedRequests = 0;
#endif
    // Merge together all the queued data to save bandwidth
    foreach( Request *queuedRequest, queuedRequests_ )
    {
      // Delete polling requests eventually in the list
      if( queuedRequest->isPollingRequest )
      {
        queuedRequests_.removeAll( queuedRequest );
        delete queuedRequest;
        continue;
      }

      // Stop merging requests if we get a special one
      if( queuedRequest->isSpecialRequest )
      {
        break;
      }

#ifdef KMESSDEBUG_CONNECTION_SOCKET_HTTP
      mergedRequests++;
#endif

      // Merge the data
      data.append( queuedRequest->data );

      // Remove the request from the queue, since it is going to be sent now
      queuedRequests_.removeAll( queuedRequest );
      delete queuedRequest;
    }
#ifdef KMESSDEBUG_CONNECTION_SOCKET_HTTP
    kDebug() << "Merged" << mergedRequests << "requests out from the queue.";
#endif

    // Set up the size of the data in the header
    header.setContentLength( data.size() );
    if( data.size() == 0 )
    {
      header.removeValue( "Content-Type" );
    }
  }

  // Send it
  int requestId = http_->request( header, data );

#ifdef KMESSDEBUG_CONNECTION_SOCKET_HTTP
  kDebug() << "Sent request" << requestId << ".";
#endif
#ifdef KMESSDEBUG_CONNECTION_SOCKET_HTTP_DUMPS
  kDebug() << "Full dump:" << endl
           << header.toString()
           << QString::fromUtf8( request->data );
#endif

  // Add the message to the list of messages without response
  unansweredRequests_[ requestId ] = request;

  // Set the last request type
  lastRequestWasPolling_ = request->isPollingRequest;

  // Stop polling, the response to this message will carry new messages along
  pollingTimer_.stop();
  lastRequestTime_ = now;
}



/**
 * @brief Change the server connection gateway
 *
 * Set the address of the gateway, from this point on, all queued data will
 * be sent to the new gateway.
 *
 * @param  host  The new gateway hostname or IP address.
 */
void MsnSocketHttp::setGateway( QString host )
{
#ifdef KMESSDEBUG_CONNECTION_SOCKET_HTTP
  kDebug() << "Using server at" << host << ":" << DEFAULT_GATEWAY_PORT << "as gateway.";
#endif

  // Set the host
  http_->setHost( host );

  // Save the server address in our requests' fields
  requestHeader_.setValue( "Host", host );

  // The setHost() call starts establishing the connection with a request, so
  // while it's on air we can't send any other message
  canSendRequests_ = false;

  gatewayAddress_ = host;
}



/**
 * @brief Set whether to send pings or not
 *
 * When sending pings, the application receives a periodic signal
 * which is used to perform timed updates (like receiving display pics).
 *
 * @param  sendPings  True to send pings.
 */
void MsnSocketHttp::setSendPings( bool sendPings )
{
  sendPings_ = sendPings;
}



/**
 * @brief Send a polling message to the server, to retrieve pending commands
 *
 * This method gets called every two seconds: each time it checks if there still are
 * messages queued to be sent, and if not, it polls the gateway for new messages.
 */
void MsnSocketHttp::slotQueryServer()
{
  QDateTime now( QDateTime::currentDateTime() );

  // Emit a "ping" signal every about 30 seconds, to let the app perform scheduled tasks
  if( sendPings_ && ( ! lastPing_.isValid() || lastPing_.secsTo( now ) > 30 ) )
  {
    lastPing_ = now;
    emit pingSent();
  }

  // Don't send polling messages when: 1) we are disconnected; 2) a message is still being
  // responded at (we'll receive new messages along with the response); 3) there still are
  // messages to send (same reason as point 2).
  if( ! connected_ || ! canSendRequests_ )
  {
    return;
  }

  // If the queue isn't empty send the next message
  if( ! queuedRequests_.isEmpty() )
  {
    sendNextRequest();
    return;
  }

  // Prepare the new request
  Request *newRequest = new Request();
  newRequest->data = QByteArray();
  newRequest->isPollingRequest = true;

  // Queue it
  queuedRequests_.append( newRequest );

#ifdef KMESSDEBUG_CONNECTION_SOCKET_HTTP
  kDebug() << "Queued a new polling request";
#endif

  // The last response had data; maybe there's more so keep checking
  if( ! lastResponseWasEmpty_ )
  {
    sendNextRequest();
  }
  else
  {
    // Send a poll five seconds after the last command and ten after the last polling request
    pollingTimer_.start( lastRequestWasPolling_ ? 10000 : 5000 );
  }
}



/**
 * @brief Read received data and pass along the contained commands
 *
 * This method parses the incoming raw data for the syntax used by the MSN protocol.
 * Data arrives here only when it's been received completely. This may make it appear
 * slower than usual TCP.
 * Incoming messages contain all new messages from the server, so we need to bounce
 * them back individually to MsnConnection.
 *
 * @param  requestId  ID of the request for which we've received a response
 * @param  isError    True if it's an error response
 */
void MsnSocketHttp::slotDataReceived( int requestId, bool isError )
{
  // Get the message header and contents
  QByteArray          rawResponseData   = http_->readAll();
  QHttpResponseHeader rawResponseHeader = http_->lastResponse();
  bool                hasMsnHeader      = rawResponseHeader.hasKey( "X-MSN-Messenger" );

  // Discard unwanted messages (after you disconnect, all pending messages become unwanted)
  // Also ignore the first response, which is not valid and just used to establish the connection.
  if( ! unansweredRequests_.contains( requestId ) || ! rawResponseHeader.isValid() )
  {
    return;
  }

  // Catch errors. Also, catch responses without the "X-MSN-Messenger" header:
  // responses without it aren't coming from an MSN server, so maybe there is
  // some HTTP redirection happening, like an authentication web page
  if( isError || rawResponseHeader.statusCode() != 200 || ! hasMsnHeader )
  {
    kWarning() << "Received a type" << rawResponseHeader.statusCode() << "response for request" << requestId << ":" << http_->errorString();
    kWarning() << "Error number:" << http_->error() << "is error?" << isError;
#ifdef KMESSDEBUG_CONNECTION_SOCKET_HTTP_DUMPS
    kDebug() << "Full dump:" << endl
             << rawResponseHeader.toString()
             << QString::fromUtf8( rawResponseData );
#endif

    // Report the right type of error
    ErrorType errorType;
    switch( http_->error() )
    {
      case QHttp::NoError:
        // Invalid MSN responses (web authentication redirects often replace the
        // original response contents). Treat like a connection error.
      case QHttp::HostNotFound:
      case QHttp::ConnectionRefused:
        errorType = ERROR_CONNECTING;
        break;

      case QHttp::UnexpectedClose:
        errorType = ERROR_DROP;
        break;

      default:
        errorType = ERROR_DATA;
        break;
    }

    // Ignore errors when disconnected: avoids duplicate error dialogs when two
    // following requests are answered with an error.
    if( connected_ )
    {
      emit error( http_->errorString(), errorType );
    }

    // Allow the connection to continue working: if the error is fatal, the
    // connection will be disconnected so no more messages will be sent
    canSendRequests_ = true;
    sendNextRequest();
    return;
  }

  // Extract the MSN informative header from the response
  QStringList msnInfo = rawResponseHeader.value( "X-MSN-Messenger" ).split( ";" );

  // Delete the responded request from the list
  Request *doneRequest = unansweredRequests_.take( requestId );
#ifdef KMESSDEBUG_CONNECTION_SOCKET_HTTP
  kDebug() << "Received response to request" << requestId << "(" << doneRequest->data.simplified() << ")";
#endif
#ifdef KMESSDEBUG_CONNECTION_SOCKET_HTTP_DUMPS
  kDebug() << "Full dump:" << endl
           << rawResponseHeader.toString()
           << QString::fromUtf8( rawResponseData );
#endif
  delete doneRequest;

  // Extract the MSN session information from the headers
  foreach( QString item, msnInfo )
  {
    item = item.trimmed();
    int keyEnd = item.indexOf( '=' );
    QString key   = item.left( keyEnd );
    QString value = item.mid ( keyEnd + 1 );

    if( key == "SessionID" )
    {
      // Every message has a different session ID, which needs to be sent along
      // with the next message.
      nextRequestId_ = value;
    }
    else if( key == "GW-IP" )
    {
      // Change the gateway, if we didn't do it already
      if( value != gatewayAddress_ )
      {
#ifdef KMESSDEBUG_CONNECTION_SOCKET_HTTP
        kDebug() << "Received gateway change command";
#endif
        setGateway( value );
      }
    }
    else if( key == "Session" )
    {
      // Session close request. We've been kicked or the connection was closed by us first.
      // Do nothing if we are disconnected already.
      // TODO Behavior when receiving this command needs to be tested.
      if( value == "close" && connected_ )
      {
#ifdef KMESSDEBUG_CONNECTION_SOCKET_HTTP
        kDebug() << "Received disconnection command";
#endif

        cleanUp();
        disconnectFromServer();
      }
    }
    else
    {
#ifdef KMESSDEBUG_CONNECTION_SOCKET_HTTP
      kDebug() << "Unknown attribute received in the MSN Header:" << item;
#endif
    }
  }

  lastResponseWasEmpty_ = true;

  // Extract all new messages from the response data
  if( ! rawResponseData.isEmpty() )
  {
    int         startingPos = 0;
    QStringList command;
    QString     commandLine;
    int         commandLineSize;
    QByteArray  payloadData;

    // Loop until the entire response has been parsed
    while( true )
    {
      if( startingPos >= rawResponseData.size() )
      {
        break;
      }

      // Retrieve the first available command from the response body
      commandLineSize = rawResponseData.indexOf( "\r\n", startingPos ) - startingPos + 2; // Add the newline characters to the command size
      commandLine     = QString::fromUtf8( rawResponseData.mid( startingPos, commandLineSize ) );
      command         = commandLine.trimmed().split(" ");

      // Stop if there's nothing more to parse (unlikely, most times the first
      // 'if' statement above ends the loop)
      if( command[0].isEmpty() )
      {
        break;
      }

      // See if it's a payload command, and extract the payload from the response data
      if( isPayloadCommand( command[0] ) )
      {
        int payloadSize = command[ command.count() - 1 ].toInt();  // Last arg is payload length.
        payloadData = rawResponseData.mid( startingPos + commandLineSize, payloadSize );

        startingPos += commandLineSize + payloadSize;
      }
      else // Not a payload message, just go on from right after the command
      {
        payloadData.clear();

        startingPos += commandLineSize;
      }

      // We've received a message, hand it over to MsnConnection for interpretation
      emit dataReceived( command, payloadData );
      lastResponseWasEmpty_ = false;
    }
  }

  // Allow sending new messages, now that this has been done
  canSendRequests_ = true;

  // Keep querying for more
  slotQueryServer();
}



/**
 * @brief Called when an SSL error happens
 *
 * Reports the error to MsnConnection.
 *
 * @param  errorList  List of SSL errors received.
 */
void MsnSocketHttp::slotError( const QList<QSslError> &errorList )
{
  kWarning() << "Received SSL errors:" << errorList;

  emit error( i18n( "SSL Error" ), ERROR_DATA );
}



/**
 * @brief Write data to the gateway without any conversion
 *
 * This method creates a new request and queues it to be sent when possible.
 * See sendNextRequest() for more info on sending.
 *
 * @param  data  Contents of the message which will be sent
 * @return -1 on error, or else always the exact size of the sent data.
 */
qint64 MsnSocketHttp::writeBinaryData( const QByteArray &data )
{
  if( ! connected_ )
  {
    kWarning() << "Attempted to write data to a disconnected interface, aborting.";
    return -1;
  }

  // Remove any polling messages from the queue, because the response to this
  // new message will carry new messages too.
#ifdef KMESSDEBUG_CONNECTION_SOCKET_HTTP
  int deletedPollingRequests = 0;
#endif
  foreach( Request *req, queuedRequests_ )
  {
    if( req->isPollingRequest )
    {
      queuedRequests_.removeAll( req );
      delete req;

#ifdef KMESSDEBUG_CONNECTION_SOCKET_HTTP
      deletedPollingRequests++;
#endif
    }
  }
#ifdef KMESSDEBUG_CONNECTION_SOCKET_HTTP
  kDebug() << "Deleted" << deletedPollingRequests << "polling requests from the queue.";
#endif

  // Create the new request with the data
  Request *newRequest = new Request();
  newRequest->data = data;
  newRequest->isPollingRequest = false;

  // If there's some special parameters to be sent, override the default ones
  if( ! nextRequestParameters_.isEmpty() )
  {
    newRequest->isSpecialRequest = true;
    newRequest->requestParameters = nextRequestParameters_;
    // Clear up the special parameters for the next request
    nextRequestParameters_ = QString();
  }
  else // Send the default parameters
  {
    newRequest->isSpecialRequest = false;
    newRequest->requestParameters = "SessionID=" + nextRequestId_;
  }

  // Queue the new request
  queuedRequests_.append( newRequest );

#ifdef KMESSDEBUG_CONNECTION_SOCKET_HTTP
  kDebug() << "Queued a new request";
#endif

  // Try to immediately send the new message, if possible
  sendNextRequest();

  return data.size();
}



/**
 * @brief Write a string to the gateway
 *
 * This is a convenience method for writeBinaryData, which can be used to directly send a QString.
 *
 * @param  data  The message which will be sent
 * @return -1 on error, or else always the exact size of the sent data.
 */
qint64 MsnSocketHttp::writeData( const QString &data )
{
  return writeBinaryData( data.toUtf8() );
}



#include "msnsockethttp.moc"
