/***************************************************************************
                          httpsoapconnection.cpp -  description
                             -------------------
    begin                : Sun Sep 25 2005
    copyright            : (C) 2005 by Diederik van der Boor
    email                : vdboor --at-- codingdomain.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "httpsoapconnection.h"

#include <QAuthenticator>
#include <QHttp>
#include <QSslError>

#include <KLocale>

#include "../../utils/xmlfunctions.h"
#include "../../kmessdebug.h"
#include "../mimemessage.h"
#include "soapmessage.h"

#include "config-kmess.h"


#ifdef KMESSDEBUG_HTTPSOAPCONNECTION
  #define KMESSDEBUG_HTTPSOAPCONNECTION_GENERAL
//   #define KMESSDEBUG_HTTPSOAPCONNECTION_HTTPDUMP
#endif


#define XMLNS_XSD  "http://www.w3.org/2001/XMLSchema"
#define XMLNS_XSI  "http://www.w3.org/2001/XMLSchema-instance"
#define XMLNS_SOAP "http://schemas.xmlsoap.org/soap/envelope/"



/**
 * @brief The constructor
 *
 * Initializes the class.
 * The actual connection is established when a SOAP call is made.
 *
 * @param  endpoint  The endpoint of the webservice, should be a http:// or https:// URL.
 * @param  parent    The Qt parent object. The object is deleted when the parent is deleted.
 */
HttpSoapConnection::HttpSoapConnection( const QString &endpoint, QObject *parent )
  : QObject( parent )
  , currentRequestId_( 0 )
{
#ifdef KMESSDEBUG_HTTPSOAPCONNECTION_GENERAL
  kDebug() << "CREATED. (endpoint=" << endpoint << ")";
#endif

  // Create the http socket
  http_ = new QHttp( this );

  // Make signal connections
  connect( http_, SIGNAL(                requestFinished(int,bool) ),
           this,  SLOT  (        slotHttpRequestFinished(int,bool) ));
  connect( http_, SIGNAL(         responseHeaderReceived(const QHttpResponseHeader&) ),
           this,  SLOT  ( slotHttpResponseHeaderReceived(const QHttpResponseHeader&) ));

  // Also connect to signals for error reporting
  connect( http_, SIGNAL(         authenticationRequired(const QString&, quint16, QAuthenticator*) ),
           this,  SLOT  ( slotHttpAuthenticationRequired(const QString&, quint16, QAuthenticator*) ));
  connect( http_, SIGNAL(              sslErrors(const QList<QSslError>&) ),
           this,  SLOT  (          slotSslErrors(const QList<QSslError>&) ));

  // Set endpoint setting
  setEndpoint( endpoint );
}



/**
 * @brief The destructor.
 *
 * Closes the sockets if these are still open.
 */
HttpSoapConnection::~HttpSoapConnection()
{
#ifdef KMESSDEBUG_HTTPSOAPCONNECTION_GENERAL
  kDebug() << "DESTROYED.";
#endif

  closeConnection();
  delete http_;
}



/**
 * @brief Delete any user data associated with the request.
 *
 * This method is called when a request is completely handled.
 * Overwrite this method to delete any call-specific data stored for the request ID.
 *
 * @param requestId  ID of the request, as returned by sendRequest().
 */
void HttpSoapConnection::clearRequestData( int requestId )
{
  Q_UNUSED( requestId );
  // Empty default.
  // can be overwritten;
}



/**
 * @brief Close the connection.
 */
void HttpSoapConnection::closeConnection()
{
  // Close the connection
  http_->close();
}



/**
 * @brief Escape a string for inclusion in XML.
 * @param  value  The value to escape for inclusion in an XML message.
 * @return  The escaped value, with the &lt;, &gt;, &amp; and " tokens escaped.
 */
QString HttpSoapConnection::escapeString( QString value ) const
{
  // Only escape basic properties, nothing else!
  return value.replace("<", "&lt;")
              .replace(">", "&gt;")
              .replace("\"", "&quot;")
              .replace("&", "&amp;");
}



/**
 * @brief Return the currently active request ID.
 *
 * This value is only valid inside the parseSoapFault() and parseSoapResult() slots.
 * When this method is called outside these slots, the value is 0.
 *
 * @return The ID of the request which is currently being processed.
 */
int HttpSoapConnection::getCurrentRequestId() const
{
#ifdef KMESSTEST
  KMESS_ASSERT( currentRequestId_ == http_->currentId() );
#endif

  return currentRequestId_;
}



/**
 * @brief Return the endpoint
 * @return  Returns the endpoint given with the constructor.
 */
QString HttpSoapConnection::getEndpoint() const
{
  return endpoint_.toString();
}



/**
 * @brief Return the last SOAP action
 * @return  The last SOAP action sent.
 */
const QString & HttpSoapConnection::getSoapAction() const
{
  return soapAction_;
}



/**
 * @brief Return whether the connection is idle.
 * @return  Returns true when the connection is idle, false when a SOAP request is pending / being processed.
 */
bool HttpSoapConnection::isIdle()
{
  return http_->state() == QHttp::Unconnected
      || http_->state() == QHttp::Closing;
}



/**
 * @brief Parse the received server response.
 *
 * This body extracts the XML message from the body.
 * It detects SOAP Faults automatically.
 * Other responses are returned to parseSoapResult().
 *
 * @param   responseBody  The full message body received from the HTTP server.
 * @return  Whether there is a parsing error that should be emitted.
 */
bool HttpSoapConnection::parseHttpBody( const QByteArray &responseBody )
{
#ifdef KMESSDEBUG_HTTPSOAPCONNECTION_GENERAL
  kDebug() << "Request finished, pasing body.";
#endif

  // Attempt to convert to XML
  // see http://doc.trolltech.com/4.3/qdomdocument.html#setContent
  QString xmlError;
  QDomDocument xml;
  if( ! xml.setContent( responseBody, true, &xmlError ) )   // true for namespace processing.
  {
    kWarning() << "Unable to parse XML "
                  "(error='" << xmlError << "' endpoint=" << endpoint_ << ")" << endl;
    return false;
  }

  // See if there is a header element
  QDomElement headerNode = XmlFunctions::getNode(xml, "/Envelope/Header").firstChild().toElement();

  // See if there is a Fault in the main Envelope (happens with passport RST service).
  QDomElement faultNode = XmlFunctions::getNode(xml, "/Envelope/Fault").toElement();
  if( ! faultNode.isNull() )
  {
#ifdef KMESSDEBUG_HTTPSOAPCONNECTION_GENERAL
    kDebug() << "Found soap fault outside the body!";
#endif
    closeConnection();
    parseSoapFault( faultNode, headerNode );
    return true;   // no parsing error, it's a valid response.
  }

  // Get the main body node
  QDomElement resultNode = XmlFunctions::getNode(xml, "/Envelope/Body").firstChild().toElement();
  if( resultNode.isNull() )
  {
    kWarning() << "Unable to parse response: result node is null "
                << "(root=" << xml.nodeName()  << " endpoint=" << endpoint_ << ")." << endl;
    return false;
  }

  // See if it's a soap error message, fault in the body.
  // NOTE: the test is done with localName() instead of nodeName(), because the prefix may differ.
  if( resultNode.localName() == "Fault"
  &&  resultNode.namespaceURI() == XMLNS_SOAP )
  {
#ifdef KMESSDEBUG_HTTPSOAPCONNECTION_GENERAL
    kDebug() << "Found soap fault.";
#endif
    closeConnection();
    parseSoapFault( resultNode, headerNode );
    return true;   // no parsing error, it's a valid response.
  }

  // WS-I compliant webservices wrap the response in another object so the response has one root element.
  // So you get a <MethodNameResponse><MethodNameResult>...</MethodNameResult></MethodNameResponse>
  // This object wrapper is stripped transparently in .Net, also do this here:
  QString resultNodeName = resultNode.nodeName();
  if( resultNodeName.endsWith("Response")
  &&  resultNode.childNodes().length()   == 1
  &&  resultNode.firstChild().nodeName() == resultNodeName.replace("Response", "Result") )
  {
    resultNode = resultNode.firstChild().toElement();
  }

  // Request completed, indicate derived class and listeners.
#ifdef KMESSDEBUG_HTTPSOAPCONNECTION_GENERAL
  kDebug() << "Request finished, signalling to start parsing of the result node "
              "(name:"      << resultNode.nodeName()     <<
              " prefix:"    << resultNode.prefix()       <<
              " localName:" << resultNode.localName()    <<
              " namespace:" << resultNode.namespaceURI() << ")" << endl;
#endif

  // Notify listeners
  parseSoapResult( resultNode, headerNode );
  return true;
}



/**
 * @brief  Parse the SOAP fault.
 *
 * This method extracts the node data, and emits a requestFailed() signal.
 * Overwrite this method to implement custom error handling.
 *
 * @param  faultNode   The XML node which contains the SOAP fault.
 * @param  headerNode  The XML node which contains the SOAP header.
 */
void HttpSoapConnection::parseSoapFault( const QDomElement &faultNode, const QDomElement &/*headerNode*/ )
{
  QString faultCode   = XmlFunctions::getNodeValue( faultNode, "/faultcode" );
  QString faultString = XmlFunctions::getNodeValue( faultNode, "/faultstring" );

#ifdef KMESSDEBUG_HTTPSOAPCONNECTION_GENERAL
  kDebug() << "Signalling the request failed, "
           << faultCode << ": " << faultString << endl;
#endif

  emit requestFailed( this, faultCode, faultString, faultNode );
}



/**
 * @brief  Send the a raw POST request to the server.
 *
 * This method is used internally by sendRequest().
 *
 * @param  content     The POST data to send.
 * @param  soapAction  The SOAPAction header to send, may be QString::null to omit the header.
 * @return  The request ID.
 */
int HttpSoapConnection::sendRawPostRequest( const QString &content, const QString &soapAction )
{
#ifdef KMESSTEST
  if( ! isIdle() )
  {
    kDebug() << "NOTICE: A new SOAP request started while the previous request is not finished yet.";
  }
#endif

  // Create HTTP request
  QHttpRequestHeader header( "POST", endpoint_.path() );
  QByteArray rawPostData = content.toUtf8();

  if( ! soapAction.isNull() )
  {
    header.setValue( "SOAPAction", "\"" + soapAction + "\"" );
  }

  header.setValue( "Host", endpoint_.host() );
  header.setValue( "Content-Type", "text/xml; charset=utf-8" );
  header.setValue( "Content-Length", QString::number( rawPostData.length() ) );
  header.setValue( "User-Agent", "KMess/" KMESS_VERSION );
  header.setValue( "Accept", "*/*" );
  header.setValue( "Cache-Control", "no-cache" );

  // Initialize state vars
  soapAction_    = soapAction_;

#ifdef KMESS_NETWORK_WINDOW
  QUrl soapActionUrl( soapAction );
  if( ! soapAction.isEmpty() && soapActionUrl.isValid() )
  {
    KMESS_NET_INIT(this, "SOAP " + soapActionUrl.path() );
  }
  else
  {
    KMESS_NET_INIT(this, "SOAP " + endpoint_.path() );
  }

  KMESS_NET_SENT( this, header.toString().toUtf8() );
  KMESS_NET_SENT( this, rawPostData );
#endif

#ifdef KMESSDEBUG_HTTPSOAPCONNECTION_HTTPDUMP
  kDebug() << "Sending HTTP request:\n" << header.toString().toUtf8() + "\n\n" + content;
#endif

  // Send the request
  int httpRequestId = http_->request( header, rawPostData );
  return httpRequestId;
}



/**
 * @brief Send a SOAP request to the webservice.
 *
 * This is a convenience function for sendRawPostRequest().
 * The message is generated by SoapMessage::getMessage().
 *
 * @param  soapAction  The SOAPAction header to send, may be QString::null to omit the header.
 * @param  message     The SOAP message to send.
 * @return  The request ID.
 */
int HttpSoapConnection::sendRequest(const QString &soapAction, const SoapMessage &message )
{
  return sendRawPostRequest( message.getMessage(), soapAction );
}



/**
 * @brief Send a SOAP request to the webservice.
 *
 * This method constructs the <code>&lt;soap:Envelope&gt;</code>,
 * <code>&lt;soap:Header&gt;</code> and <code>&lt;soap:Body&gt;</code> fields
 * to embedding the message. The generated message is sent with sendRawPostRequest().
 *
 * When an example defines additional XML namespaces in the <code>&lt;soap:Envelope&gt;</code> header,
 * include those namespaces in the sub elements instead. Make sure the XML namespace is always
 * declared in the parent element which uses the XML namespace. 
 *
 * @param  soapAction     The SOAPAction header to send, may be QString::null to omit the header.
 * @param  messageBody    The raw XML body to send. For a readable formatting in the network window, use 4 spaces indent and no trailing newline.
 * @param  messageHeader  The raw XML header to send.
 * @return  The request ID.
 */
int HttpSoapConnection::sendRequest( const QString &soapAction, const QString &messageBody, const QString &messageHeader )
{
#ifdef KMESSDEBUG_HTTPSOAPCONNECTION_GENERAL
  kDebug() << "Connecting to endpoint for SOAP action '" << soapAction << "'.";
#endif

  // Build the SOAP request
  QString postData = "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n"
                     "<soap:Envelope"
                     " xmlns:xsi=\"" XMLNS_XSI "\""
                     " xmlns:xsd=\"" XMLNS_XSD "\""
                     " xmlns:soap=\"" XMLNS_SOAP "\">\n";

  // Append header
  if( ! messageHeader.isEmpty() )
  {
    postData += "  <soap:Header>\n" + messageHeader + "\n  </soap:Header>\n";
  }

  // Append body
  postData += "  <soap:Body>\n" + messageBody + "\n  </soap:Body>\n</soap:Envelope>\n";

  return sendRawPostRequest( postData, soapAction );
}



/**
 * @brief Change the SOAP endpoint.
 *
 * This method is available for the rare occasion
 * a webservice wants to change it's endpoint (e.g. passport login).
 */
void HttpSoapConnection::setEndpoint( const QString &endpoint )
{
#ifdef KMESSDEBUG_HTTPSOAPCONNECTION_GENERAL
  QString oldEndpoint = endpoint_.toString();
#endif
#ifdef KMESSTEST
  KMESS_ASSERT( endpoint.startsWith("http://") || endpoint.startsWith("https://") );
  KMESS_ASSERT( http_->state() == QHttp::Unconnected );
#endif

  // Avoid a http host request if this is not needed.
  if( endpoint_ == endpoint )
  {
    return;
  }

  // Parse the endpoint
  endpoint_.setUrl( endpoint );
  QString protocol = endpoint_.scheme();
  QHttp::ConnectionMode mode = ( protocol == "https" ) ? QHttp::ConnectionModeHttps : QHttp::ConnectionModeHttp;

  
  // Make host request, tricky bits:
  // - this starts a request so slotHttpRequestFinished() will be called.
  // - if mode is an int, it will call setHost( name, port ) instead of setHost( name, mode, port = 0 )
  //   so appending the default 3rd argument so it's more difficult to break.
  int hostRequestId = http_->setHost( endpoint_.host(), mode, 0 );
  hostRequestIds_.append( hostRequestId );

#ifdef KMESSDEBUG_HTTPSOAPCONNECTION_GENERAL
  if( endpoint_.isValid() )
  {
    kDebug() << "Changing HttpSoapConnection endpoint from " << oldEndpoint << " to " << endpoint << ", request ID" << hostRequestId;
  }
#endif
}



/**
 * @brief Called when the remote server requires authentication.
 */
void HttpSoapConnection::slotHttpAuthenticationRequired( const QString &hostname, quint16 port, QAuthenticator *authenticator )
{
  Q_UNUSED(hostname);
  Q_UNUSED(port);

  kWarning() << "Got http authentication request for" << authenticator->realm() <<
                "while connecting to" << endpoint_;
}



// The request to the remote server finished
void HttpSoapConnection::slotHttpRequestFinished( int requestId, bool error )
{
  // Get the error code
  QHttp::Error errorCode = http_->error();

  // Very annoying. QHttp also generates a request for the setHost() and close() calls.
  if( hostRequestIds_.contains( requestId ) )
  {
#ifdef KMESSDEBUG_HTTPSOAPCONNECTION_GENERAL
    kDebug() << "completed host request (id=" << requestId << "error=" << errorCode << ").";
#endif

    // Abort next requests if the host request failed.
    if( error )
    {
      kWarning() << "Failed to resolve host" << endpoint_;

      // It turns out that calling abort() from a requestFinished() slot causes crashes later (Qt 4.3.3)
//    http_->abort();
      http_->clearPendingRequests();
    }

    return;
  }

  // Catch other QHttp requests we don't want to parse.
  if( ! http_->currentRequest().isValid() )
  {
#ifdef KMESSDEBUG_HTTPSOAPCONNECTION_GENERAL
    kDebug() << "completed close request (id=" << requestId << "error=" << errorCode << ").";
#endif

    return;
  }

#ifdef KMESSDEBUG_HTTPSOAPCONNECTION_GENERAL
  QHttpResponseHeader header = http_->lastResponse();
  kDebug() << "start processing http response (id=" << requestId << "error=" << errorCode << "status=" << header.statusCode() << header.reasonPhrase() << ").";
#endif

  // Set current request ID for all slots.
  currentRequestId_ = requestId;


  // See if an error occurred
  if( error )
  {
    // Avoid double errors by our abort() calls.
    if( errorCode == QHttp::Aborted )
    {
      return;
    }

    // Display at the console for logging
    QString httpError = http_->errorString();
    kWarning() << "Got HTTP error '" << httpError << "' while connecting to '" << endpoint_ << "'.";

    // Inform listeners the request failed.
    emit requestFailed( this, i18n( "Unable to make a connection: %1", httpError ) );

    // Reset all data.
    clearRequestData( currentRequestId_ );
    currentRequestId_ = 0;
    return;
  }


  // Read the body, make copy
  bool responseOk = true;
  QByteArray body = http_->readAll();

  // Avoid XML errors if there is no response
  if( body.isEmpty() )
  {
    kWarning() << "Received an empty response "
                  "(state="    << http_->state() <<
                  " error="    << http_->error() <<
                  " endpoint=" << endpoint_      << ")";
    responseOk = false;
  }
  else
  {
    // Parse the body
#ifdef KMESS_NETWORK_WINDOW
    KMESS_NET_RECEIVED( this, body );
#endif

#ifdef KMESSDEBUG_HTTPSOAPCONNECTION_HTTPDUMP
    kDebug() << "Received the HTTP response:\n" << body;
#endif

    responseOk = parseHttpBody( body );
  }

  // emit later, so the caller can destroy this class
  if( ! responseOk )
  {
    // It turns out that calling abort() from a requestFinished() slot causes crashes later (Qt 4.3.3)
//    http_->abort();
    http_->clearPendingRequests();

    emit requestFailed( this, i18n( "There was an internal error in KMess: %1",
                                    i18n("Unable to parse SOAP response") ) );
  }
  else
  {
#ifdef KMESSDEBUG_HTTPSOAPCONNECTION_GENERAL
    kDebug() << "completed response handling (id=" << requestId << ")";
#endif
  }

  // Reset
  currentRequestId_ = 0;
}



/**
 * Called when the response header is received
 */
void HttpSoapConnection::slotHttpResponseHeaderReceived( const QHttpResponseHeader &header )
{
  // Log it in the network window.
#ifdef KMESS_NETWORK_WINDOW
  KMESS_NET_RECEIVED( this, header.toString().toLatin1() );
#endif

  // TODO: generate custom error message, and call http_->abort().

  // Check the response header
  if( ! header.isValid() )
  {
    kWarning() << "could not parse last http response from '" << endpoint_ << "'.";
    return;
  }

  // Check response code
  int statusCode = header.statusCode();
  if( statusCode != 200 && statusCode != 500 )   // 500 might happen with SOAP faults.
  {
    kWarning() << "received unexpected header code: " << header.statusCode() << " " << header.reasonPhrase() << ".";
    return;
  }

  // Check Content-Type
  QString contentType = header.contentType();
  if( contentType.contains(";") )
  {
    contentType = contentType.section(';', 0, 0);
  }

  if( contentType != "text/xml"
  &&  contentType != "text/html" )  // Needed for PassportLoginService, amazing!
  {
    kWarning() << "received unexpected response type '" << contentType << "' from '" << endpoint_ << "'.";
    return;
  }

#ifdef KMESSDEBUG_HTTPSOAPCONNECTION_HTTPDUMP
  kDebug() << "Received the HTTP response header:\n" << header.toString().toLatin1();
#endif
}



/**
 * @brief Called when an ssl error occurred while setting up the connection.
 */
void HttpSoapConnection::slotSslErrors( const QList<QSslError> &sslErrors )
{
  foreach( QSslError error, sslErrors )
  {
    kWarning() << "Got SSL error '" << error.errorString() << "' while connecting to '" << endpoint_ << "'.";
  }

  // It's needed to ignore SSL errors because rsi.hotmail.com uses an invalid certificate.
  http_->ignoreSslErrors();
}



#include "httpsoapconnection.moc"
