/***************************************************************************
                          offlineimservice.cpp -  description
                             -------------------
    begin                : Thu Sep 07 2006
    copyright            : (C) 2006 by Diederik van der Boor
    email                : "vdboor" --at-- "codingdomain.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "offlineimservice.h"

#include "../../utils/xmlfunctions.h"
#include "../../utils/kmessshared.h"
#include "../../currentaccount.h"
#include "../../kmessdebug.h"
#include "../applications/p2papplication.h"
#include "../chatmessage.h"
#include "../mimemessage.h"
#include "../msnchallengehandler.h"

#include <QTextDocument>

#include <KDateTime>


#ifdef KMESSDEBUG_OFFLINE_IM
#define KMESSDEBUG_OFFLINE_IM_GENERAL
#endif


#define PRODUCT_KEY "ILTXC!4IXB5FB*PX"
#define PRODUCT_ID  "PROD0119GSJUC$18"



/**
 * @brief  The receiving mode constructor.
 *
 * Initializes the client to receive offline messages using the Offline-IM webservice.
 *
 * @param  authT   The <code>t</code> value of the passport cookie.
 * @param  authP   The <code>p</code> value of the passport cookie.
 * @param  parent  The Qt parent object, when it's destroyed this class will be cleaned up automatically too.
 */
OfflineImService::OfflineImService( const QString &authT, const QString &authP,
                                     QObject *parent )
  : HttpSoapConnection( "https://rsi.hotmail.com/rsi/rsi.asmx", parent )
  , authT_(authT)
  , authP_(authP)
  , nextSequenceNum_(0)
  , passportService_(0)
{
#ifdef KMESSDEBUG_OFFLINE_IM_GENERAL
  kDebug() << "CREATED receiver service.";
#endif

  setObjectName( "OfflineImService[receiver]" );

  // Initialize the header once. 
  passportCookieHeader_ =
      "    <PassportCookie xmlns=\"http://www.hotmail.msn.com/ws/2004/09/oim/rsi\">\n"
      "      <t>" + escapeString(authT) + "</t>\n"
      "      <p>" + escapeString(authP) + "</p>\n"
      "    </PassportCookie>";
}



/**
 * @brief  The sending mode constructor.
 *
 * Initializes the client to send offline messages using the Offline-IM webservice.
 *
 * @param  parent  The Qt parent object, when it's destroyed this class will be cleaned up automatically too.
 */
OfflineImService::OfflineImService( QObject *parent )
  : HttpSoapConnection( "https://ows.messenger.msn.com/OimWS/oim.asmx", parent )
  , nextSequenceNum_(0)
  , passportService_(0)
{
#ifdef KMESSDEBUG_OFFLINE_IM_GENERAL
  kDebug() << "CREATED sender service.";
#endif

  setObjectName( "OfflineImService[sender]" );
}



/**
 * @brief  The destructor.
 */
OfflineImService::~OfflineImService()
{
  delete passportService_;

  // Delete stored request data
  qDeleteAll( getMessageCallData_ );
  qDeleteAll( storeMessageCallData_ );

#ifdef KMESSDEBUG_OFFLINE_IM_GENERAL
  kDebug() << "DESTROYED.";
#endif
}



/**
 * @brief Delete the temporary data associated with the request id.
 *
 * This method is called by the base class when a request is completely processed.
 *
 * @param requestId  ID of the request, as returned by sendRequest().
 */
void OfflineImService::clearRequestData( int requestId )
{
#ifdef KMESSDEBUG_OFFLINE_IM_GENERAL
  kDebug() << "deleting call data of request" << requestId;
#endif

  // Delete stored objects
  delete getMessageCallData_.take( requestId );
  delete storeMessageCallData_.take( requestId );

  // Reset other state tracking vars.
  passportRequestIds_.removeAll( requestId );
}



/**
 * @brief Create a passport login service on demand.
 *
 * This service is used to authenticate the client to send offline messages.
 */
PassportLoginService * OfflineImService::createPassportLoginService()
{
  // Create the new passport service
  PassportLoginService *passportService = new PassportLoginService( this );

  // Connect the signals to continue sending the offline message
  connect( passportService, SIGNAL(         loginIncorrect() ),
           this,            SLOT  ( slotLoginRequestFailed() ) );
  connect( passportService, SIGNAL(         loginSucceeded(QString, QString, QString)  ),
           this,            SLOT  (     slotLoginSucceeded(QString, QString, QString)  ) );

  // Also relay the standard signals.
  connect( passportService, SIGNAL(          requestFailed(HttpSoapConnection*, const QString&) ),
           this,            SLOT  ( slotLoginRequestFailed(HttpSoapConnection*, const QString&) ) );
  connect( passportService, SIGNAL(          requestFailed(HttpSoapConnection*, const QString&, const QString&, QDomElement) ),
           this,            SLOT  ( slotLoginRequestFailed(HttpSoapConnection*, const QString&, const QString&, QDomElement) ) );

  return passportService;
}



/**
 * @brief  SOAP call to delete messages from the remote storage.
 * @param  messageIds  List of messages to delete.
 */
void OfflineImService::deleteMessages( const QStringList &messageIds )
{
#ifdef KMESSDEBUG_OFFLINE_IM_GENERAL
  kDebug() << "requesting deletion of messages: '" << messageIds.join("', '") << "'";
#endif
#ifdef KMESSTEST
  KMESS_ASSERT( ! messageIds.isEmpty() );
#endif

  // Initialize request
  QString soapAction = "http://www.hotmail.msn.com/ws/2004/09/oim/rsi/DeleteMessages";
  QString soapBody =
      "    <DeleteMessages xmlns=\"http://www.hotmail.msn.com/ws/2004/09/oim/rsi\">\n"
      "      <messageIds>\n"
      "        <messageId>" + messageIds.join("</messageId>\n        <messageId>") + "</messageId>\n"
      "      </messageIds>\n"
      "    </DeleteMessages>";

  // Make the request
  sendRequest( soapAction, soapBody, passportCookieHeader_ );
}



/**
 * @brief Internal function to extract the e-mail address from an RFC822 formatted string.
 *
 * For a value such as '<code>"Contactname" &lt;contact@hotmail.com&gt;</code>',
 * this method returns '<code>contact@hotmail.com</code>'.
 *
 * @param    address  The string with an e-mail address.
 * @returns  The e-mail address of the contact.
 */
QString OfflineImService::extractRFC822Address( const QString &address )
{
  if( address.endsWith(">") )
  {
    int pos = address.lastIndexOf('<');
    if( pos != -1 )
    {
      return address.mid( pos + 1, address.length() - pos - 2 );
    }
  }

  return address;
}



/**
 * @brief SOAP call to download an offline message.
 *
 * The message ID can be extracted from
 * the following messages types received at the notification connection:
 * - <code>text/x-msmsgsoimnotification</code>
 * - <code>text/x-msmsgsinitialmdatanotification</code>
 *
 * The messageReceived() signal is fired when the webservice returns the message.
 *
 * @param  messageId  The ID of the message.
 * @param  markAsRead  Whether the message should be marked as read.
 */
void OfflineImService::getMessage( const QString &messageId, bool markAsRead )
{
#ifdef KMESSDEBUG_OFFLINE_IM_GENERAL
  kDebug() << "requesting message '" << messageId << "'";
#endif

  // Initialize the call data.
  // Will be used in the slots to know which session this was.
  GetMessageCallData *callData = new GetMessageCallData;
  callData->messageId = messageId;

  // Initialize request
  QString soapAction = "http://www.hotmail.msn.com/ws/2004/09/oim/rsi/GetMessage";
  QString soapBody =
      "    <GetMessage xmlns=\"http://www.hotmail.msn.com/ws/2004/09/oim/rsi\">\n"
      "      <messageId>" + escapeString( messageId ) + "</messageId>\n"
      "      <alsoMarkAsRead>" + ( markAsRead ? "true" : "false" ) + "</alsoMarkAsRead>\n"
      "    </GetMessage>";

  // Send the request.
  int requestId = sendRequest( soapAction, soapBody, passportCookieHeader_ );
  getMessageCallData_.insert( requestId, callData );
}



/**
 * @brief SOAP call to download the value of the Mail-Data field.
 *
 * The value of the <code>Mail-Data</code> field is normally received by the notification server.
 * If there are many offline-IM messages, the value of this field is literally <code>too-large</code>.
 * In that case, the message data can be downloaded using this method.
 *
 * The metaDataReceived() signal is fired when the webservice returns the data.
 */
void OfflineImService::getMetaData()
{
#ifdef KMESSDEBUG_OFFLINE_IM_GENERAL
  kDebug() << "requesting Mail-Data field";
#endif

  QString soapAction = "http://www.hotmail.msn.com/ws/2004/09/oim/rsi/GetMetadata";
  QString soapBody   = "    <GetMetadata xmlns=\"http://www.hotmail.msn.com/ws/2004/09/oim/rsi\" />";
  sendRequest( soapAction, soapBody, passportCookieHeader_ );
}



/**
 * @brief Process the SOAP fault returned when sending an offline message.
 */
void OfflineImService::parseSoapFault( const QDomElement &faultNode, const QDomElement &headerNode )
{
  // Get the type of the error
  int currentRequestId = getCurrentRequestId();
  QString faultCode = XmlFunctions::getNodeValue( faultNode, "faultcode" );

  // Get calling data.
  StoreMessageCallData *callData = storeMessageCallData_[ currentRequestId ];

  // See which fault we received
  if( faultCode == "q0:AuthenticationFailed" )
  {
    if( KMESS_NULL(callData) ) return;

    QString lockKeyChallenge = XmlFunctions::getNodeValue( faultNode, "detail/LockKeyChallenge" );
    QString twnChallenge     = XmlFunctions::getNodeValue( faultNode, "detail/TweenerChallenge" );

    // See if a lock key challenge was requested.
    if( ! lockKeyChallenge.isEmpty() )
    {
      // In the response from server there is a LockKeyChallenge that allows
      // to calculate hash for LockKey (together the Product Key and Product id of MSN).

      // Compute hash with MSN product key/id for the identification
      MSNChallengeHandler handler( PRODUCT_KEY, PRODUCT_ID );

      // Compute the lock key using the MSNP11 Challenge algorithm
      CurrentAccount::instance()->setOfflineImKey( handler.computeHash( lockKeyChallenge ) );

#ifdef KMESSDEBUG_OFFLINE_IM_GENERAL
      kDebug() << "Lock key to compute:" << lockKeyChallenge;
      kDebug() << "Computed lock key:"   << CurrentAccount::instance()->getOfflineImKey();
#endif
      // Check if there is also TweenerChallenge
      if( twnChallenge.isEmpty() )
      {
        // Re-send the message with the new lock key
        storeMessage( callData->to, callData->message, callData->sequenceNum );
        return;
      }
    }


    // See if a tweaner challenge was requested.
    if( ! twnChallenge.isEmpty() )
    {
      // Initialize login service on demand.
      if( passportService_ == 0 )
      {
        passportService_ = createPassportLoginService();
      }

      // Start the login attempt.
      CurrentAccount *currentAccount = CurrentAccount::instance();
      passportService_->login( twnChallenge, currentAccount->getHandle(), currentAccount->getPassword() );
      passportRequestIds_.append( currentRequestId );
      return;
    }

    // Any other situation means something is wrong.
    kWarning() << "Unknown challenge type received";
    emit sendMessageFailed( callData->to, callData->message );
  }
  else if( faultCode == "q0:SystemUnavailable"
       ||  faultCode == "q0:DeliveryFailed"                 // happens with unverified passport accounts
       ||  faultCode == "q0:InvalidParameter"               // seen with SOAP syntax error
       ||  faultCode == "q0:SenderThrottleLimitExceeded" )  // TODO: what to for that one??
  {
    if( KMESS_NULL(callData) ) return;

    // failed, can't send offline IM's
    emit sendMessageFailed( callData->to, callData->message );

    // Also report at the console.
    kWarning() << "Received SOAP fault" << faultCode << "when sending an offline-im message to" << callData->to;
  }
  else
  {
    // Any other situation means something is wrong.
    kWarning() << "Unknown response type received "
                  "(faultCode=" << faultCode << " endpoint=" << getEndpoint() << ")" << endl;

    // Notify the sender if this error was
    // in response of a sent offline im.
    if( callData != 0 )
    {
      emit sendMessageFailed( callData->to, callData->message );
    }

    // Do standard SOAP error handling.
    // (currently will relay to a kWarning() for the msn swichboard connection).
    HttpSoapConnection::parseSoapFault( faultNode, headerNode );
  }
}



/**
 * @brief  Internal function to process the response of the webservice.
 *
 * This handles the following response nodes:
 * - <code>GetMessageResult</code> is handled by processGetMessageResult().
 * - <code>GetMetadataResponse</code> is sent with the metaDataReceived() signal.
 * - <code>DeleteMessagesResponse</code> is ignored because it's always empty.
 *
 * Other unrecognized response types result in a warning message at the console.
 *
 * @param  resultRoot  The result root of the SOAP response.
 * @param  headerNode  The header node of the SOAP response.
 */
void OfflineImService::parseSoapResult( const QDomElement &resultRoot, const QDomElement &/*headerNode*/ )
{
#ifdef KMESSDEBUG_OFFLINE_IM_GENERAL
  kDebug() << "parsing SOAP response.";
#endif

  int currentRequestId = getCurrentRequestId();

  QString resultName = resultRoot.localName();
  if( resultName == "GetMessageResult" )
  {
    processGetMessageResult( resultRoot );
  }
  else if( resultName == "GetMetadataResponse" )
  {
    emit metaDataReceived( resultRoot.namedItem("MD").toElement() );
  }
  else if( resultName == "DeleteMessagesResponse" )
  {
    // do nothing
  }
  else if( resultName == "StoreResponse" )
  {
#ifdef KMESSDEBUG_OFFLINE_IM_GENERAL
    StoreMessageCallData *callData = storeMessageCallData_[ currentRequestId ];
    if( ! KMESS_NULL(callData) )
    {
      kDebug().nospace() << "Stored offline message for " << callData->to << ": " << callData->message;
    }
#endif

    // Reset the session ID
    runID_ = QString();
  }
  else
  {
    kWarning().nospace() << "Unknown response type received "
                            "(rootnode=" << resultName << " endpoint=" << getEndpoint() << ")" << endl;
  }
}



/**
 * @brief  Internal function to process the response of the getMessage() call.
 *
 * This method parses the XML data, extracts the relevant fields,
 * and fires the messageReceived() signal with the results.
 *
 * @param  resultRoot  The <code>GetMessageResult</code> node of the response.
 */
void OfflineImService::processGetMessageResult( const QDomElement &resultRoot )
{
#ifdef KMESSDEBUG_OFFLINE_IM_GENERAL
  kDebug() << "parsing downloaded Offline-IM message.";
#endif

  // Get the call data with original message ID.
  int currentRequestId = getCurrentRequestId();
  GetMessageCallData *callData = getMessageCallData_[ currentRequestId ];
  if( KMESS_NULL(callData) ) return;

  // Parse as MIME message.
  MimeMessage message( resultRoot.text() );

  // Extract the fields
  QString from        = message.decodeRFC2047String( message.getValue("From").toUtf8() );
  QString to          = message.decodeRFC2047String( message.getValue("To").toUtf8() );
  QString dateString  = message.getValue("Date");
  QString runId       = message.getValue("X-OIM-Run-Id");
  int     sequenceNum = message.getValue("X-OIM-Sequence-Num").toInt();

  // Convert the 'From' and 'To' lines, it's something like "Contactname <contact@hotmail.com>"
  from = extractRFC822Address( from );
  to   = extractRFC822Address( to );

  // Convert the date from RFC 2822 format, e.g. "15 Nov 2005 14:24:27 -0800"
  QDateTime date;
  date.setTime_t( KDateTime::fromString( dateString, KDateTime::RFCDate ).toTime_t() );

  // Validate content type
  QString contentType = message.getValue("Content-Type");
  if( contentType.toLower() != "text/plain; charset=utf-8" )
  {
    kWarning() << "received unexpected content type: " << contentType;
  }

  // Validate transfer encoding
  QString transferEncoding = message.getValue("Content-Transfer-Encoding");
  if( transferEncoding != "base64" )
  {
    kWarning() << "received unexpected transfer encoding: " << transferEncoding;
  }

  // Validate message type
  QString messageType  = message.getValue("X-OIM-Message-Type");
  if( messageType != "OfflineMessage" )
  {
    kWarning() << "received unexpected message type: " << messageType;
  }

  // Process the body
  QByteArray decodedBody = QByteArray::fromBase64( message.getBody().toUtf8() );
  QString body = QString::fromUtf8( decodedBody.data(), decodedBody.length() );

  // Return to caller
  // The MessageID was stored in the "RequestId" field of the sendRawRequest() method
  emit messageReceived( callData->messageId, from, to, date, body, runId, sequenceNum );
}



/**
 * @brief  Send an offline message.
 *
 * Sends an Offline Message. With the message the lockKey is empty, so the server will respond with a SOAP fault.
 * At parseSoapFault(), this error will be detected and the message is sent again with the proper lock key.
 *
 * @param  to       Handle of the message sender.
 * @param  message  Message body.
 */
void OfflineImService::sendMessage( const QString &to, const QString &message )
{
  // Increment sequence counter.
  nextSequenceNum_++;

  // Call internal version with sequence number
  storeMessage( to, message, nextSequenceNum_ );
}



/**
 * @brief  Internal method to store a message in the offline-im storage.
 *
 * This method is used internally by sendMessag() and others to send the message.
 * It accepts a third sequenceNum parameter so retries are stored with the same sequence number.
 *
 * @param  to           Handle of the message sender.
 * @param  message      Message body.
 * @param  sequenceNum  The sequence number.
 */
void OfflineImService::storeMessage( const QString &to, const QString &message, int sequenceNum )
{
  CurrentAccount *currentAccount = CurrentAccount::instance();

  QString auth                = Qt::escape( currentAccount->getAuthenticationTicket() );
  const QString &from         = currentAccount->getHandle();
  const QString &friendlyName = currentAccount->getFriendlyName();
  const QString &offlineImKey = currentAccount->getOfflineImKey();

  // Initialize the call data.
  // Will be used in the slots to know which session this was.
  StoreMessageCallData *callData = new StoreMessageCallData;
  callData->to          = to;
  callData->message     = message;
  callData->sequenceNum = sequenceNum;

  // Store the GUID for this sending session
  if( runID_.isEmpty() )
  {
    runID_ = KMessShared::generateGUID();
  }

  // Build the request
  QString soapAction = "http://messenger.msn.com/ws/2004/09/oim/Store";

  QString soapHeader = 
      "<From"
      " xmlns=\"http://messenger.msn.com/ws/2004/09/oim/\""
      " xml:lang=\"en-US\" proxy=\"MSNMSGR\" msnpVer=\"MSNP13\" buildVer=\"8.0.0812\""
      " memberName=\"" + from + "\""
      " friendlyName=\"=?utf-8?B?" + friendlyName.toUtf8().toBase64() + "?=\" />\n"
      "<To xmlns=\"http://messenger.msn.com/ws/2004/09/oim/\" memberName=\"" + to + "\" />\n"
      "<Ticket"
      " xmlns=\"http://messenger.msn.com/ws/2004/09/oim/\"\n"
      " appid=\"" PRODUCT_ID "\""
      " lockkey=\"" + offlineImKey + "\""
      " passport=\"" + auth + "\" />"
      "<Sequence xmlns=\"http://schemas.xmlsoap.org/ws/2003/03/rm\">\n"
      "  <Identifier xmlns=\"http://schemas.xmlsoap.org/ws/2002/07/utility\">http://messenger.msn.com</Identifier>\n"
      "  <MessageNumber>" + QString::number( sequenceNum ) + "</MessageNumber>\n"
      "</Sequence>";

  QString soapBody =
      "<MessageType xmlns=\"http://messenger.msn.com/ws/2004/09/oim/\">text</MessageType>\n"
      "<Content xmlns=\"http://messenger.msn.com/ws/2004/09/oim/\">MIME-Version: 1.0\r\n"
        "Content-Type: text/plain; charset=UTF-8\r\n"
        "Content-Transfer-Encoding: base64\r\n"
        "X-OIM-Message-Type: OfflineMessage\r\n"
        "X-OIM-Run-Id:" + runID_ + "\r\n"
        "X-OIM-Sequence-Num: " + QString::number( sequenceNum ) + "\r\n"
        "\r\n" +
        callData->message.replace( "\n" ,"\r\n" ).toUtf8().toBase64() +
      "</Content>";

  // Send the request.
  int requestId = sendRequest( soapAction, soapBody, soapHeader );
  storeMessageCallData_.insert( requestId, callData );
}



/**
 * @brief Called when the passport login was succesful.
 *
 * Send the offline IM message again.
 */
void OfflineImService::slotLoginSucceeded( QString authenticationKey,  QString passportToken,
                                           QString proofToken )
{
  // See which messages were waiting for passport authentication.
  QList<int> loginRequestIds = passportRequestIds_;  // copy so clearRequestData() can also clear this list.
  foreach( int loginRequestId, loginRequestIds )
  {
    // Don't delete yet, might get another call.
    StoreMessageCallData *callData = storeMessageCallData_[ loginRequestId ];
    if( ! KMESS_NULL(callData) )  // yes, warn for this!
    {
      // Update the authentication ticket
      CurrentAccount *currentAccount = CurrentAccount::instance();
      currentAccount->setAuthenticationTicket( authenticationKey );
      currentAccount->setPassportToken       ( passportToken  );
      currentAccount->setProofToken          ( proofToken     );

      // Send the OIM message again.
      storeMessage( callData->to, callData->message, callData->sequenceNum );
    }
  }
}



/**
 * @brief Called when the login of the passport login request failed.
 *
 * Notify the calling class that sending the offline IM message failed.
 */
void OfflineImService::slotLoginRequestFailed()
{
  // See which messages were waiting for passport authentication.
  QList<int> loginRequestIds = passportRequestIds_;  // copy so clearRequestData() can also clear this list.
  foreach( int loginRequestId, loginRequestIds )
  {
    // Notify the sender this message can't be delivered.
    // Delete call data manually here.
    StoreMessageCallData *callData = storeMessageCallData_.take( loginRequestId );
    if( ! KMESS_NULL(callData) )   // yes, warn for this!
    {
      emit sendMessageFailed( callData->to, callData->message );
      delete callData;
    }
  }

  passportRequestIds_.clear();
}



/**
 * @brief Called when the sub request of the passport login failed.
 */
void OfflineImService::slotLoginRequestFailed( HttpSoapConnection *passportService, const QString &reason )
{
  // Display a standard "message could not be sent" notice.
  slotLoginRequestFailed();

  // Display the exact detailed reason at the console, as always.
  kWarning() << "A offline message could not be sent. The passport login request failed with the following error message: " << reason;

  // TODO: Also relay to caller?
  Q_UNUSED( passportService );
  Q_UNUSED( reason );
//  emit requestFailed( this, reason );
}



/**
 * @brief Called when the sub request of the passport login failed.
 */
void OfflineImService::slotLoginRequestFailed( HttpSoapConnection *passportService, const QString &faultCode,
                                               const QString &faultString, QDomElement faultNode )
{
  Q_UNUSED( faultNode );

  // Reform to standard message.
  // Just like MsnConnection::slotSoapFault() does.

  slotLoginRequestFailed( passportService, faultCode + ": " + faultString );
}



#include "offlineimservice.moc"
