/***************************************************************************
                          chatnotification.cpp -  notifies when a
                            contact communicates with the user
                             -------------------
    begin                : Tuesday April 10 2007
    copyright            : (C) 2007 by Valerio Pilo
    email                : valerio@kmess.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "chatnotification.h"

#include "../chat/chat.h"
#include "../contact/contactbase.h"
#include "../network/chatmessage.h"
#include "../kmessdebug.h"
#include "../currentaccount.h"
#include "notificationmanager.h"
#include "passivepopup.h"

#include <QFile>
#include <QPixmap>
#include <QTextDocument>

#include <KLocale>
#include <KNotification>


#define MAX_BALLOON_MESSAGE_SIZE      128



// Class constructor
ChatNotification::ChatNotification( NotificationManager* manager )
  : enableChatStart_(true)
  , enableChatMessages_(false)
  , manager_(manager)
{
}



// Trigger this event's action (Raise the chat window for the calling contact)
void ChatNotification::activate( QStringList /*stringList*/, QObject *object )
{
  Chat *chatWindow = static_cast<Chat*>( object );
  if ( chatWindow != 0 )
  {
    // Force the window to be raised to front and focused
    emit raiseChat( chatWindow, true );
  }
  else
  {
#ifdef KMESSDEBUG_CHATNOTIFICATION
    kDebug() << "The given chat window was not valid.";
#endif
  }
}



// Notify the user about this event (a contact has contacted the user)
void ChatNotification::notify( const ChatMessage &message, Chat *chatWindow )
{
  // End if the event is coming from us instead of from a contact
  if( ! message.isIncoming() )
  {
#ifdef KMESSDEBUG_CHATNOTIFICATION
    kDebug() << "Not displaying a popup for an outgoing message.";
#endif
    return;
  }

  // Check the chat window pointer, we need it
  if( ! chatWindow )
  {
#ifdef KMESSDEBUG_CHATNOTIFICATION
    kDebug() << "Invalid pointer to chat window.";
#endif
    return;
  }

  // First send the sound notification; even if the balloon is disabled.
  if( chatWindow->isChatFirstMessage() )
  {
    KNotification::event( "chat start", QString(), QPixmap(), chatWindow );
  }
  else
  {
    KNotification::event( "new message", QString(), QPixmap(), chatWindow );
  }

  // If all possible notifications for this event are disabled, do nothing.
  if( ! enableChatStart_ && ! enableChatMessages_ )
  {
#ifdef KMESSDEBUG_CHATNOTIFICATION
    kDebug() << "Start and messages notifications are off, skipping.";
#endif
    return;
  }

  int messageType = message.getType();

  // If we should only notify chat starts, skip notificating further messages
  if( ! enableChatMessages_ && enableChatStart_ &&
      messageType == ChatMessage::TYPE_INCOMING && ! chatWindow->isChatFirstMessage() )
  {
#ifdef KMESSDEBUG_CHATNOTIFICATION
  kDebug() << "Messages notification is off and it's not the first message, skipping.";
#endif
    return;
  }


  QString text;
  QString friendlyName;
  Status status;
  QString body = Qt::escape( message.getBody() );
  const QString &handle = message.getContactHandle();
  CurrentAccount *currentAccount = CurrentAccount::instance();
  ContactBase *contact = currentAccount->getContactByHandle( handle );

#ifdef KMESSDEBUG_CHATNOTIFICATION
  kDebug() << "Contact " << message.getContactHandle();
#endif

  // Check for a contact handle, we need it for the popup
  if( handle.isEmpty() )
  {
    kWarning() << "Cannot notify chat event, no contact handle given!";
    return;
  }

  if( contact )
  {
    // Use rich text for friendlyname
    friendlyName = contact->getFriendlyName( USE_LIST_SETTING );
    status = contact->getStatus();
  }
  else
  {
    friendlyName = handle;
    status = STATUS_ONLINE; // Invalid status code, will be shown as Online
  }

  // Truncate the message so it doesn't make the balloon too big
  if( body.length() > MAX_BALLOON_MESSAGE_SIZE )
  {
    body.truncate( MAX_BALLOON_MESSAGE_SIZE );
    body += " ...";
  }


  // Set the balloon text to a generic message, which will be overridden with more useful messages, if available
  switch( messageType )
  {
    case ChatMessage::TYPE_NOTIFICATION:
      text = i18n( "<html>In <b>%1</b>'s chat: %2</html>", friendlyName, body );
      break;

    case ChatMessage::TYPE_INCOMING:
      text = i18n( "<html><b>%1</b> says:<br>'%2'</html>", friendlyName, body );
      break;

    case ChatMessage::TYPE_OFFLINE_INCOMING:
      text = i18n( "<html><b>%1</b> has sent you an offline message:<br>'%2'</html>", friendlyName, body );
      break;

    case ChatMessage::TYPE_APPLICATION:
    case ChatMessage::TYPE_APPLICATION_WEBCAM:
    case ChatMessage::TYPE_APPLICATION_FILE:
    case ChatMessage::TYPE_APPLICATION_AUDIO:
      text = i18n( "<html>%1's chat requests attention!</html>", friendlyName );
      break;

    case ChatMessage::TYPE_SYSTEM:
    default:
      text = i18n( "<html><b>%1</b>:<br>%2</html>", friendlyName, body );
      break;
  }

  switch( message.getContentsClass() )
  {
    case ChatMessage::CONTENT_MESSAGE_INK:
      text = i18n( "<html><b>%1</b> has sent you a handwritten message!</html>", friendlyName );
      break;

    case ChatMessage::CONTENT_NOTIFICATION_NUDGE:
      text = i18n( "<html><b>%1</b> has sent you a nudge!</html>", friendlyName );
      break;

    case ChatMessage::CONTENT_NOTIFICATION_WINK:
      text = i18n( "<html><b>%1</b> has sent you a wink!</html>", friendlyName );
      break;

    case ChatMessage::CONTENT_APP_INVITE:
      switch( messageType )
      {
        case ChatMessage::TYPE_APPLICATION_WEBCAM: text = i18n( "<html><b>%1</b> wants to use the webcam!</html>", friendlyName ); break;
        case ChatMessage::TYPE_APPLICATION_FILE: text = i18n( "<html><b>%1</b> is sending you a file!</html>", friendlyName ); break;
        default: text = i18n( "<html><b>%1</b> has sent you an invitation!</html>", friendlyName ); break;
      }
      break;

    case ChatMessage::CONTENT_APP_CANCELED:
      switch( messageType )
      {
        case ChatMessage::TYPE_APPLICATION_WEBCAM: text = i18n( "<html><b>%1</b> has canceled the webcam session!</html>", friendlyName ); break;
        case ChatMessage::TYPE_APPLICATION_FILE: text = i18n( "<html><b>%1</b> has canceled the file transfer!</html>", friendlyName ); break;
        default: text = i18n( "<html><b>%1</b>'s activity has been canceled!</html>", friendlyName ); break;
      }
      break;

    case ChatMessage::CONTENT_APP_STARTED:
      switch( messageType )
      {
        case ChatMessage::TYPE_APPLICATION_WEBCAM: text = i18n( "<html><b>%1</b> has accepted to use the webcam!</html>", friendlyName ); break;
        case ChatMessage::TYPE_APPLICATION_FILE: text = i18n( "<html><b>%1</b> has accepted the file transfer!</html>", friendlyName ); break;
        default: text = i18n( "<html><b>%1</b> has accepted your invitation!</html>", friendlyName ); break;
      }
      break;

    case ChatMessage::CONTENT_APP_ENDED:
      switch( messageType )
      {
        case ChatMessage::TYPE_APPLICATION_WEBCAM: text = i18n( "<html><b>%1</b> has ended the webcam session!</html>", friendlyName ); break;
        case ChatMessage::TYPE_APPLICATION_FILE: text = i18n( "<html>The file transfer with <b>%1</b> is done!</html>", friendlyName ); break;
        default: text = i18n( "<html><b>%1</b>'s activity has ended!</html>", friendlyName ); break;
      }
      break;

    case ChatMessage::CONTENT_APP_FAILED:
      switch( messageType )
      {
        case ChatMessage::TYPE_APPLICATION_WEBCAM: text = i18n( "<html><b>%1</b>'s webcam session has failed!</html>", friendlyName ); break;
        case ChatMessage::TYPE_APPLICATION_FILE: text = i18n( "<html>The file transfer with <b>%1</b> has failed!</html>", friendlyName ); break;
        default: text = i18n( "<html><b>%1</b>'s activity has ended with an error!</html>", friendlyName ); break;
      }
      break;

    default:
      break;
  }


  // Display a popup for this event
  PassivePopup::PopupProperties props;
  props.source  = message.getContactHandle();
  props.message = text;
  props.icon    = message.getContactPicturePath();
  props.status  = status;
  props.destMessages = QStringList();
  props.destObject = chatWindow;

  manager_->notify( this, props );
}



// Enable or disable the chat notifications
void ChatNotification::setEnabled( bool enableChatStart, bool enableChatMessages )
{
  enableChatStart_ = enableChatStart;
  enableChatMessages_ = enableChatMessages;
}



#include "chatnotification.moc"
