/***************************************************************************
                          contactstatusnotification.cpp - notifies when a
                            contact changes their msn status
                             -------------------
    begin                : Saturday July 28 2007
    copyright            : (C) 2007 by Valerio Pilo
    email                : valerio@kmess.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "contactstatusnotification.h"

#include "../contact/contact.h"
#include "../contact/contactextension.h"
#include "../currentaccount.h"
#include "../kmessdebug.h"
#include "notificationmanager.h"
#include "passivepopup.h"

#include <QFile>
#include <QTextDocument>
#include <QPixmap>

#include <KLocale>
#include <KNotification>
#include <Phonon/MediaObject>



// Class constructor
ContactStatusNotification::ContactStatusNotification( NotificationManager* manager )
  : enableNotifyOnline_(true)
  , enableNotifyOffline_(false)
  , enableNotifyStatus_(false)
  , manager_(manager)
{
}



// Trigger this event's action
void ContactStatusNotification::activate( QStringList stringList, QObject */*object*/ )
{
#ifdef KMESSDEBUG_CONTACTSTATUSNOTIFICATION
  kDebug() << "Starting chat.";
#endif

  // Do not try to open a chat with a contact which has gone offline
  if( stringList[1] == "FLN" )
  {
    return;
  }

  emit startChat( stringList[0], ChatInformation::CONNECTION_CHAT );
}



// Obtains a status string like "contact has come online" for a MSN status code
QString ContactStatusNotification::getStatusString( const Status status, const QString& friendlyName ) const
{
  switch( status )
  {
    case STATUS_ONLINE:        return i18n( "<html><b>%1</b> is now online</html>",        friendlyName );
    case STATUS_AWAY:          return i18n( "<html><b>%1</b> has gone away</html>",        friendlyName );
    case STATUS_BE_RIGHT_BACK: return i18n( "<html><b>%1</b> will be right back</html>",   friendlyName );
    case STATUS_BUSY:          return i18n( "<html><b>%1</b> is now busy</html>",          friendlyName );
    case STATUS_INVISIBLE:     return i18n( "<html><b>%1</b> has become invisible</html>", friendlyName ); // Is this one even possible?!
    case STATUS_IDLE:          return i18n( "<html><b>%1</b> has gone idle</html>",        friendlyName );
    case STATUS_OFFLINE:       return i18n( "<html><b>%1</b> has logged out</html>",       friendlyName );
    case STATUS_ON_THE_PHONE:  return i18n( "<html><b>%1</b> is on the phone</html>",      friendlyName );
    case STATUS_OUT_TO_LUNCH:  return i18n( "<html><b>%1</b> is out for lunch</html>",     friendlyName );
    default:
      kWarning() << "Invalid status" << status << "!";
      return getStatusString( STATUS_ONLINE, friendlyName );
  }
}



// Notify the user about this event (a contact has changed its status)
void ContactStatusNotification::notify( Contact *contact, bool isNeeded )
{
  if( contact == 0 )
  {
    kWarning() << "Contact is null!";
    return;
  }

  if( ! KMESS_NULL(contact->getExtension()) && ! contact->getExtension()->getNotifyOnline() )
  {
#ifdef KMESSDEBUG_CONTACTSTATUSNOTIFICATION
    kDebug() << "Skipping, notification disabled for " << contact->getHandle() << ".";
#endif
    return;
  }

  // If notification for this event is disabled, don't show a balloon.
  if( ! isNeeded )
  {
#ifdef KMESSDEBUG_CONTACTSTATUSNOTIFICATION
    kDebug() << "Not needed, skipping.";
#endif
    return;
  }


  QString text;
  Status  status = contact->getStatus();
  Status  lastStatus = contact->getLastStatus();

#ifdef KMESSDEBUG_CONTACTSTATUSNOTIFICATION
  kDebug() << "Contact: " << contact->getHandle() << ", status: " << MsnStatus::getCode( lastStatus ) << "->" << status << " - "
            << "Online: "  << ( enableNotifyOnline_  ? "true" : "false" ) << ", "
            << "Offline: " << ( enableNotifyOffline_ ? "true" : "false" ) << ", "
            << "Status: "  << ( enableNotifyStatus_  ? "true" : "false" ) << endl;
#endif

  // When contacts go online, whichever may be their status, just show an 'is now online'.
  if( lastStatus == STATUS_OFFLINE && status != STATUS_OFFLINE && enableNotifyOnline_ )
  {
    text = getStatusString( STATUS_ONLINE, contact->getFriendlyName( USE_LIST_SETTING ) );
  }
  else
  {
    text = getStatusString( status, contact->getFriendlyName( USE_LIST_SETTING ) );
  }


  // Play a sound
  QString soundPath = contact->getExtension()->getContactSoundPath();
  QFile   soundFile( soundPath );

  // Find what kind of event is this, and skip the notification if it's a disabled event
  bool notifyUser = true;
       if( lastStatus == STATUS_OFFLINE && status != STATUS_OFFLINE ) notifyUser = enableNotifyOnline_;
  else if( lastStatus != STATUS_OFFLINE && status == STATUS_OFFLINE ) notifyUser = enableNotifyOffline_;
  else                                                                notifyUser = enableNotifyStatus_;

  if( ! notifyUser )
  {
#ifdef KMESSDEBUG_CONTACTSTATUSNOTIFICATION
    kDebug() << "Event popup disabled, skipping.";
#endif

    return;
  }

  // If a custom audio file exists, play it
  if ( soundFile.exists() )
  {
#ifdef KMESSDEBUG_CONTACTSTATUSNOTIFICATION
    kDebug() << "Play sound " << soundPath;
#endif

    // Play the custom sound instead of the generic sound
    Phonon::MediaObject* media = Phonon::createPlayer( Phonon::NotificationCategory );
    media->setCurrentSource( soundPath );
    media->play();
  }
  else
  {
#ifdef KMESSDEBUG_CONTACTSTATUSNOTIFICATION
    kDebug() << "Play the default sound.";
#endif

    QString statusText;

    if( lastStatus == STATUS_OFFLINE )
    {
      statusText = "contact online";
    }
    else if( status == STATUS_OFFLINE )
    {
      statusText = "contact offline";
    }
    else
    {
      statusText = "contact status";
    }

    KNotification::event( statusText, text );
  }

  // Display a popup for this event
  PassivePopup::PopupProperties props;
  props.source = contact->getHandle();
  props.message = text;
  props.icon = contact->getContactPicturePath();
  props.status = status;
  props.destMessages << contact->getHandle() << MsnStatus::getCode( status );
  props.destObject = 0;

  manager_->notify( this, props );
}



// Enable or disable the contact offline notifications
void ContactStatusNotification::setEnabled( bool enableNotifyOnline, bool enableNotifyOffline, bool enableNotifyStatus )
{
  enableNotifyOnline_  = enableNotifyOnline;
  enableNotifyOffline_ = enableNotifyOffline;
  enableNotifyStatus_  = enableNotifyStatus;
}



#include "contactstatusnotification.moc"
