/***************************************************************************
                          notificationmanager.cpp - manage the notifications
                            stack and updates them as needed
                             -------------------
    begin                : Thursday April 19 2007
    copyright            : (C) 2007 by Valerio Pilo
    email                : valerio@kmess.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "notificationmanager.h"

#include "../utils/xautolock.h"
#include "../currentaccount.h"
#include "../kmessdebug.h"

#include <QSystemTrayIcon>

#include <KWindowSystem>

#ifdef Q_WS_WIN
#include <windows.h>
#endif


// Initialize the instance to zero
NotificationManager* NotificationManager::instance_(0);



// Class constructor
NotificationManager::NotificationManager()
{
  // Initialize the interfaces to screen saver and fullscreen apps checks
  autoLock_ = new XAutoLock();
}



// Class destructor
NotificationManager::~NotificationManager()
{
  qDeleteAll( stack_ );
  stack_.clear();

  delete autoLock_;
}



// Return the current tray object
QSystemTrayIcon *NotificationManager::getTrayObject()
{
  return trayObject_;
}


// Insert a new popup in the stack if needed, or change an existing one
void NotificationManager::notify( QObject *caller, PassivePopup::PopupProperties props )
{
  // Check if the screen saver is running
  if( autoLock_->isScreenSaverActive() )
  {
#ifdef KMESSDEBUG_NOTIFICATIONMANAGER
    kDebug() << "Screen saver/locker is active, not showing balloons.";
#endif
    return;
  }

  bool isFullscreen = false;

#ifdef Q_WS_X11
  // Check if a full screen application is running by querying the window manager and asking
  // the state of the currently active window
  KWindowInfo info = KWindowSystem::windowInfo( KWindowSystem::activeWindow(), NET::WMState | NET::FullScreen );
  isFullscreen = ( info.valid() && info.hasState( NET::FullScreen ) );
#else
  #ifdef Q_WS_WIN
  // Windows version
  HWND hwnd = GetForegroundWindow();
  RECT rcWindow;
  GetWindowRect( hwnd, &rcWindow );
  int screenWidth  = GetSystemMetrics( SM_CXSCREEN );
  int screenHeight = GetSystemMetrics( SM_CYSCREEN );

  int windowWidth  = ( rcWindow.right  - rcWindow.left );
  int windowHeight = ( rcWindow.bottom - rcWindow.top );

  isFullscreen = ( screenWidth == windowWidth && screenHeight == windowHeight );
  #else
    #warning Fullscreen detection is not implemented for this platform.
  #endif
#endif

  if( isFullscreen )
  {
    kDebug() << "Active window is full screen, do not show the balloon.";
    return;
  }

#ifdef KMESSDEBUG_NOTIFICATIONMANAGER
  #ifdef Q_WS_X11
  kDebug() << "Active window was valid?" << info.valid() <<". If valid, it is not fullscreen: showing balloons.";
  #endif
#endif

  // We have to show a popup

  PassivePopup *popup;

  if( stack_[ props.source ] )
  {
#ifdef KMESSDEBUG_NOTIFICATIONMANAGER
    kDebug() << "a popup for " << props.source << " is already in the stack, updating it";
#endif

    popup = stack_[ props.source ];
    popup->change( props );

    popup->show();

    return;
  }

#ifdef KMESSDEBUG_NOTIFICATIONMANAGER
  kDebug() << "creating new popup for " << props.source << " in the stack";
#endif

  int minY = 99999999;
  int current;
  QHashIterator<QString,PassivePopup*> it( stack_ );

  // Set the new popup's position
  while( it.hasNext() )
  {
    PassivePopup *currentPopup = it.next().value();

    if( ! currentPopup )
    {
      continue;
    }

    current = currentPopup->getPosition().y();
    if( current < minY )
    {
      minY = current;
    }
  }

  props.position = trayObject_->geometry().topLeft();

  if( minY != 99999999 )
  {
    // TODO Adding like, 10px to this would make the balloons touch. But I haven't found a method to obtain the distance
    // value that is actually used by KDE to space them.
    props.position.ry() = minY;
  }

  props.timeout = CurrentAccount::instance()->getNotificationDuration() * 1000;

  popup = new PassivePopup( 0, props );
  stack_.insert( props.source, popup );

  connect( popup,  SIGNAL( activate( QStringList, QObject * ) ),
           caller,   SLOT( activate( QStringList, QObject * ) ) );
  connect( popup,  SIGNAL( popupDestroyed( QString ) ),
           this,     SLOT( remove( QString ) ) );

  popup->show();
}



// Return a singleton instance of the current account
NotificationManager* NotificationManager::instance()
{
  // If the instance is null, create a new manager and return that.
  if ( instance_ == 0 )
  {
    instance_ = new NotificationManager();
  }
  return instance_;
}



// Remove a timed out popup from the stack
void NotificationManager::remove( QString source )
{
  delete stack_.take( source );

#ifdef KMESSDEBUG_NOTIFICATIONMANAGER
  kDebug() << "deleted popup for " << source << " from the stack";
#endif
}



// Set the tray object that will be used
void NotificationManager::setTrayObject( QSystemTrayIcon *trayObject )
{
  trayObject_ = trayObject;
}



#include "notificationmanager.moc"
