/***************************************************************************
                          chatstylepage.cpp -  description
                             -------------------
    begin                : Thu Jan 30 2008
    copyright            : (C) 2008 by Diederik van der Boor
    email                : "vdboor" --at-- "codingdomain.com"
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "chatstylepage.h"

#include "../chat/chatmessagestyle.h"
#include "../chat/chatmessageview.h"
#include "../currentaccount.h"
#include "../kmessdebug.h"

#include <QDir>

#include <KFontDialog>
#include <KHTMLView>
#include <KStandardDirs>



/**
 * The constructor
 */
ChatStylePage::ChatStylePage( QWidget *parent )
: QWidget( parent )
, Ui::ChatStylePage()
{
  setupUi( this );

  QPixmap        pixmap;
  QString        emoticonDir = CurrentAccount::instance()->getEmoticonStyle();
  KStandardDirs *dirs        = KGlobal::dirs();

  // Set up the emoticons which explain what the graphics elements in chat are
  pixmap.load( dirs->findResource( "emoticons", emoticonDir + "/teeth.png" ) );
  pixmapLabel1_->setPixmap( pixmap );
  pixmap.load( dirs->findResource( "emoticons", emoticonDir + "/rainbow.png" ) );
  pixmapLabel2_->setPixmap( pixmap );
  pixmap.load( dirs->findResource( "emoticons", emoticonDir + "/dog.png" ) );
  pixmapLabel3_->setPixmap( pixmap );

  // Connect our signals
  connect( contactFontColorButton_,SIGNAL(               changed(const QColor&)  ),
           this,                   SLOT  (         updatePreview()               ) );
  connect( fontColorButton_,       SIGNAL(               changed(const QColor&)  ),
           this,                   SLOT  (         updatePreview()               ) );
  connect( useContactFontCheckBox_,SIGNAL(               toggled(bool)           ),
           this,                   SLOT  ( useContactFontToggled(bool)           ) );
  connect( fontButton_,            SIGNAL(               clicked()               ),
           this,                   SLOT  (           fontClicked()               ) );
  connect( contactFontButton_,     SIGNAL(               clicked()               ),
           this,                   SLOT  (    contactFontClicked()               ) );
  connect( chatStyle_,             SIGNAL(             activated(const QString&) ) ,
           this,                   SLOT  (  slotChatStyleChanged(const QString&) ) );
  connect( useEmoticonsCheckBox_,  SIGNAL(               toggled(bool)           ),
           this,                   SLOT  (         updatePreview()               ) );
  connect( useFontEffectsCheckBox_,SIGNAL(               toggled(bool)           ),
           this,                   SLOT  (         updatePreview()               ) );
  connect( showTimeCheckbox_,      SIGNAL(               toggled(bool)           ),
           this,                   SLOT  (         updatePreview()               ) );
  connect( groupFollowupCheckbox_, SIGNAL(               toggled(bool)           ),
           this,                   SLOT  (         updatePreview()               ) );
  connect( enableChatFormattingCheckBox_, SIGNAL(        toggled(bool)           ),
           this,                   SLOT  (         updatePreview()               ) );


  // Configure the preview form of the "chatting" widget

  // Insert a KHTMLPart in the placeholder
  chatMessageView_ = new ChatMessageView( khtmlPlaceholder_ );

  // Create a layout to maximize the KHTMLPart
  QBoxLayout *layout = new QHBoxLayout( khtmlPlaceholder_ );
  layout->addWidget( chatMessageView_->view() );

  QStringList styleNames;

  // Get all available chat styles for the "chatting" widget, avoiding duplicate entries
  QStringList stylesDirectories( dirs->findDirs( "data", "kmess/styles" ) );
  foreach( QString dir, stylesDirectories )
  {
    QDir styleDir( dir );
    QStringList styles = styleDir.entryList( QDir::Dirs, QDir::Name | QDir::IgnoreCase );

    foreach( QString styleName, styles )
    {
      QString xslFile = styleDir.path() + "/" + styleName + "/" + styleName + ".xsl";  // e.g. share/apps/kmess/styles/Demo/Demo.xsl
      if( QFile::exists( xslFile ) && ! styleNames.contains( styleName ) )
      {
        styleNames.append( styleName );
      }
    }
  }

  // Add all items to the chat styles list
  chatStyle_->addItems( styleNames );

  // Prepare the XSL transformation
  chatStyleTransform_ = new ChatMessageStyle();
}



// Destructor
ChatStylePage::~ChatStylePage()
{
  delete chatStyleTransform_;
}



// The contact font button was pressed.  Show a font dialog to get a new font.
void ChatStylePage::contactFontClicked()
{
  // Get a new font and font family from the user
  QFont font     = contactFontButton_->font();
  bool  accepted = getFont( font );

  // If it was accepted, set the font to the button
  if ( accepted )
  {
    // Reset size, as it can't be used in the chat message view.
    font.setPointSize(10);

    contactFontButton_->setText( font.family().replace("&", "&&") );
    contactFontButton_->setFont( font );
    contactFontButton_->setShortcut( QKeySequence() );

    // Update the preview
    chatStyleTransform_->setContactFont( font );
    updatePreview();
  }
}



// The font button was pressed.  Show a font dialog to get a new font.
void ChatStylePage::fontClicked()
{
  // Get a new font and font family from the user
  QFont font     = fontButton_->font();
  bool  accepted = getFont( font );

  // If it was accepted, set the font to the button
  if ( accepted )
  {
    // Reset size, as it can't be used in the chat message view.
    font.setPointSize(10);

    fontButton_->setText( font.family().replace("&", "&&") );
    fontButton_->setFont( font );
    fontButton_->setShortcut( QKeySequence() );

    // Update the preview
    updatePreview();
  }
}



// Get a font and cleaned-up font family from a dialog
bool ChatStylePage::getFont( QFont &font ) const
{
  bool  accepted;
  int   result;
  QFont selection;

  selection = font;

  result = KFontDialog::getFont( selection );
  accepted = ( result == KFontDialog::Accepted );

  if ( accepted )
  {
    font = selection;
  }
  return accepted;
}



/**
 * @brief Load the widget state from an account
 *
 * Reads an Account's info and sets our internal widgets accordingly
 *
 * @param account The account where to pick settings from
 */
void ChatStylePage::loadSettings( const Account *account )
{
  QColor color;
  QFont  font;

  // Set up the values
  color.setNamedColor( account->getFontColor() );
  font = account->getFont();
  font.setPointSize(10);  // Reset size, as it can't be used in the chat message view.

  fontButton_->setText( font.family().replace("&", "&&") );
  fontButton_->setFont( font );
  fontButton_->setShortcut( QKeySequence() );
  fontColorButton_->setColor( color );

  color.setNamedColor( account->getContactFontColor() );
  font = account->getContactFont();
  font.setPointSize(10);  // Reset size, as it can't be used in the chat message view.

  contactFontButton_->setText( font.family().replace("&", "&&") );
  contactFontButton_->setFont( font );
  contactFontButton_->setShortcut( QKeySequence() );
  contactFontColorButton_->setColor( color );

  useContactFontCheckBox_->setChecked( account->getUseContactFont() );
  useContactFontToggled( account->getUseContactFont() );
  useEmoticonsCheckBox_->setChecked( account->getUseEmoticons() );
  showTimeCheckbox_->setChecked( account->getShowMessageTime() );
  useFontEffectsCheckBox_->setChecked( account->getUseFontEffects() );
  groupFollowupCheckbox_->setChecked( account->getGroupFollowupMessages() );
  enableListFormattingCheckBox_->setChecked( account->getUseListFormatting() );
  enableChatFormattingCheckBox_->setChecked( account->getUseChatFormatting() );
  tabbedChatMode_->setCurrentIndex( account->getTabbedChatMode() );

  // Select the correct chat style,
  int item = chatStyle_->findText( account->getChatStyle(), Qt::MatchExactly | Qt::MatchCaseSensitive );
  if( item == -1 )
  {
    kWarning() << "Current chat style was not found, attempting to revert to the default setting.";
    item = chatStyle_->findText( "Default", Qt::MatchExactly | Qt::MatchCaseSensitive );
  }

  if( item != -1 )
  {
    chatStyle_->setCurrentIndex( item );
    slotChatStyleChanged( chatStyle_->currentText() );
  }

  // Also load contact font settings into the chatStyle object,
  // other settings are set in updatePreview()
  chatStyleTransform_->setContactFont( font );
  chatStyleTransform_->setContactFontColor( color.name() );
}



/**
 * @brief Save account information from the chat style widget
 *
 * Calls the set-up methods from Account to change its properties, based on how
 * the user has set the widgets of this page.
 */
void ChatStylePage::saveSettings( Account *account )
{
   QFont font, contactFont;

  font = fontButton_->font();
  font.setFamily( fontButton_->font().family() );

  contactFont = contactFontButton_->font();
  contactFont.setFamily( contactFontButton_->font().family() );

  account->setChatInformation( useContactFontCheckBox_->isChecked(),
                               useEmoticonsCheckBox_->isChecked(),
                               useFontEffectsCheckBox_->isChecked(),
                               enableListFormattingCheckBox_->isChecked(),
                               enableChatFormattingCheckBox_->isChecked(),
                               showTimeCheckbox_->isChecked(),
                               groupFollowupCheckbox_->isChecked(),
                               tabbedChatMode_->currentIndex(),
                               chatStyle_->currentText() );


  // After setChatInformation, due the way ChatView changes updates in font and chat.
  // If the ordering is changed, it will re-generate all messages twice;
  // once for the changedFontSettings(), and once for the changedChatStyleSettings() signal.
  account->setFontInformation( font, fontColorButton_->color().name(),
                               contactFont, contactFontColorButton_->color().name() );
}



// The chat style was changed.
void ChatStylePage::slotChatStyleChanged(const QString &style)
{
  // Update the chat style object
  chatStyleTransform_->setStyle( style );

  // Update the preview
  updatePreview();
}



// Force the page onto a specific tab
void ChatStylePage::switchToTab( int tabIndex )
{
  if( tabIndex < 0 || tabIndex > tabWidget_->count() )
  {
    kWarning() << "Invalid tab index" << tabIndex << "selected!";
    return;
  }

  tabWidget_->setCurrentIndex( tabIndex );
}



// Update the chat style preview
void ChatStylePage::updatePreview()
{
  QString htmlBody;

  if( chatStyleTransform_ == 0 || ! chatStyleTransform_->hasStyle() )
  {
    // Don't bother updating the preview for every event fired by the load...Settings() methods.
    return;
  }

  // Update the boolean settings (other settings are updated in their slots)
  chatStyleTransform_->setContactFontColor( contactFontColorButton_->color().name()    );
  chatStyleTransform_->setShowTime        ( showTimeCheckbox_->isChecked()             );
  chatStyleTransform_->setUseContactFont  ( useContactFontCheckBox_->isChecked()       );
  chatStyleTransform_->setUseEmoticons    ( useEmoticonsCheckBox_->isChecked()         );
  chatStyleTransform_->setUseFontEffects  ( useFontEffectsCheckBox_->isChecked()       );
  chatStyleTransform_->setUseFormatting   ( enableChatFormattingCheckBox_->isChecked() );

  // Create three fake messages
  ChatMessage fake1( ChatMessage::TYPE_INCOMING, ChatMessage::CONTENT_MESSAGE, true, i18n("Hi, how are you doing? :)"), "test@kmess.org",
                     i18n("Stacy") );
  ChatMessage fake2( ChatMessage::TYPE_OUTGOING, ChatMessage::CONTENT_MESSAGE, false, i18n("Great!"), previewHandle_,
                     previewUsername_, previewPicture_, fontButton_->font(), fontColorButton_->color().name() );
  ChatMessage fake3( ChatMessage::TYPE_OUTGOING, ChatMessage::CONTENT_MESSAGE, false, i18n("I /just/ got back from my vacation in Italy!"), previewHandle_,
                     previewUsername_, previewPicture_, fontButton_->font(), fontColorButton_->color().name() );

  // Display the sample chat
  htmlBody += "<div class='messageContainer'>" + chatStyleTransform_->convertMessage(fake1) + "</div>\n";

  if( groupFollowupCheckbox_->isChecked() )
  {
    QList<ChatMessage*> messageGroup;
    messageGroup.append(&fake2);
    messageGroup.append(&fake3);
    htmlBody += "<div class='messageContainer'>" + chatStyleTransform_->convertMessageList(messageGroup) + "</div>\n";
  }
  else
  {
    htmlBody += "<div class='messageContainer'>" + chatStyleTransform_->convertMessage(fake2) + "</div>\n";
    htmlBody += "<div class='messageContainer'>" + chatStyleTransform_->convertMessage(fake3) + "</div>\n";
  }

  QString messageRoot = chatStyleTransform_->convertMessageRoot();
  if( messageRoot.isNull() )
  {
    // Assign standard HTML
    chatMessageView_->setStandardHtml( htmlBody, chatStyleTransform_->getCssFile(), chatStyleTransform_->getBaseFolder() );
  }
  else
  {
    // Assign the custom HTML root
    chatMessageView_->setHtml( messageRoot, htmlBody );
  }
}



// Update the data shown in the message preview
void ChatStylePage::updatePreviewDetails( const QString &handle, const QString &name, const QString picture )
{
  previewHandle_   = handle;
  previewPicture_  = picture;

  if( name.isEmpty() )
  {
    previewUsername_ = handle;
  }
  else
  {
    previewUsername_ = name;
  }
}



// Enable or disable the contact font and color selectors
void ChatStylePage::useContactFontToggled( bool checked )
{
  // Enable/disable buttons
  contactFontButton_->setEnabled( checked );
  contactFontColorButton_->setEnabled( checked );

  // Update preview
  chatStyleTransform_->setUseContactFont( checked );
  updatePreview();
}


#include "chatstylepage.moc"
