/***************************************************************************
                          nowlisteningclient.cpp -  description
                             -------------------
    begin                : Sat Nov 4 2006
    copyright            : (C) 2006 by Diederik van der Boor
    email                : "vdboor" --at-- "codingdomain.com"
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "nowlisteningclient.h"

#include "kmessdebug.h"

#include <QDBusInterface>
#include <QDBusReply>

#include <KApplication>
#include <KProcess>
#include <KStandardDirs>


#ifdef KMESSDEBUG_NOWLISTENINGCLIENT
  #define KMESSDEBUG_NOWLISTENINGCLIENT_GENERAL
#endif


// This code is loosely inspired by the "now listening" plugin of Kopete.
// Therefore some parts are also
// (c) 2002-2006 the by Kopete developers <kopete-devel@kde.org>

bool NowListeningClient::alreadyQuery = false;

/**
 * Constructor
 */
NowListeningClient::NowListeningClient()
 : playing_(false)
{
  // Connect the update timer event
  connect( &timer_, SIGNAL(timeout()), this, SLOT(slotUpdate()) );
}



/**
 * Enable or disable the update interval timer.
 */
void NowListeningClient::setEnabled( bool enable )
{
  if( enable )
  {
    if( ! timer_.isActive() )
    {
      // Reset the fields to avoid problems when the "now listening information" is
      // re-enabled.
      artist_ = "";
      album_ = "";
      track_ = "";
      playing_ = false;

      // Query directly so the GUI is up to date
      slotUpdate();

      // check every 8 seconds
      timer_.start( 8000 );
    }
  }
  else
  {
    // Emit a null changedSong signal to clean the GUI
    emit changedSong( QString(), QString(), QString(), false );
    timer_.stop();
  }
}



/**
 * Update the current song
 */
void NowListeningClient::slotUpdate()
{
#ifdef KMESSDEBUG_NOWLISTENINGCLIENT_GENERAL
    kDebug() << "Updating Now Playing status";
#endif

  // Detect changes to reduce signal calls.
  QString prevArtist = artist_;
  QString prevAlbum  = album_;
  QString prevTrack  = track_;
  bool prevPlaying   = playing_;

  // Reset until proven otherwise.
  playing_ = false;

  // Query all applications: the first found stops the search
  if(    queryAmarok1()
      || queryAmarok2()
      || queryJuk()
      || queryKsCD()   )
  {
    // Found active media player!

#ifdef KMESSDEBUG_NOWLISTENINGCLIENT_GENERAL
    kDebug() << "found song" << artist_ << "-" << track_ << "(extra info: album=" << album_ << ", playing=" << playing_ << ")";
#endif

    // App found and playing, detect change.
    // Check if player is playing ( and if the song was changed ) or if the player was stopped/paused
    // and now play again.
    if( ( playing_ &&
        (   prevArtist != artist_
        ||  prevAlbum  != album_
        ||  prevTrack  != track_ ) ) || ( ! prevPlaying && playing_ )
      )
    {
#ifdef KMESSDEBUG_NOWLISTENINGCLIENT_GENERAL
      kDebug() << "playing information changed, emitting changedSong().";
#endif
      emit changedSong( artist_, album_, track_, playing_ );
    }
  }

  // Emit a signal when the player was stopped.
  if( prevPlaying && ! playing_ )
  {
#ifdef KMESSDEBUG_NOWLISTENINGCLIENT_GENERAL
    kDebug() << "player was stopped, emitting changedSong().";
#endif
    emit changedSong( QString(), QString(), QString(), false );
  }
}



/**
 * Query Amarok 1.x for track information.
 */
bool NowListeningClient::queryAmarok1()
{
  // Check for executable to avoid problem of defunct fork
  if( KStandardDirs::findExe( "dcop" ).isEmpty() )
  {
    return false;
  }

  QStringList defaultParams( QStringList() << "dcop" << "amarok" << "default" );

  KProcess proc;
  proc.setOutputChannelMode( KProcess::OnlyStdoutChannel );

  // Read the 'is playing' answer. This will filter out the case of a missing "dcop" command, and of
  // a "amarok is not playing/paused" answer
  proc.setProgram( defaultParams + QStringList( "isPlaying" ) );
  proc.execute( 250 );
  if( ! proc.readAllStandardOutput().contains( "true" ) )
  {
    proc.kill();
    return false;
  }

  playing_ = true;

  // Run the DCOP client and selectively ask the artist, album and title
  proc.setProgram( defaultParams + QStringList( "artist" ) );
  proc.execute( 250 );
  artist_ = QString::fromUtf8( proc.readAllStandardOutput() ).simplified();

  proc.setProgram( defaultParams + QStringList( "album" ) );
  proc.execute( 250 );
  album_  = QString::fromUtf8( proc.readAllStandardOutput() ).simplified();

  proc.setProgram( defaultParams + QStringList( "title" ) );
  proc.execute( 250 );
  track_ = QString::fromUtf8( proc.readAllStandardOutput() ).simplified();

  proc.kill();
  return true;
}



/**
 * Query Amarok for track information.
 */
bool NowListeningClient::queryAmarok2()
{
  if( alreadyQuery )
  {
    return false;
  }

  alreadyQuery = true;

  // Start the d-bus interface, needed to check the application status and make calls to it
  QDBusInterface dbusAmarok( "org.mpris.amarok", "/Player", "org.freedesktop.MediaPlayer" );

  // See if the application is registered.
  if( ! dbusAmarok.isValid() )
  {
    return false;
  }

#ifdef KMESSDEBUG_NOWLISTENINGCLIENT_GENERAL
  kDebug() << "querying Amarok for now listening information...";
#endif

  // Call Amarok to know whether it's playing, and what
  QDBusReply<QVariantMap> metadata = dbusAmarok.call( QDBus::BlockWithGui, "GetMetadata" );
  QDBusReply<int> isPlaying        = dbusAmarok.call( QDBus::BlockWithGui, "GetStatus"   );

  if( ! isPlaying.isValid() || ! metadata.isValid() )
  {
#ifdef KMESSDEBUG_NOWLISTENINGCLIENT_GENERAL
    kDebug() << "Invalid response from Amarok!";
#endif

    return false;
  }

  // Get the list of properties from the map
  QVariantMap map( metadata.value() );

  // Save the returned values
  album_   = map[ "album"  ].toString();
  track_   = map[ "title"  ].toString();
  artist_  = map[ "artist" ].toString();
  playing_ = ( isPlaying == 0 ); // 0 playing, 1 paused, 2 stopped

  alreadyQuery = false;

  return true;
}



/**
 * Query KsCD for track information.
 */
bool NowListeningClient::queryKsCD()
{
  if( alreadyQuery )
  {
    return false;
  }

  alreadyQuery = true;

  // Start the d-bus interface, needed to check the application status and make calls to it
  QDBusInterface dbusKsCD( "org.kde.kscd", "/CDPlayer", "org.kde.KSCD" );

  // See if the application is registered.
  if( ! dbusKsCD.isValid() )
  {
    return false;
  }

#ifdef KMESSDEBUG_NOWLISTENINGCLIENT_GENERAL
  kDebug() << "querying KsCD for now listening information...";
#endif

  // Call KsCD to know whether it's playing, and what
  QDBusReply<QString> album  = dbusKsCD.call( QDBus::BlockWithGui, "currentAlbum"      );
  QDBusReply<QString> track  = dbusKsCD.call( QDBus::BlockWithGui, "currentTrackTitle" );
  QDBusReply<QString> artist = dbusKsCD.call( QDBus::BlockWithGui, "currentArtist"     );
  QDBusReply<bool> isPlaying = dbusKsCD.call( QDBus::BlockWithGui, "playing"           );

  if( ! isPlaying.isValid() || ! album.isValid() || ! track.isValid() || ! artist.isValid() )
  {
#ifdef KMESSDEBUG_NOWLISTENINGCLIENT_GENERAL
    kDebug() << "Invalid response from KsCD!";
#endif

    return false;
  }

  // Save the returned values
  album_   = album;
  track_   = track;
  artist_  = artist;
  playing_ = isPlaying;

  alreadyQuery = false;

  return true;
}



/**
 * Query Juk for track information.
 */
bool NowListeningClient::queryJuk()
{
  if( alreadyQuery )
  {
    return false;
  }

  alreadyQuery = true;

  // Start the d-bus interface, needed to check the application status and make calls to it
  QDBusInterface dbusJuk( "org.kde.juk", "/Player", "org.kde.juk.player" );

  // See if the application is registered.
  if( ! dbusJuk.isValid() )
  {
    return false;
  }

#ifdef KMESSDEBUG_NOWLISTENINGCLIENT_GENERAL
  kDebug() << "querying Juk for now listening information...";
#endif

  // Call KsCD to know whether it's playing, and what
  QDBusReply<QString> album  = dbusJuk.call( QDBus::BlockWithGui, "trackProperty", "Album"  );
  QDBusReply<QString> track  = dbusJuk.call( QDBus::BlockWithGui, "trackProperty", "Title"  );
  QDBusReply<QString> artist = dbusJuk.call( QDBus::BlockWithGui, "trackProperty", "Artist" );
  QDBusReply<bool> isPlaying = dbusJuk.call( QDBus::BlockWithGui, "playing"                 );

  if( ! isPlaying.isValid() || ! album.isValid() || ! track.isValid() || ! artist.isValid() )
  {
#ifdef KMESSDEBUG_NOWLISTENINGCLIENT_GENERAL
    kDebug() << "Invalid response from Juk!";
#endif

    return false;
  }

  // Save the returned values
  album_   = album;
  track_   = track;
  artist_  = artist;
  playing_ = isPlaying;

  alreadyQuery = false;

  return true;
}



#include "nowlisteningclient.moc"
